/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!

@file 
    $RCSfile: register.h,v $

@brief
    This file abstracts the register read/write routines used in the storport driver.
    
@author 
        
@notes
    
CVS Header Information:
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/initiator/acit/register.h,v 1.3 2006/01/30 23:09:31 prabalm Exp $
*/    
#ifndef __register_h__
#define __register_h__

// 
// Converts a 32 bit structure to a u32. This allows us to use
// srcgen defined 32-bit structs as the value to the register write
// functions below.
//
#define TOU32(_struct_) *((u32*)(&_struct_))
#define TOPU32(_struct_) ((u32*)(&_struct_))

#define ZERO_DESCRIPTOR(_name_) zero_descriptor (TOPU32(_name_), sizeof (_name_) / 4)

STATIC INLINE void zero_descriptor (u32* dw, u32 dwords) {
    OSM_ASSERT (dwords > 0);
    while (dwords--) { *dw++ = 0; }
}


// 
// Macros for reading/writing a protection domain or CSR registers in BladeEngine.
//
#define ACIT_PD_READ(_pcontroller_, _field_ )                                        \
    PdRead (_pcontroller_, STRUCT_FIELD_OFFSET (PROTECTION_DOMAIN_DBMAP, _field_));  \

#define ACIT_PD_WRITE(_pcontroller_, _field_, _value_ )                                                 \
    PdWrite (_pcontroller_, STRUCT_FIELD_OFFSET ( PROTECTION_DOMAIN_DBMAP, _field_ ), TOU32(_value_) )  \

#define ACIT_CSR_READ(_pcontroller_, _field_ )                                      \
    CsrRead( _pcontroller_, STRUCT_FIELD_OFFSET ( BLADE_ENGINE_CSRMAP, _field_ ) )  \

#define ACIT_CSR_WRITE(_pcontroller_, _field_, _value_ )                                            \
    CsrWrite(_pcontroller_, STRUCT_FIELD_OFFSET ( BLADE_ENGINE_CSRMAP, _field_ ), TOU32(_value_) )  \

#define ACIT_CSR_WRITE_CONST(_pcontroller_, _field_, _value_ )                                 \
    CsrWrite(_pcontroller_, STRUCT_FIELD_OFFSET ( BLADE_ENGINE_CSRMAP, _field_ ), (_value_) )  \

#define ACIT_PCICFG0_READ(_pcontroller_, _field_)                                 \
    PciCfgRead (_pcontroller_, STRUCT_FIELD_OFFSET (PCICFG_PF0_CSRMAP, _field_))  \

#define ACIT_PCICFG1_READ(_pcontroller_, _field_)                                 \
    PciCfgRead (_pcontroller_, STRUCT_FIELD_OFFSET (PCICFG_PF1_PF7_CSRMAP, _field_))  \

#define ACIT_PCICFG0_WRITE(_pcontroller_, _field_, _value_)                                        \
    PciCfgWrite (_pcontroller_, STRUCT_FIELD_OFFSET (PCICFG_PF0_CSRMAP, _field_), TOU32(_value_))  \

#define ACIT_PCICFG1_WRITE(_pcontroller_, _field_, _value_)                                        \
    PciCfgWrite (_pcontroller_, STRUCT_FIELD_OFFSET (PCICFG_PF1_PF7_CSRMAP, _field_), TOU32(_value_))  \

/*!
@brief
    This routine reads from a register located within the protection domain
    doorbell space for a given function object.
@param    
    pcontroller     - Pointer to controller object
@param    
    Offset      - The Offset (in bytes) to read from within the function's PD space.
@return
    The value read. 
@note
    IRQL: any
*/
STATIC INLINE
u32 PdRead( PACIT_CONTROLLER_WS pcontroller, u32 Offset )
{
    OSM_ASSERT (pcontroller);
    OSM_ASSERT (Offset < sizeof(PROTECTION_DOMAIN_DBMAP));    

    return OSM_READ32 (pcontroller->pcontroller_key, 
                       pcontroller->bar2_handle, 
                       Offset);
}

/*!
@brief
    This routine writes to a register located within the protection domain
    doorbell space for a given function object.
@param    
    pcontroller     - Pointer to controller object
@param    
    Offset      - The Offset (in bytes) to write to within the function's PD space.
@param    
    Value       - The value to write to the register.
@return
    none
@note
    IRQL: any
*/
STATIC INLINE
void PdWrite( PACIT_CONTROLLER_WS pcontroller, u32 Offset, u32 Value )
{
    OSM_ASSERT (pcontroller);
    OSM_ASSERT (Offset < sizeof(PROTECTION_DOMAIN_DBMAP));    

    OSM_WRITE32 (pcontroller->pcontroller_key, 
                 pcontroller->bar2_handle, 
                 Offset, 
                 Value);
}

/*!
@brief
    This routine reads from a register located within the CSR space for a 
    given function object.
@param    
    pcontroller     - Pointer to controller object
@param    
    Offset      - The Offset (in bytes) to read from within the function's CSR space.
@return
    The value read.    
@note
    IRQL: any
*/
STATIC INLINE
u32 CsrRead( PACIT_CONTROLLER_WS pcontroller, u32 Offset )
{
    OSM_ASSERT (pcontroller);
    OSM_ASSERT (Offset < sizeof(BLADE_ENGINE_CSRMAP));    

    return OSM_READ32 (pcontroller->pcontroller_key, 
                       pcontroller->bar1_handle, 
                       Offset);
}

/*!
@brief
    This routine writes to a register located within the CSR
    space for a given function object.
@param    
    pcontroller     - Pointer to controller object
@param    
    Offset      - The Offset (in bytes) to write to within the function's CSR space.
@param    
    Value       - The value to write to the register.
@return
    none
@note
    IRQL: any
*/
STATIC INLINE
void CsrWrite( PACIT_CONTROLLER_WS pcontroller, u32 Offset, u32 Value )
{
    OSM_ASSERT (pcontroller);
    OSM_ASSERT (Offset < sizeof(BLADE_ENGINE_CSRMAP));    

    OSM_WRITE32 (pcontroller->pcontroller_key, 
                 pcontroller->bar1_handle, 
                 Offset, 
                 Value);
}


STATIC INLINE
u32 PciCfgRead( PACIT_CONTROLLER_WS pcontroller, u32 Offset )
{
    OSM_ASSERT (pcontroller);
    ASSERT( Offset < sizeof(PCICFG0_CSRMAP) );

    return OSM_READ32 (pcontroller->pcontroller_key, 
                       pcontroller->bar0_handle, 
                       Offset);
}

STATIC INLINE
void PciCfgWrite( PACIT_CONTROLLER_WS pcontroller, u32 Offset, u32 Value )
{
    OSM_ASSERT (pcontroller);
    ASSERT( Offset < sizeof(PCICFG0_CSRMAP) );

    OSM_WRITE32 (pcontroller->pcontroller_key, 
                 pcontroller->bar0_handle, 
                 Offset, 
                 Value);
}

#endif

