/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!
  @file
  be_gen_fat.h

  @brief

  This file is meant to be included in other source and not compiled 
  stand-alone. It is expected this file can be shared by drivers
  as well as ARM firmware to collect Fault Analysis Tool log data.

  Source which includes this file must include srcgen bmap structures
  (no-postfix version) with the SG_BMAP_UNION literal defined. 

  The be_gen_check_defines.h file lists a group of macros that must be
  defined before including this file.

  The macros used by ARM firmware are defined in be_utils.h and be_utils_priv.h.

 */
#ifndef __be_gen_fat_h__
#define __be_gen_fat_h__

#include "be2_gen_check_defines.h"
#include "version.h"
#include "fat_header_bmap.h"

#ifndef BE_GEN_FAT_LOG
#error BE_GEN_FAT_LOG not defined
#endif

#ifndef BE_GEN_FAT_PRINTF
#error BE_GEN_FAT_PRINTF not defined
#endif

#ifndef BE_GEN_FAT_MSG_LEVEL
#error BE_GEN_FAT_MSG_LEVEL not defined
#endif

/* Globals that are required for FAT to work accross platforms */
u32 g_GEN_FAT_bytes_used;
u32 g_BE_GEN_FAT_max_buf_size;
u32 *g_BE_GEN_FAT_buffer;

#define FAT_PRINTF BE_GEN_FAT_PRINTF

// Defining various data types such as FAT_VOID, FAT_VOID_STAR, FAT_UINT32 and FAT_INT32.

typedef long int FAT_INT32;
typedef unsigned long int FAT_UINT32;
typedef void* FAT_VOID_STAR;

#define FAT_VOID void
#define FAT_INST_STREAM FAT_UINT32
#define FAT_OUT_TYPE u32

BE_GEN_STATIC_INLINE u32 
FAT_LOG( u32 value ) {
    if(g_GEN_FAT_bytes_used <= (g_BE_GEN_FAT_max_buf_size - 4)) {
        BE_GEN_FAT_LOG(g_BE_GEN_FAT_buffer, g_GEN_FAT_bytes_used, value);
        g_GEN_FAT_bytes_used += 4;
    }
    return value;
}

BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Write_Block(
        u32 *pData,
        u32  size,
        u32  offset)
{
    u32 i;

    for(i = 0; i < size; i += sizeof(u32), offset += sizeof(u32), pData++) {
        BE_GEN_FAT_LOG(g_BE_GEN_FAT_buffer, offset, *pData);
    }

    return size;
}

// FAT read 
#define FAT_CSR_READ(__bus__, __reg__)  FAT_LOG(BE_GEN_CSR_READ(__bus__, __reg__))
#define FAT_PCI0_READ(__bus__, __reg__)  FAT_LOG(BE_GEN_PCI0_READ(__bus__, __reg__))
#define FAT_PCI1_READ(__bus__, __reg__)  FAT_LOG(BE_GEN_PCI1_READ(__bus__, __reg__))
#define FAT_PCI_READ(__bus__, __function__, __reg__)  FAT_LOG(BE_GEN_PCI_READ(__bus__, __function__, __reg__))

// FAT write
#define FAT_CSR_WRITE(__bus__, __reg__, __value__)   \
        BE_GEN_CSR_WRITE(__bus__, __reg__, __value__)
#define FAT_PCI0_WRITE(__bus__, __reg__, __value__)  \
        BE_GEN_PCI0_WRITE(__bus__, __reg__, __value__)
#define FAT_PCI1_WRITE(__bus__, __reg__, __value__)  \
        BE_GEN_PCI1_WRITE(__bus__, __reg__, __value__)
#define FAT_PCI_WRITE(__bus__, __function__, __reg__, __value__)  \
        BE_GEN_PCI_WRITE(__bus__, __function__, __reg__, __value__)

#include "fat_proto.h"             // FAT protos

#define JTAG_EnOCD 0x11
#define JTAG_DbgInt 0x12
#define JTAG_RetDbgInt 0x13
#define JTAG_DisRetOCD 0x14
#define JTAG_ExecDI 0x15
#define JTAG_LoadDIR 0x16
#define JTAG_ScanDDR 0x17
#define JTAG_ReadDOSR 0x18
#define JTAG_ScanDCR 0x19

BE_GEN_STATIC_INLINE void
BE_Gen_FAT_Write_JTAG_Pulse(MPU_JTAG_INSERTION  data)
{
    data.enable_ocd = 1;
    data.override   = 1;
    data.jtck       = 0;
    BE_GEN_CSR_WRITE(0, mpu.jtag_insertion, data.dw);
    BE_GEN_CSR_READ (0, mpu.jtag_insertion);
    data.jtck       = 1;
    BE_GEN_CSR_WRITE(0, mpu.jtag_insertion, data.dw);
    BE_GEN_CSR_READ (0, mpu.jtag_insertion);

    return;
}

BE_GEN_STATIC_INLINE void
BE_Gen_FAT_Write_JTAG_Instruction(
        u32  ep_id,
        u32  data)
{
    u32                     i;
    MPU_JTAG_INSERTION jtag_insertion = {0};

    jtag_insertion.chain  = ep_id;
    jtag_insertion.jtrst  = 1;

    // Select data register
    jtag_insertion.jtms   = 1;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Select instruction register
    jtag_insertion.jtms  = 1;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Capture data register
    jtag_insertion.jtms  = 0;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Shift instruction register
    jtag_insertion.jtms  = 0;
    for(i = 0; i < 5; i++) {
        BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
        jtag_insertion.jtdi = ((data >> i) & 0x01);
        jtag_insertion.jtms = 0;
    }

    // Exit instruction register
    jtag_insertion.jtms  = 1;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Update instruction register
    jtag_insertion.jtms  = 1;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Idle
    jtag_insertion.jtms = 0;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    return;
}

BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Write_JTAG_Data(
        u32  ep_id,
        u32  data,
        u32  len)
{
    u32  i;
    u32  shift_data;
    MPU_JTAG_INSERTION jtag_insertion = {0};

    jtag_insertion.chain  = ep_id;
    jtag_insertion.jtrst  = 1;

    // Select data register
    jtag_insertion.jtms   = 1;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Capture data register
    jtag_insertion.jtms   = 0;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Shift instruction register
    jtag_insertion.jtms   = 0;

    // Write data
    for(i = 0; i < len; i++) {
        // Shift data register
        BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
        jtag_insertion.jtdi = ((data >> i) & 0x01);
        jtag_insertion.jtms = 0;
    }

    // Exit data register
    jtag_insertion.jtms   = 1;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Update data register
    jtag_insertion.jtms   = 1;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Idle
    jtag_insertion.jtms   = 0;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Read data shifted out.
    shift_data = BE_GEN_CSR_READ(0, mpu.jtag_shift);

    return shift_data;
}

BE_GEN_STATIC_INLINE void
BE_Gen_FAT_JTAG_Init(
        u32  ep_id )
{
    MPU_JTAG_INSERTION jtag_insertion = {0};

    jtag_insertion.chain = ep_id;

    // JTAG Init
    jtag_insertion.jtrst = 0;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    jtag_insertion.jtrst = 1;
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);
    BE_Gen_FAT_Write_JTAG_Pulse(jtag_insertion);

    // Put EP into a mode ready to accept instructions.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_EnOCD);
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_DbgInt);

    return;
}

BE_GEN_STATIC_INLINE void
BE_Gen_FAT_JTAG_Exit_OCD(
        u32  ep_id )
{
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_RetDbgInt);
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_DisRetOCD);

    return;
}

BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Get_An(
        u32  ep_id,
        u32  which_reg,
        u32  new_value)
{
    u32            reg_val = 0x00;
    FAT_EP_PROC_OP operation = {0};

    // Write contents of a3 to DDR.
    operation.addrreg   = (u8)which_reg;
    operation.operation = (u8)FAT_EP_WSR;
    operation.altreg    = (u8)FAT_EP_DDR;
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Get original contents of a3 into local register.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
    reg_val = BE_Gen_FAT_Write_JTAG_Data(ep_id, new_value, 32);  // Use anything for new_value if you only want reg_val.

    return reg_val;
}

// This function assumes that a3 has been preserved prior this function being called.
BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Get_UR_Reg(
        u32  ep_id,
        u32  which_reg )
{
    u32 reg_value = 0x00;
    FAT_EP_PROC_OP operation = {0};

    // Load the desired register into a3.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.addrreg   = (which_reg < 0x10) ? which_reg : which_reg - 0x10;
    operation.altreg    = (which_reg < 0x10) ? FAT_EP_A3 : FAT_EP_A3 | 0x10;
    operation.operation = FAT_EP_RUR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Write a3 into DDR.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.addrreg   = FAT_EP_A3;
    operation.altreg    = FAT_EP_DDR;
    operation.operation = FAT_EP_WSR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Get current contents of a3, put garbage in ddr, we'll restore later, at higher level.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
    reg_value = BE_Gen_FAT_Write_JTAG_Data(ep_id, 0xfefefefe, 32);

    return reg_value;
}

BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Get_Special_Register(
        u32  ep_id,
        u8   reg_to_get,
        u32  preserve_restore_a3)
{
    u32            reg_val = 0x00;
    FAT_EP_PROC_OP operation = {0};
    u32            original_a3_contents = 0xffffffff ;

    if(preserve_restore_a3)
        original_a3_contents = BE_Gen_FAT_Get_An(ep_id, FAT_EP_A3, 0xffffffff);  // Use junk value for new_value since we don't care what's in a3 right now.

    // Load the desired register into a3.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.addrreg   = FAT_EP_A3;
    operation.altreg    = reg_to_get;
    operation.operation = FAT_EP_RSR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Write a3 into DDR.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.altreg    = FAT_EP_DDR;
    operation.operation = FAT_EP_WSR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Get current contents of a3, restoring previous value (-1 if not preserved)
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
    reg_val = BE_Gen_FAT_Write_JTAG_Data(ep_id, original_a3_contents, 32);

    // Copy ddr back into a3.
    if(preserve_restore_a3) {
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
        operation.altreg    = FAT_EP_DDR;
        operation.operation = FAT_EP_RSR;

        BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);
    }

    return reg_val;
}

// This turns out not to be particularly useful, in that you are now in OCD
// mode, so the pc is not what was running when you entered OCD but is actually
// in the debug exception handler.

BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Get_PC(
        u32  ep_id)
{
    u32            pcval = 0;
    u32            original_a0 = 0;
    FAT_EP_PROC_OP operation = {0};

    // Write contents of a0 to DDR.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.addrreg   = (u8)FAT_EP_A0;                          // Use a0
    operation.operation = (u8)FAT_EP_WSR;
    operation.altreg    = (u8)FAT_EP_DDR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Get original contents of a0 into local register.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
    original_a0 = BE_Gen_FAT_Write_JTAG_Data(ep_id, 0xffffffff, 32);  // Any old old value will do.

    // Execute call0
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.dw = 0x500000;                                                 // Execute call.
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Write contents of a0 to DDR.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.addrreg   = (u8)FAT_EP_A0;                          // Use a0
    operation.operation = (u8)FAT_EP_WSR;
    operation.altreg    = (u8)FAT_EP_DDR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Get current contents of a0, restoring previous value.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
    pcval = BE_Gen_FAT_Write_JTAG_Data(ep_id, original_a0, 32);

    // Put pcval back so for use by return.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, pcval, 32);

    // Execute ret
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.dw = 0x00d00f;                                                 // Execute ret.n.
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    pcval &= 0x0000ffff;
    pcval -= 3;
    BE_GEN_DEBUG_PRINT("pcval = 0x%08x\n", pcval);

    // Now put the original value of a0 back in DDR.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, original_a0, 32);

    // Copy ddr back into a0.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.addrreg   = (u8)FAT_EP_A0;                          // Use a0
    operation.altreg    = FAT_EP_DDR;
    operation.operation = FAT_EP_RSR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    return pcval;
}

BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Print_EP_JTAG(
        u32  ep_id)
{
    FAT_EP_BE3_PROCESSOR_STATE state;
    FAT_EP_PROC_OP operation = {0};
    u32 i, k;
    u32 original_a3_contents;

    memset(&state, 0x00, sizeof(FAT_EP_BE3_PROCESSOR_STATE));
    state.version = FAT_EP_PROCESSOR_COLLECTION_VERSION_CURRENT;
    state.number_of_address_regs = 32;

    // Put the ep into OCD (On Chip Debugger) state.
    BE_Gen_FAT_JTAG_Init(ep_id);

    state.windowbase        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_WINDOWBASE, 1);
    state.windowstart       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_WINDOWSTART, 1);

    // Only print the current window (16 regs).
    // Get 16 address regs.
    for(i = 0; i < 16; i++) {
        // Move the register value into DDR.
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
        operation.addrreg   = (u8)i;
        operation.operation = (u8)FAT_EP_WSR;
        operation.altreg    = (u8)FAT_EP_DDR;
        BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

        // This is how sw/host/drivers/beklib/fw_util.c finished the read.
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
        state.a[i] = BE_Gen_FAT_Write_JTAG_Data(ep_id, 0xffffffff, 32);

        // Now put the address value back in DDR before we move it back into the addr reg.
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
        BE_Gen_FAT_Write_JTAG_Data(ep_id, state.a[i], 32);

        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
        operation.operation = (u8)FAT_EP_RSR;
        BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);
    }

    original_a3_contents = BE_Gen_FAT_Get_An(ep_id, FAT_EP_A3, 0xffffffff);  // Use junk value for new_value since we don't care what's in a3 right now.

    state.ccount            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_CCOUNT, 0);
    state.debug_cause       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_DEBUGCAUSE, 0);
    state.depc              = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_DEPC, 0);
    state.epc[0]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPC1, 0);
    state.epc[1]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPC2, 0);
    state.epc[2]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPC3, 0);
    state.epc[3]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPC4, 0);
    state.eps[0]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPS2, 0);
    state.eps[1]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPS3, 0);
    state.eps[2]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPS4, 0);
    state.exc_cause         = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCCAUSE, 0);
    state.exc_save[0]       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCSAVE1, 0);
    state.exc_save[1]       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCSAVE2, 0);
    state.exc_save[2]       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCSAVE3, 0);
    state.exc_save[3]       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCSAVE4, 0);
    state.instruction_count = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_ICOUNT, 0);
    state.icountlevel       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_ICOUNTLEVEL, 0);
    state.intenable         = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_INTENABLE, 0);
    state.loop_begin        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_LBEG, 0);
    state.loop_count        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_LCOUNT, 0);
    state.loop_end          = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_LEND, 0);
    state.pr_id             = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_PRID, 0);
    state.ps                = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_PS, 0);
    state.sar               = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_SAR, 0);

    // Get current contents of ddr, restoring previous value of a3.  Throw away current ddr contents.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, original_a3_contents, 32);

    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.addrreg   = (u8)FAT_EP_A3;
    operation.operation = (u8)FAT_EP_RSR;
    operation.altreg    = (u8)FAT_EP_DDR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    state.program_counter   = BE_Gen_FAT_Get_PC(ep_id);

    Print_Msg("\n!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");

    for(k = 0; k < 16; k++) {
        Print_Msg("a%d : 0x%08x\n", k, state.a[k]);
    }
    Print_Msg("window base  : 0x%08x\n", state.windowbase);
    Print_Msg("window start : 0x%08x\n", state.windowstart);
    Print_Msg("ccount       : 0x%08x\n", state.ccount);
    Print_Msg("debug_cause  : 0x%08x\n", state.debug_cause);
    Print_Msg("depc         : 0x%08x\n", state.depc);
    Print_Msg("current pc   : 0x%08x (debug mode pc)\n", state.program_counter);
    for(i = 0, k = 1; i < 4; i++, k++) {
        Print_Msg("epc%d         : 0x%08x\n", k, state.epc[i]);
    }
    for(i = 0, k = 2; i < 3; i++, k++) {
        Print_Msg("eps%d         : 0x%08x\n", k, state.eps[i]);
    }
    Print_Msg("exc_cause    : 0x%08x\n", state.exc_cause);
    for(i = 0, k = 1; i < 4; i++, k++) {
        Print_Msg("exc_save%d    : 0x%08x\n", k, state.exc_save[i]);
    }
    Print_Msg("icount       : 0x%08x\n", state.instruction_count);
    Print_Msg("icountlevel  : 0x%08x\n", state.icountlevel);
    Print_Msg("lbegin       : 0x%08x\n", state.loop_begin);
    Print_Msg("lcount       : 0x%08x\n", state.loop_count);
    Print_Msg("lend         : 0x%08x\n", state.loop_end);
    Print_Msg("prid         : 0x%08x\n", state.pr_id);
    Print_Msg("ps           : 0x%08x\n", state.ps);
    Print_Msg("sar          : 0x%x\n", state.sar);

    // Exit OCD state.
    BE_Gen_FAT_JTAG_Exit_OCD(ep_id);

    return sizeof(FAT_EP_BE3_PROCESSOR_STATE);
}

#define COLLECT_PROGRAM_COUNTER
//#define FAT_RESTORE_DDR
#define FAT_1ST_16
#define FAT_2ND_16
#define FAT_GET_UR

BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Collect_EP_JTAG(
        u32  ep_id,
        u32 *pData)
{
    PFAT_EP_BE3_PROCESSOR_STATE pState = (PFAT_EP_BE3_PROCESSOR_STATE)pData;
    FAT_EP_PROC_OP operation = {0};
    u32 i;
    u32 original_a3_contents, tmp;
    int wb, ws;

    memset(pState, 0xfe, sizeof(FAT_EP_BE3_PROCESSOR_STATE));
    pState->version = FAT_EP_PROCESSOR_COLLECTION_VERSION_CURRENT;
    pState->number_of_address_regs = 32;

    wb = (int)pState->windowbase        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_WINDOWBASE, 1);
    ws = (int)pState->windowstart       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_WINDOWSTART, 1);

    // Get the windowed address regs.
    for(i = 0; i < 16; i++) {
        // Move the register value into DDR.
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
        operation.addrreg   = (u8)i;
        operation.operation = (u8)FAT_EP_WSR;
        operation.altreg    = (u8)FAT_EP_DDR;
        BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
        pState->a[i] = BE_Gen_FAT_Write_JTAG_Data(ep_id, 0xffffffff, 32);

#ifdef FAT_RESTORE_DDR
        // Now put the address value back in DDR before we move it back into the addr reg.
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
        BE_Gen_FAT_Write_JTAG_Data(ep_id, pState->a[i], 32);

        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
        operation.operation = (u8)FAT_EP_RSR;
        BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);
#endif
    }

    original_a3_contents = BE_Gen_FAT_Get_An(ep_id, FAT_EP_A3, 0xffffffff);  // Use junk value for new_value since we don't care what's in a3 right now.

    // OK, now we need to get the AR regs.
    // First, we rotate the window based on wb.
    operation.operation  = 0x04;
    operation.altreg     = 0x08;
    operation.addrreg    = 16 - wb;
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Then get the first 16 registers.
    for(i = 0; i < 16; i++) {
        operation.addrreg   = (u8)i;
        operation.operation = (u8)FAT_EP_WSR;
        operation.altreg    = (u8)FAT_EP_DDR;
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
        BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

        // Swap ddr with junk.
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
        pState->ar[i] = BE_Gen_FAT_Write_JTAG_Data(ep_id, 0xffffffff, 32);
    }

    // Now rotate the window back.
    operation.operation  = 0x04;
    operation.altreg     = 0x08;
    operation.addrreg    = (u8)wb;
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Then rotate the window for the second 16.
    operation.operation  = 0x04;
    operation.altreg     = 0x08;
    operation.addrreg    = (u8)((wb - 2 < 0) ?  wb + 2 : ((wb == 2) ? 2 : wb - 2));

    tmp = (u32)operation.addrreg;
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Get the second 16 registers.
    for(i = 16; i < 32; i++) {
        operation.addrreg   = (u8)i - 16;
        operation.operation = (u8)FAT_EP_WSR;
        operation.altreg    = (u8)FAT_EP_DDR;
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
        BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

        // Swap ddr with junk.
        BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
        pState->ar[i] = BE_Gen_FAT_Write_JTAG_Data(ep_id, 0xffffffff, 32);
    }

    // Rotate the window back.
    operation.operation  = 0x04;
    operation.altreg     = 0x08;
    operation.addrreg    = (u8)(16 - tmp);
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Now get the UR regs.
    for(i = 0; i < 22; i++) {
        pState->ur[i] = BE_Gen_FAT_Get_UR_Reg(ep_id, i);
    }

    pState->ccount            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_CCOUNT, 0);
    pState->debug_cause       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_DEBUGCAUSE, 0);
    pState->depc              = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_DEPC, 0);
    pState->epc[0]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPC1, 0);
    pState->epc[1]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPC2, 0);
    pState->epc[2]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPC3, 0);
    pState->epc[3]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPC4, 0);
    pState->eps[0]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPS2, 0);
    pState->eps[1]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPS3, 0);
    pState->eps[2]            = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EPS4, 0);
    pState->exc_cause         = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCCAUSE, 0);
    pState->exc_save[0]       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCSAVE1, 0);
    pState->exc_save[1]       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCSAVE2, 0);
    pState->exc_save[2]       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCSAVE3, 0);
    pState->exc_save[3]       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_EXCSAVE4, 0);
    pState->instruction_count = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_ICOUNT, 0);
    pState->icountlevel       = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_ICOUNTLEVEL, 0);
    pState->intenable         = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_INTENABLE, 0);
    pState->loop_begin        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_LBEG, 0);
    pState->loop_count        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_LCOUNT, 0);
    pState->loop_end          = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_LEND, 0);
    pState->pr_id             = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_PRID, 0);
    pState->ps                = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_PS, 0);
    pState->sar               = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_SAR, 0);

    pState->ccompare0         = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_CCOMPARE0, 0);
    pState->dbreaka[0]        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_DBREAKA0, 0);
    pState->dbreaka[1]        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_DBREAKA1, 0);
    pState->ibreaka[0]        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_IBREAKA0, 0);
    pState->ibreaka[1]        = BE_Gen_FAT_Get_Special_Register(ep_id, FAT_EP_IBREAKA1, 0);

    // Get current contents of ddr, restoring previous value of a3.  Throw away current ddr contents.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_ScanDDR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, original_a3_contents, 32);

    // Read the contents of DDR into A3.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.addrreg   = (u8)FAT_EP_A3;
    operation.operation = (u8)FAT_EP_RSR;
    operation.altreg    = (u8)FAT_EP_DDR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

#ifdef COLLECT_PROGRAM_COUNTER
    pState->program_counter   = BE_Gen_FAT_Get_PC(ep_id);
#endif

    return sizeof(FAT_EP_BE3_PROCESSOR_STATE);
}

// This routine assumes that a4 and a5 have already been saved.
BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Get_MMR_Value(
        u32 ep_id,
        u32 offset)
{
    FAT_EP_PROC_OP operation = {0};

    if(offset < 0x1000)
        offset +=  0x3000;

    BE_Gen_FAT_Get_An(ep_id, FAT_EP_A5, offset);

    // BE_Gen_FAT_Get_An has moved the value of offset into DDR.  Now actually place that in A5.
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    operation.addrreg   = (u8)FAT_EP_A5;
    operation.altreg    = (u8)FAT_EP_DDR;
    operation.operation = (u8)FAT_EP_RSR;
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // Do memw based on address in a5.
    operation.addrreg   = 0x0c;
    operation.altreg    = 0x05;       //
    operation.operation = 0x00;       //
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // So we have the address of the MMR we want in A5.  l32i.n the value at that address into a4.
    operation.addrreg   = 0x84;       // 8 : l32i.n, 4 : a4.
    operation.altreg    = 0x50;       // 5 : a5 has the address, 0 : additional offset is 0.
    operation.operation = 0x00;
    BE_Gen_FAT_Write_JTAG_Instruction(ep_id, JTAG_LoadDIR);
    BE_Gen_FAT_Write_JTAG_Data(ep_id, operation.dw, 24);

    // MMR value is now in A4.  Get it out while putting junk value back into DDR.
    return BE_Gen_FAT_Get_An(ep_id, FAT_EP_A4, 0xffffffff);
}

//#define INCLUDE_FAT_IRAM
#define COLLECT_FIXED_RAM
#define COLLECT_BANKED_RAM
#define COMPARE_BANKS
#define COLLECT_STAT_RAM
//#define COLLECT_MMRs
#define COLLECT_EP_REGS
#define COLLECT_BANK_FROM_1000

BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Get_EP_Data(
        u32  ep_id,
        u32 *pData,
        u32  size,
        PFAT_EP_IMAGE pep_image,
        u32  address_register,
        u32  data_register,
        u32  stat_control_addr,
        u32  stat_data_addr,
        u32  stat_count)
{
    u32 bytes_processed = 0;

    u32 i = 0, j, k;
    PFAT_EP_SUBSECTION pSubsection;
    u32 start_centi_seconds = 0, last_centi_seconds = 0;
    u32 *pBanks[5] = {NULL};

    j = 0;
    k = 0;
    memset(pBanks, 0, sizeof(pBanks));

#ifdef COLLECT_FIXED_RAM
    if(pep_image->size_of_fixed_ram > 0) {
        // If the size of the fixed ram is > 0, then collect the fixed ram.
        for(i = 0; i < pep_image->number_of_ep_cores; i++) {
            // Fill in the subsection identifier.
            start_centi_seconds = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
            pSubsection = (PFAT_EP_SUBSECTION)pData;
            memset(pSubsection, 0x00, sizeof(FAT_EP_SUBSECTION));
            pSubsection->ep_subsection_signature = BE_FAT_EP_SUBSECTION_SIGNATURE;
            pSubsection->ep_subsection_version   = FAT_EP_SUBSECTION_VERSION_CURRENT;
            pSubsection->ep_which_of_type        = i + 1;
            pSubsection->ep_total_of_type        = pep_image->number_of_ep_cores;
            pSubsection->ep_subsection_type      = FAT_EP_FIXED_ID;
            pSubsection->ep_subsection_size      = sizeof(FAT_EP_SUBSECTION);
            pData += (sizeof(FAT_EP_SUBSECTION) / sizeof(u32));
            bytes_processed += sizeof(FAT_EP_SUBSECTION);

            // Get the fixed ram for each ep core.
            BE_GEN_CSR_WRITE_ADDR(NULL, address_register, pep_image->starting_address_of_fixed_ram );
            for(k = 0; k < pep_image->size_of_fixed_ram; k += sizeof(u32), pData++, bytes_processed += 4, pSubsection->ep_subsection_size += 4) {
                // Get each dword of fixed ram.  Address expected to auto-increment.
                *pData = BE_GEN_CSR_READ_ADDR( NULL, data_register );
            }

            last_centi_seconds = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
            pSubsection->ep_sub_delta_deciseconds = (last_centi_seconds - start_centi_seconds) / 10;
            start_centi_seconds = last_centi_seconds;

            // Banked ram immediately follows the fixed ram.  Get it if we have a size specified.
#ifdef COLLECT_BANKED_RAM
            if(pep_image->size_of_banked_ram > 0) {
                // Init uses a 256 bit gap between fixed and banked ram.  (From there its contiguous.)  Use that model.
#ifndef COLLECT_BANK_FROM_1000
                BE_GEN_CSR_WRITE_ADDR(NULL, address_register, 0x1100 );
#else
                BE_GEN_CSR_WRITE_ADDR(NULL, address_register, 0x1000 );
#endif

                for(j = 0; j < pep_image->number_of_data_banks; j++) {
                    // Fill in the subsection header.
                    pSubsection = (PFAT_EP_SUBSECTION)pData;
                    memset(pSubsection, 0x00, sizeof(FAT_EP_SUBSECTION));
                    pSubsection->ep_subsection_signature = BE_FAT_EP_SUBSECTION_SIGNATURE;
                    pSubsection->ep_subsection_version   = FAT_EP_SUBSECTION_VERSION_CURRENT;
                    pSubsection->ep_which_of_type        = (i * pep_image->number_of_ep_cores) + (j + 1);
                    pSubsection->ep_total_of_type        = pep_image->number_of_ep_cores * pep_image->number_of_data_banks;
                    pSubsection->ep_subsection_type      = FAT_EP_BANKED_ID;
                    pSubsection->ep_subsection_size      = sizeof(FAT_EP_SUBSECTION);
                    pData += (sizeof(FAT_EP_SUBSECTION) / sizeof(u32));
                    bytes_processed += sizeof(FAT_EP_SUBSECTION);
                    pBanks[(i * 2) + j] = pData;

                    for(k = 0; k < pep_image->size_of_banked_ram; k += sizeof(u32), pData++, bytes_processed += 4, pSubsection->ep_subsection_size += 4) {
                        // Get each dword of fixed ram.  Address expected to auto-increment.
                        *pData = BE_GEN_CSR_READ_ADDR( NULL, data_register );
                    }

                    last_centi_seconds = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
                    pSubsection->ep_sub_delta_deciseconds = (last_centi_seconds - start_centi_seconds) / 10;
                    start_centi_seconds = last_centi_seconds;
                }
            }
#endif  // COLLECT_BANKED_RAM
        }
    }
#endif  // COLLECT_FIXED_RAM

#ifdef COMPARE_BANKS
    // Check out banked ram consistency.
    pep_image->banked_ram_consistent = 1;

    for(i = 0; pBanks[i] != NULL; i += 2) {
        // If the banks are the same, then they are in an inconsistent state.
        if(memcmp(pBanks[i], pBanks[i + 1], pep_image->size_of_banked_ram) == 0) {
            pep_image->banked_ram_consistent = 0;
        }
    }
#endif // COMPARE_BANKS

#ifdef COLLECT_STAT_RAM
    // Get ep statistics ram.
    {
        EP_STATS_RW_CONTROL_CSR ep_stats;

        ep_stats.dw = 0;
        ep_stats.ep_stats_read   = 1;
        ep_stats.ep_stats_be     = 0xf;

        pSubsection = (PFAT_EP_SUBSECTION)pData;
        memset(pSubsection, 0x00, sizeof(FAT_EP_SUBSECTION));
        pSubsection->ep_subsection_signature = BE_FAT_EP_SUBSECTION_SIGNATURE;
        pSubsection->ep_subsection_version   = FAT_EP_SUBSECTION_VERSION_CURRENT;
        pSubsection->ep_which_of_type        = 1;
        pSubsection->ep_total_of_type        = 1;
        pSubsection->ep_subsection_type      = FAT_EP_STATISTICS_ID;
        pSubsection->ep_subsection_size      = sizeof(FAT_EP_SUBSECTION) + (stat_count * sizeof(u32));
        pData += (sizeof(FAT_EP_SUBSECTION) / sizeof(u32));

        for(i = 0; i < stat_count; i++, pData++) {
            ep_stats.ep_stats_addr = i;      // Actually the index into the 256 stat area, not the address.
            BE_CSR_WRITE_ADDR(stat_control_addr, ep_stats.dw);
            *pData = BE_CSR_READ_ADDR(stat_data_addr);
        }

        bytes_processed                     += pSubsection->ep_subsection_size;

        last_centi_seconds = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
        pSubsection->ep_sub_delta_deciseconds = (last_centi_seconds - start_centi_seconds) / 10;
        start_centi_seconds = last_centi_seconds;
    }
#endif // COLLECT_STAT_RAM

    // Put the ep into OCD (On Chip Debugger) state.
    BE_Gen_FAT_JTAG_Init(ep_id);

#ifdef COLLECT_EP_REGS
    if(pep_image->size_of_processor_image > 0) {
        // If the size of the image to collect is > 0, then collect data from the processor.
        for(i = 0; i < pep_image->number_of_ep_cores; i++) {
            // Get the data for each ep core.
            pSubsection = (PFAT_EP_SUBSECTION)pData;
            memset(pSubsection, 0x00, sizeof(FAT_EP_SUBSECTION));
            pSubsection->ep_subsection_signature = BE_FAT_EP_SUBSECTION_SIGNATURE;
            pSubsection->ep_subsection_version   = FAT_EP_SUBSECTION_VERSION_CURRENT;
            pSubsection->ep_which_of_type        = i + 1;
            pSubsection->ep_total_of_type        = pep_image->number_of_ep_cores;
            pSubsection->ep_subsection_type      = FAT_EP_PROC_ID;
            pData += (sizeof(FAT_EP_SUBSECTION) / sizeof(u32));

            j = BE_Gen_FAT_Collect_EP_JTAG(ep_id, pData);
            pData += (j / sizeof(u32));
            pSubsection->ep_subsection_size      = sizeof(FAT_EP_SUBSECTION) + j;
            bytes_processed                     += pSubsection->ep_subsection_size;

            last_centi_seconds = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
            pSubsection->ep_sub_delta_deciseconds = (last_centi_seconds - start_centi_seconds) / 10;
            start_centi_seconds = last_centi_seconds;

            FAT_PRINTF (outputfile, "    Collected processor state = %lu bytes\n ", pep_image->size_of_processor_image);
        }
    }
#endif // COLLECT_EP

#ifdef COLLECT_MMRs
    // FAT_EP_MMR_ID
    {
        u32 size = BE_Gen_FAT_Get_MMRs(ep_id, pData);
        pData += (size / sizeof(u32));
        bytes_processed += size;
    }
#endif //COLLECT_MMRs

    // Exit OCD state.
    BE_Gen_FAT_JTAG_Exit_OCD(ep_id);

#ifdef INCLUDE_FAT_IRAM

    if(pep_image->size_of_iram > 0) {
        // Get iram.
        // Fill in the subsection identifier.
        start_centi_seconds = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
        pSubsection = (PFAT_EP_SUBSECTION)pData;
        memset(pSubsection, 0x00, sizeof(FAT_EP_SUBSECTION));
        pSubsection->ep_subsection_signature = BE_FAT_EP_SUBSECTION_SIGNATURE;
        pSubsection->ep_subsection_version   = FAT_EP_SUBSECTION_VERSION_CURRENT;
        pSubsection->ep_which_of_type        = i + 1;
        pSubsection->ep_total_of_type        = pep_image->number_of_ep_cores;
        pSubsection->ep_subsection_type      = FAT_EP_IRAM_ID;
        pSubsection->ep_subsection_size      = sizeof(FAT_EP_SUBSECTION);
        pData += (sizeof(FAT_EP_SUBSECTION) / sizeof(u32));
        bytes_processed += sizeof(FAT_EP_SUBSECTION);

        // Get the fixed ram for each ep core.
        BE_GEN_CSR_WRITE_ADDR(NULL, address_register, 0x8000 );
        for(k = 0; k < pep_image->size_of_iram; k += sizeof(u32), pData++, bytes_processed += 4, pSubsection->ep_subsection_size += 4) {
            // Get each dword of fixed ram.  Address expected to auto-increment.
            *pData = BE_GEN_CSR_READ_ADDR( NULL, data_register );
        }

        last_centi_seconds = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
        pSubsection->ep_sub_delta_deciseconds = (last_centi_seconds - start_centi_seconds) / 10;
        start_centi_seconds = last_centi_seconds;
    }

#endif

    return bytes_processed;
}

#define FAT_COLLECT_BUFFER_SIZE  (32 * 1024)

u32 g_FAT_collect_buffer[ (FAT_COLLECT_BUFFER_SIZE / sizeof(u32)) ];

BE_GEN_STATIC_INLINE u32
BE_Gen_FAT_Get_Processor(
        u32           *pData,
        u32            size,
        u32            ep_id)
{
    PFAT_EP_IMAGE pep_image;
#if defined(BLADE_ENGINE_3) || defined (BLADE_ENGINE_2)
    u32 fixed_size = 4 * 1024, bank_size = 2 * 1024;
    u32 ep_cores = 1, data_banks = 2;
#elif defined(SKYHAWK)
    u32 fixed_size = 6 * 1024, bank_size = 3 * 1024;
    u32 ep_cores = 2, data_banks = 2;
#endif
#ifdef BLADE_ENGINE_3
    u32 proc_size = sizeof(FAT_EP_BE3_PROCESSOR_STATE);
#ifdef INCLUDE_FAT_IRAM
    u32 iram_size = 24 * 1024;
#else
    u32 iram_size = 0;
#endif
#else
    u32 proc_size = sizeof(FAT_EP_BE3_PROCESSOR_STATE);
    u32 iram_size = 0;
#endif
    u32 total_size = 0;
    u32 fixed_address = 0, fixed_address_register, fixed_data_register;
    u32 *pTmp;
    u32 data_bytes;
    u32 stat_control_addr;
    u32 stat_data_addr;
    u32 stat_count = 256;

    FAT_PRINTF (outputfile, "    EP_ID = %lu\n ", ep_id);

    switch(ep_id) {
    case FAT_EP_ID_TPOST:
        fixed_address_register = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, tpost.ep.ep_data_ram_addr);
        fixed_data_register    = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, tpost.ep.ep_data_ram_rwdata);
        stat_control_addr      = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, tpost.ep.ep_stats_control);
        stat_data_addr         = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, tpost.ep.ep_stats_read_data);
        break;

    case FAT_EP_ID_TPRE:
        fixed_address_register = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, tpre.ep.ep_data_ram_addr);
        fixed_data_register    = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, tpre.ep.ep_data_ram_rwdata);
        stat_control_addr      = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, tpre.ep.ep_stats_control);
        stat_data_addr         = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, tpre.ep.ep_stats_read_data);
        break;

    case FAT_EP_TXULP0:
        fixed_address_register = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, txulp0.ep.ep_data_ram_addr);
        fixed_data_register    = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, txulp0.ep.ep_data_ram_rwdata);
        stat_control_addr      = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, txulp0.ep.ep_stats_control);
        stat_data_addr         = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, txulp0.ep.ep_stats_read_data);
        break;

    case FAT_EP_TXULP1:
        fixed_address_register = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, txulp1.ep.ep_data_ram_addr);
        fixed_data_register    = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, txulp1.ep.ep_data_ram_rwdata);
        stat_control_addr      = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, txulp1.ep.ep_stats_control);
        stat_data_addr         = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, txulp1.ep.ep_stats_read_data);
        break;

    case FAT_EP_RXULP0:
        fixed_address_register = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, rxulp0.ep.ep_data_ram_addr);
        fixed_data_register    = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, rxulp0.ep.ep_data_ram_rwdata);
        stat_control_addr      = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, rxulp0.ep.ep_stats_control);
        stat_data_addr         = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, rxulp0.ep.ep_stats_read_data);
        //stat_count             = 0x20;
        break;

    case FAT_EP_RXULP1:
        fixed_address_register = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, rxulp1.ep.ep_data_ram_addr);
        fixed_data_register    = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, rxulp1.ep.ep_data_ram_rwdata);
        stat_control_addr      = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, rxulp1.ep.ep_stats_control);
        stat_data_addr         = BE_GEN_FIELD_OFFSET(BLADE_ENGINE_CSRMAP, rxulp1.ep.ep_stats_read_data);
        //stat_count             = 0x20;
        break;

    default:
        return 0;
        break;
    }

    total_size = (ep_cores * (fixed_size + sizeof(FAT_EP_SUBSECTION))) +
            (ep_cores * (proc_size + sizeof(FAT_EP_SUBSECTION))) +
            (ep_cores * data_banks * (bank_size + sizeof(FAT_EP_SUBSECTION)));

    pep_image = (PFAT_EP_IMAGE)pData;
    memset(pData, 0x00, total_size + sizeof(SG_PACK FAT_EP_IMAGE));
    pep_image->number_of_data_banks          = data_banks;
    pep_image->number_of_ep_cores            = ep_cores;
    pep_image->size_of_processor_image       = proc_size;
    pep_image->size_of_fixed_ram             = fixed_size;
    pep_image->starting_address_of_fixed_ram = fixed_address;
    pep_image->size_of_banked_ram            = bank_size;
    pep_image->size_of_iram                  = iram_size;

    pTmp = pData + (sizeof(SG_PACK FAT_EP_IMAGE) / sizeof(u32));

    data_bytes = BE_Gen_FAT_Get_EP_Data(ep_id, pTmp, total_size, pep_image, fixed_address_register, fixed_data_register, stat_control_addr, stat_data_addr, stat_count);

    FAT_PRINTF (outputfile, "    data_bytes = %lu\n ", data_bytes);

    return (data_bytes + sizeof(SG_PACK FAT_EP_IMAGE));
}

/*!
  @brief
  This is the entrypoint routine for collecting FAT data.

  @param
  function_context - generic function context used in macros (may be NULL if unused)

  @param
  buffer - pointer to buffer memory for storing data (may be NULL if unused)

  @param
  buffer_size - number of bytes available in the buffer

  @return
  This routine returns the number of bytes collected.

  @note
  For ARMfw use, function context and buffers are normally unused. 
  Logging of data is done directly into FLASH memory at a fixed 
  starting offset. However, users can specify however they want to 
  collect data via the BE_GEN_FAT_LOG() macro.

 */
BE_GEN_STATIC_INLINE u32 
BE_Gen_FAT_Data_Collect(
        PVOID function_context,
        PVOID buffer,
        u32 buffer_size
)
{
    SG_PACK FAT_HEADER       fatHeader  = {0};
    SG_PACK FAT_DATA_SECTION fatData    = {0};
    u32                      fatHeader_offset = sizeof(u32);    // Put this immediately after the data collection size.
    u32                      dataHeaderOffset, dataStartWrite;
    u32                      start_centi_seconds = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
    u32                      current_centi_seconds, last_centi_seconds = start_centi_seconds;
    u32                      delta_centi_seconds = 0, final_centi_seconds;
    u32                      eps_to_collect[] = {
            FAT_EP_ID_TPOST,
            FAT_EP_ID_TPRE,
            FAT_EP_TXULP0,
            FAT_EP_TXULP1,
            FAT_EP_RXULP0,
            FAT_EP_RXULP1,
            FAT_EP_END_MARKER
    };
    u32                      ep_data_size = FAT_COLLECT_BUFFER_SIZE;
    int                      i;

    g_BE_GEN_FAT_buffer = (u32 *)buffer;
    g_BE_GEN_FAT_max_buf_size = buffer_size;
    g_GEN_FAT_bytes_used = 0;

    // First four bytes are reserved for writing the total bytes logged.
    // Then comes the header.
    g_GEN_FAT_bytes_used += (sizeof(u32) + sizeof(fatHeader));
    dataHeaderOffset      = g_GEN_FAT_bytes_used;

    // We also need to skip past the location where we will write the data header.
    g_GEN_FAT_bytes_used += (sizeof(fatData) - sizeof(u32));
    dataStartWrite        = g_GEN_FAT_bytes_used;

    BE_Gen_UE_Analyze( function_context, function_context, BE_GEN_FAT_MSG_LEVEL);

    fatData.signature            = BE_FAT_DATA_SIGNATURE;
    fatData.ep_id                = FAT_EP_ID_ALL_CSR;
    fatData.data_type            = FAT_DATA_ID_EP_CSR;
    fatData.version              = FAT_CSR_DATA_VERSION_CURRENT;
    fatData.data_header_version  = FAT_DATA_SECTION_VERSION_CURRENT;
    fatData.section_total_size   = (g_GEN_FAT_bytes_used - dataHeaderOffset);
    current_centi_seconds        = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
    fatData.delta_deciseconds    = (current_centi_seconds - last_centi_seconds) / 10;
    last_centi_seconds           = current_centi_seconds;

    // Now write the data header.
    BE_Gen_FAT_Write_Block((u32 *)&fatData, (sizeof(fatData) - sizeof(u32)), dataHeaderOffset);

    // Now collect the data on the various eps.
    for(i = 0; eps_to_collect[i] != FAT_EP_END_MARKER; i++) {
        // Collect data.
        ep_data_size = BE_Gen_FAT_Get_Processor(g_FAT_collect_buffer, FAT_COLLECT_BUFFER_SIZE, eps_to_collect[i]);

        if(ep_data_size > 0) {
            // Fill out header.
            dataHeaderOffset            = g_GEN_FAT_bytes_used;
            fatData.ep_id               = eps_to_collect[i];
            fatData.data_type           = FAT_DATA_ID_EP_DUMP;
            fatData.version             = FAT_EP_DUMP_VERSION_CURRENT;
            fatData.data_header_version  = FAT_DATA_SECTION_VERSION_CURRENT;
            fatData.section_total_size  = sizeof(fatData) - sizeof(u32) + ep_data_size;

            current_centi_seconds       = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
            fatData.delta_deciseconds   = (current_centi_seconds - last_centi_seconds) / 10;
            last_centi_seconds          = current_centi_seconds;

            // Write header.
            BE_Gen_FAT_Write_Block((u32 *)&fatData, sizeof(fatData) - sizeof(u32), dataHeaderOffset);
            dataHeaderOffset += (sizeof(fatData) - sizeof(u32));

            // Write data (which includes small data preamble SG_PACK FAT_EP_IMAGE).
            BE_Gen_FAT_Write_Block((u32 *)g_FAT_collect_buffer, ep_data_size, dataHeaderOffset);
            g_GEN_FAT_bytes_used       += fatData.section_total_size;
        }
    }

    // Now process the log header.

    memset(&fatHeader, 0, sizeof(fatHeader));

    // Load the header.
    fatHeader.fixed_signature   = (u32)BE_FAT_HDR_SIGNATURE;
    fatHeader.total_file_size   = g_GEN_FAT_bytes_used;
    fatHeader.header_version    = FAT_HDR_VERSION_CURRENT;
    fatHeader.header_size       = sizeof(fatHeader);
    fatHeader.build_version[0]  = BE_MAJOR;
    fatHeader.build_version[1]  = BE_MINOR;
    fatHeader.build_version[2]  = BE_BUILD;
    fatHeader.build_version[3]  = BE_DOT;
    memset(fatHeader.driver_version, 0, sizeof(fatHeader.driver_version));
#ifdef BE_ARM_FIRMWARE
    fatHeader.board_id          = g_seeprom_data.board_id;
#endif
    fatHeader.be_config         = BE_CONFIG;
    // Get the Unique IDentifier (UID).
#if defined(BLADE_ENGINE_2)
    /* Add the UID bytes */
    fatHeader.be_chip_uid[0] = BE_MMR_READ(registers.mpu_local_regmap.uid0);
    fatHeader.be_chip_uid[1] = BE_MMR_READ(registers.mpu_local_regmap.uid1);
#endif
#ifdef BE_ARM_FIRMWARE
    BE_Utils_Get_PCI_IDs_From_PF(0, &fatHeader.pci_ids[0], &fatHeader.pci_ids[1], &fatHeader.pci_ids[2], &fatHeader.pci_ids[3]);
    fatHeader.be_chip_revision  = BE_Utils_Chip_Revision();
#endif

    fatHeader.compressed_size = g_GEN_FAT_bytes_used;

    // Record current timestamp.  Currently, this only indicates how long the card has been running as we do
    // not yet have an ioctl to synchronize the card clock with the system clock.
    final_centi_seconds               = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
    fatHeader.failure_seconds         = BE_GEN_CSR_READ (0, tpre.timestamp_10ms);
    if(fatHeader.failure_seconds >= start_centi_seconds)
        delta_centi_seconds           = fatHeader.failure_seconds - start_centi_seconds;
    else  // Wrapped.  Should not happen.
        delta_centi_seconds           = (0xffffffff - start_centi_seconds) + fatHeader.failure_seconds + 1;

#ifdef FAT_DEBUG_TIME
    Print_Msg ("\n!!!!!   failure_seconds        = %lu\n", fatHeader.failure_seconds);
    Print_Msg ("!!!!!   start_centi_seconds    = %lu\n", start_centi_seconds);
#endif

    fatHeader.completion_deciseconds  = (u16)(delta_centi_seconds / 10);
    fatHeader.failure_seconds        /= 100;

#ifdef FAT_DEBUG_TIME
    Print_Msg ("!!!!!   delta_centi_seconds    = %lu\n", delta_centi_seconds);
#endif
    TRACE_LOG(MSG_LEVEL_ALWAYS, "         Time to complete       = %u.%u\n", TRC_FUNCTION_UNDEF, TRC_PORT_UNDEF, TRC_DW_PARAM2(fatHeader.completion_deciseconds / 10, fatHeader.completion_deciseconds % 10));

    // Now write the header.
    BE_Gen_FAT_Write_Block((u32 *)&fatHeader, sizeof(fatHeader), fatHeader_offset);

    // Indicate the total bytes logged at offset 0
    BE_GEN_FAT_LOG((u32 *)g_BE_GEN_FAT_buffer, 0, g_GEN_FAT_bytes_used);

    ASSERT(g_GEN_FAT_bytes_used <= buffer_size);
    return (g_GEN_FAT_bytes_used);
}   /* end BE_Gen_FAT_Data_Collect */

#endif  // __be_gen_fat_h__
