/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
//
#ifndef __be2_gen_misc_h__
#define __be2_gen_misc_h__

#include "be2_gen_check_defines.h"
#include "be2_gen_api_enum.h"

// Return minimum sized value
#define BE2_GEN_MIN(__a__, __b__) (((__a__)<(__b__)) ? (__a__) : (__b__))
#define BE2_GEN_MAX(__a__, __b__) (((__a__)>(__b__)) ? (__a__) : (__b__))

// Macro to determine if the revision is an ASIC or not
#define BE2_GEN_IS_ASIC(__revid__) (!!((__revid__) < BE2_GEN_ASIC_REV_MAX))

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Chip_Rev
//
//   Returns chip revision number. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type - Chip revision number (0 | 1). 
//---------------------------------------------------------------------------------------
STATIC INLINE u32
BE2_Gen_Chip_Rev (
   IN void* function_context )
{
    u32 revid_value = BE_GEN_CSR_READ(function_context, tpre.rev_id);
    if (!BE2_GEN_IS_ASIC(revid_value))
    {
        /* Convert the FPGA revision number to the equivalent ASIC value */
        revid_value = (revid_value < BE2_GEN_FPGA_A1_REV_MIN) ? 0 : 1;
    }
    return revid_value;
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Is_ASIC
//
//   Determines if the adapter is an FPGA or ASIC. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type - Returns TRUE if an asic, FALSE otherwise.
//---------------------------------------------------------------------------------------
STATIC INLINE boolean
BE2_Gen_Is_ASIC (
   IN void* function_context )
{
#if defined(SIMULATION)
    return TRUE;
#else
    return (BE2_GEN_IS_ASIC(BE_GEN_CSR_READ(function_context, tpre.rev_id)));
#endif
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Turn_Online
//
//   Updates either the online0 or online1 register to turn modules online. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// online1            - TRUE if the online1 PCICFG register should be written. FALSE if
//                      the online0 PCICFG register should be written. 
// value              - The 32-bit value to OR into the requested online register. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Turn_Online (
   IN void*   function_context,
   IN boolean online1,
   IN u32     value )
{
    u32 pci_online = value;

    // Assume register is in same location for each function.
    BE_GEN_C_ASSERT(BE_GEN_FIELD_OFFSET(PCICFG_PF0_CSRMAP, online1) ==
                    BE_GEN_FIELD_OFFSET(PCICFG_PF1_PF7_CSRMAP, online1));
    BE_GEN_C_ASSERT(BE_GEN_FIELD_OFFSET(PCICFG_PF0_CSRMAP, online0) ==
                    BE_GEN_FIELD_OFFSET(PCICFG_PF1_PF7_CSRMAP, online0));

    if (0 != online1)
    {
        pci_online |= BE_GEN_PCI0_READ( function_context, online1 );
        BE_GEN_PCI0_WRITE( function_context, online1, pci_online );
    }
    else
    {
        pci_online |= BE_GEN_PCI0_READ( function_context, online0 );
        BE_GEN_PCI0_WRITE( function_context, online0, pci_online );
    }
}

//---------------------------------------------------------------------------------------
// Function: _BE2_Gen_Wait_Ctx_Done
//
//   Private routine that waits for context memory interface to be free.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
_BE2_Gen_Wait_Ctx_Done (
   IN void*   function_context )
{
    u32 retry_count = 0;
    CTX_CMD_CSR ctx_cmd;

    ctx_cmd.busy = 1;
    while (ctx_cmd.busy)
    {
        ctx_cmd.dw = BE_GEN_CSR_READ(function_context, ctx.ctxcmd);
        if (++retry_count > 0x200)
        {
            BE_GEN_DEBUG_PRINT("Waiting for context interface.\n");
        }
    }
}

//---------------------------------------------------------------------------------------
// Function: _BE2_Gen_Access_Ctx_Mem_Single
//
//   Private routine to perform a single context memory access less than 16 bytes.
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// byte_address       - Context memory byte address. 
// p_data             - Pointer to source or destination data. 
// byte_count         - Number of bytes to read or write. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
_BE2_Gen_Access_Ctx_Mem_Single (
   IN void*   function_context,
   IN u32     byte_address,
   IN OUT u8* p_data,
   IN u32     byte_count,
   IN boolean write_flag )
{
    union {
        u8 bytes[16];
        u32 dw[4];
    } real_data;
    u32 offset = byte_address % 16;
    u32 aligned_address = byte_address / 16;
    u32 byte_enables = 0;
    u32 i;
    CTX_CMD_CSR ctx_cmd;

    BE_GEN_ASSERT (p_data);
    BE_GEN_ASSERT (byte_count > 0);
    BE_GEN_ASSERT ((offset + byte_count) <= 16);

    real_data.dw[0] = 0;
    real_data.dw[1] = 0;
    real_data.dw[2] = 0;
    real_data.dw[3] = 0;

    /* Copy the data to write into the correct byte locations and
     * set the byte enables.
     */  
    for ( i = 0; i < byte_count; i++ ) 
    {
        real_data.bytes[i+offset] = p_data[i];
        byte_enables |= (1 << (i + offset));
    }

    BE_GEN_CSR_WRITE(function_context, ctx.ctxaddr, aligned_address);

    ctx_cmd.dw = 0;
    ctx_cmd.be = byte_enables;
    ctx_cmd.wr = write_flag;
    ctx_cmd.busy = 1;

    if (TRUE == write_flag) 
    {
        BE_GEN_CSR_WRITE(function_context, ctx.ctxdata0, real_data.dw[0]);
        BE_GEN_CSR_WRITE(function_context, ctx.ctxdata1, real_data.dw[1]);
        BE_GEN_CSR_WRITE(function_context, ctx.ctxdata2, real_data.dw[2]);
        BE_GEN_CSR_WRITE(function_context, ctx.ctxdata3, real_data.dw[3]);
        BE_GEN_CSR_WRITE(function_context, ctx.ctxcmd, ctx_cmd.dw);
        _BE2_Gen_Wait_Ctx_Done(function_context);
    } 
    else 
    {
        BE_GEN_CSR_WRITE(function_context, ctx.ctxcmd, ctx_cmd.dw);
        _BE2_Gen_Wait_Ctx_Done(function_context);
        real_data.dw[0] = BE_GEN_CSR_READ(function_context, ctx.ctxdata0);
        real_data.dw[1] = BE_GEN_CSR_READ(function_context, ctx.ctxdata1);
        real_data.dw[2] = BE_GEN_CSR_READ(function_context, ctx.ctxdata2);
        real_data.dw[3] = BE_GEN_CSR_READ(function_context, ctx.ctxdata3);

        /* Copy the read data into the correct byte locations */
        for ( i = 0; i < byte_count; i++ ) 
        {
            p_data[i] = real_data.bytes[i+offset]; 
        }
    }
}

//---------------------------------------------------------------------------------------
// Function: BE2_Gen_Access_Context_Memory
//
//   Accesses chip context memory. 
//
// function_context   - Pointer passed in to client defined GEN macros. Not actively used
//                      by GEN source.
// byte_address       - Context memory byte address. 
// p_data             - Pointer to source or destination data. 
// byte_count         - Number of bytes to read or write. 
// write_flag         - TRUE for write access, FALSE for read access. 
// return return_type - 
//---------------------------------------------------------------------------------------
STATIC INLINE void
BE2_Gen_Access_Context_Memory (
   IN void*   function_context,
   IN u32     byte_address,
   IN OUT u8* p_data,
   IN u32     byte_count,
   IN boolean write_flag )
{
    u32 offset = 0;
    u32 length = 0;
    const u32 access_size = 16;

    if ((byte_address % access_size) != 0) 
    {
        /* Access starting unaligned entry */
        length = BE2_GEN_MIN(byte_count, access_size - (byte_address % access_size));
        _BE2_Gen_Access_Ctx_Mem_Single(function_context, byte_address + offset, &p_data[offset], length, write_flag);
        offset += length;
    }

    /* Write aligned entries */
    while ((byte_count - offset) >= access_size) 
    {
        BE_GEN_ASSERT(offset % access_size == 0);
        _BE2_Gen_Access_Ctx_Mem_Single(function_context, byte_address + offset, &p_data[offset], access_size, write_flag);
        offset += access_size;
    }

    if ((byte_count - offset) > 0) 
    {
        /* Write ending unaligned entry */
        BE_GEN_ASSERT(offset % access_size == 0);
        length = byte_count - offset;
        _BE2_Gen_Access_Ctx_Mem_Single(function_context, byte_address + offset, &p_data[offset], length, write_flag);
        offset += length;
    }

    BE_GEN_ASSERT(offset == byte_count);
}

#endif  // __be2_gen_misc_h__

