/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   beclib_ll.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/beclib_ll.h,v 1.49 2006/04/17 01:35:05 menz Exp $
*/
#ifndef __beclib_lower_h__
#define __beclib_lower_h__

#include "beclib_common.h"  // High and low level API commonalities
#include "emulation_switch_ll.h" // MPU emulation or nothing, depending on BE_EMULATION

SA_CPP_HEADER

typedef struct _BE_ERROR_OBJECT {

    SA_DEV *sa_dev;
    u32 pci_function_number;    
    u32 chip_version; 
    u32 function_type; 

    boolean supports_recovery; // Recovery is supported based on chip revision.
    boolean master;            // Elected master.
    boolean crash_dump;         // 

    u32 recovery_state; // Current recovery state.
    u64 start_time;     // Time at the last state transition during recovery.
    u64 original_time;  // Time at the start of recovery.

    SA_SPINLOCK recovery_lock;  // Protects the recovery_state variable.
    SA_SPINLOCK semaphore_lock; // Held entire duration of semaphore owning.
    SA_IRQ irq; // for the semaphore lock

    // Timeouts in microseconds    
    u64 fat_timeout;            // wait for FAT to complete timeout    
    u64 pci_reset_poll_timeout; // master polling timeout for soft reset
    u64 semaphore_timeout;      // timeout to poll for semaphore
    u64 slave_timeout;          // timeout for slaves to remain quiet
    u64 master_timeout;         // timeout for master to wait for slaves
    u64 after_reset_timeout;    // timeout for master after reset
    
    // Cached UE mask, and final UE that is present at start of recovery.
    u32 ue0, ue1, ue0mask, ue1mask;

    // Optional debug tiny log pointer.
    SA_TLOG *tlog;

} BE_ERROR_OBJECT, *PBE_ERROR_OBJECT;

/*!
@brief
    Internal memory descriptor for memory shared by both the host driver and 
    the controller
*/
typedef struct _BE_SHARED_MEM_DESCRIPTOR {
    PVOID        va;                     //! virtual address
    PSA_SGL      sgl;                    //! memory descriptor list
    u32          tag;                    //! memory tracking tag
} BE_SHARED_MEM_DESCRIPTOR, *PBE_SHARED_MEM_DESCRIPTOR;

/*!
@brief
    Locking structure for beklib.
*/
typedef struct _BE_LOCK
{
    SA_SPINLOCK      lock;
    SA_IRQ           irql;
} BE_LOCK, *PBE_LOCK;

typedef SA_FAST_MUTEX BE_BLOCKING_LOCK, *PBE_BLOCKING_LOCK;


/*!
@brief
    Represents a Blade Engine chip. This is used to manage 
    global chip resources.  Only the ARM uses this from the low
    level API.
*/
typedef struct _BE_CHIP_OBJECT {
    u32              magic;
    u32              ref_count;
    BE_LOCK          lock;
    SA_LIST_ENTRY    function_list_head;

    // High Level
    #if defined(_BE_HIGH_LEVEL_BUILD)
        SA_LIST_ENTRY            chip_list;
        BE_SHARED_MEM_DESCRIPTOR jell;        
    #endif

    #ifdef BE_ENABLE_ULP_CHAOS
        u32 ulpchaos_refcount;
        BE_SHARED_MEM_DESCRIPTOR ulpchaos;
    #endif

    // Emulated    
    #if defined(BE_EMULATION) || defined(BE_SELECTIVE_EMULATION)
        PVOID          pd0_bar_va[8]; // indexed by pci function
        boolean        mpu_initialized;
        BE_MCC_EMUL    mcc_emul;
    #endif
    
} BE_CHIP_OBJECT, *PBE_CHIP_OBJECT;

typedef struct _BE_MCC_WRB_RESPONSE_COPY {

    u16   length;       // bytes in response
    u16   ioctl_offset; // offset within the wrb of the response
    PVOID va;           // user's va to copy response into

} BE_MCC_WRB_RESPONSE_COPY, *PBE_MCC_WRB_RESPONSE_COPY;

typedef struct _BE_MCC_WRB_CONTEXT {

    MCC_WRB_CQE_CALLBACK     internal_callback;             // Function to call on completion
    PVOID                    internal_callback_context;     // Parameter to pass to completion function

    MCC_WRB_CQE_CALLBACK     callback;             // Function to call on completion
    PVOID                    callback_context;     // Parameter to pass to completion function
    
    volatile BESTATUS        *users_final_status; // pointer to a local variable for synchronous commands
    MCC_WRB                  *wrb;                 // pointer to original wrb for embedded commands only
    SA_LIST_ENTRY            next;                 // links context structs together in free list
        
    BE_MCC_WRB_RESPONSE_COPY copy;                 // Optional parameters to copy embedded response to user's va

    #if defined(SA_DEBUG)
        u16 subsystem, opcode;                     // Track this IOCTL for debug builds.
        MCC_WRB *ring_wrb;
        u32 consumed_count;
    #endif
} BE_MCC_WRB_CONTEXT, *PBE_MCC_WRB_CONTEXT;

/*!
@brief
    Represents a function object for network or storage.  This
    is used to manage per-function resources like MCC CQs, etc.
*/
typedef struct _BE_FUNCTION_OBJECT {
    
    u32             magic;      //!< magic for detecting memory corruption.
    PBE_CHIP_OBJECT parent_chip;

    //! Generic system abstraction (SA) device object for accessing registers, physical 
    //! memory, etc.
    SA_DEV *sa_dev;

    u32             emulate;    // if set, MPU is not available.  Emulate everything.    
    u32             pend_queue_driving; // if set, drive the queued WRBs after releasing the WRB lock

    SA_SPINLOCK     post_lock;   // lock for verifying one thread posting wrbs
    SA_SPINLOCK     cq_lock;     // lock for verifying one thread processing cq
    SA_SPINLOCK     mcc_context_lock; // lock for protecting mcc context free list
    SA_IRQ          post_irq;
    SA_IRQ          cq_irq;

    u32             ref_count;  //!< number objects referencing this funcobj
    BE_LOCK         lock;       //!< synchronization lock        
    u32             type;
    u32             pci_function_number;

    SA_LIST_ENTRY   function_list; //!< for linking these objects together

    struct {
        SA_LIST_ENTRY   pd_list_head;    // protection domain list
        SA_LIST_ENTRY   cq_list_head;    // completion queue list
        SA_LIST_ENTRY   eq_list_head;    // event queue list    
        SA_LIST_ENTRY   cxn_list_head;   // list of active TOE connections                
        SA_LIST_ENTRY   eth_sq_list_head; // list of ethernet send rings
        SA_LIST_ENTRY   eth_rq_list_head; // list of ethernet receive rings
        SA_LIST_ENTRY   lro_rq_list_head; // list of LRO receive rings
        SA_LIST_ENTRY   rdma_qp_list_head;// queue pair list
        SA_LIST_ENTRY   rxif_list_head;  // receive interface objects

        struct _BE_MCC_OBJECT*   mcc;    // mcc rings.
        struct _BE_PD_OBJECT*    pd_object; // perceived PD 0        

        union {
            struct {
                SA_LIST_ENTRY wrbq_list_head;
                SA_LIST_ENTRY defq_list_head; 
            } iscsi;

            struct {                
                SA_LIST_ENTRY                   dq_list_head;
                SA_LIST_ENTRY                   sq_list_head;   
                SA_LIST_ENTRY                   rq_list_head;
            } networking;
 
            struct {
                SA_LIST_ENTRY                   wrbq_list_head;
            } fcoe;
        };

        struct _BE_ERROR_OBJECT *errobj;
    } links;
   
    struct {
        PMCC_MAILBOX             va;  // VA to the mailbox
        SA_PHYSICAL_ADDRESS      pa;  // PA to the mailbox
        u32                      length; // byte length of mailbox

        // One default context struct used for posting at least one MCC_WRB
        BE_MCC_WRB_CONTEXT       default_context;
        boolean                  default_context_allocated;
    } mailbox;
        
    struct {
        // Page table    
        u32 num_page_table;
        u32 num_sgl;        

        // Wake on lans configured.
        u32 wol_bitmask;  // bits 0,1,2,3 are set if corresponding index is enabled
        u32 num_jell;
        u32 num_vlan;
        u32 num_template; 
        u32 num_multicast;
        u32 rss_type;
        u32 num_zero;
        u32 num_ooo;        
    } config;

    // High Level API only
    #if defined(_BE_HIGH_LEVEL_BUILD)

        struct {
            PVOID   mailbox_allocated_va; // high level only
            PSA_SGL mailbox_sgl;

            //PSA_SGL template_headers_sgl;
            
            struct _NT_ADAPTER_CONTEXT* AdapterContext;
            
            BE_SHARED_MEM_DESCRIPTOR iscsi_pdu_buffer; // default PDU header for all cxn
            BE_SHARED_MEM_DESCRIPTOR templates; // only used by highlevel API
            BE_SHARED_MEM_DESCRIPTOR template_scratch;                

            struct _CHIPIF_UPCALL *upcall; // Functions for calling back into upper driver    
            BE_SHARED_MEM_DESCRIPTOR ooo_buffers; //!< descriptor for internal out-of-order (ooo) buffer allocation

            PSA_SGL page_table_sgl_array[32];
            u32 num_page_table_sgl;                        
        } hl;
    #endif

    #if defined(BE_EMULATION) || defined(BE_SELECTIVE_EMULATION)
        // Only used during emulation to determine which PD is sending the ioctl.
        u32 pd_id;
        DA_TABLE da_table;
    #endif

    u32 own_semaphore;
    SA_IRQ current_irql;

    boolean is_asic; // Initialized to TRUE if we are running on the ASIC.
    BE_FW_CFG fw_cfg;
    BECLIB_FUNCTION_STATS stats;    

    u32 PendingWRBs;
    SA_LIST_ENTRY InProgressList;
    
    boolean ue;      // UE detected variable causes new ioctls to fail.
    boolean aborted; // Aborted variable tracks if the upper driver has aborted all existing IOCTLs yet.

    //
    // Add LANCER specific variables at the end so that cache/structure
    // alignments will not get messed up.
    //

    // Top level variable to distinguish between BE and LANCER family of adapters
    u32 adapter_family;

    // SLI4 Interface Type. Provides info about BAR and Register layout
    u32 sli_intf_type;

    // Provides info about ASIC Revision
    u32 sli_intf_asic;

} BE_FUNCTION_OBJECT, *PBE_FUNCTION_OBJECT;


/*!
@brief
    Represents a protection domain
*/
typedef struct _BE_PD_OBJECT {
    u32                 magic;
    u32                 ref_count;

    PBE_FUNCTION_OBJECT parent_function;
    SA_LIST_ENTRY       pd_list;

    u32                 pd_id;
    u8*                 pd_page_va;
    u64                 pd_page_pa;

} BE_PD_OBJECT, *PBE_PD_OBJECT;



/*! 
  @brief
      Represents an Event Queue
*/
typedef struct _BE_EQ_OBJECT {
    u32                 magic;
    u32                 ref_count;
    BE_LOCK             lock;

    PBE_FUNCTION_OBJECT parent_function;

    SA_LIST_ENTRY       eq_list;
    SA_LIST_ENTRY       cq_list_head;

    u32                 eq_id;
    EQ_CALLBACK         callback;
    PVOID               callback_context;

    #if defined(_BE_HIGH_LEVEL_BUILD)
        PSA_SGL             sgl;
    #endif

} BE_EQ_OBJECT, *PBE_EQ_OBJECT;


/*!
@brief
    Manages a completion queue
*/
typedef struct _BE_CQ_OBJECT {
    u32                 magic;
    u32                 ref_count;

    PBE_FUNCTION_OBJECT parent_function;
    PBE_EQ_OBJECT       eq_object;
    PBE_EQ_OBJECT       new_eq_object;

    SA_LIST_ENTRY       cq_list;
    SA_LIST_ENTRY       cqlist_for_eq;

    PVOID va;
    u32 num_entries;
   
    CQ_CALLBACK         callback;
    PVOID               callback_context;

    u32                 cq_id;

    #if defined(_BE_HIGH_LEVEL_BUILD)
        PSA_SGL sgl;
    #endif

} BE_CQ_OBJECT, *PBE_CQ_OBJECT;

/*!
@brief
    Manages an ethernet send queue
*/
typedef struct _BE_ETHSQ_OBJECT {
    u32                 magic;

    SA_LIST_ENTRY       list;

    PBE_FUNCTION_OBJECT parent_function;
    PBE_CQ_OBJECT       cq_object;    
    u32                 bid;
    u32                 rid;

    #if defined(_BE_HIGH_LEVEL_BUILD)
        PSA_SGL             sgl;
    #endif

    u32                 doorbell_offset;
    u32                 class_queue_id;

} BE_ETHSQ_OBJECT, *PBE_ETHSQ_OBJECT;


/*!
@brief
    Manages an ethernet receive queue
*/
typedef struct _BE_ETHRQ_OBJECT {
    u32                 magic;
    SA_LIST_ENTRY       list;
    PBE_FUNCTION_OBJECT parent_function;
    u32                 rid;    
    u32                 page_size;    
    PBE_CQ_OBJECT       cq_object;
    PBE_CQ_OBJECT       rss_cq_object[4];
    u8                  rss_cpu_index;

} BE_ETHRQ_OBJECT, *PBE_ETHRQ_OBJECT;


/*!
@brief
    Manages a LRO queue
*/
typedef struct _BE_LRORQ_OBJECT {
    u32                 magic;
    SA_LIST_ENTRY       list;
    PBE_FUNCTION_OBJECT parent_function;
    u32                 rid;    
    u32                 frag_size; 
    PBE_CQ_OBJECT       cq_object;

} BE_LRORQ_OBJECT, *PBE_LRORQ_OBJECT;


/*!
@brief
    Manages an MCC
*/
typedef struct _BE_MCC_OBJECT {
    u32                   magic;
                                
    PBE_FUNCTION_OBJECT   parent_function;
    SA_LIST_ENTRY         mcc_list;

    PBE_CQ_OBJECT         cq_object;

    // Async event callback for MCC CQ.
    MCC_ASYNC_EVENT_CALLBACK async_callback;
    PVOID                    async_context;

    struct {
        struct _BE_MCC_WRB_CONTEXT* base;
        u32 num;
        SA_LIST_ENTRY list_head;
    } wrb_context;

    struct {
        PSA_SGL              sgl;
        BE_SHARED_MEM_DESCRIPTOR sm_descriptor;
        SA_RING ring;
    } sq;

    struct {        
        BE_SHARED_MEM_DESCRIPTOR sm_descriptor;
        SA_RING ring;
    } cq;

    u32 processing; // flag indicating that one thread is processing CQ
    u32 rearm;      // doorbell rearm setting to make sure the active processing thread
                    // rearms the CQ if any of the threads requested it.

    SA_LIST_ENTRY   backlog;
    u32 backlog_length;
    volatile u32 driving_backlog;
    u32 consumed_index;
                            
} BE_MCC_OBJECT, *PBE_MCC_OBJECT;


/*!
@brief
   Manages an iSCSI default PDU
*/
typedef struct _BE_DEFAULT_PDU_QUEUE_OBJECT {
    u32                            magic;

    PBE_FUNCTION_OBJECT            parent_function;
    SA_LIST_ENTRY                  func_list;
    PBE_CQ_OBJECT                  cq_object;
    PSA_SGL                        sgl;
    u32                            rid;
} BE_DEFAULT_PDU_QUEUE_OBJECT, *PBE_DEFAULT_PDU_QUEUE_OBJECT;


typedef struct _BE_ISCSI_WRB_QUEUE_OBJECT {

    u32 magic;

    PBE_FUNCTION_OBJECT parent_function;

    SA_LIST_ENTRY wrbq_list;

    //u32 cra, length;
    u32 cid;
    PSA_SGL sgl;
    u32 length;

    // High Level
    #if defined(_BE_HIGH_LEVEL_BUILD)
        PSA_SGL ioctl_sgl;    
        MCC_WRB_CQE_CALLBACK         callback;
        PVOID                        callback_context;
        PVOID                        se_tcp_socket; // deprecate eventually
    #endif

} BE_ISCSI_WRB_QUEUE_OBJECT, *PBE_ISCSI_WRB_QUEUE_OBJECT;

typedef struct _BE_FCOE_WRB_QUEUE_OBJECT {
        u32 magic;

        PBE_FUNCTION_OBJECT parent_function;

        SA_LIST_ENTRY        wrbq_list;

        //u32 cra, length
        u32  cid;
        PSA_SGL sgl;
        u32 length;

        // High Level
        #if defined(_BE_HIGH_LEVEL_BUILD)
                PSA_SGL ioctl_sgl;
                MCC_WRB_CQE_CALLBACK    callback;
                PVOID                   callback_context;
                PVOID                   se_tcp_socket;
        #endif

} BE_FCOE_WRB_QUEUE_OBJECT, *PBE_FCOE_WRB_QUEUE_OBJECT;




/*!
@brief
   Manages iSCSI connection resources like CID, WRB ring, etc.
*/
typedef struct _BE_ISCSI_CONNECTION_OBJECT
{
    u32                          magic;
    PBE_FUNCTION_OBJECT          parent_function;    

    PBE_CQ_OBJECT                cq;
    PBE_ISCSI_WRB_QUEUE_OBJECT   wrbq;
    PBE_DEFAULT_PDU_QUEUE_OBJECT defq;

    SA_LIST_ENTRY                connection_list;
}
BE_ISCSI_CONNECTION_OBJECT, *PBE_ISCSI_CONNECTION_OBJECT;

typedef struct _BE_FCOE_INTERFACE_OBJECT
{
        u32             magic;
        PBE_FUNCTION_OBJECT     parent_function;

        PBE_CQ_OBJECT                   cq;
        PBE_FCOE_WRB_QUEUE_OBJECT       wrbq;

        SA_LIST_ENTRY           interface_list;
}
BE_FCOE_INTERFACE_OBJECT, *PBE_FCOE_INTERFACE_OBJECT;

/*!
@brief
   Manages a TOE default queue
*/
typedef struct _BE_TOE_DQ_OBJECT
{
    u32                 magic;
    SA_LIST_ENTRY       list;
    PBE_FUNCTION_OBJECT parent_function;
    PSA_SGL             sgl;
    u32                 id;
    u32                 default_buffer_length; // bytes per buffer
    u32                 num_offloads;

}
BE_TOE_DQ_OBJECT, *PBE_TOE_DQ_OBJECT;

/*!
@brief
   Manages a TOE send queue. this is an internal object only
*/
typedef struct _BE_TOE_SQ_OBJECT {
    u32              magic;
    u32 cid;

    SA_LIST_ENTRY list;

    struct _BE_FUNCTION_OBJECT*       parent_function;
    struct _BE_TOE_CONNECTION_OBJECT* parent_connection;

    PBE_CQ_OBJECT  cq_object;
    PSA_SGL        sgl;
    u32 size_encoding;
    boolean offloaded;

    // High Level
    #if defined(_BE_HIGH_LEVEL_BUILD)
        PSA_SGL ioctl_sgl;    
    #endif

} BE_TOE_SQ_OBJECT, *PBE_TOE_SQ_OBJECT;

/*!
@brief
   Manages a TOE Receive queue. This is an internal object only
*/
typedef struct _BE_TOE_RQ_OBJECT
{
    u32           magic;
    u32           cid;
    struct        _BE_FUNCTION_OBJECT*       parent_function;
    struct        _BE_TOE_CONNECTION_OBJECT* parent_connection;
    PBE_CQ_OBJECT cq_object;
    PSA_SGL       sgl;
    u32           size_encoding;
    boolean       offloaded;
    SA_LIST_ENTRY list;
}
BE_TOE_RQ_OBJECT, *PBE_TOE_RQ_OBJECT;

/*!
@brief
   Manages TOE connection resources like CID, send ring, etc.
*/
typedef struct _BE_TOE_CONNECTION_OBJECT
{
    u32                 magic;
    PBE_FUNCTION_OBJECT parent_function;
    PBE_TOE_DQ_OBJECT   dq;
    PBE_TOE_SQ_OBJECT   sq;
    PBE_TOE_RQ_OBJECT   rq;
    SA_LIST_ENTRY       connection_list;    
}
BE_TOE_CONNECTION_OBJECT, *PBE_TOE_CONNECTION_OBJECT;

/*!
@brief
   Manages a RDMA Queue Pair. This is an internal object only.
*/
typedef struct _BE_RDMA_QP_OBJECT {
    u32                                 magic;
    u32                                 qp_id;
    boolean                             offloaded;

    SA_LIST_ENTRY                       list;
    struct _BE_FUNCTION_OBJECT*         parent_function;
    struct _BE_RDMA_CXN_OBJECT*         parent_connection;
    
    // Transmit part of QP
    PBE_CQ_OBJECT                       tx_cq_object;
    PSA_SGL                             tx_sgl;
    u32                                 tx_size_encoding;

    // Receive part of QP
    PBE_CQ_OBJECT                       rx_cq_object;
    PSA_SGL                             rx_sgl;
    u32                                 rx_size_encoding;
   
    #if defined(_BE_HIGH_LEVEL_BUILD)
    // High Level
    PSA_SGL ioctl_sgl;    
    #endif

} BE_RDMA_QP_OBJECT, *PBE_RDMA_QP_OBJECT;

/*!
@brief
   RDMA connection object. This is an internal object only!
*/
typedef struct _BE_RDMA_CXN_OBJECT
{
    u32                 magic;
    PBE_FUNCTION_OBJECT parent_function;
    SA_LIST_ENTRY       connection_list;    
} BE_RDMA_CXN_OBJECT, *PBE_RDMA_CXN_OBJECT;





// VLAN object
typedef struct _BE2_VLAN_OBJECT 
{
    u32 magic;
    PBE_FUNCTION_OBJECT parent_function;
    u32 handle;

    u16 mode; // VLAN_OBJECT_QINQ_MODE_ENUM
    u16 tag; // tag value
    
} BE2_VLAN_OBJECT, *PBE2_VLAN_OBJECT;

// MAC address object
typedef struct _BE2_MAC_ADDR_OBJECT 
{
    u32 magic;
    PBE_FUNCTION_OBJECT parent_function;
    
    u32 handle; // Opaque handle
    SA_MAC_ADDRESS mac_address; // MAC
    boolean unicast; // TRUE = unicast, FALSE = multicast
    boolean perfect; // TRUE = perfect multicast, FALSE = hash multicast
} BE2_MAC_ADDR_OBJECT, *PBE2_MAC_ADDR_OBJECT;


// Receive interface object 
typedef struct _BE_RXIF_OBJECT 
{
    u32 magic;
        
    u32 domain;
    u32 handle;
    u32 mac_id;

    u32 capability_flags;
    u32 enabled_flags;
    PBE_FUNCTION_OBJECT parent_function;
    SA_LIST_ENTRY list;
} BE_RXIF_OBJECT, *PBE_RXIF_OBJECT;


// Queue context header -- the required software information for queueing a WRB.
typedef struct _BE_QUEUE_DRIVER_CONTEXT {
    MCC_WRB_CQE_CALLBACK     internal_callback;             // Function to call on completion
    PVOID                    internal_callback_context;     // Parameter to pass to completion function

    MCC_WRB_CQE_CALLBACK     callback;             // Function to call on completion
    PVOID                    callback_context;     // Parameter to pass to completion function
        
    BE_MCC_WRB_RESPONSE_COPY copy;                 // Optional parameters to copy embedded response to user's va
    
    PVOID optional_ioctl_va;

    SA_LIST_ENTRY list;

    u32 bytes;

} BE_QUEUE_DRIVER_CONTEXT, *PBE_QUEUE_DRIVER_CONTEXT;

/*lint -e413 */
/*lint -e831 */
/*lint -e30 */
/*lint -e84 */
// Common MCC WRB header that all commands require.
typedef struct _BE_MCC_WRB_HEADER {
    u8 rsvd[(SA_FIELD_OFFSET (MCC_WRB, payload))];    
} BE_MCC_WRB_HEADER, *PBE_MCC_WRB_HEADER;
SA_GLOBAL_C_ASSERT (queue_header, sizeof (BE_MCC_WRB_HEADER) == SA_FIELD_OFFSET (MCC_WRB, payload));
/*lint -restore */

// All non embedded commands supported by beclib functions only allow 1 SGE.  This
// queue context handles them all.
typedef struct _BE_NONEMBEDDED_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT context;
    BE_MCC_WRB_HEADER       wrb_header;
    MCC_SGE                 sge[1];
} BE_NONEMBEDDED_QUEUE_CONTEXT, *PBE_NONEMBEDDED_QUEUE_CONTEXT;

// ------------------------------------------------------------------------
//  This section contains the specific queue struct for each command.
//  The user could always provide a BE_GENERIC_QUEUE_CONTEXT but this is a 
//  rather large struct.  By using the specific struct, memory consumption 
//  can be reduced.
// ------------------------------------------------------------------------
typedef BE_NONEMBEDDED_QUEUE_CONTEXT   BE_TOE_OFFLOAD_QUEUE_CONTEXT, *PBE_TOE_OFFLOAD_QUEUE_CONTEXT;
typedef BE_NONEMBEDDED_QUEUE_CONTEXT   BE_TOE_UPDATE_QUEUE_CONTEXT,  *PBE_TOE_UPDATE_QUEUE_CONTEXT;
typedef BE_NONEMBEDDED_QUEUE_CONTEXT   BE_TOE_QUERY_QUEUE_CONTEXT,   *PBE_TOE_QUERY_QUEUE_CONTEXT;
typedef BE_NONEMBEDDED_QUEUE_CONTEXT   BE_ETH_STATS_QUEUE_CONTEXT,   *PBE_ETH_STATS_QUEUE_CONTEXT;
typedef BE_NONEMBEDDED_QUEUE_CONTEXT   BE_ETH_NETWORK_STATS_QUEUE_CONTEXT, *PBE_ETH_NETWORK_STATS_QUEUE_CONTEXT;
typedef BE_NONEMBEDDED_QUEUE_CONTEXT   BE_TOE_STATS_QUEUE_CONTEXT,   *PBE_TOE_STATS_QUEUE_CONTEXT;
typedef BE_NONEMBEDDED_QUEUE_CONTEXT   BE_TCP_STATS_QUEUE_CONTEXT,   *PBE_TCP_STATS_QUEUE_CONTEXT;
typedef BE_NONEMBEDDED_QUEUE_CONTEXT   BE_FLASHACCESS_QUEUE_CONTEXT, *PBE_FLASHACCESS_QUEUE_CONTEXT;
typedef BE_NONEMBEDDED_QUEUE_CONTEXT   BE_MGMTIOCTLS_QUEUE_CONTEXT,  *PBE_MGMTIOCTLS_QUEUE_CONTEXT;

typedef struct _BE_SET_TCP_PARAMETERS_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT         context;
    BE_MCC_WRB_HEADER               wrb_header;
    IOCTL_COMMON_SET_TCP_PARAMETERS ioctl;
} BE_SET_TCP_PARAMETERS_QUEUE_CONTEXT, *PBE_SET_TCP_PARAMETERS_QUEUE_CONTEXT; 

typedef struct _ENABLE_DISABLE_BEACON_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT             context;
    BE_MCC_WRB_HEADER                   wrb_header;
    IOCTL_COMMON_ENABLE_DISABLE_BEACON  ioctl;
} BE_ENABLE_DISABLE_BEACON_QUEUE_CONTEXT, *PBE_ENABLE_DISABLE_BEACON_QUEUE_CONTEXT;

typedef struct _GET_BEACON_STATE_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT              context;
    BE_MCC_WRB_HEADER                    wrb_header;
    IOCTL_COMMON_GET_BEACON_STATE        ioctl;
} BE_GET_BEACON_STATE_QUEUE_CONTEXT, *PBE_GET_BEACON_STATE_QUEUE_CONTEXT; 

typedef struct _GET_TEMPERATURE_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT              context;
    BE_MCC_WRB_HEADER                    wrb_header;
    IOCTL_COMMON_GET_CNTL_ADDITIONAL_ATTRIBUTES         ioctl;
} BE_GET_TEMPERATURE_QUEUE_CONTEXT, *PBE_GET_TEMPERATURE_QUEUE_CONTEXT; 

typedef struct _BE_QUERY_DCN_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT         context;
    BE_MCC_WRB_HEADER               wrb_header;
    IOCTL_DCBX_GET_DCBX_CFG         ioctl;
} BE_QUERY_DCN_QUEUE_CONTEXT, *PBE_QUERY_DCN_QUEUE_CONTEXT; 

typedef struct _SET_EXIT_LOOPBACK_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT              context;
    BE_MCC_WRB_HEADER                    wrb_header;
    IOCTL_LOWLEVEL_SET_LOOPBACK_MODE     ioctl;
}BE_SET_EXIT_LOOPBACK_CONTEXT, *PBE_SET_EXIT_LOOPBACK_CONTEXT;

typedef struct _LOOPBACK_TEST_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT              context;
    BE_MCC_WRB_HEADER                    wrb_header;
    IOCTL_LOWLEVEL_LOOPBACK_TEST         ioctl;
}BE_LOOPBACK_TEST_CONTEXT, *PBE_LOOPBACK_TEST_CONTEXT;

/*lint -e413 */
/*lint -e30 */
/*lint -e831 */
typedef struct _BE_TOE_UPLOAD_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT     context;
    BE_MCC_WRB_HEADER           wrb_header;
    u8                          ioctl[SA_FIELD_OFFSET (IOCTL_COMMON_TCP_UPLOAD, params.request) + 
                                      SA_SIZEOF_FIELD (IOCTL_COMMON_TCP_UPLOAD, params.request)];
} BE_TOE_UPLOAD_QUEUE_CONTEXT, *PBE_TOE_UPLOAD_QUEUE_CONTEXT;
/*lint -restore */

//<no_inbox>
typedef struct _BE_TOE_RQ_CREATE_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT    context;
    BE_MCC_WRB_HEADER          wrb_header;
    IOCTL_TOE_RQ_CREATE ioctl;
} BE_TOE_RQ_CREATE_QUEUE_CONTEXT, *PBE_TOE_RQ_CREATE_QUEUE_CONTEXT;

typedef struct _BE_TOE_SQ_CREATE_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_TOE_WRBQ_CREATE ioctl;
} BE_TOE_SQ_CREATE_QUEUE_CONTEXT, *PBE_TOE_SQ_CREATE_QUEUE_CONTEXT;
//</no_inbox>

typedef struct _BE_LINK_STATUS_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT             context;
    BE_MCC_WRB_HEADER                   wrb_header;
    IOCTL_COMMON_NTWK_LINK_STATUS_QUERY ioctl;
} BE_LINK_STATUS_QUEUE_CONTEXT, *PBE_LINK_STATUS_QUEUE_CONTEXT;

typedef struct _BE_MULTICAST_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT         context;
    BE_MCC_WRB_HEADER               wrb_header;
    IOCTL_COMMON_NTWK_MULTICAST_SET ioctl;
} BE_MULTICAST_QUEUE_CONTEXT, *PBE_MULTICAST_QUEUE_CONTEXT;

typedef struct _BE_WAKE_ON_LAN_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT context;
    BE_MCC_WRB_HEADER       wrb_header;
    IOCTL_ETH_ACPI_CONFIG   ioctl;
} BE_WAKE_ON_LAN_QUEUE_CONTEXT, *PBE_WAKE_ON_LAN_QUEUE_CONTEXT;

typedef struct _BE_VLAN_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT       context;
    BE_MCC_WRB_HEADER             wrb_header;
    IOCTL_COMMON_NTWK_VLAN_CONFIG ioctl;
} BE_VLAN_QUEUE_CONTEXT, *PBE_VLAN_QUEUE_CONTEXT;

typedef struct _BE_PROMISCUOUS_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT context;
    BE_MCC_WRB_HEADER       wrb_header;
    IOCTL_ETH_PROMISCUOUS   ioctl;
} BE_PROMISCUOUS_QUEUE_CONTEXT, *PBE_PROMISCUOUS_QUEUE_CONTEXT;

typedef struct _BE_RSS_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT context;
    BE_MCC_WRB_HEADER       wrb_header;
    IOCTL_ETH_RSS_CONFIG    ioctl;
} BE_RSS_QUEUE_CONTEXT, *PBE_RSS_QUEUE_CONTEXT;

typedef struct _BE_NOP_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT context;
    BE_MCC_WRB_HEADER       wrb_header;
    IOCTL_COMMON_NOP        ioctl;
} BE_NOP_QUEUE_CONTEXT, *PBE_NOP_QUEUE_CONTEXT;

typedef struct _BE_RXF_FILTER_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT         context;
    BE_MCC_WRB_HEADER               wrb_header;
    IOCTL_COMMON_NTWK_RX_FILTER ioctl;
} BE_RXF_FILTER_QUEUE_CONTEXT, *PBE_RXF_FILTER_QUEUE_CONTEXT;

#ifdef FUNCTION_RDMA
typedef struct _BE_RDMA_QP_CREATE_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_RDMA_QP_CREATE         ioctl;
} BE_RDMA_QP_CREATE_QUEUE_CONTEXT, *PBE_RDMA_QP_CREATE_QUEUE_CONTEXT;

typedef struct _BE_RDMA_QP_DESTROY_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_RDMA_QP_DESTROY        ioctl;
} BE_RDMA_QP_DESTROY_QUEUE_CONTEXT, *PBE_RDMA_QP_DESTROY_QUEUE_CONTEXT;

typedef struct _BE_RDMA_QP_QUERY_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_RDMA_QP_QUERY          ioctl;
} BE_RDMA_QP_QUERY_QUEUE_CONTEXT, *PBE_RDMA_QP_QUERY_QUEUE_CONTEXT;

typedef struct _BE_RDMA_QP_MODIFY_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_RDMA_QP_QUERY          ioctl;
} BE_RDMA_QP_MODIFY_QUEUE_CONTEXT, *PBE_RDMA_QP_MODIFY_QUEUE_CONTEXT;

typedef struct _BE_RDMA_CXN_OFFLOAD_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_RDMA_CXN_OFFLOAD       ioctl;
} BE_RDMA_CXN_OFFLOAD_QUEUE_CONTEXT, *PBE_RDMA_CXN_OFFLOAD_QUEUE_CONTEXT;

typedef struct _BE_RDMA_CXN_UPDATE_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_RDMA_CXN_UPDATE        ioctl;
} BE_RDMA_CXN_UPDATE_QUEUE_CONTEXT, *PBE_RDMA_CXN_UPDATE_QUEUE_CONTEXT;

typedef struct _BE_RDMA_CXN_UPLOAD_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_RDMA_CXN_UPLOAD        ioctl;
} BE_RDMA_CXN_UPLOAD_QUEUE_CONTEXT, *PBE_RDMA_CXN_UPLOAD_QUEUE_CONTEXT;
#endif

typedef struct _BE_EQ_MODIFY_DELAY_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_COMMON_MODIFY_EQ_DELAY ioctl;
} BE_EQ_MODIFY_DELAY_QUEUE_CONTEXT, *PBE_EQ_MODIFY_DELAY_QUEUE_CONTEXT;

typedef struct _BE_CEV_MODIFY_MSI_MESSAGES_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_COMMON_CEV_MODIFY_MSI_MESSAGES ioctl;
} BE_CEV_MODIFY_MSI_MESSAGES_QUEUE_CONTEXT, *PBE_CEV_MODIFY_MSI_MESSAGES_QUEUE_CONTEXT;

//TODO_BE2: Review
typedef struct _BE_COMMON_SET_QOS_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    IOCTL_COMMON_SET_QOS         ioctl;
} BE_COMMON_SET_QOS_QUEUE_CONTEXT, *PBE_COMMON_SET_QOS_QUEUE_CONTEXT;

//TODO_BE2: Review
typedef struct _BE_PD_READ_WRITE_MAC_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT      context;
    BE_MCC_WRB_HEADER            wrb_header;
    union {
        IOCTL_COMMON_NTWK_MAC_QUERY query;
        IOCTL_COMMON_NTWK_MAC_SET   set;
    } ioctl;
} BE_PD_READ_WRITE_MAC_QUEUE_CONTEXT, *PBE_PD_READ_WRITE_MAC_QUEUE_CONTEXT;

typedef struct _BE_MANAGE_FAT_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT context;    
    BE_MCC_WRB_HEADER       wrb_header;
    IOCTL_COMMON_GET_FAT    ioctl;    
} BE_MANAGE_FAT_QUEUE_CONTEXT, *PBE_MANAGE_FAT_QUEUE_CONTEXT;

typedef struct _BE_IF_ADD_MAC_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT    context;    
    BE_MCC_WRB_HEADER          wrb_header;
    IOCTL_COMMON_NTWK_PMAC_ADD ioctl;        
} BE_IF_ADD_MAC_QUEUE_CONTEXT, *PBE_IF_ADD_MAC_QUEUE_CONTEXT;

// The generic context is the largest size that would be required.  It is the software
// context plus an entire WRB.
typedef struct _BE_GENERIC_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT context;    
    BE_MCC_WRB_HEADER       wrb_header;
    MCC_WRB_PAYLOAD         payload;    
} BE_GENERIC_QUEUE_CONTEXT, *PBE_GENERIC_QUEUE_CONTEXT;

SA_GLOBAL_C_ASSERT (generic_queue_context, 
                    sizeof (BE_GENERIC_QUEUE_CONTEXT) >= sizeof (BE_QUEUE_DRIVER_CONTEXT) + sizeof (MCC_WRB));

typedef BE_GENERIC_QUEUE_CONTEXT BE_CONFIG_RED_QUEUE_CONTEXT, *PBE_CONFIG_RED_QUEUE_CONTEXT;
typedef BE_GENERIC_QUEUE_CONTEXT BE_CONFIG_PORT_EQUALIXATION_QUEUE_CONTEXT, *PBE_CONFIG_PORT_EQUALIXATION_QUEUE_CONTEXT;
typedef BE_GENERIC_QUEUE_CONTEXT BE_FUNCTION_PRIVILEGES_CONTEXT, *PBE_FUNCTION_PRIVILEGES_CONTEXT;

//
// Types for the BE_QUEUE_CONTEXT object.
//
#define BE_QUEUE_INVALID        (0)
#define BE_QUEUE_TOE_OFFLOAD    (0xA001)
#define BE_QUEUE_TOE_UPLOAD     (0xA002)
#define BE_QUEUE_TOE_UPDATE     (0xA003)
#define BE_QUEUE_TOE_RQ_CREATE  (0xA004)
#define BE_QUEUE_TOE_SQ_CREATE  (0xA005)
#define BE_QUEUE_LINK_STATUS    (0xA006)
#define BE_QUEUE_ETH_STATS      (0xA007)
#define BE_QUEUE_TOE_STATS      (0xA008)
#define BE_QUEUE_TCP_STATS      (0xA009)
#define BE_QUEUE_MULTICAST      (0xA00A)
#define BE_QUEUE_VLAN           (0xA00B)
#define BE_QUEUE_RSS            (0xA00C)

#define BE_QUEUE_PROMISCUOUS    (0xA00E)
#define BE_QUEUE_WAKE_ON_LAN    (0xA00F)
#define BE_QUEUE_NOP            (0xA010)


// These define constant 32-bit values that are stored as the magic number in each
// data type.  The magic number is a debugging tool. In debug mode we use asserts to
// constantly check the integrity of the magic value, which spots double frees and some
// memory corruption.  In release mode, the magic value can be used for post-mortem 
// debugging without symbols to locate data types in memory -- this is easier using
// 4 character strings.
#define BE_CHIP_MAGIC              SA_TAG("beco")
#define BE_FUNCTION_MAGIC          SA_TAG("befo")
#define BE_PD_MAGIC                SA_TAG("bepd")
#define BE_CQ_MAGIC                SA_TAG("becq")
#define BE_EQ_MAGIC                SA_TAG("beeq")
#define BE_ETHSQ_MAGIC             SA_TAG("bees")
#define BE_ETHRQ_MAGIC             SA_TAG("beer")
#define BE_LRORQ_MAGIC             SA_TAG("boob")
#define BE_MCC_MAGIC               SA_TAG("bemc")
#define BE_DEFAULT_PDU_QUEUE_MAGIC SA_TAG("bedp")
#define BE_ISCSI_CONNECTION_MAGIC  SA_TAG("beic")
#define BE_FCOE_INTERFACE_MAGIC    SA_TAG("befi")
#define BE_TOE_CONNECTION_MAGIC    SA_TAG("betc")
#define BE_TOE_SQ_MAGIC            SA_TAG("besq")
#define BE_TOE_RQ_MAGIC            SA_TAG("berq")
#define BE_TOE_DQ_MAGIC            SA_TAG("bedq")
#define BE_ISCSI_WRB_QUEUE_MAGIC   SA_TAG("beiq")
#define BE_FCOE_WRB_QUEUE_MAGIC    SA_TAG("befq")
#define BE_RDMA_QP_MAGIC           SA_TAG("beqp")
#define BE_RDMA_CXN_MAGIC          SA_TAG("brcx")
#define BE_RXIF_MAGIC              SA_TAG("brif")

// Macro to declare an inline function that asserts on both a pointer and 
// the magic value.  Inline function nicely forces type safety at compile time.
#define MAGIC_ASSERT_DECLARE(_short_type_)                                      \
    STATIC INLINE void _short_type_##_ASSERT (BE_##_short_type_##_OBJECT *p) {  \
        ASSERT (p);                                                             \
        ASSERT (p->magic == BE_##_short_type_##_MAGIC);                         \
    }                                                                           \

// Declare inline functions, e.g. CHIP_ASSERT, that check the
// pointer for non-NULL and verify the magic constant.  Since they are functions,
// they also verify the type of the pointer at compile time.
MAGIC_ASSERT_DECLARE (CHIP)
MAGIC_ASSERT_DECLARE (FUNCTION)
MAGIC_ASSERT_DECLARE (PD)
MAGIC_ASSERT_DECLARE (CQ)
MAGIC_ASSERT_DECLARE (EQ)
MAGIC_ASSERT_DECLARE (ETHSQ)
MAGIC_ASSERT_DECLARE (ETHRQ)
MAGIC_ASSERT_DECLARE (LRORQ)
MAGIC_ASSERT_DECLARE (MCC)
MAGIC_ASSERT_DECLARE (DEFAULT_PDU_QUEUE)
MAGIC_ASSERT_DECLARE (FCOE_INTERFACE)
MAGIC_ASSERT_DECLARE (ISCSI_CONNECTION)
MAGIC_ASSERT_DECLARE (TOE_CONNECTION)
MAGIC_ASSERT_DECLARE (TOE_SQ)
MAGIC_ASSERT_DECLARE (TOE_RQ)
MAGIC_ASSERT_DECLARE (TOE_DQ)
MAGIC_ASSERT_DECLARE (ISCSI_WRB_QUEUE)
MAGIC_ASSERT_DECLARE (RDMA_QP)
MAGIC_ASSERT_DECLARE (RDMA_CXN)
MAGIC_ASSERT_DECLARE (FCOE_WRB_QUEUE)
MAGIC_ASSERT_DECLARE (RXIF)

// Shorter names
#define ISCSI_CXN_ASSERT  ISCSI_CONNECTION_ASSERT
#define TOE_CXN_ASSERT    TOE_CONNECTION_ASSERT
#define ISCSI_WRBQ_ASSERT ISCSI_WRB_QUEUE_ASSERT
#define ISCSI_DEFQ_ASSERT DEFAULT_PDU_QUEUE_ASSERT
#define FCOE_WRBQ_ASSERT  FCOE_WRB_QUEUE_ASSERT

// Include the srcgen API.
#include "beclib_ll_enum.h"
#include "beclib_ll_bmap.h"

// ----------------------------------------------------------------------
//   API MACROS
// ----------------------------------------------------------------------

#define BE_IOCTL_NAME(_short_name_)     IOCTL_##_short_name_
#define BE_OPCODE_NAME(_short_name_)    OPCODE_##_short_name_
#define BE_SUBSYSTEM_NAME(_short_name_) SUBSYSTEM_##_short_name_

#define BE_FUNCTION_PREPARE_EMBEDDED_IOCTL(_function_object_, _wrb_, _short_name_)  \
    (BE_IOCTL_NAME(_short_name_)*)                                                  \
        be_function_prepare_embedded_ioctl (                                        \
            _function_object_,                                                      \
            _wrb_,                                                                  \
            sizeof (BE_IOCTL_NAME(_short_name_)),                                   \
            SA_SIZEOF_FIELD (BE_IOCTL_NAME(_short_name_), params.request),          \
            SA_SIZEOF_FIELD (BE_IOCTL_NAME(_short_name_), params.response),         \
            BE_OPCODE_NAME(_short_name_),                                           \
            BE_SUBSYSTEM_NAME(_short_name_));                                       \

#define BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL(_function_object_, _wrb_, _ioctl_va_, _ioctl_pa_, _short_name_, _version_)  \
    (BE_IOCTL_NAME(_short_name_)*)                                           \
        be_function_prepare_nonembedded_ioctl (                              \
            (_function_object_),                                             \
            (_wrb_),                                                         \
            (_ioctl_va_),                                                    \
            (_ioctl_pa_),                                                    \
            sizeof (BE_IOCTL_NAME(_short_name_)),                            \
            SA_SIZEOF_FIELD (BE_IOCTL_NAME(_short_name_), params.request),   \
            SA_SIZEOF_FIELD (BE_IOCTL_NAME(_short_name_), params.response),  \
            BE_OPCODE_NAME(_short_name_),                                    \
            BE_SUBSYSTEM_NAME(_short_name_),                                 \
            (_version_));                                                    \

// ----------------------------------------------------------------------
//   API Inline Functions
// ----------------------------------------------------------------------

// Returns TRUE for the ISCSI function
STATIC INLINE
boolean be_function_is_iscsi (
    PBE_FUNCTION_OBJECT function_object) { 
    FUNCTION_ASSERT (function_object);
    return (function_object->type == BE_FUNCTION_TYPE_ISCSI);
}

//RETURNES TRUE for FCoE function
//Bipul : Need to change it later. right
//now for simulation assume FUNCTION_TYPE_ISCSI and then
//run
STATIC INLINE
boolean be_function_is_fcoe (
    PBE_FUNCTION_OBJECT function_object) {
    FUNCTION_ASSERT(function_object);
    return (function_object->type == BE_FUNCTION_TYPE_FCOE);
}


// Returns TRUE for the networking function
STATIC INLINE
boolean be_function_is_networking (
    PBE_FUNCTION_OBJECT function_object) { 
    FUNCTION_ASSERT (function_object);
    return (function_object->type == BE_FUNCTION_TYPE_NETWORK);
}

STATIC INLINE void
be_function_copy_stats (
   IN PBE_FUNCTION_OBJECT     function_object,
   OUT PBECLIB_FUNCTION_STATS stats
   ) {
    FUNCTION_ASSERT (function_object);
    ASSERT (stats);
    sa_memcpy (stats, &function_object->stats, sizeof (function_object->stats));

    #ifdef SA_SPINLOCK_DEBUG
        stats->cq_lock_acquires   = (u32)function_object->cq_lock.acquires;
        stats->cq_lock_ticks      = function_object->cq_lock.spin_ticks;
        stats->post_lock_acquires = (u32)function_object->post_lock.acquires;
        stats->post_lock_ticks    = function_object->post_lock.spin_ticks;
    #endif
}

// Returns a pointer to the parent chip object for this function.
STATIC INLINE
PBE_CHIP_OBJECT be_function_get_chip_object (
    PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    return (function_object->parent_chip);
}


SA_CPP_TRAILER

#endif // __beclib_lower__

