/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

#ifndef __lancer_temp_h__
#define __lancer_temp_h__


///////////////////////////////////
//////////    Macros    ///////////
///////////////////////////////////

// BE's scratchpad register (PCI Config Space, Offset 0x58) is replaced by
// SLI Interface Definition Register in LANCER. So the actual change should
// be in PCICFG_PF0_CSRMAP structure.
#define sli_intf  scratchpad

// Global defines which should be visible to all layers. Help in identifying
// the adapter type, BAR/Register layout, ASIC Revision
//
// Should go to beclib_ll_enum.h?
#define ADAPTER_FAMILY_BE      0x00
#define ADAPTER_FAMILY_LANCER  0x01
#define ADAPTER_FAMILY_MAX     0x02

#define SLI_INTF_TYPE_0        0x00
#define SLI_INTF_TYPE_1        0x01
#define SLI_INTF_TYPE_2        0x02
#define SLI_INTF_TYPE_3        0x03
#define SLI_INTF_TYPE_MAX      0x04

#define LANCER_ASIC_A0         0x0A
#define LANCER_ASIC_B0         0x0B

// SLI Port Status Register
#define SLI_PORT_STATUS_REG    0x404

// SLI Port Control Register
#define SLI_PORT_CONTROL_REG   0x408

// SLI Port Error Data Registers
#define SLI_PORT_ERROR1_REG    0x40C
#define SLI_PORT_ERROR2_REG    0x410

// SLI Port Physical Device Control Register
#define SLI_PORT_PHYSDEV_REG   0x414

// SRB_AXI_XXX Registers
#define SRB_AXI_ADDR_REG       0x870
#define SRB_AXI_DATA_REG       0x874
#define SRB_AXI_CMD_REG        0x878

// LANCER specific new Opcodes
#define SUBSYSTEM_ETH_GET_PPORT_STATISTICS       0x03
#define OPCODE_ETH_GET_PPORT_STATISTICS          0x12
#define SUBSYSTEM_COMMON_SET_DIAG_REGISTERS      0x01
#define OPCODE_COMMON_SET_DIAG_REGISTERS         0x9E
#define SUBSYSTEM_COMMON_ADD_IFACE_QUEUE_FILTER  0x01
#define OPCODE_COMMON_ADD_IFACE_QUEUE_FILTER     0x91
#define SUBSYSTEM_COMMON_DEL_IFACE_QUEUE_FILTER  0x01
#define OPCODE_COMMON_DEL_IFACE_QUEUE_FILTER     0x92
#define SUBSYSTEM_COMMON_MODIFY_CQ               0x01
#define OPCODE_COMMON_MODIFY_CQ                  0x95
#define SUBSYSTEM_COMMON_GET_FUNCTION_CONFIG     0x01
#define OPCODE_COMMON_GET_FUNCTION_CONFIG        0xA0
#define SUBSYSTEM_ETH_GET_VPORT_STATISTICS       0x03
#define OPCODE_ETH_GET_VPORT_STATISTICS          0x13
#define SUBSYSTEM_COMMON_WRITE_OBJECT            0x01
#define OPCODE_COMMON_WRITE_OBJECT               0xAC
#define SUBSYSTEM_COMMON_GET_IFACE_MACLIST       0x1
#define OPCODE_COMMON_GET_IFACE_MACLIST          0x93
#define SUBSYSTEM_COMMON_SET_IFACE_MACLIST       0x1
#define OPCODE_COMMON_SET_IFACE_MACLIST          0x94
#define SUBSYSTEM_COMMON_GET_PROFILE_CONFIG      0x1
#define OPCODE_COMMON_GET_PROFILE_CONFIG         0xA4
#define SUBSYSTEM_COMMON_SET_PROFILE_CONFIG      0x1
#define OPCODE_COMMON_SET_PROFILE_CONFIG         0xA5


// New interface capabilities supported by LANCER
#define LANCER_NTWK_RX_FILTER_IF_UNICAST  0x8000
#define LANCER_NTWK_RX_FILTER_IF_VMQ      0x10000
#define LANCER_CAPACITY_IF_FLAG_RSS       0x4

// Definitions for new command (COMMON_ADD_IFACE_QUEUE_FILTER)
#define FILTER_TYPE_MAC        0x00
#define FILTER_TYPE_VLAN       0x01
#define FILTER_TYPE_MAC_VLAN   0x02
#define FILTER_TYPE_PRIORITY   0x03
#define FILTER_TYPE_ETHERTYPE  0x04

// Definitions for new command (COMMON_MODIFY_CQ)
#define WM_INTR_DISABLE                    0x00
#define WM_INTR_DISABLE_SET_NEW_THRESHOLD  0x01
#define WM_INTR_ENABLE                     0x02
#define WM_INTR_ENABLE_SET_NEW_THRESHOLD   0x03

// Definitions for new command (COMMON_WRITE_OBJECT)
#define NO_RESET_NEEDED        0x00
#define PHYS_DEV_RESET_NEEDED  0x01
#define FW_RESET_NEEDED        0x02
#define PROTOCOL_RESET_NEEDED  0x03


// Status codes associated with IOCTL_CODE_LANCER_REGISTER_ACCESS
#define LANCER_STATUS_SUCCESS            0  // Read/Write succeeded
#define LANCER_STATUS_FAILURE            1  // Generic failure
#define LANCER_STATUS_NOT_SUPPORTED      2  // Command not supported (if you are trying on a BE card)
#define LANCER_STATUS_INVALID_PARAMETER  3  // Check the address provided (not 4 byte aligned)
#define LANCER_STATUS_TIMEOUT            4  // AXI not coming out of busy state (timeout of 10ms)
#define LANCER_STATUS_PRIVILEGES         5  // BAR mapping failed. Insufficient privileges (may be trying on function other than function 0 or from a VF)
#define LANCER_STATUS_NEED_REBOOT        6  // BAR mapping done. Host reboot required (throw a notification to user)


///////////////////////////////////
//////////  Structures  ///////////
///////////////////////////////////

// SLI Interface Definition Register at offset 0x58 in PCI Config Space.
typedef struct _PCICFG_SLI_INTF_CSR {
   union {
      struct { 
         u32 funcType  : 1; /* Function Type
                               0 - Physical fcuntion
                               1 - Virtual function
                               */
         u32 rsvd      : 3;
         u32 sliRev    : 4; /* This field indicates the SLI interface revision.
                               0x03: SLI-3
                               0x04: SLI-4
                               */
         u32 sliFamily : 4; /* This field specifies the implementation family of
                               the SLI interface.
                               0x0:    BE3
                               0x1:    BE3
                               0xA:    Lancer A0
                               0xB:    Lancer B0
                               */
         u32 ifType    : 4; /* This field indicates the Function register layout type
                               used in conjuction with the SLI Revision field.
                               0x0: SLI-4 interface 0
                               0x1: SLI-4 interface 1
                               0x2: SLI-4 interface 2
                               0x3: SLI-4 interface 3
                               */
         u32 sliHint1  : 8; /* This field specifies additional information
                               about the SLI-Interface supported by the
                               function required by reset or initialization.
                               If SLI-3:
                               0x0:    No hints specified
                               0x1:    Function is SLI-3.5 based
                               If SLI-4:
                               0x0:    No hints specified
                               0x1:    Function initialization via the SLIPORT_SEMAPHORE
                                       register and the INITIALIZE mailbox command followed
                                       by COMMON_FUNCTION_RESET via the BMBX register.
                                       Function termination should be via the
                                       COMMON_FUNCTION_RESET and the DEINITIALIZE mailbox
                                       command via the BMBX register.
                               0x2:    Function initialization via the SLIPORT_STATUS and
                                       SLIPORT_CONTROL registers. Function termination via
                                       SLIPORT_CONTROL register.
                               */
         u32 sliHint2  : 5; /* This field specifies additional information
                               about the SLI-Interface supported by the
                               function required by reset or initialization */
         u32 valid     : 3; /* This field must contain the bit pattern 110b in
                               order for the contents of this register to be
                               considered valid */
      } SG_PACK ; // unnamed struct
      u32 dw; // dword union
   }; // unnamed union
} SG_PACK PCICFG_SLI_INTF_CSR, *PPCICFG_SLI_INTF_CSR;

SG_C_ASSERT(__sizeof__PCICFG_SLI_INTF_CSR, sizeof (PCICFG_SLI_INTF_CSR) == 4);


// SLI Port Status Register at offset 0x404 in CSR BAR
typedef struct _SLI_PORT_STATUS_CSR {
   union {
      struct { 
         u32 rsvd1    : 22;
         u32 idis     : 1; /* 1 means all interrupts on function are disabled
                              0 means all interrupts on function are enabled
                              */
         u32 ready    : 1; /* This field specifies whether the SLI Port is
                              operational and ready to access new requests
                              from the SLI Host. This field is read-only and
                              ignored on writes. When set to 0, the SLI Post
                              is non-operational. The SLI Port will not respond
                              to writes to the PORT_CONTROL register, the BMBX
                              register, doorbell registers, or to submission of
                              commands or requests in any other fashion. When
                              set to 1, the SLI Port is operational and ready to
                              accept register writes and new commands or requests.
                              */
         u32 need_rst : 1; /* This field specifies whether the SLI Port has
                              encountered a condition in which it is no longer
                              fully functional and requires a SLI Port reset via
                              the SLIPORT_CONTROL register IP bit (Bit 27) to
                              resume full operation. When set to 0, no reset is
                              needed. When set to 1, the SLI Port requires a
                              reset to resume operation.
                              */
         u32 dip      : 1; /* 1 means dump image present */
         u32 dl       : 1; /* Dump location. 0 means dump image exists internally
                              and can be retrieved via COMMON_READ_OBJECT mailbox.
                              1 means dump image is written to previously registered
                              host memory (COMMON_SET_DUMP_LOCATION mailbox)
                              */
         u32 rsvd2    : 2;
         u32 oti      : 1; /* For SLI Port implementations that support
                              over-temperature reporting, this bit is set to 1
                              when an over-temperature condition has occurred.
                              When set to 0, the feature is unimplemented or no
                              over-temperature condition has occurred.
                              */
         u32 endian   : 1; /* When set to 0, the SLI Host endian-ness is Little
                              Endian. When set to 1, the SLI Host endian-ness is
                              Big Endian.
                              */
         u32 error    : 1; /* The SLI Port sets this bit to 1 when it encounters
                              a condition that has halted SLI Port operation.
                              Further information about the cause of the condition
                              can be obtained from the SLIPORT_ERROR1 and
                              SLIPORT_ERROR2 registers. If the error condition is
                              recoverable by re-initialization of the SLI Port,
                              the RN bit (Bit 24) will be set to 1. If the RN bit
                              is 0, the error condition is unrecoverable and the
                              SLI Host must discontinue interaction with the SLI Port.
                              */
      } SG_PACK ; // unnamed struct
      u32 dw; // dword union
   }; // unnamed union
} SG_PACK SLI_PORT_STATUS_CSR, *PSLI_PORT_STATUS_CSR;

SG_C_ASSERT(__sizeof__SLI_PORT_STATUS_CSR, sizeof (SLI_PORT_STATUS_CSR) == 4);


// SLI Port Control Register at offset 0x408 in CSR BAR
typedef struct _SLI_PORT_CONTROL_CSR {
   union {
      struct { 
         u32 rsvd1     :22;
         u32 idis      : 1; /* Write 1 to disable all interrupts on function
                               Write 0 to enable all interrupts on function
                               */
         u32 rsvd2     : 4;
         u32 init_port : 1; /* This field instructs the SLI Port to perform
                               a reset (it is functionally similar to the
                               FW_INITIALIZE and COMMON_FUNCTION_RESET
                               mailbox commands). This field is write-only
                               and must be ignored on reads. When set to 1,
                               the SLI Port will be reset, and the endian-ness
                               of the SLI Host is set based on the value of
                               the END (bit 30) field. All SLI Port state will
                               be deleted and cleared, requiring the SLI Port
                               to be reinitialized. When the reset first occurs,
                               the SLIPORT_STATUS register RDY bit (Bit 23)
                               will clear to 0. Upon completion of the reset,
                               the SLIPORT_STATUS register RDY bit (Bit 23) will
                               transition to 1. When set as 0, the SLI Port is
                               unaffected.
                               */
         u32 rsvd3     : 2;
         u32 endian    : 1; /* This field specifies the endian-ness of the SLI
                               Host. The endian-ness of the SLI Host is set at
                               the same time the SLI Host request to Initialize
                               the SLI Port via the IP bit (Bit 27). This field
                               is write-only and is only interpreted by the SLI
                               Port when the IP bit (Bit 27) is set to 1. The SLI
                               Port will use the endian-ness setting to determine
                               whether it should automatically byte-swap data
                               structures that DMA.d over the PCI bus for control
                               paths. Examples of such data structures are Queue
                               Entries (MQEs, EQEs, CQEs, WQEs, or RQEs). Data
                               corresponding to payload transmit or received on
                               the physical link will remain in the physical link
                               network byte order. By default, the SLI Host
                               endian-ness is set to Little Endian. The default
                               endian-ness will be set after device power cycles,
                               PCI fundamental resets, PCI FLRs, or VH Resets.
                               When set to 0, the SLI Host endian-ness is Little Endian.
                               When set to 1, the SLI Host endian-ness is Big Endian.
                               */
         u32 rsvd4     : 1;
      } SG_PACK ; // unnamed struct
      u32 dw; // dword union
   }; // unnamed union
} SG_PACK SLI_PORT_CONTROL_CSR, *PSLI_PORT_CONTROL_CSR;

SG_C_ASSERT(__sizeof__SLI_PORT_CONTROL_CSR, sizeof (SLI_PORT_CONTROL_CSR) == 4);


// SLI Port Physical Device Control Register at offset 0x414 in CSR BAR
typedef struct _SLI_PORT_PHYSDEV_CSR {
   union {
      struct {
         u32 drst     : 1; /* 1 means reset physical device */
         u32 frst     : 1; /* 1 means reset firmware */
         u32 dd       : 1; /* 1 means diagnostic dump has to be taken */
         u32 lc       : 1; /* 1 means physical links are not shutdown during fw reset */
         u32 frl      : 4; /* Firmware Reset Level:
                              0x00: Reset all fw
                              0x01: Reset protocol fw (FC/FCoE)
                              0x02: Reset protocol fw (Ethernet Networking)
                              */
         u32 rc       : 3; /* Restart Configuration:
                              0x00: Unspecified. Default values.
                              0x01: Persistent settings.
                              0x02: Factory defaults.
                              */
         u32 rsvd1    : 2;
         u32 gp       : 1; /* 1 means place all function in general pause state */
         u32 gpc      : 1; /* General pause complete */
         u32 ipld     : 1; /* 1 means ignore PCI link down condition. Used with gp */
         u32 rsvd2    :14;
         u32 inp      : 1; /* Previous operation in progress */
         u32 err      : 1; /* Current write to PHYSDEV register failed */
      } SG_PACK ; // unnamed struct
      u32 dw; // dword union
   }; // unnamed union
} SG_PACK SLI_PORT_PHYSDEV_CSR, *PSLI_PORT_PHYSDEV_CSR;

SG_C_ASSERT(__sizeof__SLI_PORT_PHYSDEV_CSR, sizeof (SLI_PORT_PHYSDEV_CSR) == 4);


// SRB AXI CMD Register
typedef struct _SRB_AXI_CMD {
   union {
      struct { 
         u32 byte_enables : 4;
         u32 rsvd         : 26;
         u32 write        : 1;
         u32 busy         : 1;
      } SG_PACK ; // unnamed struct
      u32 dw; // dword union
   }; // unnamed union
} SG_PACK SRB_AXI_CMD, *PSRB_AXI_CMD;

SG_C_ASSERT(__sizeof__SRB_AXI_CMD, sizeof (SRB_AXI_CMD) == 4);


// RX CQE for LANCER
typedef struct _ETH_RX_COMPL_LANCER {
   union {
      struct { 
         u32 vlan_tag : 16;                       // DWORD 0
         u32 pktsize : 14;                        // DWORD 0
         u32 rsvd0 : 1;                           // DWORD 0
         u32 ip_opt : 1;                          // DWORD 0
         u32 err : 1;                             // DWORD 1
         u32 rsshp : 1;                           // DWORD 1
         u32 ipf : 1;                             // DWORD 1
         u32 tcpf : 1;                            // DWORD 1
         u32 udpf : 1;                            // DWORD 1
         u32 ipcksm : 1;                          // DWORD 1
         u32 l4_cksm : 1;                         // DWORD 1
         u32 ip_version : 1;                      // DWORD 1
         u32 filter_hi : 6;                       // DWORD 1
         u32 vtp : 1;                             // DWORD 1
         u32 ipfrag : 1;                          // DWORD 1
         u32 fragndx : 10;                        // DWORD 1
         u32 rsvd1 : 1;                           // DWORD 1
         u32 rh_present : 1;                      // DWORD 1
         u32 sw : 1;                              // DWORD 1
         u32 numfrags : 3;                        // DWORD 1
         u32 rss_flush : 1;                       // DWORD 2
         u32 cast_enc : 2;                        // DWORD 2
         u32 QnQ : 1;                             // DWORD 2
         u32 rss_bank : 1;                        // DWORD 2
         u32 hdr_fragndx : 10;                    // DOWRD 2
         u32 hdrsize_filter_lo : 12;              // DOWRD 2
         u32 toe : 1;                             // DWORD 2
         u32 lro_pkt : 1;                         // DWORD 2
         u32 hds_type : 2;                        // DWORD 2
         u32 valid : 1;                           // DWORD 2
         u32 rsshash;                             // DWORD 3
      } SG_PACK ; // unnamed struct
      u32 dw[4]; // dword union
   }; // unnamed union
} SG_PACK ETH_RX_COMPL_LANCER, *PETH_RX_COMPL_LANCER;

SG_C_ASSERT(__sizeof__ETH_RX_COMPL_LANCER, sizeof (ETH_RX_COMPL_LANCER) == 16);


// LANCER doesn't support version 0 of IOCTL_COMMON_MCC_CREATE_EXT command.
// So instead version 1 of the same command should be used.
typedef struct _MCC_RING_CONTEXT_V1 {
   union {
      struct { 
         u32 async_cq_id : 16;                    // DWORD 0
         u32 ring_size : 4;                       // DWORD 0
         u32 rsvd0 : 12;                          // DWORD 0
         u32 rsvd1 : 31;                          // DWORD 1
         u32 valid : 1;                           // DWORD 1
         u32 async_cq_valid : 1;                  // DWORD 2
         u32 rsvd2 : 31;                          // DWORD 2
         u32 rsvd3[1];                            // DWORDS 3 to 3
      } SG_PACK ; // unnamed struct
      u32 dw[4]; // dword union
   }; // unnamed union
} SG_PACK MCC_RING_CONTEXT_V1, *PMCC_RING_CONTEXT_V1;

SG_C_ASSERT(__sizeof__MCC_RING_CONTEXT_V1, sizeof (MCC_RING_CONTEXT_V1) == 16);


typedef struct _IOCTL_COMMON_MCC_CREATE_EXT_V1 {
   union {
      struct { 
         IOCTL_HEADER header;
         union {
            struct {
               UEXACT16 num_pages;
               UEXACT16 cq_id;
               UEXACT32 async_event_bitmap[1];
               MCC_RING_CONTEXT_V1 context;
               PHYS_ADDR pages[8];
            } request;
            struct {
               UEXACT16 id;
               UEXACT16 rsvd0;
            } response;
         } params;
      } SG_PACK ; // unnamed struct
      u32 dw[26]; // dword union
   }; // unnamed union
} SG_PACK IOCTL_COMMON_MCC_CREATE_EXT_V1, *PIOCTL_COMMON_MCC_CREATE_EXT_V1;

SG_C_ASSERT(__sizeof__IOCTL_COMMON_MCC_CREATE_EXT_V1, sizeof (IOCTL_COMMON_MCC_CREATE_EXT_V1) == 104);


// LANCER doesn't support version 0 of IOCTL_COMMON_CQ_CREATE command.
// So instead version 2 of the same command should be used.
typedef struct _CQ_CONTEXT_V2 {
   union {
      struct { 
         u32 rsvd0 : 12;                          // DWORD 0
         u32 CoalescWM : 2;                       // DWORD 0
         u32 NoDelay : 1;                         // DWORD 0
         u32 AutoValid : 1;                       // DWORD 0
         u32 rsvd1 : 11;                          // DWORD 0
         u32 Count : 2;                           // DWORD 0
         u32 valid : 1;                           // DWORD 0
         u32 rsvd2 : 1;                           // DWORD 0
         u32 Eventable : 1;                       // DWORD 0
         u32 EQID : 16;                           // DWORD 1
         u32 rsvd3 : 15;                          // DWORD 1
         u32 Armed : 1;                           // DWORD 1
         u32 CqeCount : 16;                       // DWORD 2
         u32 rsvd4 : 16;                          // DWORD 2
         u32 rsvd5[1];                            // DWORDS 3 to 3
      } SG_PACK ; // unnamed struct
      u32 dw[4]; // dword union
   }; // unnamed union
} SG_PACK CQ_CONTEXT_V2, *PCQ_CONTEXT_V2;

SG_C_ASSERT(__sizeof__CQ_CONTEXT_V2, sizeof (CQ_CONTEXT_V2) == 16);


typedef struct _IOCTL_COMMON_CQ_CREATE_V2 {
   union {
      struct { 
         IOCTL_HEADER header;
         union {
            struct {
               UEXACT16 num_pages;
               UEXACT16 page_size : 8;
               UEXACT16 rsvd0 : 8;
               CQ_CONTEXT_V2 context;
               PHYS_ADDR pages[8];
            } request;
            struct {
               UEXACT16 cq_id;
               UEXACT16 rsvd0;
            } response;
         } params;
      } SG_PACK ; // unnamed struct
      u32 dw[25]; // dword union
   }; // unnamed union
} SG_PACK IOCTL_COMMON_CQ_CREATE_V2, *PIOCTL_COMMON_CQ_CREATE_V2;

SG_C_ASSERT(__sizeof__IOCTL_COMMON_CQ_CREATE_V2, sizeof (IOCTL_COMMON_CQ_CREATE_V2) == 100);


// This is a temporary structure. Actual change has to be done in
// BE_ETH_SQ_PARAMETERS present in beclib_ll_bmap.h
typedef struct _BE_ETH_SQ_PARAMETERS_TEMP {
   // In LANCER IOV mode, every TXQ must be associated with an interface
   PBE_RXIF_OBJECT  rxif_object;
   u32 rsvd0[2];
} SG_PACK BE_ETH_SQ_PARAMETERS_TEMP, *PBE_ETH_SQ_PARAMETERS_TEMP;


// LANCER doesn't support version 0 of IOCTL_ETH_TX_CREATE command.
// So instead version 1 of the same command should be used.
typedef struct _ETX_CONTEXT_V1_LANCER {
   union {
      struct { 
         u32 iface_id : 16;                       // DWORD 0
         u32 tx_ring_size : 4;                    // DWORD 0
         u32 rsvd0 : 12;                          // DWORD 0
         u32 rsvd1 : 31;                          // DWORD 0
         u32 ctx_valid : 1;                       // DWORD 1
         u32 cq_id_send : 16;                     // DWORD 2
         u32 rsvd2 : 16;                          // DWORD 2
         u32 rsvd3[13];                             // DWORD 3 to DWORD 15
      } SG_PACK ; // unnamed struct
      u32 dw[16]; // dword union
   }; // unnamed union
} SG_PACK ETX_CONTEXT_V1_LANCER, *PETX_CONTEXT_V1_LANCER;

SG_C_ASSERT(__sizeof__ETX_CONTEXT_V1_LANCER, sizeof (ETX_CONTEXT_V1_LANCER) == 64);


typedef struct _IOCTL_ETH_TX_CREATE_V1_LANCER {
   union {
      struct { 
         IOCTL_HEADER header;
         union {
            struct {
               UEXACT8 num_pages;
               UEXACT8 rsvd0;
               UEXACT8 type;
               UEXACT8 rsvd1;
               ETX_CONTEXT_V1_LANCER context;
               PHYS_ADDR pages[8];
            } request;
            struct {
               UEXACT16 cid;
               UEXACT16 rid;
               UEXACT32 db_offset;
               UEXACT8  class_id;
               UEXACT8  rsvd0[3];
            } response;
         } params;
      } SG_PACK ; // unnamed struct
      u32 dw[37]; // dword union
   }; // unnamed union
} SG_PACK IOCTL_ETH_TX_CREATE_V1_LANCER, *PIOCTL_ETH_TX_CREATE_V1_LANCER;

SG_C_ASSERT(__sizeof__IOCTL_ETH_TX_CREATE_V1_LANCER, sizeof (IOCTL_ETH_TX_CREATE_V1_LANCER) == 148);


#define MAX_MAC_IDS 64 
typedef struct _MACLIST_CONTEXT {
    u32 mac_id: 31;
    u32 active_mac: 1;
} MACLIST_CONTEXT, *PMACLIST_CONTEXT;

typedef struct _IOCTL_COMMON_GET_IFACE_MACLIST {
    union {
        struct { 
            IOCTL_HEADER header;
            union {
                struct {
                    u32 rsvd;
                } request;
                struct {
                    u32 maclist_len : 8;                  // DWORD 0
                    u32 rsvd0 : 24;                          // DWORD 0
                    MACLIST_CONTEXT maclist[MAX_MAC_IDS]; //DWORD 1 to 65
                } response;
            } params;
        } SG_PACK ; // unnamed struct
        u32 dw[69]; // dword union
    }; // unnamed union
} SG_PACK IOCTL_COMMON_GET_IFACE_MACLIST, *PIOCTL_COMMON_GET_IFACE_MACLIST;


typedef struct _IOCTL_COMMON_SET_IFACE_MACLIST {
    union {
        struct { 
            IOCTL_HEADER header;
            union {
                struct {
                    u32 maclist_len : 8;                  // DWORD 0
                    u32 rsvd :24;
                    u8 mac_list[MAX_MAC_IDS][6]; //DWORD 1 to 65
                } request;
                struct {
                    u32 rsvd;
                } response;
            } params;
        } SG_PACK ; // unnamed struct
        u32 dw[101]; // dword union
    }; // unnamed union
} SG_PACK IOCTL_COMMON_SET_IFACE_MACLIST, *PIOCTL_COMMON_SET_IFACE_MACLIST;


typedef struct _BE_IF_SET_MACLIST_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT             context;
    BE_MCC_WRB_HEADER                   wrb_header;
    IOCTL_COMMON_SET_IFACE_MACLIST      ioctl;        
} BE_IF_SET_MACLIST_QUEUE_CONTEXT, *PBE_IF_SET_MACLIST_QUEUE_CONTEXT;


typedef struct _BE_IF_GET_MACLIST_QUEUE_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT             context;    
    BE_MCC_WRB_HEADER                   wrb_header;
    IOCTL_COMMON_GET_IFACE_MACLIST      ioctl;        
} BE_IF_GET_MACLIST_QUEUE_CONTEXT, *PBE_IF_GET_MACLIST_QUEUE_CONTEXT;

#define MAX_RES_DESCRIPTORS         256 

#define RES_DESC_PCIE          0x40
#define RES_DESC_NIC           0x41
#define RES_DESC_ISCSI         0x42
#define RES_DESC_FCFCOE        0x43
#define RES_DESC_RDMA          0x44


typedef struct _NIC_RES_DESCRIPTOR {
    u32 desc_type:8; //0x40 for PCIe descriptor, 0x41 for NIC descriptor
    u32 rsvd:15;
    u32 lld_valid:1;
    u32 if_flags_valid:1;
    u32 mac_valid:1;
    u32 vlan_valid:1;
    u32 vf_template:1;
    u32 qos_units:1; //0 indiates bandwidth min/max in percentage, 1 indicates actual values
    u32 del:1;
    u32 immediate:1; //used for set for immediate/running config or not
    u32 no_save:1;
    u32 vf_number:8;
    u32 rsvd1:8;
    u32 pf_number:10;
    u32 rsvd2:6;
    u16 mac_unicast_count;
    u8 mac_address[6];
    u16 mq_count;
    u16 vlan_count;
    u16 mac_mcast_count;
    u16 wq_count;
    u16 rq_count;
    u16 rss_queue_count;
    u16 lro_queue_count;
    u16 cq_count;
    u16 toe_conn_count;
    u16 eq_count;
    u16 vlan_id;
    u16 rsvd4;
    u32 iface_caps_flags;
    u32 link_number:6;
    u32 link_type:2;
    u32 lmc:1;
    u32 lld:1;
    u32 rsvd5:22;
    u32 bw_min;
    u32 bw_max;
    u32 acpi_patrn_count:8;
    u32 oob_wol:2;
    u32 wol:2;
    u32 rsvd6:20;
    u32 rsvd7[3];
}  SG_PACK NIC_RES_DESCRIPTOR, *PNIC_RES_DESCRIPTOR;

typedef struct _PCIE_RES_DESCRIPTOR
{
    u32 desc_type:8;
    u32 rsvd:22;
    u32 immediate:1;
    u32 no_save:1;
    u32 rsvd1:16;
    u32 pf_number:10;
    u32 rsvd2:6;
    u32 rsvd3;
    u8 sriov_state;
    u8 pf_state;
    u8 pf_type;
    u8 rsvd4;
    u16 num_vfs;
    u16 rsvd5;
    u32 mission_roles:8; //0x00 for NIC/TOE
    u32 rsvd6:19;
    u32 presentation_chg:1;
    u32 sriov_state_chg:1;
    u32 xrom_chg:1;
    u32 xrom:2;
    u32 rsvd7[12];
}SG_PACK PCIE_RES_DESCRIPTOR, *PPCIE_RES_DESCRIPTOR;

//used for get/set profile config commands
typedef struct _PROFILE_CONFIG {
    union {
       u32     desc_count; //used in resonse
       struct { // struct is used while sending request
            u32 profile_id :8; // 0 for current active profile
            u32 cfg_type :2;  //0x0 for persistent_cfg, 0x1 for factory-default, 0x2 for current active cfg.
            u32 rsvd :22;
       };
    };
    union {
        NIC_RES_DESCRIPTOR  nic_desc[MAX_RES_DESCRIPTORS];
        PCIE_RES_DESCRIPTOR pcie_desc[MAX_RES_DESCRIPTORS];
    }; // unnamed union
} SG_PACK PROFILE_CONFIG, *PPROFILE_CONFIG;

typedef struct _IOCTL_COMMON_GET_PROFILE_CONFIG {
    IOCTL_HEADER header;
    union {
        struct {
            u32 profile_id :8; // 0 for current active profile
            u32 cfg_type :2;  //0x0 for persistent_cfg, 0x1 for factory-default, 0x2 for current active cfg.
            u32 rsvd :22;
        } request;
        struct {
            PROFILE_CONFIG desc;
        } response;
    } params;
} SG_PACK IOCTL_COMMON_GET_PROFILE_CONFIG, *PIOCTL_COMMON_GET_PROFILE_CONFIG;

typedef struct _IOCTL_COMMON_SET_PROFILE_CONFIG {
    IOCTL_HEADER header;
    union {
        struct {
            u32 profile_id : 8;
            u32 rsvd :24;
            PROFILE_CONFIG desc;
        } request;
        struct {
            u32 rsvd;
        } response;
    } params;
} SG_PACK IOCTL_COMMON_SET_PROFILE_CONFIG, *PIOCTL_COMMON_SET_PROFILE_CONFIG;


typedef struct _BE_GET_PROFILE_CONFIG_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT             context;    
    BE_MCC_WRB_HEADER                   wrb_header;
    IOCTL_COMMON_GET_PROFILE_CONFIG     ioctl;        
} BE_GET_PROFILE_CONFIG_CONTEXT, *PBE_GET_PROFILE_CONFIG_CONTEXT;


typedef struct _BE_SET_PROFILE_CONFIG_CONTEXT {
    BE_QUEUE_DRIVER_CONTEXT             context;    
    BE_MCC_WRB_HEADER                   wrb_header;
    IOCTL_COMMON_SET_PROFILE_CONFIG     ioctl;        
} BE_SET_PROFILE_CONFIG_CONTEXT, *PBE_SET_PROFILE_CONFIG_CONTEXT;


typedef struct _IOCTL_COMMON_GET_FUNCTION_CONFIG {
    IOCTL_HEADER header;
    union {
        struct {
            u32 desc_count;
             u8 desc[2 * sizeof (NIC_RES_DESCRIPTOR)];  // Max 2 desc for NIC + RDMA
        } response;
    } params;
} SG_PACK IOCTL_COMMON_GET_FUNCTION_CONFIG, *PIOCTL_COMMON_GET_FUNCTION_CONFIG;

// NIC_GET_PPORT_STATS Mailbox command request/response structures
// Should be moved to be2nic_bmap.h??
typedef struct _LANCER_PPORT_STATS {
   union {
      struct { 
         /* Tx */
         u32 txNumPktsLo;
         u32 txNumPktsHi;
         u32 txNumUnicastPktsLo;
         u32 txNumUnicastPktsHi;
         u32 txNumMulticastPktsLo;
         u32 txNumMulticastPktsHi;
         u32 txNumBroadcastPktsLo;
         u32 txNumBroadcastPktsHi;
         u32 txNumBytesLo;
         u32 txNumBytesHi;
         u32 txNumUnicastBytesLo;
         u32 txNumUnicastBytesHi;
         u32 txNumMulticastBytesLo;
         u32 txNumMulticastBytesHi;
         u32 txNumBroadcastBytesLo;
         u32 txNumBroadcastBytesHi;
         u32 txNumDiscardsLo;
         u32 txNumDiscardsHi;
         u32 txNumErrorsLo;
         u32 txNumErrorsHi;

         u32 txNumPauseFrmsLo;
         u32 txNumPauseFrmsHi;
         u32 txNumPauseOnFrmsLo;
         u32 txNumPauseOnFrmsHi;
         u32 txNumPauseOffFrmsLo;
         u32 txNumPauseOffFrmsHi;
         u32 txNumIntMacErrorsLo;
         u32 txNumIntMacErrorsHi;
         u32 txNumControlFrmsLo;
         u32 txNumControlFrmsHi;

         u32 txNumPkts64BytesLo;
         u32 txNumPkts64BytesHi;
         u32 txNumPkts65to127BytesLo;
         u32 txNumPkts65to127BytesHi;
         u32 txNumPkts128to255BytesLo;
         u32 txNumPkts128to255BytesHi;
         u32 txNumPkts256to511BytesLo;
         u32 txNumPkts256to511BytesHi;
         u32 txNumPkts512to1023BytesLo;
         u32 txNumPkts512to1023BytesHi;
         u32 txNumPkts1024to1518BytesLo;
         u32 txNumPkts1024to1518BytesHi;
         u32 txNumPkts1519to2047BytesLo;
         u32 txNumPkts1519to2047BytesHi;
         u32 txNumPkts2048to4095BytesLo;
         u32 txNumPkts2048to4095BytesHi;
         u32 txNumPkts4096to8191BytesLo;
         u32 txNumPkts4096to8191BytesHi;
         u32 txNumPkts8192to9216BytesLo;
         u32 txNumPkts8192to9216BytesHi;
         u32 txNumLsoPktsLo;
         u32 txNumLsoPktsHi;

         /* Rx */
         u32 rxNumPktsLo;
         u32 rxNumPktsHi;
         u32 rxNumUnicastPktsLo;
         u32 rxNumUnicastPktsHi;
         u32 rxNumMulticastPktsLo;
         u32 rxNumMulticastPktsHi;
         u32 rxNumBroadcastPktsLo;
         u32 rxNumBroadcastPktsHi;
         u32 rxNumBytesLo;
         u32 rxNumBytesHi;
         u32 rxNumUnicastBytesLo;
         u32 rxNumUnicastBytesHi;
         u32 rxNumMulticastBytesLo;
         u32 rxNumMulticastBytesHi;
         u32 rxNumBroadcastBytesLo;
         u32 rxNumBroadcastBytesHi;

         u32 rxNumUnknownProtos;         // u32
         u32 rsvd_69;

         u32 rxNumDiscardsLo;
         u32 rxNumDiscardsHi;
         u32 rxNumErrorsLo;
         u32 rxNumErrorsHi;
         u32 rxNumCrcErrorsLo;
         u32 rxNumCrcErrorsHi;
         u32 rxNumAlignErrorLo;
         u32 rxNumAlignErrorHi;
         u32 rxNumSymbolErrorLo;
         u32 rxNumSymbolErrorHi;
         u32 rxNumPauseFrmsLo;
         u32 rxNumPauseFrmsHi;
         u32 rxNumPauseOnFrmsLo;
         u32 rxNumPauseOnFrmsHi;
         u32 rxNumPauseOffFrmsLo;
         u32 rxNumPauseOffFrmsHi;
         u32 rxNumFrmTooLongLo;
         u32 rxNumFrmTooLongHi;
         u32 rxNumIntMacErrorLo;
         u32 rxNumIntMacErrorHi;

         u32 rxNumUndersizePkts;
         u32 rxNumOversizePkts;
         u32 rxNumFragmentPkts;
         u32 rxNumJabbers;

         u32 rxNumControlFrmsLo;
         u32 rxNumControlFrmsHi;
         u32 rxNumCtrlFrmUnknownOpcodeLo;
         u32 rxNumCtrlFrmUnknownOpcodeHi;

         u32 rxNumInRangeErrors;
         u32 rxNumOutRangeErrors;
         u32 rxNumAddrMatchErrors;
         u32 rxNumVlanMismatchErrors;
         u32 rxNumDroppedTooSmall;
         u32 rxNumDroppedTooShort;
         u32 rxNumDroppedHeaderTooSmall;
         u32 rxNumDroppedInvalidTcpLength;
         u32 rxNumDroppedRunt;
         u32 rxNumIpCheckSumErrors;
         u32 rxNumTcpCheckSumErrors;
         u32 rxNumUdpCheckSumErrors;
         u32 rxNumNonRssPkts;
         u32 rsvd_111;

         u32 rxNumIpv4PktsLo;
         u32 rxNumIpv4PktsHi;
         u32 rxNumIpv6PktsLo;
         u32 rxNumIpv6PktsHi;
         u32 rxNumIpv4BytesLo;
         u32 rxNumIpv4BytesHi;
         u32 rxNumIpv6BytesLo;
         u32 rxNumIpv6BytesHi;
         u32 rxNumNicPktsLo;
         u32 rxNumNicPktsHi;
         u32 rxNumTcpPktsLo;
         u32 rxNumTcpPktsHi;
         u32 rxNumIscsiPktsLo;
         u32 rxNumIscsiPktsHi;
         u32 rxNumMgmtPktsLo;
         u32 rxNumMgmtPktsHi;
         u32 rxNumSwUnicastPktsLo;
         u32 rxNumSwUnicastPktsHi;
         u32 rxNumSwMulticastPktsLo;
         u32 rxNumSwMulticastPktsHi;
         u32 rxNumSwBroadcastPktsLo;
         u32 rxNumSwBroadcastPktsHi;
         u32 rxNumForwardsLo;
         u32 rxNumForwardsHi;

         u32 rxNumFifoOverflow;
         u32 rxNumInputFifoOverflow;

         u32 rxNumDropTooManyFragsLo;
         u32 rxNumDropTooManyFragsHi;

         u32 rxNumDropInvalidQueue;
         u32 rsvd_141;

         u32 rxNumDropMtuLo;
         u32 rxNumDropMtuHi;

         u32 rxNumPkts64BytesLo;
         u32 rxNumPkts64BytesHi;
         u32 rxNumPkts65to127BytesLo;
         u32 rxNumPkts65to127BytesHi;
         u32 rxNumPkts128to255BytesLo;
         u32 rxNumPkts128to255BytesHi;
         u32 rxNumPkts256to511BytesLo;
         u32 rxNumPkts256to511BytesHi;
         u32 rxNumPkts512to1023BytesLo;
         u32 rxNumPkts512to1023BytesHi;
         u32 rxNumPkts1024to1518BytesLo;
         u32 rxNumPkts1024to1518BytesHi;
         u32 rxNumPkts1519to2047BytesLo;
         u32 rxNumPkts1519to2047BytesHi;
         u32 rxNumPkts2048to4095BytesLo;
         u32 rxNumPkts2048to4095BytesHi;
         u32 rxNumPkts4096to8191BytesLo;
         u32 rxNumPkts4096to8191BytesHi;
         u32 rxNumPkts8192to9216BytesLo;
         u32 rxNumPkts8192to9216BytesHi;
      } SG_PACK ; // unnamed struct
      u32 dw[164]; // dword union
   }; // unnamed union
} SG_PACK LANCER_PPORT_STATS, *PLANCER_PPORT_STATS;

SG_C_ASSERT(__sizeof__LANCER_PPORT_STATS, sizeof (LANCER_PPORT_STATS) == 656);


typedef struct _ETH_GET_PPORT_STATISTICS_PARAMS {
   union {
      union { 
         struct {
             UEXACT16  port_number;
             UEXACT8   rsvd0;
             UEXACT8   reset_stats;
         } request;
         LANCER_PPORT_STATS response;
      } SG_PACK ; // unnamed struct
      u32 dw[164]; // dword union
   }; // unnamed union
} SG_PACK ETH_GET_PPORT_STATISTICS_PARAMS, *PETH_GET_PPORT_STATISTICS_PARAMS;

SG_C_ASSERT(__sizeof__ETH_GET_PPORT_STATISTICS_PARAMS, sizeof (ETH_GET_PPORT_STATISTICS_PARAMS) == 656);


typedef struct _IOCTL_ETH_GET_PPORT_STATISTICS {
   union {
      struct { 
         IOCTL_HEADER header;
         ETH_GET_PPORT_STATISTICS_PARAMS params;
      } SG_PACK ; // unnamed struct
      u32 dw[168]; // dword union
   }; // unnamed union
} SG_PACK IOCTL_ETH_GET_PPORT_STATISTICS, *PIOCTL_ETH_GET_PPORT_STATISTICS;

SG_C_ASSERT(__sizeof__IOCTL_ETH_GET_PPORT_STATISTICS, sizeof (IOCTL_ETH_GET_PPORT_STATISTICS) == 672);


typedef BE_NONEMBEDDED_QUEUE_CONTEXT  LANCER_PPORT_STATS_QUEUE_CONTEXT, *PLANCER_PPORT_STATS_QUEUE_CONTEXT;


// NIC_GET_VPORT_STATS Mailbox command request/response structures for vNIC

typedef struct _LANCER_VPORT_STATS {
   union {
      struct { 
         /* Tx */
         u32 txNumPktsLo;
         u32 txNumPktsHi;
         u32 txNumUnicastPktsLo;
         u32 txNumUnicastPktsHi;
         u32 txNumMulticastPktsLo;
         u32 txNumMulticastPktsHi;
         u32 txNumBroadcastPktsLo;
         u32 txNumBroadcastPktsHi;
         u32 txNumBytesLo;
         u32 txNumBytesHi;
         u32 txNumUnicastBytesLo;
         u32 txNumUnicastBytesHi;
         u32 txNumMulticastBytesLo;
         u32 txNumMulticastBytesHi;
         u32 txNumBroadcastBytesLo;
         u32 txNumBroadcastBytesHi;
         u32 txNumDiscardsLo;
         u32 txNumDiscardsHi;
         u32 txNumErrorsLo;
         u32 txNumErrorsHi;
         u32 txNumPkts64BytesLo;
         u32 txNumPkts64BytesHi;
         u32 txNumPkts65to127BytesLo;
         u32 txNumPkts65to127BytesHi;
         u32 txNumPkts128to255BytesLo;
         u32 txNumPkts128to255BytesHi;
         u32 txNumPkts256to511BytesLo;
         u32 txNumPkts256to511BytesHi;
         u32 txNumPkts512to1023BytesLo;
         u32 txNumPkts512to1023BytesHi;
         u32 txNumPkts1024to1518BytesLo;
         u32 txNumPkts1024to1518BytesHi;
         //below two stats are new in VPORT
         u32 txNumPkts1519to9699BytesLo;
         u32 txNumPkts1519to9699BytesHi;

         u32 txNumLsoPktsLo;
         u32 txNumLsoPktsHi;

         /* Rx */
         u32 rxNumPktsLo;
         u32 rxNumPktsHi;
         u32 rxNumUnicastPktsLo;
         u32 rxNumUnicastPktsHi;
         u32 rxNumMulticastPktsLo;
         u32 rxNumMulticastPktsHi;
         u32 rxNumBroadcastPktsLo;
         u32 rxNumBroadcastPktsHi;
         u32 rxNumBytesLo;
         u32 rxNumBytesHi;
         u32 rxNumUnicastBytesLo;
         u32 rxNumUnicastBytesHi;
         u32 rxNumMulticastBytesLo;
         u32 rxNumMulticastBytesHi;
         u32 rxNumBroadcastBytesLo;
         u32 rxNumBroadcastBytesHi;
         u32 rxNumDiscardsLo;
         u32 rxNumDiscardsHi;
         u32 rxNumErrorsLo;
         u32 rxNumErrorsHi;
         u32 rxNumPkts64BytesLo;
         u32 rxNumPkts64BytesHi;
         u32 rxNumPkts65to127BytesLo;
         u32 rxNumPkts65to127BytesHi;
         u32 rxNumPkts128to255BytesLo;
         u32 rxNumPkts128to255BytesHi;
         u32 rxNumPkts256to511BytesLo;
         u32 rxNumPkts256to511BytesHi;
         u32 rxNumPkts512to1023BytesLo;
         u32 rxNumPkts512to1023BytesHi;
         u32 rxNumPkts1024to1518BytesLo;
         u32 rxNumPkts1024to1518BytesHi;
         //below 4 stats are new in VPORT
         u32 rxNumPkts1519to9699BytesLo;
         u32 rxNumPkts1519to9699BytesHi;
         u32 rxNumLroPktsLo;
         u32 rxNumLroPktsHi;

      } SG_PACK ; // unnamed struct
      u32 dw[72]; // dword union
   }; // unnamed union
} SG_PACK LANCER_VPORT_STATS, *PLANCER_VPORT_STATS;



typedef struct _ETH_GET_VPORT_STATISTICS_PARAMS {
   union {
      union { 
         struct {
             UEXACT16  if_id;
             UEXACT8   rsvd0;
             UEXACT8   reset_stats;
         } request;
         LANCER_VPORT_STATS response;
      } SG_PACK ; // unnamed struct
      u32 dw[73]; // dword union
   }; // unnamed union
} SG_PACK ETH_GET_VPORT_STATISTICS_PARAMS, *PETH_GET_VPORT_STATISTICS_PARAMS;


typedef struct _IOCTL_ETH_GET_VPORT_STATISTICS {
   union {
      struct { 
         IOCTL_HEADER header;
         ETH_GET_VPORT_STATISTICS_PARAMS params;
      } SG_PACK ; // unnamed struct
      u32 dw[77]; // dword union
   }; // unnamed union
} SG_PACK IOCTL_ETH_GET_VPORT_STATISTICS, *PIOCTL_ETH_GET_VPORT_STATISTICS;

typedef BE_NONEMBEDDED_QUEUE_CONTEXT  LANCER_VPORT_STATS_QUEUE_CONTEXT, *PLANCER_VPORT_STATS_QUEUE_CONTEXT;

typedef struct _IOCTL_COMMON_ADD_IFACE_QUEUE_FILTER {
   union {
      struct { 
         IOCTL_HEADER header;
         union {
            struct {
               UEXACT32 if_id;
               UEXACT16 filter_type;
               UEXACT16 cq_id;
               UEXACT16 filter_val;
               UEXACT8  mac_lo[2];
               UEXACT8  mac_hi[4];
            } request;
            struct {
               UEXACT32 filter_id;
            } response;
         } params;
      } SG_PACK ; // unnamed struct
      u32 dw[8]; // dword union
   }; // unnamed union
} SG_PACK IOCTL_COMMON_ADD_IFACE_QUEUE_FILTER, *PIOCTL_COMMON_ADD_IFACE_QUEUE_FILTER;

SG_C_ASSERT(__sizeof__IOCTL_COMMON_ADD_IFACE_QUEUE_FILTER, sizeof (IOCTL_COMMON_ADD_IFACE_QUEUE_FILTER) == 32);


typedef struct _IOCTL_COMMON_DEL_IFACE_QUEUE_FILTER {
   union {
      struct { 
         IOCTL_HEADER header;
         union {
            struct {
               UEXACT32 if_id;
               UEXACT32 filter_id;
            } request;
            struct {
               UEXACT32 rsvd1;
               UEXACT32 rsvd2;
            } response;
         } params;
      } SG_PACK ; // unnamed struct
      u32 dw[6]; // dword union
   }; // unnamed union
} SG_PACK IOCTL_COMMON_DEL_IFACE_QUEUE_FILTER, *PIOCTL_COMMON_DEL_IFACE_QUEUE_FILTER;

SG_C_ASSERT(__sizeof__IOCTL_COMMON_DEL_IFACE_QUEUE_FILTER, sizeof (IOCTL_COMMON_DEL_IFACE_QUEUE_FILTER) == 24);


typedef struct _IOCTL_COMMON_MODIFY_CQ {
   union {
      struct { 
         IOCTL_HEADER header;
         union {
            struct {
               UEXACT32 cq_id : 16;     // DWORD 0
               UEXACT32 rsvd0 : 13;     // DWORD 0
               UEXACT32 cwm_cntrl : 2;  // DWORD 0
               UEXACT32 eq_id_val : 1;  // DWORD 0
               UEXACT32 cwm_val : 8;    // DWORD 1
               UEXACT32 rsvd1 : 8;      // DWORD 1
               UEXACT32 eq_id : 16;     // DWORD 1
            } request;
            struct {
               UEXACT32 rsvd0;
               UEXACT32 rsvd1;
            } response;
         } params;
      } SG_PACK ; // unnamed struct
      u32 dw[6]; // dword union
   }; // unnamed union
} SG_PACK IOCTL_COMMON_MODIFY_CQ, *PIOCTL_COMMON_MODIFY_CQ;

SG_C_ASSERT(__sizeof__IOCTL_COMMON_MODIFY_CQ, sizeof (IOCTL_COMMON_MODIFY_CQ) == 24);


typedef struct _IOCTL_COMMON_WRITE_OBJECT {
   union {
      struct { 
         IOCTL_HEADER header;
         union {
            struct {
               UEXACT32 write_len : 24;    // DWORD 0
               UEXACT32 rsvd0 : 6;         // DWORD 0
               UEXACT32 noc : 1;           // DWORD 0
               UEXACT32 eof : 1;           // DWORD 0
               UEXACT32 write_offset;      // DWORD 1
               UEXACT32 obj_name[26];      // DWORD 2 to 27
               UEXACT8  write_data[4];     // DWORD 28. Size extends till write_len.
            } request;
            struct {
               UEXACT32 actual_write_len;
               UEXACT32 change_status : 8;
               UEXACT32 rsvd1 : 24;
            } response;
         } params;
      } SG_PACK ; // unnamed struct
   }; // unnamed union
} SG_PACK IOCTL_COMMON_WRITE_OBJECT, *PIOCTL_COMMON_WRITE_OBJECT;


///////////////////////////////////
////////// Declarations ///////////
///////////////////////////////////

// Function implemented in funcobj_ll.c
//
// Should be moved to beclib_private_ll.h? 
void
be_clear_function_object (
    IN PBE_FUNCTION_OBJECT function_object);

// Function implemented in mpu_ll.c
//
// Should be moved to beclib_ll_bmap.h?
BESTATUS
lancer_drive_POST (
   IN PBE_FUNCTION_OBJECT function_object
);

// Function implemented in funcobj_ll.c
//
// Should be moved to beclib_private_ll.h? 
BESTATUS
lancer_function_set_adapter_type (
    IN PBE_FUNCTION_OBJECT function_object,
    IN u32                 adapter_family,
    IN u32                 sli_intf_type,
    IN u32                 sli_intf_asic);

// Function implemented in rxf_ll.c
//
// Should be moved to beclib_ll_bmap.h?
BESTATUS
lancer_query_pport_statistics (
    IN PBE_FUNCTION_OBJECT               function_object,
    IN IOCTL_ETH_GET_PPORT_STATISTICS*   va_for_ioctl,
    IN u64                               pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK              callback,
    IN PVOID                             callback_context,
    IN PLANCER_PPORT_STATS_QUEUE_CONTEXT queue_context);

BESTATUS
lancer_query_vport_statistics (
    IN PBE_FUNCTION_OBJECT               function_object,
    IN IOCTL_ETH_GET_VPORT_STATISTICS*   va_for_ioctl,
    IN u64                               pa_for_ioctl,
    IN PBE_RXIF_OBJECT                   rxif_object,  // rxif object
    IN MCC_WRB_CQE_CALLBACK              callback,
    IN PVOID                             callback_context,
    IN PLANCER_VPORT_STATS_QUEUE_CONTEXT queue_context);


BESTATUS
lancer_function_set_diag_registers (
    IN PBE_FUNCTION_OBJECT        function_object,
    u8                            enable,
    IN MCC_WRB_CQE_CALLBACK       callback,
    IN PVOID                      callback_context,
    IN PBE_GENERIC_QUEUE_CONTEXT  queue_context);

BESTATUS
be_rxf_link_status_v1 (
    IN  PBE_FUNCTION_OBJECT           function_object,
    OUT BE_LINK_STATUS*               link_status,
    IN  MCC_WRB_CQE_CALLBACK          callback,
    IN  PVOID                         callback_context,
    IN  PBE_LINK_STATUS_QUEUE_CONTEXT queue_context);

BESTATUS
be_get_maclist (
    IN PBE_RXIF_OBJECT                      rxif_object,
    OUT u32                                 *mac_id,
    IN u8                                   domain,
    IN MCC_WRB_CQE_CALLBACK                 callback,
    IN PVOID                                callback_context,
    IN PBE_IF_GET_MACLIST_QUEUE_CONTEXT     queue_context,
    IN IOCTL_COMMON_GET_IFACE_MACLIST      *va_for_ioctl,
    IN u64                                  pa_for_ioctl
    );

BESTATUS
be_set_maclist (
    IN PBE_FUNCTION_OBJECT                  function_object,
    IN u8                                   *mac,
    IN u8                                   num_macs,
    IN u8                                   domain,
    IN MCC_WRB_CQE_CALLBACK                 callback,
    IN PVOID                                callback_context,
    IN PBE_IF_SET_MACLIST_QUEUE_CONTEXT     queue_context,
    IN IOCTL_COMMON_SET_IFACE_MACLIST       *va_for_ioctl,
    IN u64                                  pa_for_ioctl
    );

BESTATUS
be_rxf_get_mac_with_id (
    IN PBE_FUNCTION_OBJECT      function_object,
    IN     u32                  mac_id,
    IN     PBE_RXIF_OBJECT      rxif_object,
    IN OUT PSA_MAC_ADDRESS      mac_address,
    IN     MCC_WRB_CQE_CALLBACK callback,
    IN     PVOID                callback_context
    );

BESTATUS 
lancer_rxf_add_queue_filter (
    IN PBE_FUNCTION_OBJECT        function_object,
    IN PBE_RXIF_OBJECT            rxif_object,
    IN u16                        cq_id,
    IN u16                        filter_type,
    IN u16                        filter_val,
    IN PSA_MAC_ADDRESS            mac_address,
    OUT u32                       *filter_id,
    IN MCC_WRB_CQE_CALLBACK       callback,
    IN PVOID                      callback_context,
    IN PBE_GENERIC_QUEUE_CONTEXT  queue_context);

BESTATUS 
lancer_rxf_del_queue_filter (
    IN PBE_FUNCTION_OBJECT        function_object,
    IN PBE_RXIF_OBJECT            rxif_object,
    IN u32                        filter_id,
    IN MCC_WRB_CQE_CALLBACK       callback,
    IN PVOID                      callback_context,
    IN PBE_GENERIC_QUEUE_CONTEXT  queue_context);

BESTATUS
lancer_cq_modify (
   IN PBE_FUNCTION_OBJECT       function_object,
   IN PBE_CQ_OBJECT             cq_object,   
   IN PBE_EQ_OBJECT             eq_object,
   IN u8                        cwm_cntrl,
   IN u8                        cwm_val,
   IN MCC_WRB_CQE_CALLBACK      callback,
   IN PVOID                     callback_context,
   IN BE_GENERIC_QUEUE_CONTEXT  *queue_context);

BESTATUS
lancer_wait_for_physdev_inp (
   IN  PBE_FUNCTION_OBJECT function_object);

#endif // __lancer_temp_h__
