/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#include "pch.h"

// LANCER TODO: Remove once LANCER specific defs find their right place
#include "lancer_temp.h"

BESTATUS
be_function_reset(
   IN PBE_FUNCTION_OBJECT  function_object);

void
be_function_abort_in_progress_wrbs (
    IN PBE_FUNCTION_OBJECT       function_object);

//============================================================================
//                            F I L E   S C O P E
//============================================================================

//<no_inbox>
/*!
@brief 
    This routine allocates iscsi ooo buffers & then registers them with the controller.
@param function_object - the iscsi function object to create these for
@param page_count      - number of pages to allocate for i_sc_si ooo buffers
@return 
    BE_SUCCESS if successful, otherwise a useful BESTATUS is returned.
*/
BESTATUS
be_function_iscsi_ooo_buffers_post (
    PBE_FUNCTION_OBJECT function_object,
    u32                 page_count,
    PSA_SGL             sgl)
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_ISCSI_CFG_POST_OOO_BUFFERS *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    u32 num_pages = 0;
    u32 page_offset = 0;
    u32 i;

    be_lock_wrb_post (function_object); 

    ASSERT (sa_sgl_get_offset (sgl) == 0); // need page aligned buffers
    ASSERT (page_count <= sa_sgl_get_page_count (sgl));

    // sw tracking
    function_object->config.num_ooo += page_count;

    TRACE (DL_INFO, "post ooo buffers. num:%d", page_count);

    // Post numerous ioctls to keep each one small enuf to be embedded in 
    // the WRB.
    do {
        wrb = be_function_peek_mcc_wrb (function_object);
        if (!wrb) {
            TRACE (DL_ERR, "MCC wrb peek failed.");
            status = BE_STATUS_NO_MCC_WRB;
            goto error;
        }

        // Prepares an embedded ioctl, including request/response sizes.
        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ISCSI_CFG_POST_OOO_BUFFERS);
        
        num_pages = MIN (page_count, SA_NUMBER_OF_FIELD (IOCTL_COMMON_ISCSI_CFG_POST_OOO_BUFFERS, params.request.pages));
        ioctl->params.request.num_pages = num_pages; 
        
        // Create a page list for the IOCTL.
        for (i=0; i<ioctl->params.request.num_pages; i++) {
            u64 pa = sa_sgl_get_page (sgl, page_offset + i);
            ioctl->params.request.pages[i].lo = sa_lo (pa);
            ioctl->params.request.pages[i].hi = sa_hi (pa);            
        }

        page_offset += num_pages;
        page_count -= num_pages;

        ioctl->params.request.final = (page_count == 0 ? 1 : 0);

        status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
        if (status != BE_SUCCESS) {
            TRACE (DL_ERR, "MCC to add iscsi ooo buffers failed.");
            goto error;
        }

    } while (page_count > 0);


error:
    be_unlock_wrb_post (function_object); 
    return status;
}

/*!
@brief
    This routine allocates iSCSI ooo buffers & then registers them with the controller.
@param    
    FunctionObject      - The iSCSI function object to create these for
@param    
    PageCount           - Number of pages to allocate for iSCSI ooo buffers
@return   
    BE_SUCCESS if successfull, otherwise a useful BESTATUS is returned.
*/
BESTATUS
be_function_iscsi_ooo_buffers_remove (
    PBE_FUNCTION_OBJECT     function_object)
{
    BESTATUS status = BE_SUCCESS;
    IOCTL_COMMON_ISCSI_CFG_REMOVE_OOO_BUFFERS *ioctl = NULL;
    MCC_WRB *wrb = NULL;

    be_lock_wrb_post (function_object);

    // sw tracking
    function_object->config.num_ooo = 0;
    TRACE (DL_INFO, "remove ooo buffers.");

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ISCSI_CFG_REMOVE_OOO_BUFFERS);
        
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "MCC to remove iscsi ooo buffers failed.");
        goto error;
    }

error:

    be_unlock_wrb_post (function_object);    
    return status;
}
//</no_inbox>



/*!

@brief
    This allocates an initializes a function object based on the information
    provided by upper layer drivers.         

@param        
     BarLocations     - Location & information on PCI BARs contained by this PCI function
@param    
    Upcall            - Dispatch table of calls into the upper layer drivers
@param    
    ppFunctionObject  - New function object

@return   
    Returns BE_SUCCESS on success and an appropriate BESTATUS on failure.

@note
    IRQL < DISPATCH_LEVEL    
    
    A function object represents a single
    BladeEngine (logical) PCI function. That is a function object either represents
    the networking side of BladeEngine or the iSCSI side of BladeEngine. Each
    function object has a chip object that is shared by up to 1 other function
    object.
    
    This routine will also detect and create an appropriate PD object for the
    PCI function as needed.

*/
BESTATUS
be_function_create_ex (
    IN      SA_DEV*                 sa_dev, 
    IN      u32                     function_type,
    IN      PSA_SGL                 mailbox,
    IN OUT SET_DRIVER_FUNCTION_CAPABILITIES_PARAMS_V1* caps,
    OUT     PBE_FUNCTION_OBJECT     function_object)
{
    BESTATUS status;
    PCICFG_MEMBAR_CTRL_INT_CTRL_CSR membar_ctrl_int_ctrl;

    SA_TRACE_ENTRY ();

    ASSERT (sa_dev);
    ASSERT (function_object); // not a magic assert
    ASSERT (function_type <= 3);

    TRACE (DL_INFO, 
           "Create function object. type:%s sa_dev:0x%p object:0x%p", 
           (function_type == BE_FUNCTION_TYPE_ISCSI ?   "iSCSI" : 
           (function_type == BE_FUNCTION_TYPE_NETWORK ? "Network" : 
           (function_type == BE_FUNCTION_TYPE_FCOE ? "FCoE" : 
                                                        "Arm"))), 
           sa_dev,
           function_object);
        
    function_object->magic     = BE_FUNCTION_MAGIC;
    function_object->ref_count = 0;
    function_object->type      = function_type;    
    function_object->sa_dev    = sa_dev;

    be_lock_init (&function_object->lock);

    sa_init_spinlock (&function_object->cq_lock);
    sa_init_spinlock (&function_object->post_lock);
    sa_init_spinlock (&function_object->mcc_context_lock);

    sa_initialize_list_head (&function_object->links.pd_list_head);
    sa_initialize_list_head (&function_object->links.cq_list_head);
    sa_initialize_list_head (&function_object->links.eq_list_head);
    sa_initialize_list_head (&function_object->links.cxn_list_head);
    sa_initialize_list_head (&function_object->links.eth_sq_list_head);
    sa_initialize_list_head (&function_object->links.eth_rq_list_head);    
    sa_initialize_list_head (&function_object->links.lro_rq_list_head);    
    sa_initialize_list_head (&function_object->links.rxif_list_head);
    sa_initialize_list_head (&function_object->InProgressList);

    if (be_function_is_iscsi (function_object)) {
        sa_initialize_list_head (&function_object->links.iscsi.wrbq_list_head);
        sa_initialize_list_head (&function_object->links.iscsi.defq_list_head);
    } else if (be_function_is_networking (function_object)){

        sa_initialize_list_head (&function_object->links.networking.dq_list_head);
        sa_initialize_list_head (&function_object->links.networking.sq_list_head);
        sa_initialize_list_head (&function_object->links.networking.rq_list_head);        
    } else if (be_function_is_fcoe (function_object)){

        sa_initialize_list_head (&function_object->links.fcoe.wrbq_list_head);
        sa_initialize_list_head (&function_object->links.iscsi.defq_list_head);
    }

/*    membar_ctrl_int_ctrl.dw = PCICFG_READ (function_object, membar_ctrl_intr_ctrl);*/
    // Do not need function number in BECLIB for LANCER
    if (function_object->adapter_family != ADAPTER_FAMILY_LANCER) {
        BeSetDword( &membar_ctrl_int_ctrl, PCICFG_READ (function_object, membar_ctrl_intr_ctrl));
    
        function_object->pci_function_number = membar_ctrl_int_ctrl.PFNUM;
    }

    function_object->config.rss_type = RSS_ENABLE_NONE;

    #if defined(BE_EMULATION)
        function_object->emulate = TRUE;
        TRACE (DL_NOTE, "Emulation mode");
    #else
        function_object->emulate = FALSE;
        TRACE (DL_NOTE, "Non-emulation mode");

        status = be_download_management_firmware (function_object);
        if (status != BE_SUCCESS) {
            TRACE (DL_ERR, "Management firmware download failed.");
            goto error;
        }

        if (function_object->adapter_family == ADAPTER_FAMILY_LANCER) {
            status = lancer_drive_POST (function_object);
            if (status != BE_SUCCESS) {
                TRACE (DL_ERR, "LANCER POST failed.");
                goto error;
            }
        } else if (BeCsrBarAvailable (function_object->sa_dev)) {
            status = be_drive_POST (function_object);
            if (status != BE_SUCCESS) {
               TRACE (DL_ERR, "BladeEngine POST failed.");
               goto error;
            }
        } else {
            #ifdef SA_DEBUG
            u32 postStatus = BePdRead (function_object->sa_dev, 0x180, NULL);
            TRACE (DL_ERR, "Skip post for VF. Post:0x%x", postStatus);
            #endif
        }
    #endif

    // Initialize the mailbox
    status = be_mpu_init_mailbox (function_object, mailbox);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "Failed to initialize mailbox.");
        goto error;
    }

    // BUG -- Hack to avoid be_function_reset on V1 until it works for
    // PDs > 0. Remove the "if" statement once this works.
    if ((function_object->adapter_family == ADAPTER_FAMILY_LANCER) || !(BeCsrBarAvailable (function_object->sa_dev))){
        (void)be_function_reset(function_object);
    } else if (!caps || (caps->request.sbz[111] != 7)) {
        (void)be_function_reset(function_object);
    } else {
        caps->request.sbz[111] = 0;
    }

    if (caps) {
        if (function_object->adapter_family != ADAPTER_FAMILY_LANCER) {
            status = be_function_internal_driver_capabilities(
                        function_object, 
                        caps);
            if (status != BE_SUCCESS) {
                memset (caps, 0, sizeof (*caps));
                status = 0;
            }
        } else {
            // LANCER TODO: Command not present in SLI4.
            memset (caps, 0, sizeof (*caps));
            status = 0;
        }
    }

    // Cache the firmware config for ASSERTs in beclib and later driver queries.
    status = be_function_internal_query_fw_cfg (
                function_object,
                &function_object->fw_cfg);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "Failed to query firmware config.");
        goto error;
    }

    // LACNER -> BE schematics. Avoids lots of checks further down the road
    if (function_object->adapter_family == ADAPTER_FAMILY_LANCER) {
        function_object->fw_cfg.asic_revision = (function_object->sli_intf_asic << 8);
    }

    //
    // LACNER TODO: FW not returning function capabilities properly.
    // As per SLI4, COMMON_QUERY_FW_CONFIG command will be deprecated. Need to
    // implement alternative command
    //
    if (function_object->adapter_family == ADAPTER_FAMILY_LANCER) {
        // Set what is appropriate for LANCER
        function_object->fw_cfg.function_caps =
            (BE_FUNCTION_CAPS_RSS |
             BE_FUNCTION_CAPS_PROMISCUOUS);
    }

    // TODO -- Consider getting this from the firmware config IOCTL.
    // Asumming tpre.rev_id < 10 covers up to the max number of ASICs we may spin.
    if ((function_object->adapter_family != ADAPTER_FAMILY_LANCER) &&
        BeCsrBarAvailable(function_object->sa_dev))
    {
        function_object->is_asic = (BeCsrRead (function_object->sa_dev, 0x910, "") < 10);
    }
    else
    {
        function_object->is_asic = 1;
    }

error:

    if (status != BE_SUCCESS) {
        // No cleanup necessary
        TRACE (DL_ERR, "Failed to create function.");
        SA_ZERO_MEM (function_object);
    }

    return status;
}


BESTATUS
be_function_create (
    IN      SA_DEV*                 sa_dev, 
    IN      u32                     function_type,
    IN      PSA_SGL                 mailbox,
    OUT     PBE_FUNCTION_OBJECT     function_object)
{
    return be_function_create_ex (sa_dev, function_type, mailbox, NULL, function_object);
}

//---------------------------------------------------------------------------------------
// Function: be_function_is_asic
//
//   Returns TRUE if the driver is currently running on the ASIC, rather than the FPGA. 
//
// function_object - 
// return is_asic  - TRUE, if the driver is currently running on the ASIC.
//---------------------------------------------------------------------------------------
boolean
be_function_is_asic (
    BE_FUNCTION_OBJECT *function_object)
{    
    return function_object->is_asic;
}

//---------------------------------------------------------------------------------------
// Function: be_function_chip_revision
//
//   Returns the overall revision number of the adapter. 
//
// function_object      - 
// return chip_revision - the overall revision number of the adapter
//---------------------------------------------------------------------------------------
u32
be_function_chip_revision (
   BE_FUNCTION_OBJECT* function_object)
{    
    return function_object->fw_cfg.asic_revision;
}

/*!
@brief
    This routine drops the reference count on a given function object. Once
    the reference count falls to zero, the function object is destroyed and all
    resources held are freed.

@param    
    FunctionObject      - The function object to drop the reference to.
*/
BESTATUS
be_function_destroy (
    IN PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    
    TRACE (DL_INFO, "Destroy function_object. Object:0x%p", function_object);

    if (be_function_is_iscsi (function_object)) {
        ASSERT (sa_is_list_empty (&function_object->links.cxn_list_head));
        ASSERT (sa_is_list_empty (&function_object->links.iscsi.wrbq_list_head));
        ASSERT (sa_is_list_empty (&function_object->links.iscsi.defq_list_head));
        ASSERT (function_object->config.num_ooo == 0);
        ASSERT (function_object->config.num_sgl == 0);
    } else {
        ASSERT (be_function_is_networking (function_object));
        ASSERT (sa_is_list_empty (&function_object->links.cxn_list_head));
        ASSERT (sa_is_list_empty (&function_object->links.networking.sq_list_head));
        ASSERT (sa_is_list_empty (&function_object->links.networking.rq_list_head));
        ASSERT (sa_is_list_empty (&function_object->links.networking.dq_list_head));
        ASSERT (function_object->config.num_multicast == 0);
        //ASSERT (function_object->config.rss_type == RSS_ENABLE_NONE);        //BE2-TODO
        // Reset VF, to free any HW resources
        if ((function_object->adapter_family == ADAPTER_FAMILY_LANCER) || !(BeCsrBarAvailable (function_object->sa_dev))){
            (void)be_function_reset(function_object);
        }
    }

    ASSERT (function_object->config.num_template == 0);
    ASSERT (function_object->config.num_jell == 0);
    ASSERT (function_object->config.num_zero == 0);
    ASSERT (sa_is_list_empty (&function_object->links.eth_sq_list_head));
    ASSERT (sa_is_list_empty (&function_object->links.eth_rq_list_head));
    //ASSERT (function_object->config.num_vlan == 0);
    ASSERT (function_object->links.mcc == NULL);
    ASSERT (sa_is_list_empty (&function_object->links.cq_list_head));
    ASSERT (sa_is_list_empty (&function_object->links.eq_list_head));
    ASSERT (function_object->config.num_page_table == 0);
    ASSERT (sa_is_list_empty (&function_object->links.rxif_list_head));

    // Superfluous additional ref count.    
    ASSERT (function_object->ref_count == 0);

    return BE_SUCCESS;
}

// Create chip object. 
BESTATUS
be_chip_object_create (
    PBE_CHIP_OBJECT     chip) // object to initialize
{
    BESTATUS status = 0;
    boolean chip_created = FALSE;


    status = be_chip_create (chip);
    if (status != BE_SUCCESS) {        
        goto Error;
    }

    chip_created = TRUE;

Error:

    if (status != BE_SUCCESS) {

        // Cleanup everything in case of failure.

        if (chip_created) {
            be_chip_destroy (chip);
        }

    }

    return status;

}

// Create function object. 
BESTATUS
be_function_object_create_ex (
    SA_DEV*             sa_dev,           // previously created device
    u32                 function_type,    // e.g FUNCTION_TYPE_ISCSI
    PSA_SGL             mailbox_sgl,
    IN OUT SET_DRIVER_FUNCTION_CAPABILITIES_PARAMS_V1* caps,
    PBE_FUNCTION_OBJECT function_object,
    PBE_CHIP_OBJECT     chip) // object to initialize
{
    BESTATUS status = 0;
    boolean function_created = FALSE;

    status = be_function_create_ex (sa_dev, 
                                    function_type, 
                                    mailbox_sgl, 
                                    caps, 
                                    function_object);
    if (status != BE_SUCCESS) {        
        goto Error;
    }

    function_created = TRUE;


    status = be_chip_insert_function_object (chip, function_object);

Error:

    if (status != BE_SUCCESS) {

        // Cleanup everything in case of failure.

        if (function_created) {
            be_function_destroy (function_object);
        }
    }

    return status;

}


// Create function object. 
BESTATUS
be_function_object_create (
    SA_DEV*             sa_dev,           // previously created device
    u32                 function_type,    // e.g FUNCTION_TYPE_ISCSI
    PSA_SGL             mailbox_sgl,
    PBE_FUNCTION_OBJECT function_object,
    PBE_CHIP_OBJECT     chip) // object to initialize
{   
    return be_function_object_create_ex (sa_dev, function_type, mailbox_sgl, NULL, function_object, chip);

}

// Create both function and chip object with one call.  The lower level
// API will rarely use more than one function object.
BESTATUS
be_function_and_chip_create_ex (
    SA_DEV*             sa_dev,           // previously created device
    u32                 function_type,    // e.g FUNCTION_TYPE_ISCSI
    PSA_SGL             mailbox_sgl,
    PSA_SGL             emulation_sgl,
    IN OUT SET_DRIVER_FUNCTION_CAPABILITIES_PARAMS_V1* caps,
    PBE_FUNCTION_OBJECT function_object,
    PBE_CHIP_OBJECT     chip) // object to initialize
{
    BESTATUS status = 0;
    boolean function_created = FALSE, chip_created = FALSE;
#if defined(BE_EMULATION)
    u64 Physical_addr = 0;
#endif

    status = be_function_create_ex (sa_dev, function_type, mailbox_sgl, caps, function_object);
    if (status != BE_SUCCESS) {        
        goto Error;
    }

    function_created = TRUE;

    status = be_chip_create (chip);
    if (status != BE_SUCCESS) {        
        goto Error;
    }

#if defined(BE_EMULATION)
    if( emulation_sgl ) {
        chip->mcc_emul.emulation_sgl = *emulation_sgl;
    
        Physical_addr = sa_sgl_get_page( &(chip->mcc_emul.emulation_sgl), 0) ;

        chip->mcc_emul.host_addr_lo = Physical_addr & 0xFFFFFFFF;
        chip->mcc_emul.host_addr_hi = (Physical_addr >> 32)  & 0xFFFFFFFF;
    }
#endif

    chip_created = TRUE;

    status = be_chip_insert_function_object (chip, function_object);

Error:

    if (status != BE_SUCCESS) {

        // Cleanup everything in case of failure.

        if (chip_created) {
            be_chip_destroy (chip);
        }

        if (function_created) {
            be_function_destroy (function_object);
        }
    }

    return status;

}

BESTATUS
be_function_and_chip_create (
    SA_DEV*             sa_dev,           // previously created device
    u32                 function_type,    // e.g FUNCTION_TYPE_ISCSI
    PSA_SGL             mailbox_sgl,
    PSA_SGL             emulation_sgl,
    PBE_FUNCTION_OBJECT function_object,
    PBE_CHIP_OBJECT     chip) 
{
    return be_function_and_chip_create_ex (sa_dev, 
                                           function_type, 
                                           mailbox_sgl, 
                                           emulation_sgl, 
                                           NULL, 
                                           function_object, 
                                           chip);
}

BESTATUS
be_function_and_chip_destroy (
    PBE_FUNCTION_OBJECT function_object,
    PBE_CHIP_OBJECT     chip)
{
    TRACE (DL_INFO, "Destroy function & chip.");

    be_chip_remove_function_object (chip, function_object);
    
    be_chip_destroy (chip);

    be_function_destroy (function_object);    

    return BE_SUCCESS;
}

BESTATUS
be_function_object_destroy (
    PBE_FUNCTION_OBJECT function_object,
    PBE_CHIP_OBJECT     chip)
{
    TRACE (DL_INFO, "Destroy function & chip.");

    (void)be_mpu_uninit_mailbox(function_object);

    be_chip_remove_function_object (chip, function_object);
    
    be_function_destroy (function_object);    

    return BE_SUCCESS;
}


BESTATUS
be_chip_object_destroy (
    PBE_CHIP_OBJECT     chip)
{
    TRACE (DL_INFO, "Destroy function & chip.");

    
    be_chip_destroy (chip);

    return BE_SUCCESS;
}

BESTATUS
be_function_cleanup (
    PBE_FUNCTION_OBJECT function_object)
{
    PSA_LIST_ENTRY list_entry;
    BESTATUS status = 0;
    BE_RXIF_OBJECT *rxif = NULL;
    BE_CQ_OBJECT *cq = NULL;
    BE_EQ_OBJECT *eq = NULL;
    BE_ETHSQ_OBJECT *eth_sq = NULL;
    BE_ETHRQ_OBJECT *eth_rq = NULL;
    u32 isr;
    PCICFG_MEMBAR_CTRL_INT_CTRL_CSR ctrl;   
    
    FUNCTION_ASSERT (function_object);
    
    TRACE (DL_INFO, "function cleanup. function:%d", be_function_get_function_number (function_object));

    if (be_function_is_iscsi (function_object)) {
//<no_inbox>
        BE_ISCSI_CONNECTION_OBJECT *iscsi_cxn = NULL;
        BE_ISCSI_WRB_QUEUE_OBJECT *iscsi_wrbq = NULL;
        BE_DEFAULT_PDU_QUEUE_OBJECT *iscsi_defq = NULL;

        // Assert that all connections are terminated.
        ASSERT (sa_is_list_empty (&function_object->links.cxn_list_head));

        while (!sa_is_list_empty (&function_object->links.cxn_list_head)) {
            list_entry = sa_remove_tail_list (&function_object->links.cxn_list_head);
            iscsi_cxn = SA_CONTAINING_RECORD (list_entry, BE_ISCSI_CONNECTION_OBJECT, connection_list);
        
            status = be_iscsi_terminate (iscsi_cxn, NULL, NULL);
            ASSERT (status == BE_SUCCESS || function_object->ue);
        }

        // Destroy all WRB queues.
        while (!sa_is_list_empty (&function_object->links.iscsi.wrbq_list_head)) {
            list_entry = function_object->links.iscsi.wrbq_list_head.Flink;
            iscsi_wrbq = SA_CONTAINING_RECORD (list_entry, BE_ISCSI_WRB_QUEUE_OBJECT, wrbq_list);

            status = be_iscsi_wrb_queue_destroy (iscsi_wrbq);
            ASSERT (status == BE_SUCCESS || function_object->ue);
        }

        // Destroy the default PDU queues
        while (!sa_is_list_empty (&function_object->links.iscsi.defq_list_head)) {
            list_entry = function_object->links.iscsi.defq_list_head.Flink; 
            iscsi_defq = SA_CONTAINING_RECORD (list_entry, BE_DEFAULT_PDU_QUEUE_OBJECT, func_list);

            status = be_iscsi_defq_destroy (iscsi_defq);
            ASSERT (status == BE_SUCCESS || function_object->ue);
        }

        // Out-of-order buffers
        if (function_object->config.num_ooo > 0) {
            status = be_function_iscsi_ooo_buffers_remove (function_object);
            ASSERT (status == BE_SUCCESS || function_object->ue);
        }

        // SGLs
        if (function_object->config.num_sgl > 0) {
            status = be_iscsi_remove_sgl_pages (function_object);
            ASSERT (status == BE_SUCCESS || function_object->ue);
        }
//</no_inbox>
    } else {
        //<no_inbox>

        BE_TOE_CONNECTION_OBJECT *toe_cxn = NULL;
        BE_TOE_SQ_OBJECT *toe_sq = NULL;
        BE_TOE_RQ_OBJECT *toe_rq = NULL;
        BE_TOE_DQ_OBJECT *toe_dq = NULL;

        ASSERT (be_function_is_networking (function_object));        

        // Reset all TOE connections
        while (!sa_is_list_empty (&function_object->links.cxn_list_head)) {
            list_entry = function_object->links.cxn_list_head.Flink;
            toe_cxn = SA_CONTAINING_RECORD (list_entry, BE_TOE_CONNECTION_OBJECT, connection_list);

            // TODO_BE2_TOE: Revisit
            status = be_toe_cxn_upload (toe_cxn, NULL, BE_TOE_UPLOAD_FLAG_RESET, 0, NULL, NULL, NULL, NULL);
            ASSERT (status == BE_SUCCESS || function_object->ue);
        }

        // Destroy all SQs.
        while (!sa_is_list_empty (&function_object->links.networking.sq_list_head)) {
            list_entry = function_object->links.networking.sq_list_head.Flink;
            toe_sq = SA_CONTAINING_RECORD (list_entry, BE_TOE_SQ_OBJECT, list);

            status = be_toe_sq_destroy (toe_sq);
            ASSERT (status == BE_SUCCESS || function_object->ue);
        }

        // Destroy all RQs.
        while (!sa_is_list_empty (&function_object->links.networking.rq_list_head)) {
            list_entry = function_object->links.networking.rq_list_head.Flink;
            toe_rq = SA_CONTAINING_RECORD (list_entry, BE_TOE_RQ_OBJECT, list);

            status = be_toe_rq_destroy (toe_rq);
            ASSERT (status == BE_SUCCESS || function_object->ue);
        }

        // Destroy all the DQs
        while (!sa_is_list_empty (&function_object->links.networking.dq_list_head)) {
            list_entry = function_object->links.networking.dq_list_head.Flink;
            toe_dq = SA_CONTAINING_RECORD (list_entry, BE_TOE_DQ_OBJECT, list);

            status = be_toe_dq_destroy (toe_dq);
            ASSERT (status == BE_SUCCESS || function_object->ue);
        }
        
#if 0
        // don't "help" by destroying wake on lan settings when we are taking the system down.
        
        // ACPI (wake on lan)
        i = 0;
        while (function_object->config.wol_bitmask != 0) {
            u8 mask = 1 << i;
            ASSERT (mask);
            if (function_object->config.wol_bitmask & mask) {
                status = be_rxf_wake_on_lan_config (function_object, FALSE, FALSE, FALSE, i, NULL, NULL, NULL, NULL, NULL);
                ASSERT (status == BE_SUCCESS || function_object->ue);
            }
            i++;
        }
#endif
        //</no_inbox>

        // Multicast
        if (function_object->config.num_multicast > 0) {
            PBE_RXIF_OBJECT prxif_tmp;
            SA_FOR_EACH_LIST_ENTRY (prxif_tmp, function_object->links.rxif_list_head, BE_RXIF_OBJECT, list) {
                status = be_rxf_multicast_config (function_object, prxif_tmp, FALSE, 0, NULL, NULL, NULL, NULL);
                ASSERT (status == BE_SUCCESS || function_object->ue);
            }
        }
        
        if (function_object->config.rss_type != RSS_ENABLE_NONE) {
            // RSS Disable            
            PBE_RXIF_OBJECT prxif_tmp;
            SA_FOR_EACH_LIST_ENTRY (prxif_tmp, function_object->links.rxif_list_head, BE_RXIF_OBJECT, list) {
                status = be_rxf_rss_config (function_object, prxif_tmp, RSS_ENABLE_NONE, FALSE, NULL, 0, NULL, NULL, NULL, NULL);
                ASSERT (status == BE_SUCCESS || function_object->ue);            
            }        
        } else {
            ASSERT (function_object->config.rss_type == RSS_ENABLE_NONE);
        }
    }

    // Template header buffers.
    if (function_object->config.num_template > 0) {
        status = be_function_remove_template_header_memory (function_object);
        ASSERT (status == BE_SUCCESS || function_object->ue);
    }

    //<no_inbox>
    // Journal Entry Linked List
    if (function_object->config.num_jell > 0) {
        status = be_function_remove_jell (function_object);
        ASSERT (status == BE_SUCCESS || function_object->ue);
    }

    // Zero Buffer
    if (function_object->config.num_zero > 0) {
        status = be_function_remove_zero_buffer (function_object);
        ASSERT (status == BE_SUCCESS || function_object->ue);
    }
    //</no_inbox>

    // ETX
    while (!sa_is_list_empty (&function_object->links.eth_sq_list_head)) {
        list_entry = function_object->links.eth_sq_list_head.Flink;
        eth_sq = SA_CONTAINING_RECORD (list_entry, BE_ETHSQ_OBJECT, list);
    
        status = be_eth_sq_destroy (eth_sq);
        ASSERT (status == BE_SUCCESS || function_object->ue);
    }

    // ERX
    while (!sa_is_list_empty (&function_object->links.eth_rq_list_head)) {
        list_entry = function_object->links.eth_rq_list_head.Flink;
        eth_rq = SA_CONTAINING_RECORD (list_entry, BE_ETHRQ_OBJECT, list);

        status = be_eth_rq_destroy (eth_rq);
        ASSERT (status == BE_SUCCESS || function_object->ue);
    }

    // VLAN
    // Need to implement this for BE_EMULATION TODO_BE2:ganesh
    //if (function_object->config.num_vlan > 0) {
    //    status = be_rxf_vlan_config (function_object, FALSE, 0, NULL, NULL, NULL, NULL);
    //    ASSERT (status == BE_SUCCESS);
    //}

    // Destroy the rx interfaces. This cleans up the VLAN entries.
    while (!sa_is_list_empty (&function_object->links.rxif_list_head)) {
        list_entry = function_object->links.rxif_list_head.Flink;
        rxif = SA_CONTAINING_RECORD (list_entry, BE_RXIF_OBJECT, list);
        status = be2_rxf_if_destroy (rxif, NULL, NULL, NULL);
    }
    
    // MCC Queue -- Switches to mailbox mode.  May want to destroy
    // all but the MCC CQ before this call if polling CQ is much better
    // performance than polling mailbox register.
    if (function_object->links.mcc) {
        status = be_mcc_ring_destroy (function_object->links.mcc);
    }

    // CQs
    while (!sa_is_list_empty (&function_object->links.cq_list_head)) {
        list_entry = function_object->links.cq_list_head.Flink;
        cq = SA_CONTAINING_RECORD (list_entry, BE_CQ_OBJECT, cq_list);

        status = be_cq_destroy (cq);
        ASSERT (status == BE_SUCCESS || function_object->ue);
    }

    // EQs
    while (!sa_is_list_empty (&function_object->links.eq_list_head)) {
        list_entry = function_object->links.eq_list_head.Flink;
        eq = SA_CONTAINING_RECORD (list_entry, BE_EQ_OBJECT, eq_list);

        status = be_eq_destroy (eq);
        ASSERT (status == BE_SUCCESS || function_object->ue);
    }

    //
    // If interrupts are disabled, clear any CEV interrupt assertions that 
    // fired after we stopped processing EQs.
    //
/*    ctrl.dw = PCICFG_READ(function_object, membar_ctrl_intr_ctrl);*/
    if(function_object->adapter_family != ADAPTER_FAMILY_LANCER) {
        BeSetDword( &ctrl,PCICFG_READ(function_object, membar_ctrl_intr_ctrl));
    
        if (!ctrl.hostintr) {
            //FIXME TODO @@@ need to enhance this for VMs as PFs get isr0-isr7, and
            //VMs share isr8-isr11.
            if (BeCsrBarAvailable (function_object->sa_dev))
            {
                switch (function_object->pci_function_number) {
                    case 0:
                        isr = CSR_READ (function_object, cev.isr0);
                        break;
                    case 1:
                        isr = CSR_READ (function_object, cev.isr1);
                        break;
                    case 2:
                        isr = CSR_READ (function_object, cev.isr2);
                        break;
                    case 3:
                        isr = CSR_READ (function_object, cev.isr3);
                        break;
                    case 4:
                        isr = CSR_READ (function_object, cev.isr4);
                        break;
                    case 5:
                        isr = CSR_READ (function_object, cev.isr5);
                        break;
                    case 6:
                        isr = CSR_READ (function_object, cev.isr6);
                        break;
                    case 7:
                        isr = CSR_READ (function_object, cev.isr7);
                        break;
                    default:
                        ASSERT(function_object->pci_function_number < 8);
                        TRACE (DL_ERR, "Invalid function number.");
                }
            }
        } else {
            // This should never happen... except for VMQs.
            // TRACE (DL_ERR, "Calling function_cleanup with interrupts enabled.");
        }
    }

    //<no_inbox>
    // UT should be empty.  Remove all page tables
    if (function_object->config.num_page_table > 0) {
        u32 actual_num = 0;
        u32 num_page_table = function_object->config.num_page_table;

        status = be_function_remove_page_tables (function_object, num_page_table, &actual_num, NULL);
        ASSERT (status == BE_SUCCESS || function_object->ue);
        ASSERT (actual_num == num_page_table); // all page tables should be removed
    }
    //</no_inbox>

    // Function object destroy
    status = be_function_object_destroy (function_object, function_object->parent_chip);
    ASSERT (status == BE_SUCCESS || function_object->ue);

    return status;
}


/*!

@brief
    This routine adds a event queue object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object         - Function object to add the object to.
    
@param    
    EqObject        - Event queue object to add.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_eq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_EQ_OBJECT           eq_object)
{
    FUNCTION_ASSERT (function_object);   
    EQ_ASSERT (eq_object);    

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.eq_list_head, &eq_object->eq_list);

    _be_function_unlock (function_object);
}

/*!

@brief
    This routine removes a event queue object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object         - Function object to remove the object from.
    
@param    
    EqObject        - Event queue object to remove.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_eq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_EQ_OBJECT           eq_object)
{
    FUNCTION_ASSERT (function_object);
    EQ_ASSERT (eq_object);

    _be_function_lock (function_object);

    sa_remove_entry_list (&eq_object->eq_list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);

}


void 
_be_function_add_rxif (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_RXIF_OBJECT         rxif_object)
{
    _be_function_lock (function_object);
    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.rxif_list_head, &rxif_object->list);
    _be_function_unlock (function_object);
}

void 
_be_function_remove_rxif (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_RXIF_OBJECT         rxif_object)
{
    _be_function_lock (function_object);

    sa_remove_entry_list (&rxif_object->list);
    sa_atomic_decrement (&function_object->ref_count);
       
    _be_function_unlock (function_object);
}


u32
_be_function_default_rxif_id (
    IN      PBE_FUNCTION_OBJECT     function_object)
{
    u32 id = (u32)-1;

    _be_function_lock (function_object);

    if (!sa_list_is_empty (&function_object->links.rxif_list_head)) {
        SA_LIST_ENTRY *entry;
        PBE_RXIF_OBJECT rxif_object;

        entry = sa_list_head (&function_object->links.rxif_list_head);
        rxif_object = SA_CONTAINING_RECORD (entry, BE_RXIF_OBJECT, list);
        RXIF_ASSERT (rxif_object);

        id = rxif_object->handle;
        
    }

    _be_function_unlock (function_object);

    return id;
}

/*!

@brief
    This routine adds a completion queue object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object         - Function object to add the object to.
    
@param    
    CqObject        - Completion queue object to add.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_cq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_CQ_OBJECT           cq_object)
{
    FUNCTION_ASSERT (function_object);    
    CQ_ASSERT (cq_object);    

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.cq_list_head, &cq_object->cq_list);

    _be_function_unlock (function_object);
}

/*!

@brief
    This routine removes a completion queue object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object         - Function object to remove the object from.
    
@param    
    CqObject        - Completion queue object to remove.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_cq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_CQ_OBJECT           cq_object)
{
    FUNCTION_ASSERT (function_object);    
    CQ_ASSERT (cq_object);    

    _be_function_lock (function_object);

    sa_remove_entry_list (&cq_object->cq_list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);
}


/*!

@brief
    This routine adds an Ethernet Send queue object to a function object to allow
    for correct tracking and reference counting. 

@param    
    func_obj      - function object to add the object to.
    
@param    
    eth_sq        - ethernet send queue object to add.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_eth_sq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_ETHSQ_OBJECT        eth_sq)
{    
    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.eth_sq_list_head, 
                         &eth_sq->list);

    _be_function_unlock (function_object);

}
/*!

@brief
    This routine removes an Ethernet Send Queue object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object         - Function object to remove the object from.
    
@param    
    EthSq        - Ethernet Send Queue object to remove.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_eth_sq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_ETHSQ_OBJECT        eth_sq)
{
    _be_function_lock (function_object);
    
    sa_remove_entry_list (&eth_sq->list);
    sa_atomic_decrement (&function_object->ref_count);
    
    _be_function_unlock (function_object);
}

/*!

@brief
    This routine adds an ethernet receive queue object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object     - Function object to add the object to.
    
@param    
    EthRq       - Ethernet Receive Queue object to add.    
    
@param
    RssIndex    - Rss Index that this RQ will map to. Unless this RQ is for
                  a priviledged execution environment, the RssIndex must be
                  set to RSS_ID_INDEX_DQ.

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_eth_rq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_ETHRQ_OBJECT        eth_rq)
{
    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.eth_rq_list_head, 
                         &eth_rq->list);

    _be_function_unlock (function_object);
}

/*!

@brief
    This routine adds a LRO receive queue object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object     - Function object to add the object to.
    
@param    
    LroRq       - LRO Receive Queue object to add.    
    
@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_lro_rq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_LRORQ_OBJECT        lro_rq)
{
    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.lro_rq_list_head, 
                         &lro_rq->list);

    _be_function_unlock (function_object);
}

/*!

@brief
    This routine removes an Ethernet receive queue object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object     - Function object to remove the object from.
    
@param    
    EthRq       - Ethernet receive queue object to remove.    
    
@param    
    RssIndex    - Rss Index for the EthRq to remove.        

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_eth_rq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_ETHRQ_OBJECT        eth_rq)
{
    _be_function_lock (function_object);
    
    sa_remove_entry_list (&eth_rq->list);
    sa_atomic_decrement (&function_object->ref_count);
    
    _be_function_unlock (function_object);
   
}

/*!

@brief
    This routine removes a LRO receive queue object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object     - Function object to remove the object from.
    
@param    
    LroRq       - LRO receive queue object to remove.    
    
@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_lro_rq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_LRORQ_OBJECT        lro_rq)
{
    _be_function_lock (function_object);
    
    sa_remove_entry_list (&lro_rq->list);
    sa_atomic_decrement (&function_object->ref_count);
    
    _be_function_unlock (function_object);
   
}



/*!

Obrief
    This routine adds a Mcc Send queue object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object         - Function object to add the object to.
    
@param    
    MccSq           - Mcc Send Queue object to add.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_mcc (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_MCC_OBJECT          mcc)
{
    FUNCTION_ASSERT (function_object);
    
    ASSERT (function_object->links.mcc == NULL);    
    MCC_ASSERT (mcc);

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    function_object->links.mcc = mcc;

    _be_function_unlock (function_object);
}

/*!

@brief
    This routine removes an MCC Send Queue object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object     - Function object to remove the object from.
    
@param    
    MccSq      - Mcc Send Queue object to remove.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_mcc (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_MCC_OBJECT          mcc)
{
    FUNCTION_ASSERT (function_object);
    
    MCC_ASSERT (function_object->links.mcc);

    MCC_ASSERT (mcc);

    SA_NOT_USED (mcc);

    _be_function_lock (function_object);
    
    function_object->links.mcc = NULL;
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);
}

/*!

@brief
    This routine adds an iSCSI default PDU queue object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object             - Function object to add the object to.
    
@param    
    DefaultPduObject    - Default PDU object to add.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_default_pdu_queue (
    IN      PBE_FUNCTION_OBJECT             function_object, 
    IN      PBE_DEFAULT_PDU_QUEUE_OBJECT    default_pdu_object)
{
    FUNCTION_ASSERT (function_object);
    ASSERT (be_function_is_iscsi (function_object) || be_function_is_fcoe (function_object));

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.iscsi.defq_list_head, &default_pdu_object->func_list);

    _be_function_unlock (function_object);
}

/*!

@brief
    This routine removes an iSCSI default PDU Queue object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object         - Function object to remove the object from.
    
@param    
    DefaultPduObject- iSCSI Default PDU Queue object to remove.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_default_pdu_queue (
    IN      PBE_FUNCTION_OBJECT             function_object, 
    IN      PBE_DEFAULT_PDU_QUEUE_OBJECT    default_pdu_object)
{
    FUNCTION_ASSERT (function_object);
    ASSERT (be_function_is_iscsi (function_object) || be_function_is_fcoe (function_object));

    _be_function_lock (function_object);

    sa_remove_entry_list (&default_pdu_object->func_list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);
}



/*!

@brief
    This routine adds a TOE connectoin object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object             - Function object to add the object to.
    
@param    
    ConnectionObject    - TOE Connection object to add.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_toe_connection (
    IN      PBE_FUNCTION_OBJECT         function_object, 
    IN      PBE_TOE_CONNECTION_OBJECT   connection_object)
{
    FUNCTION_ASSERT (function_object);    
    ASSERT (be_function_is_iscsi (function_object)== FALSE);
    TOE_CXN_ASSERT (connection_object);    

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.cxn_list_head, 
                         &connection_object->connection_list);

    _be_function_unlock (function_object);
}

/*!

@brief
    This routine removes a TOE connection object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object         - function object to remove the object from.
    
@param    
    connection_object - TOE connection object to remove.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_toe_connection (
    IN      PBE_FUNCTION_OBJECT         function_object, 
    IN      PBE_TOE_CONNECTION_OBJECT connection_object)
{
    FUNCTION_ASSERT (function_object);    
    ASSERT (be_function_is_iscsi (function_object)== FALSE);
    TOE_CXN_ASSERT (connection_object);    

    SA_TRACE_ENTRY ();

    _be_function_lock (function_object);

    sa_remove_entry_list (&connection_object->connection_list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);

    SA_TRACE_EXIT ();

}



/*!

@brief
    This routine adds a TOE default receive queue object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object     - function object to add the object to.
    
@param    
    toe_dq       - OE default receive queue object to add.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_toe_dq (
    IN      PBE_FUNCTION_OBJECT             function_object, 
    IN      PBE_TOE_DQ_OBJECT               toe_dq)
{
    ASSERT (be_function_is_iscsi (function_object) == FALSE);
    
    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.networking.dq_list_head, 
                         &toe_dq->list);

    _be_function_unlock (function_object);
}

/*!

@brief
    this routine removes a TOE default receive queue object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object         - function object to remove the object from.
    
@param    
    toe_dq           - TOE default receive queue object to remove.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_toe_dq (
    IN      PBE_FUNCTION_OBJECT             function_object, 
    IN      PBE_TOE_DQ_OBJECT               toe_dq)
{
    SA_NOT_USED (toe_dq);
    ASSERT (be_function_is_iscsi (function_object) == FALSE);

    _be_function_lock (function_object);
    
    sa_remove_entry_list (&toe_dq->list);
    sa_atomic_decrement (&function_object->ref_count);
    
    _be_function_unlock (function_object);
}


void _be_function_add_toe_sq    (PBE_FUNCTION_OBJECT function_object, PBE_TOE_SQ_OBJECT sq)
{
    FUNCTION_ASSERT (function_object);    
    TOE_SQ_ASSERT (sq);    

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.networking.sq_list_head, &sq->list);

    _be_function_unlock (function_object);
}

void _be_function_remove_toe_sq (PBE_FUNCTION_OBJECT function_object, PBE_TOE_SQ_OBJECT sq)
{
    FUNCTION_ASSERT (function_object);
    TOE_SQ_ASSERT (sq);

    _be_function_lock (function_object);

    sa_remove_entry_list (&sq->list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);

}

void _be_function_add_toe_rq    (PBE_FUNCTION_OBJECT function_object, PBE_TOE_RQ_OBJECT rq)
{
    FUNCTION_ASSERT (function_object);    
    TOE_RQ_ASSERT (rq);    

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.networking.rq_list_head, &rq->list);

    _be_function_unlock (function_object);
}

void _be_function_remove_toe_rq (PBE_FUNCTION_OBJECT function_object, PBE_TOE_RQ_OBJECT rq)
{
    FUNCTION_ASSERT (function_object);
    TOE_RQ_ASSERT (rq);

    _be_function_lock (function_object);

    sa_remove_entry_list (&rq->list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);
}

void _be_function_add_rdma_qp (PBE_FUNCTION_OBJECT function_object, PBE_RDMA_QP_OBJECT qp_object)
{
    FUNCTION_ASSERT (function_object);    
    RDMA_QP_ASSERT (qp_object);

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.rdma_qp_list_head, &qp_object->list);

    _be_function_unlock (function_object);
}

void _be_function_remove_rdma_qp (PBE_FUNCTION_OBJECT function_object, PBE_RDMA_QP_OBJECT qp_object)
{
    FUNCTION_ASSERT (function_object);
    RDMA_QP_ASSERT (qp_object);

    _be_function_lock (function_object);

    sa_remove_entry_list (&qp_object->list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);
}


/*!

@brief
    This routine adds a event queue object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object         - Function object to add the object to.
    
@param    
    EqObject        - Event queue object to add.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_wrbq (
    IN      PBE_FUNCTION_OBJECT        function_object, 
    IN      PBE_ISCSI_WRB_QUEUE_OBJECT wrbq)
{
    FUNCTION_ASSERT (function_object);    
    ISCSI_WRBQ_ASSERT (wrbq);    

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.iscsi.wrbq_list_head, &wrbq->wrbq_list);

    _be_function_unlock (function_object);
}

/*!

@brief
    This routine removes a event queue object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object         - Function object to remove the object from.
    
@param    
    EqObject        - Event queue object to remove.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_wrbq (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PBE_ISCSI_WRB_QUEUE_OBJECT wrbq)
{
    FUNCTION_ASSERT (function_object);
    ISCSI_WRBQ_ASSERT (wrbq);

    _be_function_lock (function_object);

    sa_remove_entry_list (&wrbq->wrbq_list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);

}

void
_be_function_add_fcoe_wrbq (
    IN      PBE_FUNCTION_OBJECT   function_object,
    IN      PBE_FCOE_WRB_QUEUE_OBJECT wrbq)
{
    FUNCTION_ASSERT (function_object);
    FCOE_WRBQ_ASSERT (wrbq);

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.fcoe.wrbq_list_head, &wrbq->wrbq_list);

    _be_function_unlock (function_object);
}


void
_be_function_remove_fcoe_wrbq (
    IN     PBE_FUNCTION_OBJECT   function_object,
    IN     PBE_FCOE_WRB_QUEUE_OBJECT wrbq)
{
    FUNCTION_ASSERT (function_object);
    FCOE_WRBQ_ASSERT (wrbq);

    _be_function_lock (function_object);
   
    sa_remove_entry_list (&wrbq->wrbq_list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);
}

PVOID
be_function_prepare_embedded_ioctl (
    IN PBE_FUNCTION_OBJECT function_object, 
    IN PMCC_WRB wrb,
    IN u32 payload_length, 
    IN u32 request_length, 
    IN u32 response_length,
    IN u32 opcode,
    IN u32 subsystem)
{
    PIOCTL_REQUEST_HEADER header = NULL; 
    u32 *dw = (u32*)wrb;
    
    ASSERT (wrb);
    FUNCTION_ASSERT (function_object);

    ASSERT (payload_length <= SA_SIZEOF_FIELD (MCC_WRB, payload));

    // DWORD 0
    dw[0] = 0;
    wrb->embedded = 1;

    // DWORD 1
    SA_C_ASSERT (AMAP_WORD_OFFSET (MCC_WRB, payload_length) == 1);    
    SA_C_ASSERT (AMAP_BIT_SIZE (MCC_WRB, payload_length) == 32);
    wrb->payload_length = MIN (payload_length, SA_SIZEOF_FIELD (MCC_WRB, payload));

    // DWORD 2,3 assigned later
    SA_C_ASSERT (AMAP_WORD_OFFSET (MCC_WRB, tag) == 2);    

    // DWORD 4
    dw[4] = 0;

    // Start of payload
    SA_C_ASSERT (AMAP_WORD_OFFSET (MCC_WRB, payload) == 5);
    header = (PIOCTL_REQUEST_HEADER)&wrb->payload;

    sa_memset((PVOID)header, 0, sizeof(*header));
    header->timeout         = 0;
    header->domain          = be_function_get_pd_number (function_object);
    header->port_number     = function_object->fw_cfg.phys_port;
    header->request_length  = MAX (request_length, response_length);
    header->opcode          = opcode;
    header->subsystem       = subsystem;
    header->version         = 0;

    return header;
}

PVOID
be_function_prepare_nonembedded_ioctl (
    IN PBE_FUNCTION_OBJECT function_object,
    IN PMCC_WRB wrb,
    IN PVOID ioctl_va,
    IN u64 ioctl_pa,
    IN u32 payload_length,
    IN u32 request_length,
    IN u32 response_length,
    IN u32 opcode,
    IN u32 subsystem,
    IN u8  version)
{
    PIOCTL_REQUEST_HEADER header = NULL;
    u32 *dw = (u32*)wrb;

    ASSERT (wrb);
    ASSERT (ioctl_va);
    FUNCTION_ASSERT (function_object);

    header = (PIOCTL_REQUEST_HEADER)ioctl_va;
    
    // DWORD 0
    dw[0] = 0;
    wrb->embedded = 0;
    wrb->sge_count = 1;

    // DWORD 1
    SA_C_ASSERT (AMAP_WORD_OFFSET (MCC_WRB, payload_length) == 1);
    SA_C_ASSERT (AMAP_BIT_SIZE (MCC_WRB, payload_length) == 32);
    wrb->payload_length = payload_length;

    // DWORD 2,3 assigned later
    SA_C_ASSERT (AMAP_WORD_OFFSET (MCC_WRB, tag) == 2);
    
    // DWORD 4
    dw[4] = 0;

    // Start of payload
    SA_C_ASSERT (AMAP_WORD_OFFSET (MCC_WRB, payload) == 5);

    // Assume one fragment. The caller may override the SGL by rewriting the 0th length
    // and adding more entries.  They will also need to update the sge_count.
    wrb->payload.sgl[0].length = payload_length;
    wrb->payload.sgl[0].pa_lo = sa_lo (ioctl_pa);
    wrb->payload.sgl[0].pa_hi = sa_hi (ioctl_pa);

    sa_memset((PVOID)header, 0, sizeof(*header));
    header->timeout         = 0; // todo -- be_function_get_default_ioctl_timeout (function_object);
    header->domain          = be_function_get_pd_number (function_object);
    header->port_number     = function_object->fw_cfg.phys_port;
    header->request_length  = MAX (request_length, response_length);
    header->opcode          = opcode;
    header->subsystem       = subsystem;
    header->version         = version;

    return header;
}

PMCC_WRB
be_function_peek_mcc_wrb (
    IN PBE_FUNCTION_OBJECT function_object)
{
    PMCC_WRB wrb = NULL;
    FUNCTION_ASSERT (function_object);    
    
    if (function_object->links.mcc) {

        wrb = _be_mpu_peek_ring_wrb (function_object->links.mcc, FALSE); // Not driving the queue

    } else {
        wrb = (PMCC_WRB)&function_object->mailbox.va->wrb;
    }

    if (wrb) {
        sa_memset (wrb, 0, sizeof (MCC_WRB));
    }

    return wrb;
}

/*!
@brief
    This routine posts an MCC WRB command to a function object's MCC send
    queue (if avaiable) or posts the WRB to the MCC WRB mail box if a send
    queue is not setup.
@param    
    function_object     - function object to add the object to.
@param    
    wrb                 - wrb to post.  This may be a va from the be_function_peek_mcc_wrb
                          function, so the driver need not declare additional memory.
@param    
    callback  - address of a callback routine to invoke once the WRB
                           is completed.
@param    
    callback_context - opaque context to be passed during the call to
                                the callback.
@return   
    BE_SUCCESS if successfull, otherwise a useful error code is returned.
    
@note
    IRQL < DISPATCH_LEVEL    
*/
BESTATUS
be_function_post_mcc_wrb (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PMCC_WRB                wrb,        
    IN      MCC_WRB_CQE_CALLBACK    callback, 
    IN      PVOID                   callback_context, 
    IN      PVOID                   optional_ioctl_va)
{
    BE_MCC_WRB_RESPONSE_COPY response_copy = {0};
    return be_function_post_mcc_wrb_complete (function_object, 
                                              wrb, 
                                              NULL,  // queue context
                                              callback, 
                                              callback_context, 
                                              NULL, // internal callback
                                              NULL, // internal callback context
                                              optional_ioctl_va, 
                                              response_copy);
}


BESTATUS
be_function_post_mcc_wrb_with_queue_context (
    IN      PBE_FUNCTION_OBJECT     function_object, 
    IN      PMCC_WRB                wrb,   
    IN      PBE_GENERIC_QUEUE_CONTEXT queue_context,
    IN      MCC_WRB_CQE_CALLBACK    callback, 
    IN      PVOID                   callback_context, 
    IN      PVOID                   optional_ioctl_va)
{
    BE_MCC_WRB_RESPONSE_COPY response_copy = {0};
    return be_function_post_mcc_wrb_complete (function_object, 
                                              wrb, 
                                              queue_context,  // queue context
                                              callback, 
                                              callback_context, 
                                              NULL, // internal callback
                                              NULL, // internal callback context
                                              optional_ioctl_va, 
                                              response_copy);
}

BESTATUS
be_function_post_mcc_wrb_with_copy (
    IN PBE_FUNCTION_OBJECT       function_object,
    IN PMCC_WRB                  wrb,
    IN PBE_GENERIC_QUEUE_CONTEXT queue_context,
    IN MCC_WRB_CQE_CALLBACK      callback,
    IN PVOID                     callback_context,
    IN PVOID                     optional_ioctl_va,
    IN BE_MCC_WRB_RESPONSE_COPY  response_copy)
{
    return be_function_post_mcc_wrb_complete (function_object, 
                                               wrb, 
                                               queue_context,
                                               callback, 
                                               callback_context, 
                                               NULL, // internal callback
                                               NULL, // internal callback context
                                               optional_ioctl_va, 
                                               response_copy);
}

BESTATUS
be_function_post_mcc_wrb_with_internal_callback (
    IN PBE_FUNCTION_OBJECT       function_object,
    IN PMCC_WRB                  wrb,
    IN PBE_GENERIC_QUEUE_CONTEXT queue_context,
    IN MCC_WRB_CQE_CALLBACK      callback,    
    IN PVOID                     callback_context,
    IN MCC_WRB_CQE_CALLBACK      internal_callback,
    IN PVOID                     internal_callback_context,
    IN PVOID                     optional_ioctl_va
    )
{
    BE_MCC_WRB_RESPONSE_COPY response_copy = {0};
    return be_function_post_mcc_wrb_complete (function_object, 
                                               wrb, 
                                               queue_context,
                                               callback, 
                                               callback_context, 
                                               internal_callback,         // internal callback
                                               internal_callback_context, // internal callback context
                                               optional_ioctl_va, 
                                               response_copy);
}


#if defined(SA_DEBUG)        
    void be_function_debug_print_wrb (PBE_FUNCTION_OBJECT function_object, 
                                      PMCC_WRB wrb, 
                                      PVOID optional_ioctl_va, 
                                      PBE_MCC_WRB_CONTEXT wrb_context)
    {

        PIOCTL_REQUEST_HEADER header = NULL;
        if (wrb->embedded) {
            header = (PIOCTL_REQUEST_HEADER)&wrb->payload;
        } else {
            header = (PIOCTL_REQUEST_HEADER)optional_ioctl_va;
        }

        // Save the completed count before posting for a debug assert.        
        wrb_context->consumed_count = function_object->stats.consumed_wrbs;

        if (header) {
            wrb_context->opcode = header->opcode;
            wrb_context->subsystem = header->subsystem;
//<no_inbox>
            TRACE (DL_IOCTL,
                "Post     IOCTL. emul:%d embed:%d pd:%d sub:0x%x op:0x%x ver:0x%x ctx:%p (%s/%s)",
                be_mccemul_is_emulated (function_object, wrb, optional_ioctl_va),
                wrb->embedded,
                header->domain, // only valid for emulated postings
                header->subsystem,               
                header->opcode,
                header->version,
                wrb_context,
                ioctl_subsystems_get_name (header->subsystem), 
                be_query_opcode_name (wrb_context->subsystem, wrb_context->opcode)
                );
//</no_inbox>
        } else {
            wrb_context->opcode = 0;
            wrb_context->subsystem = 0;
            TRACE (DL_IOCTL, 
                   "Post     IOCTL. emul:0 embed:0 Unknown sub/op. ctx:%p",
                   wrb_context);
        }
    }
#else
    #define be_function_debug_print_wrb(a_, b_, c_, d_)
#endif

BESTATUS
be_function_post_mcc_wrb_complete (
    IN PBE_FUNCTION_OBJECT       function_object,
    IN PMCC_WRB                  wrb,
    IN PBE_GENERIC_QUEUE_CONTEXT queue_context,
    IN MCC_WRB_CQE_CALLBACK      callback,
    IN PVOID                     callback_context,
    IN MCC_WRB_CQE_CALLBACK      internal_callback,
    IN PVOID                     internal_callback_context,
    IN PVOID                     optional_ioctl_va,
    IN BE_MCC_WRB_RESPONSE_COPY  response_copy)
{
    BESTATUS    status;        
    PBE_MCC_WRB_CONTEXT wrb_context = NULL;
    FUNCTION_ASSERT (function_object);

    if (function_object->ue) {
        //
        // Previous UE detected. Fail this request.
        //
        status = BE_STATUS_UE_DETECTED;

        if (internal_callback) {
            // Invoke the internal callback during failure.
            internal_callback (internal_callback_context, status, NULL);
        }

        goto Error;
    }

    if (queue_context) {

        // Initialize context.        
        queue_context->context.internal_callback         = internal_callback;
        queue_context->context.internal_callback_context = internal_callback_context;
        queue_context->context.callback                  = callback;
        queue_context->context.callback_context          = callback_context;
        queue_context->context.copy                      = response_copy;
        queue_context->context.optional_ioctl_va         = optional_ioctl_va;

        // Queue this request
        status = be_function_queue_mcc_wrb (function_object, queue_context);

        if (status != BE_PENDING && internal_callback) {
            // Invoke the internal callback during failure.
            internal_callback (internal_callback_context, status, NULL);
        }

        goto Error;
    }

    // Allocate a WRB context struct to hold the callback pointers, status, etc.
    // This is required if commands complete out of order.
    wrb_context = _be_mcc_allocate_wrb_context (function_object);
    if (!wrb_context) {
        TRACE (DL_WARN, "Failed to allocate MCC WRB context.");
        status = BE_STATUS_SYSTEM_RESOURCES;

        if (internal_callback) {
            // Invoke the internal callback during failure.
            internal_callback (internal_callback_context, status, NULL);
        }

        goto Error;
    }

    // Initialize context.    
    wrb_context->internal_callback         = internal_callback;
    wrb_context->internal_callback_context = internal_callback_context;
    wrb_context->callback                  = callback;
    wrb_context->callback_context          = callback_context;
    wrb_context->copy                      = response_copy;
    wrb_context->wrb                       = wrb;

    // Copy the context pointer into the WRB opaque tag field.
    // Verify assumption of 64-bit tag with a compile time assert.
    SA_C_ASSERT (sizeof (wrb->tag) >= sizeof (u64));    
    wrb->tag[0] = (u32) SA_PTR_TO_U64 (wrb_context);         // low
    wrb->tag[1] = (u32)(SA_PTR_TO_U64 (wrb_context) >> 32);  // high

    // Print info about this IOCTL for debug builds.
    be_function_debug_print_wrb (function_object, wrb, optional_ioctl_va, wrb_context);         
//<no_inbox>
    if (be_mccemul_is_emulated (function_object, wrb, optional_ioctl_va)) {

        MCC_CQ_ENTRY fake_cqe;

        // Emulated commands are never pended.  We will never call the user's callback
        // function, since callbacks often require a different context then the original request.
        // The user must interpret BE_SUCCESS (no callback) differently from BE_PENDING 
        // (callback will be invoked later).
        wrb_context->callback = NULL;
        wrb_context->callback_context = NULL;

        //
        // Send to emulation layer.  optional_ioctl_va is required to emulate non-
        // embedded commands.
        //
        status = be_mccemul_process_wrb (function_object, wrb, optional_ioctl_va);

        //
        // Create a fake CQ entry so we can utilize the same code path to
        // call the callback function.
        //
        SA_ZERO_MEM (&fake_cqe);
        fake_cqe.completed = 1;
        fake_cqe.consumed = 1;
        fake_cqe.completion_status = (i16)status;
        fake_cqe.mcc_tag[0] = wrb->tag[0];
        fake_cqe.mcc_tag[1] = wrb->tag[1];
        fake_cqe.valid = 1;

        //
        // Process the completion, including any copy operation, callback, and freeing context.
        //
        be_mcc_process_cqe (function_object, &fake_cqe);        

    } else //</no_inbox>
    {        

        //
        // issue the WRB to the MPU as appropriate
        //
        if (function_object->links.mcc) {
            //
            // we're in WRB mode, pass to the mcc layer
            //
            status = _be_mpu_post_wrb_ring (function_object->links.mcc, wrb, wrb_context);
        } else {
            //
            // we're in mailbox mode
            //
            status = _be_mpu_post_wrb_mailbox (function_object, wrb, wrb_context);

            // mailbox mode always completes synchronously
            ASSERT (status != BE_STATUS_PENDING);
        }
    }

Error:   

    return  status;
}


/*!
@brief
    Creates and initializes common data structures needed by all offload engines.
    This includes allocating memory for use by the controller for JELL entries
    and ULP tempalates etc.    
    
@param    
    FunctionObject - Function object causing the initialization.
    
@return   
    Returns BE_SUCCESS on success and an appropriate BESTATUS on failure.
    
@note
    IRQL: < DISPATCH_LEVEL
    
    This routine should be called exactly once during chip initialization.
    
*/
BESTATUS
be_function_post_template_header_memory (
    IN      PBE_FUNCTION_OBJECT function_object, 
    IN      PSA_SGL             template_headers_sgl,
    OUT     IOCTL_COMMON_ADD_TEMPLATE_HEADER_BUFFERS_RESPONSE *response)
{
    BESTATUS  status = BE_SUCCESS;  
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_ADD_TEMPLATE_HEADER_BUFFERS *ioctl = NULL;
    u8 *va;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ADD_TEMPLATE_HEADER_BUFFERS);

    ioctl->params.request.num_pages = sa_sgl_get_page_count (template_headers_sgl);
    
    // TODO -- TEMPLATE_HEADER_TYPE_RDMA
    if (be_function_is_iscsi (function_object)) {
        ioctl->params.request.type = TEMPLATE_HEADER_TYPE_ISCSI;
    } else {
        ioctl->params.request.type = TEMPLATE_HEADER_TYPE_TOE;
    }      

    // Update software tracking
    function_object->config.num_template += ioctl->params.request.num_pages;

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (template_headers_sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // For emulation, add the VA for the template header buffers to the ioctl.
    va = sa_sgl_get_base_va (template_headers_sgl);
    ioctl->params.request.pages_va.lo = sa_pvoid_lo (va);
    ioctl->params.request.pages_va.hi = sa_pvoid_hi (va);
    ASSERT (SA_PAGE_ALIGN (va) == va);

    status = be_function_post_mcc_wrb_with_copy (
                function_object,
                wrb,
                NULL, // queue context
                NULL,
                NULL,
                ioctl,
                BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_ADD_TEMPLATE_HEADER_BUFFERS, 
                                             params.response, 
                                             response));    
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "MCC to add template header buffers failed.");
        goto error;
    } 
    
error:
    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS
be_function_remove_template_header_memory (
    IN      PBE_FUNCTION_OBJECT function_object)
{
    BESTATUS status = BE_SUCCESS;
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_REMOVE_TEMPLATE_HEADER_BUFFERS *ioctl = NULL;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_REMOVE_TEMPLATE_HEADER_BUFFERS);

    // TODO -- TEMPLATE_HEADER_TYPE_RDMA
    if (be_function_is_iscsi (function_object)) {
        ioctl->params.request.type = TEMPLATE_HEADER_TYPE_ISCSI;
    } else {
        ioctl->params.request.type = TEMPLATE_HEADER_TYPE_TOE;
    }      

    TRACE (DL_INFO, "remove template headers for function:%d type:%d", 
           be_function_get_function_number (function_object), ioctl->params.request.type);

    // Update software tracking
    function_object->config.num_template = 0;

    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "MCC to add template header buffers failed.");
        goto error;
    }    

error:
    be_unlock_wrb_post (function_object);
    return status;
}

void _be_function_lock   (PBE_FUNCTION_OBJECT fo)
{
    be_lock_acquire (&fo->lock);
}


void _be_function_unlock (PBE_FUNCTION_OBJECT fo)
{
    be_lock_release (&fo->lock);
}


u32 
be_function_get_function_number (
    PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    return function_object->pci_function_number;    
}

u32 
be_function_get_function_type (
    PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    return function_object->type;
}

u32 
be_function_get_pd_number (
    PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);

    // When the ARM is available, the choice of doorbell page determines
    // the PD.  During emulation, we explicitly set the pd_id in each IOCTL
    // so the emulation code knows which PD to use.
    #if defined(BE_EMULATION) 
        return function_object->pd_id;
    #else
        return 0;
    #endif
}


boolean
be_function_is_vm (
    PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    return be_function_get_pd_number (function_object) > 0;
}



/*!

@brief
    References the given object. The object is guaranteed to remain active until
    the reference count drops to zero.

@param    
    cq_object            - CQ handle returned from cq_object_create.

@return   
    returns the current reference count on the object
    
@note
    IRQL: any

*/
u32
_be_function_reference (
    PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    return  sa_atomic_increment (&function_object->ref_count);
}


/*!
@brief
    Dereferences the given object. The object is guaranteed to remain active until
    the reference count drops to zero.
@param    
@return   
    returns the current reference count on the object
@note
    IRQL: any
*/
u32
_be_function_dereference (
    PBE_FUNCTION_OBJECT function_object)
{
    FUNCTION_ASSERT (function_object);
    return  sa_atomic_decrement (&function_object->ref_count);
}


BESTATUS
be_function_ring_destroy_async (
    PBE_FUNCTION_OBJECT  function_object,
    u32                  id,
    u32                  ring_type,
    BE_GENERIC_QUEUE_CONTEXT *queue_context,
    MCC_WRB_CQE_CALLBACK callback,
    PVOID                callback_context,
    MCC_WRB_CQE_CALLBACK internal_callback,
    PVOID                internal_callback_context)
{
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    union {
        PIOCTL_COMMON_EQ_DESTROY eq;
        PIOCTL_COMMON_CQ_DESTROY cq;
        PIOCTL_COMMON_MCC_DESTROY mcc;
        PIOCTL_ETH_TX_DESTROY eth_tx;
        PIOCTL_ETH_RX_DESTROY eth_rx;

        //<no_inbox>
        PIOCTL_TOE_WRBQ_DESTROY toe_wrbq;
        PIOCTL_TOE_DEFQ_DESTROY toe_defq;
        PIOCTL_TOE_RQ_DESTROY toe_rq;        
        PIOCTL_COMMON_ISCSI_WRBQ_DESTROY iscsi_wrbq;
        PIOCTL_COMMON_ISCSI_DEFQ_DESTROY iscsi_defq;
        PIOCTL_LRO_RING_DESTROY lro;
#ifdef FUNCTION_RDMA
        PIOCTL_RDMA_QP_DESTROY rdma_qp;
#endif
        PIOCTL_COMMON_FCOE_WRBQ_DESTROY fcoe_wrbq;
        //</no_inbox>
    } ioctl;

    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;

    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    TRACE (DL_INFO, "Destroy ring id:%d type:%d", id, ring_type);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto Error;
        }
    }

    // Prepare an embedded ioctl, including request/response sizes.
    switch (ring_type) {
    case IOCTL_RING_TYPE_ETH_RX:
        ioctl.eth_rx = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_RX_DESTROY);
        ioctl.eth_rx->params.request.id = id;
        break;
    case IOCTL_RING_TYPE_ETH_TX:
        ioctl.eth_tx = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, ETH_TX_DESTROY);
        ioctl.eth_tx->params.request.id = id;
        break;
    case IOCTL_RING_TYPE_MCC:
        ioctl.mcc = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_MCC_DESTROY);
        ioctl.mcc->params.request.id = id;
        break;
    case IOCTL_RING_TYPE_CQ:
        ioctl.cq = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_CQ_DESTROY);
        ioctl.cq->params.request.id = id;
        break;
    case IOCTL_RING_TYPE_EQ:
        ioctl.eq = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_EQ_DESTROY);
        ioctl.eq->params.request.id = id;
        break;

    //<no_inbox>
    case IOCTL_RING_TYPE_ISCSI_WRBQ:
        ioctl.iscsi_wrbq = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ISCSI_WRBQ_DESTROY);
        ioctl.iscsi_wrbq->params.request.id = id;
        break;
    case IOCTL_RING_TYPE_ISCSI_DEFQ:
        ioctl.iscsi_defq = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ISCSI_DEFQ_DESTROY);
        ioctl.iscsi_defq->params.request.id = id;
        break;
    case IOCTL_RING_TYPE_TOE_WRBQ:
        ioctl.toe_wrbq = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, TOE_WRBQ_DESTROY);
        ioctl.toe_wrbq->params.request.id = id;
        break;
    case IOCTL_RING_TYPE_TOE_DEFQ:
        ioctl.toe_defq = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, TOE_DEFQ_DESTROY);
        ioctl.toe_defq->params.request.id = id;
        break;
    case IOCTL_RING_TYPE_TOE_RQ:
        ioctl.toe_rq = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, TOE_RQ_DESTROY);
        ioctl.toe_rq->params.request.id = id;
        break;    
#ifdef FUNCTION_RDMA
    case IOCTL_RING_TYPE_QP:
        ioctl.rdma_qp = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, RDMA_QP_DESTROY);
        ioctl.rdma_qp->params.request.id = id;
        break;
#endif
#if 0
    case IOCTL_RING_TYPE_FCOE_WRBQ:
        ioctl.fcoe_wrbq = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_FCOE_WRBQ_DESTROY);
        ioctl.fcoe_wrbq->params.request.id = id;
        break;
#endif
    case IOCTL_RING_TYPE_LRO:
        ioctl.lro = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LRO_RING_DESTROY);
        ioctl.lro->params.request.id = id;
        break;
    //</no_inbox>

    default:
        status = BE_NOT_OK;
        TRACE (DL_ERR, "Ring destroy, unsupported ring type:%d", ring_type);
        goto Error;
    }

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 internal_callback,
                 internal_callback_context,
                 &ioctl);
    if (status != BE_SUCCESS &&
        status != BE_PENDING) {
        // TRACE (DL_ERR, "Ring destroy ioctl failed. id:%d ring_type:%d", id, ring_type);
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS
be_function_ring_destroy (PBE_FUNCTION_OBJECT function_object, u32 id, u32 ring_type) {
    return be_function_ring_destroy_async (function_object, id, ring_type, NULL, NULL, NULL, NULL, NULL);
}

//<no_inbox>
BESTATUS
be_function_post_page_table_ioctl (
    PBE_FUNCTION_OBJECT function_object,
    PSA_SGL sgl,   // optional, required if num_pages > 0
    u32 num_pages, // num_pages == 0 used internally to query number of free page tables
    u32 *num_free) // returned number of free page tables on the chip
{
    IOCTL_COMMON_ADD_PAGE_TABLES *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PVOID va = NULL;
    u32 i = 0;

    ASSERT (num_free);

    be_lock_wrb_post (function_object);

    // Program hardware with these page tables.
    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in add page tables.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ADD_PAGE_TABLES);

    if (num_pages > 0) {
        va = sa_sgl_get_base_va (sgl);
        ASSERT (va);

        ioctl->params.request.va.lo = sa_pvoid_lo (va);        
        ioctl->params.request.va.hi = sa_pvoid_hi (va);                

        for (i=0; i<num_pages; i++) {
            u64 pa = sa_sgl_get_page (sgl, i);        
            ioctl->params.request.page_tables[i].hi = sa_hi (pa);
            ioctl->params.request.page_tables[i].lo = sa_lo (pa);
        }
    } 

    ioctl->params.request.num = num_pages;

    // Track the count for cleanup purposes
    ASSERT (function_object->config.num_page_table + num_pages < 256);
    function_object->config.num_page_table += num_pages;

    // This IOCTL response is copied into the num_page_table_free variable
    // for internal use in the high level API.
    status = be_function_post_mcc_wrb_with_copy  (
                function_object,
                wrb,
                NULL, // queue context
                NULL,
                NULL,
                ioctl,
                BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_ADD_PAGE_TABLES, 
                                             params.response.num_free, 
                                             num_free));
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "Failed to add page tables to chip.");
        goto error;
    }

error:
    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS
be_function_post_page_tables (
    PBE_FUNCTION_OBJECT function_object,
    PSA_SGL sgl,
    u32 num_pages)
{
    u32 num_free = 0;   
    ASSERT (sgl);
    ASSERT (num_pages > 0);
    return be_function_post_page_table_ioctl (function_object, sgl, num_pages, &num_free);
}


BESTATUS
be_function_query_num_free_page_tables (
    PBE_FUNCTION_OBJECT function_object,
    u32 *num_free)
{
    ASSERT (num_free);
    return be_function_post_page_table_ioctl (function_object, NULL, 0, num_free);
}



BESTATUS
be_function_remove_page_tables (
    PBE_FUNCTION_OBJECT function_object,
    u32 num,
    u32 *actual_num,
    u64 *pa_list_optional)
{
    IOCTL_COMMON_REMOVE_PAGE_TABLES *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;

    be_lock_wrb_post (function_object);

    // Program hardware with these page tables.
    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in remove page tables.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_REMOVE_PAGE_TABLES);
    ioctl->params.request.num = num;  

    TRACE (DL_INFO, 
           "Remove page tables. function:%d num:%d", 
           be_function_get_function_number (function_object), 
           num);

    // This IOCTL response is copied into the num_page_table_free variable
    // for internal use in the high level API.
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "Failed to remove page tables from chip.");
        goto error;
    }

    // Copy the resulting information from the WRB into user parameters.
    // This would require an internal callback to implement an async version -- the simple
    // copy is not sufficient.  For now, assume a synchronous call is okay.

    // Assign the actual number freed.
    if (actual_num) {
        *actual_num = ioctl->params.response.actual_num;
    }

    // Copy the PA list freed.
    if (pa_list_optional) {
        u32 i = 0;
        for (i=0; i<ioctl->params.response.actual_num; i++) {
            pa_list_optional[i] = sa_make_u64 (ioctl->params.response.page_tables[i].hi,
                                               ioctl->params.response.page_tables[i].lo);
        }
    }        

    // Track the count for cleanup purposes
    ASSERT (function_object->config.num_page_table >= ioctl->params.response.actual_num);
    function_object->config.num_page_table -= ioctl->params.response.actual_num;

error:
    be_unlock_wrb_post (function_object);
    return status;
}


BESTATUS
be_function_jell_config (
    PBE_FUNCTION_OBJECT function_object,
    boolean free_jell,
    PSA_SGL sgl,
    u32 num_pages)
{
    IOCTL_COMMON_JELL_CONFIG *ioctl = NULL;
    MCC_WRB *wrb = NULL;   
    u32 page_offset = 0;
    u32 i;
    BESTATUS status = 0;

    be_lock_wrb_post (function_object);

    // Post numerous ioctls to keep each one small enuf to be embedded in 
    // the WRB.
    do {

        wrb = be_function_peek_mcc_wrb (function_object);
        if (!wrb) {
            TRACE (DL_ERR, "MCC wrb peek failed.");
            status = BE_STATUS_NO_MCC_WRB;
            goto Error;
        }

        // Prepares an embedded ioctl, including request/response sizes.
        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_JELL_CONFIG);

        ioctl->params.request.free_jell = free_jell;
        if (free_jell) {
            
            TRACE (DL_INFO, "Free JELL config ioctl.");
            num_pages = 0; // exit loop            
            function_object->config.num_jell = 0; // Track number of pages posted.

        } else {

            ioctl->params.request.ddr = 0;        
            ioctl->params.request.num_pages = 
                MIN (num_pages, 
                     SA_NUMBER_OF_FIELD (IOCTL_COMMON_JELL_CONFIG, params.request.pages));

            TRACE (DL_INFO, "Create JELL config ioctl. num_pages:%d page_offset:%d",
                ioctl->params.request.num_pages, page_offset);       
            ioctl->params.request.page_offset = page_offset;

            // Track number of pages posted.
            function_object->config.num_jell += ioctl->params.request.num_pages; 

            for (i=0; i<ioctl->params.request.num_pages; i++) {
                u64 pa = sa_sgl_get_page (sgl, page_offset + i);
                ioctl->params.request.pages[i].lo = sa_lo (pa);
                ioctl->params.request.pages[i].hi = sa_hi (pa);            
            }

            // Add this group to the offset.  Subtract from remaining pages to map.
            page_offset += ioctl->params.request.num_pages;
            num_pages -= ioctl->params.request.num_pages;

            ioctl->params.request.done = (num_pages == 0); // this is final posting.
        }

        // Post the Ioctl
        status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
        if (status != BE_SUCCESS) {
            TRACE (DL_ERR, "ioctl to config JELL failed.");
            goto Error;
        }

    } while (num_pages > 0);   

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


BESTATUS
be_function_post_jell (
    PBE_FUNCTION_OBJECT function_object,
    PSA_SGL sgl,
    u32 num_pages)
{
    BESTATUS status = be_function_jell_config (function_object, FALSE, sgl, num_pages);
    if (status != BE_SUCCESS) {
        // Remove any pages that were successfully posted.
        be_function_remove_jell (function_object);
    }

    return status;
}

BESTATUS
be_function_remove_jell (
    PBE_FUNCTION_OBJECT function_object)
{
    return be_function_jell_config (function_object, TRUE, NULL, 0); // remove it
}
//</no_inbox>

BESTATUS
be_function_zero_buffer (
    PBE_FUNCTION_OBJECT function_object,
    u64 physical_address,
    boolean delete_buffer)
{    
    IOCTL_COMMON_POST_ZERO_BUFFER *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;

    FUNCTION_ASSERT (function_object);
    ASSERT (physical_address);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_POST_ZERO_BUFFER);

    ioctl->params.request.page.lo = sa_lo (physical_address);
    ioctl->params.request.page.hi = sa_hi (physical_address);
    ioctl->params.request.remove = delete_buffer;

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl);
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to post/remove 0 buffer failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


BESTATUS
be_function_post_zero_buffer (
    PBE_FUNCTION_OBJECT function_object,
    u64 pa)
{
    TRACE (DL_INFO, "post zero buffer. pa:0x%x_%08x", sa_hi (pa), sa_lo (pa));
    function_object->config.num_zero = 1;
    return be_function_zero_buffer (function_object, pa, FALSE);
}


BESTATUS
be_function_remove_zero_buffer (
    PBE_FUNCTION_OBJECT function_object)
{
    TRACE (DL_INFO, "remove zero buffer.");
    function_object->config.num_zero = 0;
    return be_function_zero_buffer (function_object, 0xbad0bad0bad0bad0ULL, TRUE);
}


void
be_sgl_to_pa_list (
    PSA_SGL sgl,
    PHYS_ADDR* pa_list,
    u32 max_num)
{
    u32 offset = sa_sgl_get_offset (sgl);
    u32 num_pages = sa_sgl_get_page_count (sgl);
    u32 i = 0;
    u64 pa;

    ASSERT (pa_list);
    ASSERT ((offset == 0) || (num_pages==1)); // MUST BE page aligned if >1 page in ring

    for (i=0; i<MIN(num_pages, max_num); i++) {
        pa = sa_sgl_get_page (sgl, i) + offset;
        offset = 0; // offset applies to first page only

        ASSERT (pa);

        pa_list[i].lo = sa_lo (pa);
        pa_list[i].hi = sa_hi (pa);

    }
}

//BE_LINK_STATUS_CONTEXT*
//be_function_allocate_link_context (
//    PBE_FUNCTION_OBJECT function_object)
//{
//    u32 final_count = 0;
//    FUNCTION_ASSERT (function_object);
//        
//    final_count = sa_atomic_increment (&function_object->link_status_context_allocated);
//    if (final_count == 1) {
//        return &function_object->link_status_context;
//    }
//    
//    // Context is in use.
//    sa_atomic_decrement (&function_object->link_status_context_allocated);
//    return NULL;    
//}
//
//void
//be_function_free_link_context (
//    PBE_FUNCTION_OBJECT function_object,
//    BE_LINK_STATUS_CONTEXT* context)
//{
//    FUNCTION_ASSERT (function_object);
//    ASSERT (context == &function_object->link_status_context);
//    ASSERT (function_object->link_status_context_allocated > 0);
//    sa_atomic_decrement (&function_object->link_status_context_allocated);
//}


/*!

@brief
    This routine adds an iSCSI connectoin object to a function object to allow
    for correct tracking and reference counting. 

@param    
    function_object             - Function object to add the object to.
    
@param    
    ConnectionObject    - iSCSI Connection object to add.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_add_iscsi_connection (
    IN      PBE_FUNCTION_OBJECT         function_object, 
    IN      PBE_ISCSI_CONNECTION_OBJECT connection_object)
{
    FUNCTION_ASSERT (function_object);
    
    ASSERT (be_function_is_iscsi (function_object)== TRUE);

    ISCSI_CXN_ASSERT (connection_object);

    _be_function_lock (function_object);

    sa_atomic_increment (&function_object->ref_count);
    sa_insert_tail_list (&function_object->links.cxn_list_head, 
                         &connection_object->connection_list);

    _be_function_unlock (function_object);
}

void
_be_function_add_fcoe_interface (
  IN    PBE_FUNCTION_OBJECT   function_object,
  IN    PBE_FCOE_INTERFACE_OBJECT interface_object)
{
    FUNCTION_ASSERT(function_object);
    ASSERT(be_function_is_fcoe(function_object) == TRUE);
    FCOE_INTERFACE_ASSERT(interface_object);
    _be_function_lock(function_object);
   sa_atomic_increment (&function_object->ref_count);
   /* TBD
    * fcoe interface is consuming one of the ring id.
    * it's been added to cxn lists for that function
    * do we need to maintain inteface list or fcoe interface
    * can be part of connection list
    */
   sa_insert_tail_list (&function_object->links.cxn_list_head,
                        &interface_object->interface_list);
   _be_function_unlock(function_object);
}


/*!

@brief
    This routine removes an iSCSI Connection object from a function object and
    drops the reference count for the given function object. 

@param    
    function_object         - Function object to remove the object from.
    
@param    
    ConnectionObject - iSCSI connection object to remove.    

@return   

@note
    IRQL < DISPATCH_LEVEL    

*/
void
_be_function_remove_iscsi_connection (
    IN      PBE_FUNCTION_OBJECT         function_object, 
    IN      PBE_ISCSI_CONNECTION_OBJECT connection_object)
{
    FUNCTION_ASSERT (function_object);
    
    ASSERT (be_function_is_iscsi (function_object)== TRUE);

    ISCSI_CXN_ASSERT (connection_object);    

    _be_function_lock (function_object);

    sa_remove_entry_list (&connection_object->connection_list);
    sa_atomic_decrement (&function_object->ref_count);

    _be_function_unlock (function_object);

}

void
_be_function_remove_fcoe_interface (
   IN   PBE_FUNCTION_OBJECT    function_object,
   IN   PBE_FCOE_INTERFACE_OBJECT interface_object)
{
    FUNCTION_ASSERT(function_object);

    ASSERT (be_function_is_fcoe (function_object) == TRUE);
    FCOE_INTERFACE_ASSERT(interface_object);

    _be_function_lock(function_object);

    sa_remove_entry_list (&interface_object->interface_list);
    sa_atomic_decrement(&function_object->ref_count);

    _be_function_unlock(function_object);
}

void
be_function_lock_mcc (
   IN PBE_FUNCTION_OBJECT function_object
)
{
    be_lock_wrb_post (function_object);
}

void
be_function_unlock_mcc (
   IN PBE_FUNCTION_OBJECT function_object
)
{
    be_unlock_wrb_post (function_object);
}

//---------------------------------------------------------------------------------------
// Function: be_function_enable_interrupts
//
//   Enables interrupts for the given PCI function. 
//
// function_object    - 
// return return_type - void
//---------------------------------------------------------------------------------------
void
be_function_enable_interrupts (
   IN PBE_FUNCTION_OBJECT function_object
)
{
    BE_EQ_OBJECT *eq = NULL;
    CQ_DB cq_db;    
    PCICFG_MEMBAR_CTRL_INT_CTRL_CSR ctrl;   
    u32 isr;

    FUNCTION_ASSERT (function_object);

    // This function should not be used for LANCER
    ASSERT (function_object->adapter_family != ADAPTER_FAMILY_LANCER);

/*    ctrl.dw = PCICFG_READ (function_object, membar_ctrl_intr_ctrl);*/
    BeSetDword( &ctrl, PCICFG_READ (function_object, membar_ctrl_intr_ctrl));
    
    if (!ctrl.hostintr) {

        //
        // Read-clear the ISR to make sure that any EQs that fired after the
        // host interrupt was masked are no longer asserted.  If the EQs still have a
        // producer-consumer mismatch, they will re-assert after they are armed.  This
        // handles the situation where software polls CQs and EQs with interrupts
        // disabled.
        //
        if (BeCsrBarAvailable (function_object->sa_dev))
        {
            switch (function_object->pci_function_number) {
                case 0:
                    isr = CSR_READ (function_object, cev.isr0);
                    break;
                case 1:
                    isr = CSR_READ (function_object, cev.isr1);
                    break;
                case 2:
                    isr = CSR_READ (function_object, cev.isr2);
                    break;
                case 3:
                    isr = CSR_READ (function_object, cev.isr3);
                    break;
                case 4:
                    isr = CSR_READ (function_object, cev.isr4);
                    break;
                case 5:
                    isr = CSR_READ (function_object, cev.isr5);
                    break;
                case 6:
                    isr = CSR_READ (function_object, cev.isr6);
                    break;
                case 7:
                    isr = CSR_READ (function_object, cev.isr7);
                    break;
                default:
                    ASSERT(function_object->pci_function_number < 8);
                    TRACE (DL_ERR, "Invalid function number.");
            }
        }
 
        //
        // Read-Modify-Write the control register
        //    
        ctrl.hostintr = 1;        
/*        PCICFG1_WRITE_CONST (function_object, membar_ctrl_intr_ctrl, ctrl.dw);        */
        PCICFG1_WRITE_CONST (function_object, membar_ctrl_intr_ctrl, (*(u32 *)&ctrl));
    }

    // clear entire doorbell
/*    cq_db.dw = 0; */
    BeSetDword( &cq_db, 0);
    
    cq_db.event = 1;
    cq_db.rearm = 1;
    cq_db.num_popped = 0;

    // Lock the function object while walking the EQ list.
    _be_function_lock (function_object);

    // Rearm each EQ.
    SA_FOR_EACH_LIST_ENTRY (eq, function_object->links.eq_list_head, BE_EQ_OBJECT, eq_list) { 
    
        EQ_ASSERT (eq);

        cq_db.qid = eq->eq_id;        
        PD_WRITE (function_object, cq_db, cq_db);
    }

    _be_function_unlock (function_object);
}

//---------------------------------------------------------------------------------------
// Function: be_function_disable_interrupts
//
//   Disables interrupts for the given PCI function. 
//
// function_object    - 
// return return_type - void
//---------------------------------------------------------------------------------------
void
be_function_disable_interrupts (
   IN PBE_FUNCTION_OBJECT function_object
)
{    
    PCICFG_MEMBAR_CTRL_INT_CTRL_CSR ctrl;

    ctrl.dw = 0;

    FUNCTION_ASSERT (function_object);

    // This function should not be used for LANCER
    ASSERT (function_object->adapter_family != ADAPTER_FAMILY_LANCER);

    //
    // Read-Modify-Write of the control register in PCI config space.
    //
    // BeSetDword( &ctrl, PCICFG_READ (function_object, membar_ctrl_intr_ctrl));
    
    ctrl.dw = PCICFG_READ (function_object, membar_ctrl_intr_ctrl);

    // Always disable interrupts, regardless of hostintr value
    ctrl.hostintr = 0;

    // Always clear membarwinen
    ctrl.membarwinen = 0;

    /*        PCICFG1_WRITE_CONST (function_object, membar_ctrl_intr_ctrl, ctrl.dw);        */
    PCICFG1_WRITE_CONST (function_object, membar_ctrl_intr_ctrl, (*(u32 *)&ctrl));
    //} else {
    //    BREAKPOINT ();
    //}

/*    ctrl.dw =  PCICFG_READ (function_object, membar_ctrl_intr_ctrl); */
    BeSetDword( &ctrl, PCICFG_READ (function_object, membar_ctrl_intr_ctrl));
    
    if (ctrl.hostintr) {
        TRACE(DL_ERR, "ERROR BE2 INTERRUPT DISABLE FAILED\n");
    }
}

//---------------------------------------------------------------------------------------
// Function: be_function_nop
//
//   Issues a NOP command to the MCC ring. The command is completed with a successful
//   status. This can be used to pend operations until a short time in the future, since
//   the callback function will be invoked upon command completion. The returned status
//   will be BE_PENDING if the command was issued successfully. 
//
// function_object    - 
// callback           - Callback function invoked when the NOP completes. 
// callback_context   - Passed to the callback function. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_nop (
    IN PBE_FUNCTION_OBJECT   function_object,
    IN MCC_WRB_CQE_CALLBACK  callback,
    IN PVOID                 callback_context,
    IN PBE_NOP_QUEUE_CONTEXT queue_context)
{
    BESTATUS  status = BE_SUCCESS;  
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_NOP *ioctl = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    FUNCTION_ASSERT (function_object);    

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_NOP);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl);

Error:
    be_unlock_wrb_post (function_object);

    return status;    
}

//---------------------------------------------------------------------------------------
// Function: be_function_get_fw_version
//
//   Retrieves the firmware version on the adpater. If the callback is NULL this call
//   executes synchronously. If the callback is not NULL, the returned status will be
//   BE_PENDING if the command was issued successfully. 
//
// function_object    - 
// fw_version         - Pointer to response buffer if callback is NULL. 
// callback           - Callback function invoked when the IOCTL completes. 
// callback_context   - Passed to the callback function. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_get_fw_version (
   IN PBE_FUNCTION_OBJECT                            function_object,
   OUT IOCTL_COMMON_GET_FW_VERSION_RESPONSE_PAYLOAD* fw_version,
   IN MCC_WRB_CQE_CALLBACK                           callback,
   IN PVOID                                          callback_context
)
{
    BESTATUS  status = BE_SUCCESS;  
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_GET_FW_VERSION *ioctl = NULL;

    FUNCTION_ASSERT (function_object);    

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    if (!callback && !fw_version) {
        TRACE (DL_ERR, "callback and response buffer NULL!");
        status = BE_NOT_OK;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_FW_VERSION);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 NULL, // queue context
                 callback,
                 callback_context,
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_GET_FW_VERSION, params.response, fw_version));

Error:
    be_unlock_wrb_post (function_object);

    return status;    
}


BESTATUS
be_function_queue_mcc_wrb (
    PBE_FUNCTION_OBJECT       function_object,
    PBE_GENERIC_QUEUE_CONTEXT queue_context)
{
    BESTATUS status;

    FUNCTION_ASSERT (function_object);
    ASSERT (queue_context);

    //
    // issue the WRB to the MPU as appropriate
    //
    if (function_object->links.mcc) {

        // We're in ring mode.  Queue this item.
        function_object->links.mcc->backlog_length++;
        sa_insert_tail_list (&function_object->links.mcc->backlog, &queue_context->context.list);

        function_object->stats.queued_wrbs++;
        function_object->stats.queue_length = function_object->links.mcc->backlog_length;
        function_object->stats.max_queue_length = MAX (function_object->stats.queue_length, function_object->stats.max_queue_length);

        status = BE_PENDING;
        
    } else {
        
        status = BE_NOT_OK;

    }

    return status;
}


//---------------------------------------------------------------------------------------
// Function: be_function_configure_arm_prints
//
//   This routine can be used to change the debug print level of the BladeEngine
//   management processor. This is only applicable in debug release builds. 
//
// function_object - 
// debug_level     - Debug print level. The lower the value the more debug info will be
//                   printed.
// idle_print      - Set to print debug info in the management firmware idle periods. 
// old_debug_level - Previous debug print level. 
// old_idle_print  - Previous idle print setting. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_configure_arm_prints (
   IN PBE_FUNCTION_OBJECT function_object,
   IN u32                 debug_level,
   IN boolean             idle_print,
   OUT u32*               old_debug_level,
   OUT boolean*           old_idle_print)
{
    IOCTL_LOWLEVEL_MGMT_DEBUG_LEVEL *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LOWLEVEL_MGMT_DEBUG_LEVEL);

    // Set parameters
    ioctl->params.request.print_level = debug_level;
    
    // TODO : handle additional parameters. For now, setting store_level the same as print_level
    // and defaulting the prints/stores as disabled.
    ioctl->params.request.store_level = debug_level;
    ioctl->params.request.enable_disable_print = 0;
    ioctl->params.request.enable_disable_store = 0;

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 NULL, // callback
                 NULL, // context 
                 ioctl);

    if (status != 0) {
        TRACE (DL_ERR, "config armfw debug prints ioctl failed.");
        goto error;
    }

    if (NULL != old_debug_level) {
        *old_debug_level = ioctl->params.response.old_print_level;
    }

error:
    be_unlock_wrb_post (function_object);

    return status;
}



//---------------------------------------------------------------------------------------
// Function: be_function_manage_FAT_log
//
//   This routine can be used to manage the BladeEngine Fault Analysis Tool (FAT) log.
//   This includes querying the FAT log size, retrieving the FAT log, and clearing the FAT
//   log. The log data can be anaylzed by BladeEngine support tools to diagnose faults.
//   Only host domains (domain 0) may issue this request. 
//
// function_object   - 
// sgl               - The SGL representing the FAT log buffer landing space. The SGL
//                     should be page aligned. It does not require a virtual address. 
// num_pages         - The number of pages in the SGL. A value of zero (0) implies no FAT
//                     log data transfer will take place. The num_pages is limited to 27
//                     pages. 
// page_offset       - The page_offset specifies the starting page offset when retrieving
//                     the FAT log. For example, a value of 0 requests data starting at
//                     byte offset 0 of the FAT log. Likewise, a value of 5 requests data
//                     starting at byte offset 20480 of the FAT log. A caller may choose
//                     to issue multiple queries when the FAT log buffer size is larger
//                     than the number of pages specified. 
// clear_log         - Set to clear the log. 
// log_size          - The size of the BladeEngine FAT log in bytes. 
// bytes_transferred - The number of FAT log data bytes transferred. 
// return status     - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_manage_FAT_log (
    IN  PBE_FUNCTION_OBJECT          function_object,
    IN  PSA_SGL                      sgl,
    IN  u32                          num_pages,
    IN  u32                          page_offset,
    IN  boolean                      clear_log,
    OUT FAT_RESPONSE*                output,
    IN  MCC_WRB_CQE_CALLBACK         callback,
    IN  PVOID                        callback_context,
    IN  PBE_MANAGE_FAT_QUEUE_CONTEXT queue_context)
{
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    IOCTL_COMMON_GET_FAT *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_FAT);

    // Set parameters
    ioctl->params.request.clear_log = clear_log;
    ioctl->params.request.page_offset = page_offset;
    ioctl->params.request.num_pages = num_pages;

    // Create a page list for the IOCTL.
    ASSERT (sa_sgl_get_page_count (sgl) >= ioctl->params.request.num_pages); 
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.buffer_addr, 
                       SA_NUMBER_OF (ioctl->params.request.buffer_addr));
    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 generic_context,  // 
                 callback,         // callback
                 callback_context, // context 
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_GET_FAT, 
                                              params.response, 
                                              output));    
error:
    be_unlock_wrb_post (function_object);

    return status;
}

//<no_inbox>
BESTATUS
be_function_measure_dma_bandwidth (
   IN PBE_FUNCTION_OBJECT                       function_object,
   IN OUT IOCTL_LOWLEVEL_MEASURE_DMA_BANDWIDTH* ioctl_input,
   IN MCC_WRB_CQE_CALLBACK                      callback,
   IN PVOID                                     callback_context
)
{
    IOCTL_LOWLEVEL_MEASURE_DMA_BANDWIDTH *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LOWLEVEL_MEASURE_DMA_BANDWIDTH);

    // Set parameters
    ioctl->params = ioctl_input->params;
   // sa_memcpy (ioctl, ioctl_input, sizeof (*ioctl_input));

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 NULL,
                 callback, // callback
                 callback_context, // context 
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_LOWLEVEL_MEASURE_DMA_BANDWIDTH, 
                                              params.response, 
                                              &ioctl_input->params.response));
   
error:
    be_unlock_wrb_post (function_object);

    return status;
}
//</no_inbox>

BESTATUS
be_function_reset(
   IN PBE_FUNCTION_OBJECT  function_object)
{
    IOCTL_COMMON_FUNCTION_RESET *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = be_function_prepare_embedded_ioctl(
        function_object, wrb, sizeof(IOCTL_COMMON_FUNCTION_RESET),
        0, 0, BE_OPCODE_NAME(COMMON_FUNCTION_RESET), 
        BE_SUBSYSTEM_NAME(COMMON_FUNCTION_RESET));

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 NULL,
                 NULL, // callback
                 NULL, // context 
                 ioctl,
                 be_create_mcc_response_copy(0, sizeof(MCC_WRB_PAYLOAD), wrb));
                 
   
error:
    be_unlock_wrb_post (function_object);

    return status;
}

BESTATUS
be_function_reset_ex (
   IN PBE_FUNCTION_OBJECT  function_object,
   u32 domain,
   IN  MCC_WRB_CQE_CALLBACK         callback,
   IN  PVOID                        context,
   IN  PBE_GENERIC_QUEUE_CONTEXT    queue_context)
{
    IOCTL_COMMON_FUNCTION_RESET *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = be_function_prepare_embedded_ioctl(
        function_object, wrb, sizeof(IOCTL_COMMON_FUNCTION_RESET),
        0, 0, BE_OPCODE_NAME(COMMON_FUNCTION_RESET), 
        BE_SUBSYSTEM_NAME(COMMON_FUNCTION_RESET));

    ioctl->header.request.domain = domain;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 generic_context, // queue context
                 callback, // callback
                 context, // context 
                 ioctl,
                 be_create_mcc_response_copy(0, 0, NULL));                 
   
Error:
    be_unlock_wrb_post (function_object);

    return status;
}

//---------------------------------------------------------------------------------------
// Function: be_function_query_fw_config
//
//   Queries the firmware configuration currently loaded. This configuration includes
//   information regarding the EP processor configuration for various Upper Layer
//   Protocols. 
//
// function_object - 
// config          - The configuration parameters currently loaded by firmware. 
// return status   - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_query_fw_config (
   IN PBE_FUNCTION_OBJECT  function_object,
   OUT BE_FW_CFG* config
)
{
    FUNCTION_ASSERT (function_object);
    ASSERT (config);

    // Copy the cached version.
    sa_memcpy (config, &function_object->fw_cfg, sizeof (*config));

    return BE_SUCCESS;
}


BESTATUS
be_function_internal_query_fw_cfg (
   IN PBE_FUNCTION_OBJECT  function_object,
   OUT BE_FW_CFG* config
)
{
    IOCTL_COMMON_QUERY_FIRMWARE_CONFIG *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_QUERY_FIRMWARE_CONFIG);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 NULL,
                 NULL, // callback
                 NULL, // context 
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_QUERY_FIRMWARE_CONFIG, 
                                              params.response, 
                                              config));
   
error:
    be_unlock_wrb_post (function_object);

    return status;
}


BESTATUS
be_function_internal_driver_capabilities (
   IN PBE_FUNCTION_OBJECT  function_object,
   IN OUT SET_DRIVER_FUNCTION_CAPABILITIES_PARAMS_V1* caps
)
{
    IOCTL_COMMON_SET_DRIVER_FUNCTION_CAPABILITIES_V1 *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_DRIVER_FUNCTION_CAPABILITIES_V1);

    ioctl->params.request = caps->request;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 NULL,
                 NULL, // callback
                 NULL, // context 
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_SET_DRIVER_FUNCTION_CAPABILITIES_V1, 
                                              params.response, 
                                              caps));
   
error:
    be_unlock_wrb_post (function_object);

    return status;
}

//---------------------------------------------------------------------------------------
// Function: be_function_config_red
//
//   This function configures global and/or ULP specific Random Early Drop (RED)
//   functionality. 
//
// function_object    - 
// parameters         - The RED parameters. Only parameters for the chutes owned by the
//                      current PCI function are applied.
// callback           - Optional callback function.
// callback_context   - Optional context for callback function.
// queue_context      - Optional context for queueing the IOCTL.
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_config_red (
   IN PBE_FUNCTION_OBJECT          function_object,
   IN PBE_RED_PARAMETERS           parameters,
   IN MCC_WRB_CQE_CALLBACK         callback,
   IN PVOID                        callback_context,
   IN PBE_CONFIG_RED_QUEUE_CONTEXT queue_context
)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    IOCTL_COMMON_RED_CONFIG *ioctl = NULL;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_RED_CONFIG);

    sa_memcpy (&ioctl->params.request,
               parameters,
               sizeof (*parameters));

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl);

error:
    be_unlock_wrb_post (function_object);

    return status;
}


#ifdef BE1
// MISSED port equalizatoin during branching. so defining it out for BE2

//---------------------------------------------------------------------------------------
// Function: be_function_config_port_equalization
//
//   This function configures or returns the XAUI port equalization parameters of
//   BladeEngine. 
//
// function_object    - 
// parameters         - XAUI port equalization parameters. 
// write              - Set (1) to write the parameters, otherwise clear (0) to read the
//                      parameters. 
// callback           - Optional callback function. 
// callback_context   - Optional context for callback function. 
// queue_context      - Optional context for queueing the IOCTL. 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_config_port_equalization (
   IN PBE_FUNCTION_OBJECT                        function_object,   
   IN PIOCTL_COMMON_GET_PORT_EQUALIZATION        parameters,
   IN boolean                                    write,   
   IN MCC_WRB_CQE_CALLBACK                       callback,
   IN PVOID                                      callback_context,
   IN PBE_CONFIG_PORT_EQUALIXATION_QUEUE_CONTEXT queue_context )
{
#ifdef BE1 //Inorder to get rid of compile errors masking the functionality
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
 
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto error;
        }
    }

    if (write) {
        IOCTL_COMMON_SET_PORT_EQUALIZATION *ioctl;
        
        // Prepare an embedded ioctl, including request/response sizes.
        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_PORT_EQUALIZATION);

        sa_memcpy (&ioctl->params.request,
                   parameters,
                   sizeof (*parameters));

        // Post the Ioctl
        status = be_function_post_mcc_wrb_with_queue_context (
                    function_object,
                    wrb,
                    generic_context,
                    callback,
                    callback_context,
                    ioctl);
    } else {
        IOCTL_COMMON_GET_PORT_EQUALIZATION *ioctl;

        // Prepare an embedded ioctl, including request/response sizes.
        ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_PORT_EQUALIZATION);

        // This IOCTL response is copied into the parameters pointer
        status = be_function_post_mcc_wrb_with_copy  (
                    function_object,
                    wrb,
                    generic_context, // queue context
                    callback,
                    callback_context,
                    ioctl,
                    BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_GET_PORT_EQUALIZATION, 
                                                 params.response, 
                                                 parameters));
    }

error:
    be_unlock_wrb_post (function_object);
    return status;
#endif

#ifndef BE1
    return BE_SUCCESS; 
#endif

}
#endif

//---------------------------------------------------------------------------------------
// Function: be_function_passthru_ioctl
//
//   This routine issues an embedded IOCTL in pass-through mode. 
//
// function_object  - 
// payload          - The embedded payload for the MCC_WRB structure. The input buffer is
//                    overwritten with response data. 
// callback         - Callback function invoked when the IOCTL completes. 
// callback_context - Callback context passed to the callback function. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_passthru_ioctl (
   IN PBE_FUNCTION_OBJECT  function_object,
   IN OUT MCC_WRB_PAYLOAD* payload,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PVOID ioctl = NULL;
    PIOCTL_REQUEST_HEADER ioctl_header = (PIOCTL_REQUEST_HEADER)payload;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl
    ioctl = be_function_prepare_embedded_ioctl(function_object,
                                               wrb,
                                               sizeof(MCC_WRB_PAYLOAD),
                                               ioctl_header->request_length,
                                               ioctl_header->request_length,
                                               ioctl_header->opcode,
                                               ioctl_header->subsystem);

    // Copy in the user's IOCTL payload
    sa_memcpy((void *)&wrb->payload, 
              (void *)payload,
              sizeof(MCC_WRB_PAYLOAD));

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 NULL,
                 callback, // callback
                 callback_context, // context 
                 ioctl,
                 be_create_mcc_response_copy(0, sizeof(MCC_WRB_PAYLOAD), payload));
  
error:
    be_unlock_wrb_post (function_object);

    return status;
}

BESTATUS
be_function_passthru_nonembedded_ioctl_with_queue_context (
   IN PBE_FUNCTION_OBJECT  function_object,
   IN IOCTL_HEADER*        ioctl_va,
   IN u64                  ioctl_pa,
   IN u32                  ioctl_len,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context,
   IN PBE_GET_TEMPERATURE_QUEUE_CONTEXT queue_context)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    IOCTL_HEADER *ioctl = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    u32 version;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            TRACE (DL_ERR, "MCC wrb peek failed.");
            status = BE_STATUS_NO_MCC_WRB;
            goto error;
        }        
    }

    ASSERT (ioctl_va->request.request_length + sizeof (IOCTL_REQUEST_HEADER) <= ioctl_len);

    version = ioctl_va->request.version;

    // Prepares an embedded ioctl
    ioctl = (IOCTL_HEADER*)be_function_prepare_nonembedded_ioctl(function_object,
                                                  wrb,
                                                  ioctl_va,
                                                  ioctl_pa,
                                                  ioctl_len,
                                                  ioctl_va->request.request_length,
                                                  ioctl_va->request.request_length,
                                                  ioctl_va->request.opcode,
                                                  ioctl_va->request.subsystem,
                                                  ioctl_va->request.version);


    // Fix the version.
    ioctl->request.version = version;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback, // callback
                 callback_context, // context 
                 ioctl);    
  
error:

    be_unlock_wrb_post (function_object);

    return status;
}

//---------------------------------------------------------------------------------------
// Function: be_function_passthru_nonembedded_ioctl
//
//   This routine issues an non-embedded IOCTL in pass-through mode. 
//
// function_object  - 
// payload          - The embedded payload for the MCC_WRB structure. The input buffer is
//                    overwritten with response data. 
// callback         - Callback function invoked when the IOCTL completes. 
// callback_context - Callback context passed to the callback function. 
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_passthru_nonembedded_ioctl (
   IN PBE_FUNCTION_OBJECT  function_object,
   IN IOCTL_HEADER*        ioctl_va,
   IN u64                  ioctl_pa,
   IN u32                  ioctl_len,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PVOID ioctl = NULL;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    ASSERT (ioctl_va->request.request_length + sizeof (IOCTL_REQUEST_HEADER) <= ioctl_len);

    // Prepares an embedded ioctl
    ioctl = be_function_prepare_nonembedded_ioctl(function_object,
                                                  wrb,
                                                  ioctl_va,
                                                  ioctl_pa,
                                                  ioctl_len,
                                                  ioctl_va->request.request_length,
                                                  ioctl_va->request.request_length,
                                                  ioctl_va->request.opcode,
                                                  ioctl_va->request.subsystem,
                                                  ioctl_va->request.version);

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 callback, // callback
                 callback_context, // context 
                 ioctl_va);
  
error:

    be_unlock_wrb_post (function_object);

    return status;
}

BESTATUS
be_get_max_link_speed(
    PBE_FUNCTION_OBJECT  function_object,
    u32 *pSpeed)
{
    IOCTL_COMMON_GET_PORT_CAPABILITY *ioctl=NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    
    ASSERT(pSpeed);
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.");
        status = BE_STATUS_NO_MCC_WRB;
        goto error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_PORT_CAPABILITY);

    // Post the Ioctl
    status = be_function_post_mcc_wrb (
                 function_object,
                 wrb,
                 NULL, // callback
                 NULL, // context 
                 ioctl);

    if (status != 0) {
        TRACE (DL_ERR, "get link speed ioctl failed.");
        goto error;
    }    
    
    if (pSpeed){
        *pSpeed = ioctl->params.response.port_speed;
    }
    
    

error:
    be_unlock_wrb_post (function_object);

    return status;

}

BESTATUS
be_function_toggle_id_led_blink_state (
    IN PBE_FUNCTION_OBJECT                      function_object,
    u8                                          port_num,
    u8                                          beacon_state,
    u8                                          beacon_duration,
    u8                                          status_duration,
    IN MCC_WRB_CQE_CALLBACK                     callback,
    IN PVOID                                    callback_context,
    IN PBE_ENABLE_DISABLE_BEACON_QUEUE_CONTEXT  queue_context)
{
    IOCTL_COMMON_ENABLE_DISABLE_BEACON *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_ENABLE_DISABLE_BEACON);

    ioctl->params.request.port_num = port_num;
    ioctl->params.request.beacon_state = beacon_state;
    ioctl->params.request.beacon_duration = beacon_duration;
    ioctl->params.request.status_duration = status_duration;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback, // callback
                 callback_context, // context 
                 ioctl);

    if ((status != 0) && (status != BE_PENDING)) {
        TRACE (DL_ERR, "enable disable beacon ioctl failed.");
        goto error;
    }    
    
error:
    be_unlock_wrb_post (function_object);

    return status;

}

BESTATUS
be_function_read_id_led_blink_state (
    IN PBE_FUNCTION_OBJECT                  function_object,
    u8                                      port_num,
    IN MCC_WRB_CQE_CALLBACK                 callback,
    IN PVOID                                callback_context,
    IN PBE_GET_BEACON_STATE_QUEUE_CONTEXT   queue_context,
    u8*                                     beacon_state)
{
    IOCTL_COMMON_GET_BEACON_STATE *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto error;
        }
    }
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_BEACON_STATE);

    ioctl->params.request.port_num = port_num;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                 function_object,
                 wrb,
                 generic_context,
                 callback, // callback
                 callback_context, // context 
                 ioctl,
                 BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_GET_BEACON_STATE, params.response.beacon_state, beacon_state));

    if ((status != 0) && (status != BE_PENDING)) {
        TRACE (DL_ERR, "get beacon state ioctl failed.");
        goto error;
    }    
    
error:
    be_unlock_wrb_post (function_object);

    return status;
}

BESTATUS
be_loopback_test(
    PBE_FUNCTION_OBJECT                     function_object,
    u8                                      port_num,    
    IN MCC_WRB_CQE_CALLBACK                 callback,
    IN PVOID                                callback_context,
    IN PBE_LOOPBACK_TEST_CONTEXT            queue_context)
{
    IOCTL_LOWLEVEL_LOOPBACK_TEST *ioctl=NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LOWLEVEL_LOOPBACK_TEST);

	ioctl->params.request.loopback_type = MAC_LOOPBACK;
	ioctl->params.request.number_packets = 1;
	ioctl->params.request.pattern = 1234;
	ioctl->params.request.src_port = port_num;
	ioctl->params.request.dest_port = port_num;
	ioctl->params.request.packet_size = 1514;
    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback, // callback
                 callback_context, // context 
                 ioctl);

	
	if ((status != 0) && (status != BE_PENDING)) {
        TRACE (DL_ERR, "loopback test ioctl failed.");
        goto error;
    }

	if (ioctl->params.response.status == LOOPBACK_DATA_MISMATCH ) {
		TRACE(DL_ERR,"miscompare offset: %d",ioctl->params.response.miscompare_offset);
	}
	if(ioctl->params.response.status == LOOPBACK_SUCCESS ) {
		TRACE(DL_ERR,"Number of packets transfered: %d",ioctl->params.response.number_transfered);
		TRACE(DL_ERR,"Number of packets received in loopback: %d",ioctl->params.response.number_received);
		TRACE(DL_ERR,"Number of ticks for transfering packets in loopback mode:%d",ioctl->params.response.num_ticks_completion);
	}

error:
    be_unlock_wrb_post (function_object);

    return status;

}
BESTATUS
be_set_or_exit_loopback_mode(
    IN PBE_FUNCTION_OBJECT                  function_object,
    u8                                      port_num,
    u8                                      set_loopback,
    IN MCC_WRB_CQE_CALLBACK                 callback,
    IN PVOID                                callback_context,
    IN PBE_SET_EXIT_LOOPBACK_CONTEXT        queue_context)
{
    IOCTL_LOWLEVEL_SET_LOOPBACK_MODE *ioctl=NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {

        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LOWLEVEL_SET_LOOPBACK_MODE);

    if (set_loopback) {
        ioctl->params.request.src_port = port_num;
        ioctl->params.request.dest_port = port_num;
        ioctl->params.request.loopback_mode = MAC_LOOPBACK;
        ioctl->params.request.loopback_state = LOOPBACK_STATE_ENABLE;
    } else {
        ioctl->params.request.src_port = port_num;
        ioctl->params.request.dest_port = port_num;
        ioctl->params.request.loopback_mode = NO_LOOPBACK;
        ioctl->params.request.loopback_state = LOOPBACK_STATE_DISABLE;
    }

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback, // callback
                 callback_context, // context 
                 ioctl);

    if ((status != 0) && (status != BE_PENDING)) {
        TRACE (DL_ERR, "set/exit loopback ioctl failed.");
        goto error;
    }   

error:
    be_unlock_wrb_post (function_object);

    return status;

}

#define DDR_ADDR_START 0x34000 

BESTATUS
be_get_armlog(
    IN PBE_FUNCTION_OBJECT                  function_object,
    IN IOCTL_LOWLEVEL_ACCESS_MEMORY        *va_for_ioctl,
    IN u64                                  pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK                 callback,
    IN PVOID                                callback_context,
    IN PBE_GENERIC_QUEUE_CONTEXT            queue_context,
    int                                     buffer_size)
{
    MCC_WRB *wrb=NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    BESTATUS status=0;
    PIOCTL_REQUEST_HEADER header = NULL;
    PIOCTL_LOWLEVEL_ACCESS_MEMORY pAccessMemoryIoctl = NULL;

    ASSERT(va_for_ioctl);
    ASSERT(pa_for_ioctl);

    header = (PIOCTL_REQUEST_HEADER)va_for_ioctl;
    pAccessMemoryIoctl  = (PIOCTL_LOWLEVEL_ACCESS_MEMORY)va_for_ioctl;

    be_lock_wrb_post(function_object);

    wrb = be_function_peek_mcc_wrb(function_object);

    if (!wrb) {
        if (queue_context && callback) {
            wrb = (MCC_WRB*)&queue_context->wrb_header;
            generic_context=(PBE_GENERIC_QUEUE_CONTEXT)queue_context;
            generic_context->context.bytes = sizeof(*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto Error;
        }
    }

    wrb->embedded = 0;
    wrb->payload_length = /*sizeof(IOCTL_LOWLEVEL_ACCESS_MEMORY)+*/buffer_size;

    // Assume one fragment. The caller may override the SGL by rewriting the 0th length
    // and adding more entries.  They will also need to update the sge_count.
    wrb->sge_count = 1;
    wrb->payload.sgl[0].length = wrb->payload_length;
    wrb->payload.sgl[0].pa_lo = sa_lo(pa_for_ioctl);
    wrb->payload.sgl[0].pa_hi = sa_hi(pa_for_ioctl);
    pAccessMemoryIoctl->params.request.address = DDR_ADDR_START; // 0x34000
    pAccessMemoryIoctl->params.request.length = buffer_size;
    pAccessMemoryIoctl->params.request.memory = MEMORY_DDR;
    pAccessMemoryIoctl->params.request.write = 0; // read the memory

    header->timeout         = 0; // todo -- be_function_get_default_ioctl_timeout (function_object);
    header->domain          = be_function_get_pd_number(function_object);
    header->request_length  = wrb->payload_length-sizeof(IOCTL_LOWLEVEL_ACCESS_MEMORY);
    header->opcode          = BE_OPCODE_NAME(LOWLEVEL_ACCESS_MEMORY);
    header->subsystem       = IOCTL_SUBSYSTEM_LOWLEVEL;

    status = be_function_post_mcc_wrb_with_queue_context(
        function_object,
        wrb,
        generic_context,
        callback,
        callback_context,
        va_for_ioctl);

Error:
    be_unlock_wrb_post(function_object);

    return status;
}


// This is called internally by beclib during a timeout/UE to avoid
// processing further ioctls.  The context may not be suitable for aborting
// pending ioctls, so the upper driver must still call be_function_abort_ioctls 
// when it is ready.
void
be_function_disable_ioctls (
    PBE_FUNCTION_OBJECT function_object)
{
    // be_lock_wrb_post (function_object);    
    function_object->ue = 1;
    // be_unlock_wrb_post_no_redrive (function_object);
}

//
// Reenable ioctl processing after UE recovery.  (May not be necessary...)
//
void
be_function_enable_ioctls (
    IN PBE_FUNCTION_OBJECT       function_object)
{
    be_lock_wrb_post (function_object);    

    function_object->ue = 0;
    function_object->aborted = 0;

    be_unlock_wrb_post_no_redrive (function_object);
}

//
// This should be called after the driver stops running be_mcc_process_cq
// to fail any pending WRBs with error status.  The ioctl callbacks will run
// from this calling context.
//
void
be_function_abort_ioctls (
    IN PBE_FUNCTION_OBJECT function_object)
{
    boolean abort = FALSE;

    be_lock_wrb_post (function_object);    
    
    // 
    // This will start failing new requests, and grabbing the wrb post
    // lock will nicely synchronize it.
    //
    function_object->ue = 1;

    //
    // This will cause beclib to abort all pending commands.
    //
    if (!function_object->aborted) {
        function_object->aborted = 1;
        abort = TRUE;
    }

    be_unlock_wrb_post_no_redrive (function_object);

    //
    // This will abort any pending requests.  The client driver must be
    // sure that be_mcc_process_cq is not running on another thread.
    //
    if (abort) {
        be_function_abort_in_progress_wrbs (function_object);
    }
}

//
// Check for a hardware UE.
//
BESTATUS
be_function_check_hardware_ue (
    BE_FUNCTION_OBJECT *function_object)
{
    BESTATUS status = BE_SUCCESS;

    if (function_object->adapter_family != ADAPTER_FAMILY_LANCER) {

        u32 ue0mask = BePciCfgRead (function_object->sa_dev, SA_FIELD_OFFSET( PCICFG_PF0_CSRMAP,  ue_status_low_mask), NULL);
        u32 ue1mask = BePciCfgRead (function_object->sa_dev, SA_FIELD_OFFSET( PCICFG_PF0_CSRMAP,  ue_status_hi_mask),  NULL);
        u32 ue0 = BePciCfgRead (function_object->sa_dev, SA_FIELD_OFFSET( PCICFG_PF0_CSRMAP,  ue_status_low), NULL);
        u32 ue1 = BePciCfgRead (function_object->sa_dev, SA_FIELD_OFFSET( PCICFG_PF0_CSRMAP,  ue_status_hi), NULL);

        if ((ue0 & ~ue0mask) || (ue1 & ~ue1mask)) {

            TRACE (DL_ERR, 
                "UE Detected. ue0:0x%x ue1:0x%x", 
                (ue0 & ~ue0mask), 
                (ue1 & ~ue1mask));

            status = BE_STATUS_UE_DETECTED;
        }

    } else {

        SLI_PORT_STATUS_CSR port_status = {{{0}}};
        u32 slip_err1 = 0;
        u32 slip_err2 = 0;

        port_status.dw = BeCsrRead(function_object->sa_dev, SLI_PORT_STATUS_REG, NULL);

        if (port_status.error) {

            slip_err1 = BeCsrRead(function_object->sa_dev, SLI_PORT_ERROR1_REG, NULL);
            slip_err2 = BeCsrRead(function_object->sa_dev, SLI_PORT_ERROR2_REG, NULL);

            TRACE (DL_ERR, "SLI Port Error Detected. Error0: 0x%x Error1: 0x%x",
                   slip_err1, slip_err2);

            status = BE_STATUS_UE_DETECTED;
        }
    }

    return status;
}

void
be_function_link_error_object (
    BE_FUNCTION_OBJECT *function_object,
    BE_ERROR_OBJECT *errobj)
{
    FUNCTION_ASSERT (function_object);
    function_object->links.errobj = errobj;
}

// This function gets the privilege mask
// For example, PRIV_FILTMGMT permits the setting of MAC/VLAN filters
//
BESTATUS
be_function_get_privileges (
    IN  PBE_FUNCTION_OBJECT             function_object,
    OUT UEXACT32*                       privilege_mask,
    IN  MCC_WRB_CQE_CALLBACK            callback,      // optional
    IN  PVOID                           callback_context,
    IN  PBE_FUNCTION_PRIVILEGES_CONTEXT queue_context) // optional
{
    IOCTL_COMMON_GET_FUNCTION_PRIVILEGES_V1 *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    ASSERT (privilege_mask);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);

    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_FUNCTION_PRIVILEGES_V1);

    // Post or queue the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                function_object,
                wrb,
                generic_context, // Queue context
                callback,
                callback_context,
                ioctl,
                BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_GET_FUNCTION_PRIVILEGES_V1, params.response, privilege_mask));

    if (status < 0) {
        TRACE (DL_ERR, "function privileges ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS
be_function_set_privileges (
    IN  PBE_FUNCTION_OBJECT             function_object,
    IN  u32                             domain,
    IN  u32                             privilege_mask,
    IN  MCC_WRB_CQE_CALLBACK            callback,      // optional
    IN  PVOID                           callback_context,
    IN  PBE_FUNCTION_PRIVILEGES_CONTEXT queue_context) // optional
{
    IOCTL_COMMON_SET_FUNCTION_PRIVILEGES_V1 *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);

    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }
    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_FUNCTION_PRIVILEGES_V1);

    ioctl->header.request.domain = domain;
    ioctl->params.request.privilege_mask = privilege_mask;

    // Post or queue the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                function_object,
                wrb,
                generic_context, // Queue context
                callback,
                callback_context,
                ioctl,
                BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_SET_FUNCTION_PRIVILEGES_V1, params.response, &privilege_mask));


    if (status < 0) {
        TRACE (DL_ERR, "function set privileges ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}
BESTATUS
be_function_dcn_query (
   IN PBE_FUNCTION_OBJECT         function_object,
   IN IOCTL_HEADER*               ioctl_va,
   IN u64                         ioctl_pa,
   IN u32                         ioctl_len,
   IN MCC_WRB_CQE_CALLBACK        callback,
   IN PVOID                       callback_context,
   IN PBE_QUERY_DCN_QUEUE_CONTEXT queue_context
)
{
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    IOCTL_HEADER *ioctl = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    u32 version;
    
    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            TRACE (DL_ERR, "MCC wrb peek failed.");
            status = BE_STATUS_NO_MCC_WRB;
            goto error;
        }        
    }

    ASSERT (ioctl_va->request.request_length + sizeof (IOCTL_REQUEST_HEADER) <= ioctl_len);

    version = ioctl_va->request.version;

    // Prepares an embedded ioctl
    ioctl = (IOCTL_HEADER*)be_function_prepare_nonembedded_ioctl(function_object,
                                                  wrb,
                                                  ioctl_va,
                                                  ioctl_pa,
                                                  ioctl_len,
                                                  ioctl_va->request.request_length,
                                                  ioctl_va->request.request_length,
                                                  ioctl_va->request.opcode,
                                                  ioctl_va->request.subsystem,
                                                  ioctl_va->request.version);


    // Fix the version.
    ioctl->request.version = version;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback, // callback
                 callback_context, // context 
                 ioctl);    
  
error:

    be_unlock_wrb_post (function_object);

    return status;
}

BESTATUS
lancer_function_set_diag_registers (
    IN PBE_FUNCTION_OBJECT        function_object,
    u8                            enable,
    IN MCC_WRB_CQE_CALLBACK       callback,
    IN PVOID                      callback_context,
    IN PBE_GENERIC_QUEUE_CONTEXT  queue_context)
{
    IOCTL_HEADER *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;
            goto error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = be_function_prepare_embedded_ioctl(
                function_object,
                wrb,
                sizeof (IOCTL_HEADER),
                0,
                0,
                OPCODE_COMMON_SET_DIAG_REGISTERS,
                SUBSYSTEM_COMMON_SET_DIAG_REGISTERS);

    ioctl->request.rsvd0[1] = (enable ? 1 : 0);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback, // callback
                 callback_context, // context 
                 ioctl);

    if ((status != 0) && (status != BE_PENDING)) {
        TRACE (DL_ERR, "enable disable diag registers failed.");
        goto error;
    }    

error:
    be_unlock_wrb_post (function_object);

    return status;
}

//---------------------------------------------------------------------------------------
// Function: be_clear_function_object
//
//   This routine clears the function object. All upper layers must call this routine
//   before be_function_and_chip_create_ex().
//
// function_object  - 
//---------------------------------------------------------------------------------------
void
be_clear_function_object (
    IN PBE_FUNCTION_OBJECT function_object)
{
    ASSERT (function_object);
    sa_zero_mem (function_object, sizeof (*function_object));

    return;
}

//---------------------------------------------------------------------------------------
// Function: lancer_function_set_adapter_type
//
//   This routine sets the adapter family and sli interface type members of function
//   object. Required to be called only for LANCER family of adapters
//
// function_object  - 
// adapter_family   - 0 for BE family of adapters
//                    1 for LANCER family of adapters 
// sli_intf_type    - A value of 0 to 3 depending on interface type
// sli_intf_asic    - 0x0 : BE2
//                  - 0x1 : BE3
//                  - 0x2 : SkyHawk
//                  - 0xA : Lancer A0
//                  - 0xB : Lancer B0
// return status    - BE_SUCCESS (0) on success. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
lancer_function_set_adapter_type (
    IN PBE_FUNCTION_OBJECT function_object,
    IN u32                 adapter_family,
    IN u32                 sli_intf_type,
    IN u32                 sli_intf_asic)
{
    ASSERT (function_object);

    TRACE (DL_INFO, "Adapter Family:%d SLI4 Interface Type:%d",
           adapter_family, sli_intf_type);

    if ((adapter_family >= ADAPTER_FAMILY_MAX) ||
        (sli_intf_type >= SLI_INTF_TYPE_MAX)) {
        return BE_NOT_OK;
    }

    function_object->adapter_family = adapter_family;
    function_object->sli_intf_type = sli_intf_type;
    function_object->sli_intf_asic = sli_intf_asic;

    return BE_SUCCESS;
}

BESTATUS
be_function_set_emi_canceller(
    IN PBE_FUNCTION_OBJECT  function_object,
    IN u32 port_number,
    IN u32 canceller_state
    )
{
    IOCTL_DIAG_SET_EMI_CANCELLER_V1 *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    u32 rsvd = 0;

    ASSERT (function_object);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);

    if (!wrb) {
        status = BE_STATUS_NO_MCC_WRB;  
        goto Error;
    }

    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, DIAG_SET_EMI_CANCELLER_V1);

    ioctl->params.request.phy_addr = port_number;
    ioctl->params.request.canceller_state = canceller_state;


    // Post or queue the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                function_object,
                wrb,
                NULL,
                NULL,
                NULL,
                ioctl,
                BE_CREATE_MCC_RESPONSE_COPY (IOCTL_DIAG_SET_EMI_CANCELLER_V1, params.response, &rsvd));

Error:
    if (status < 0) {
        TRACE (DL_ERR, "be_function_set_emi_canceller Failed: 0x%x",status);
    }

    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS
be_function_get_phy_details(
    IN PBE_FUNCTION_OBJECT  function_object,
    OUT PGET_PHY_DETAILS_PARAMS_OUT_V1 phy_details
    )
{
    IOCTL_COMMON_GET_PHY_DETAILS_V1* ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;

    ASSERT (function_object);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);

    if (!wrb) {
        status = BE_STATUS_NO_MCC_WRB;  
        goto Error;
    }

    
    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_GET_PHY_DETAILS_V1);

    // Post or queue the Ioctl
    status = be_function_post_mcc_wrb_with_copy (
                function_object,
                wrb,
                NULL,
                NULL,
                NULL,
                ioctl,
                BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_GET_PHY_DETAILS_V1, params.response, phy_details));

Error:
    if (status < 0) {
        TRACE (DL_ERR, "be_function_get_phy_details Failed: 0x%x",status);
    }

    be_unlock_wrb_post (function_object);
    return status;
}