/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!


@file
    lro_ll.c

@brief
    Shared code entry points for LRO receive queue
    setup and takedown functions. .

@author Sulanka Goonetilleke 

@note
    sulanka   6/03/2008    initial 

*/

#include "pch.h"


/*!

@brief
    This routine enables lro on chip per host interface.  
    
@param
    function_object      - handle to a function object                           
@param
    enable               - Set to 1 to enable LRO for this host interface or 0 to disable.                           
@param                          
    max_chainlen            - tcp param 
@param                          
    full_chainlen           - tcp param 
@param                          
    max_segcount            - tcp param 
@param                          
    max_chaintime           - tcp param 

@return
    BE_SUCCESS if successfull, , otherwise a useful BESTATUS error code is returned.

@note
    IRQL: < DISPATCH_LEVEL
    this function allocates a PBE_LRORQ_OBJECT object.
    there must be no more than 1 of these per function object.
    the rq_base_va must point to a buffer of exactly 8K.
    will be updated as appropriate on return
*/    
BESTATUS
be_lro_enable (
    IN  PBE_FUNCTION_OBJECT    function_object, 
    IN  u32                    enable,
    IN  u32                    max_chainlen, 
    IN  u32                    full_chainlen, 
    IN  u32                    max_segcount, 
    IN  u32                    max_chaintime, 
    IN  PBE_RXIF_OBJECT        rxif_object )
{
    BESTATUS status = 0 ;
    MCC_WRB *wrb = NULL ;
    IOCTL_LRO_CONFIG *ioctl = NULL ;

    FUNCTION_ASSERT (function_object) ;

    be_lock_wrb_post (function_object) ;

    wrb = be_function_peek_mcc_wrb (function_object) ;
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.") ;
        status = BE_STATUS_NO_MCC_WRB ;
        goto Error ;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LRO_CONFIG);

    ioctl->params.request.interface_id = rxif_object->handle ;
    ioctl->params.request.mode = enable ;

    ioctl->params.request.lro_params.max_chainlen = max_chainlen ;
    ioctl->params.request.lro_params.full_chainlen = full_chainlen ;
    ioctl->params.request.lro_params.max_segcount = max_segcount ;
    ioctl->params.request.lro_params.max_chaintime = max_chaintime ;

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl) ;
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to enable LRO failed.") ;
        goto Error ;
    }
    
    TRACE (DL_INFO, 
        "lro enable. function:%d mode:%d", 
           be_function_get_function_number (function_object), 
           enable) ;

Error:

    be_unlock_wrb_post (function_object) ;
    return status ;
}


/*!

@brief
    This routine creates a lro receive ring.  
    
@param
    function_object         - handle to a function object                           
@param
    rq_base_va              - base VA for the LRO  receive ring. this must be
                              exactly 8K in length and continguous physical memory.
@param
    cq_object               - handle to a previously created CQ to be associated
                              with the RQ.
@param                          
    page_size               - system page size

@param                          
    ring_size               - total number of buffer descriptors in the receive ring.

@param
    pp_lro_rq               - pointer to an opqaue handle where a LRO 
                              receive object is returned.                          
@return
    BE_SUCCESS if successfull, , otherwise a useful BESTATUS error code is returned.
@note
    IRQL: < DISPATCH_LEVEL
    this function allocates a PBE_LRORQ_OBJECT object.
    there must be no more than 1 of these per function object.
    the rq_base_va must point to a buffer of exactly 8K.
    will be updated as appropriate on return
*/    
BESTATUS
be_lro_rq_create (
    PBE_FUNCTION_OBJECT    function_object, 
    PSA_SGL                sgl, 
    PBE_CQ_OBJECT          cq_object, 
    u32                    page_size,
    u32                    ring_size,
    PBE_RXIF_OBJECT        rxif_object,
    PBE_LRORQ_OBJECT       lro_rq )
{
    BESTATUS status = 0 ;
    MCC_WRB *wrb = NULL ;
    IOCTL_LRO_RING_CREATE *ioctl = NULL ;
    u16 i ;
    PHYS_ADDR *pa_list ;

    // Ring size for LRO can be 512 or 1024 for now.
    ASSERT( !(ring_size & 0x1FF) ) ;

    // MPU will set the 
    ASSERT (sgl) ;
    ASSERT (lro_rq) ;

    FUNCTION_ASSERT (function_object) ;
    CQ_ASSERT (cq_object) ;

    be_lock_wrb_post (function_object) ;
    
    lro_rq->magic            = BE_LRORQ_MAGIC ;
    lro_rq->parent_function  = function_object ;
    lro_rq->cq_object        = cq_object ;

    wrb = be_function_peek_mcc_wrb (function_object) ;
    if (!wrb) {
        TRACE (DL_ERR, "MCC wrb peek failed.") ;
        status = BE_STATUS_NO_MCC_WRB ;
        goto Error ;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LRO_RING_CREATE);

    ioctl->params.request.interface_id = rxif_object->handle ;

    // The number of pages mapped is twice the actual number since the duplicate 
    // mappings are for the virtual LRO ring. 
    ioctl->params.request.num_pages = (UEXACT16)((ring_size * sizeof(WCON_PM_CC)) / page_size) ; 
    ioctl->params.request.lro_cq_id = be_cq_get_id (cq_object) ;
    ioctl->params.request.page_size = lro_rq->frag_size ;

    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages)) ;

    pa_list = ioctl->params.request.pages ; 

    // The pages are duplicated into a list.
    for (i = 0; i < ioctl->params.request.num_pages; i++) {
        pa_list[i + ioctl->params.request.num_pages].lo = pa_list[i].lo ;
        pa_list[i + ioctl->params.request.num_pages].hi = pa_list[i].hi ;
    }

    ioctl->params.request.num_pages = ioctl->params.request.num_pages * 2 ; 

    // Post the Ioctl
    status = be_function_post_mcc_wrb (function_object, wrb, NULL, NULL, ioctl) ;
    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to map lro rq frags failed.") ;
        goto Error ;
    }

    // Save the ring ID for cleanup.
    lro_rq->rid = ioctl->params.response.id ;

    be_cq_object_reference (cq_object) ;

    _be_function_add_lro_rq (function_object, lro_rq) ;
    
    TRACE (DL_INFO, 
        "lro rq created. function:%d bytes:%d cq_id:%d sys_page_size:%d lro_frag_size:%d", 
           be_function_get_function_number (function_object), 
           sa_sgl_get_page_count (sgl) * SA_PAGE_SIZE, 
           be_cq_get_id (cq_object),
           page_size,
           ioctl->params.request.page_size ) ;

Error:

    be_unlock_wrb_post (function_object) ;
    return status ;
}


/*!

@brief
    This routine destroys an LRO receive queue     

@param    
    lro_rq - LRO receive queue handle returned from lro_rq_create

@return   
    Returns BE_SUCCESS on success and an appropriate BESTATUS on failure.

@note    
    This function frees resourcs allocated by LroRqCreate.
    IRQL: < DISPATCH_LEVEL
*/
void be_lro_rq_destroy_internal_callback (PVOID context, BESTATUS status, MCC_WRB* wrb)
{
    PBE_LRORQ_OBJECT lro_rq = (PBE_LRORQ_OBJECT)context;
    LRORQ_ASSERT (lro_rq);

    if (status != BE_SUCCESS) {
        TRACE (DL_ERR, "Destroy lro rq failed in internal callback.\n");
    } else {
        // Dereference any CQs associated with this queue.
        be_cq_object_dereference (lro_rq->cq_object);

        // Remove from the function object.
        _be_function_remove_lro_rq (lro_rq->parent_function, lro_rq);
    }

    return;
}


BESTATUS
be_lro_rq_destroy_async (
    IN PBE_LRORQ_OBJECT     lro_rq,
    IN MCC_WRB_CQE_CALLBACK callback,
    IN PVOID                callback_context)
{
    PBE_FUNCTION_OBJECT function_object;
    BESTATUS status = BE_SUCCESS;
    MCC_WRB *wrb = NULL;
    IOCTL_LRO_RING_DESTROY *ioctl = NULL;
    u32 id = lro_rq->rid;

    LRORQ_ASSERT (lro_rq);

    function_object = lro_rq->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    TRACE (DL_INFO, "Destroy LRO ring id:%d", id);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in LRO destroy ring.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LRO_RING_DESTROY);

    ioctl->params.request.id = id;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (
                 function_object,
                 wrb,
                 NULL,
                 callback,
                 callback_context,
                 be_lro_rq_destroy_internal_callback,
                 lro_rq,
                 ioctl);
    if (status != BE_SUCCESS &&
        status != BE_PENDING) {
        TRACE (DL_ERR, "LRO ring destroy ioctl failed. id:%d", id);
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}



/*!

@brief
    Destroys an ethernet receive ring with finer granularity options than the standard
    be_lro_rq_destroy() API function. 

@param
    pp_lro_rq            - handle topreviously created LRO receive queue. 
    
@param
    callback             - Callback function on completion
    
@param
    callback_context     - Callback context
    
@return
    BE_SUCCESS if successfull, , otherwise a useful BESTATUS error code is returned.

@note

*/    
BESTATUS
be_lro_rq_destroy (
   IN PBE_LRORQ_OBJECT     lro_rq,
   IN MCC_WRB_CQE_CALLBACK callback,
   IN PVOID                callback_context)
{
    IOCTL_LRO_RING_DESTROY *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = BE_SUCCESS;
    PBE_FUNCTION_OBJECT function_object = NULL;

    LRORQ_ASSERT (lro_rq);

    function_object = lro_rq->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    TRACE (DL_INFO, "Destroy lro_rq ring id:%d", lro_rq->rid);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in destroy lro_rq ring.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, LRO_RING_DESTROY);

    ioctl->params.request.id = lro_rq->rid;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (
                 function_object,
                 wrb,
                 NULL,
                 callback,
                 callback_context,
                 be_lro_rq_destroy_internal_callback,
                 lro_rq,
                 ioctl);

    if (status != BE_SUCCESS &&
        status != BE_PENDING) {
        TRACE (DL_ERR, "lro_rq ring destroy ioctl failed. id:%d", lro_rq->rid);
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}





/*!

@brief
    This routine returns the RID for a LRO receive queue     

@param    
    rq - LRO receive queue handle returned from EthRqCreate

@return   
    Returns BE_SUCCESS on success and an appropriate BESTATUS on failure.

@note    
    IRQL: <= DISPATCH_LEVEL
*/
u32
be_lro_rq_get_id(
    IN      PBE_LRORQ_OBJECT        rq)
{
   LRORQ_ASSERT (rq);
   return rq->rid;
}


