/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        mgmtpvt.c

    Purpose:

        Source file for private Management functions.

 */

#include "acitpch.h"

MGMT_STATUS
mgmt_alloc_internal_resources(
    PMGMT_CONTROLLER_WS     pmgmt_controller,
    PMGMT_IOCTL_HANDLE      pioctl_handle)
{
    PMCC_WRB                pwrb;
    PBE_FUNCTION_OBJECT     pfo;

    pfo = hwi_get_be_pfo(pmgmt_controller->pacit_ws);

    /* Acquire the MCC WRB lock */
    be_function_lock_mcc(pfo);
    pwrb = be_function_peek_mcc_wrb(pfo);

    if(pwrb == ACIT_NULL)
    {
        /* Release the MCC WRB lock */
        be_function_unlock_mcc(pfo);
        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING, ("MCC WRB allocation failure - pending IOCTLs %d\n", pmgmt_controller->pacit_ws->counters.pending_ioctls));
        return MGMT_STATUS_INSUFFICIENT_BUFFER;
    }

    pioctl_handle->pmcc_wrb = pwrb;
    pioctl_handle->flags |= MGMT_IOCTLH_INTERNAL_RESOURCE;

    if(pioctl_handle->flags & MGMT_IOCTLH_INTERNAL)
    {
        pioctl_handle->pmgmt_buffer_h = &pwrb->payload.embedded;
        pioctl_handle->pmgmt_buffer = &pwrb->payload.embedded[4];
    }

    return MGMT_STATUS_SUCCESS;
}

ACIT_VOID
mgmt_free_internal_resources(
    PMGMT_CONTROLLER_WS     pmgmt_controller,
    PMGMT_IOCTL_HANDLE      pioctl_handle)
{
    PBE_FUNCTION_OBJECT     pfo;

    if(ACIT_IS_BIT_CLR(pioctl_handle->flags, MGMT_IOCTLH_SUBMITTED))
    {
        pfo = hwi_get_be_pfo(pmgmt_controller->pacit_ws);

        /* 
         * Release the MCC WRB that was allocated. Posting automatically releases
         * the lock, this is for those cases where we peeked at a WRB and changed
         * our mind submitting it
         */
        be_function_unlock_mcc(pfo);
     }
}

MGMT_STATUS
mgmt_peek_internal_resources(
    PMGMT_CONTROLLER_WS     pmgmt_controller)
{
    PMCC_WRB                pwrb;
    PBE_FUNCTION_OBJECT     pfo;

    pfo = hwi_get_be_pfo(pmgmt_controller->pacit_ws);
    be_function_lock_mcc(pfo);
    pwrb = be_function_peek_mcc_wrb(pfo);

    if(pwrb == ACIT_NULL)
    {
        be_function_unlock_mcc(pfo);
        return MGMT_STATUS_INSUFFICIENT_BUFFER;
    }

    be_function_unlock_mcc(pfo);
    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_fill_ioctl_request(
    PACIT_CONTROLLER_WS     pcontroller,
    PMGMT_IOCTL_HANDLE      pioctl_handle)
{
    PACIT_VOID              posm_reference = ACIT_NULL;
    ACIT_STATUS             status;

    OSM_ASSERT(pioctl_handle);
    OSM_ASSERT(pioctl_handle);
    OSM_ASSERT(pioctl_handle->pmcc_wrb);

    if(!(pioctl_handle->flags & MGMT_IOCTLH_INTERNAL))
    {
        OSM_ASSERT(pioctl_handle->pioctl_request);
        posm_reference = pioctl_handle->pioctl_request->posm_reference;
    }

    /*
     * There're embedded IOCTLs and non-embedded ones.
     * Former is used for internal IOCTLs when 
     * - there's no response buffer needed AND
     * - the payload to transfer can fit in MCC_WRB
     *
     * Non-embed will contain SGs in MCC_WRB payload and
     * is used for
     * - all external IOCTLs
     * - internal IOCTLs that need a response buffer
     */

    if((!(pioctl_handle->flags & MGMT_IOCTLH_INTERNAL)) ||
        (pioctl_handle->payload_length > BE_MCC_PAYLOAD_THRESHOLD) ||
        (pioctl_handle->flags & MGMT_IOCTLH_EXPLICIT_BUFFER))
    {
        status = hwi_fill_non_embedded_wrb(pcontroller, 
                                           pioctl_handle->pmcc_wrb,
                                           pioctl_handle->pmgmt_buffer,
                                           pioctl_handle->payload_length,
                                           pioctl_handle->pmgmt_buffer_h,
                                           sizeof(IOCTL_HEADER),
                                           posm_reference);

        if(status != ACIT_STATUS_SUCCESS)
        {
            PIOCTL_HEADER   ioctl_hdr;
            ACIT_UEXACT32   dw0 = 0;
            ACIT_UEXACT32   ext_error[2];

            if(pioctl_handle->pmgmt_buffer_h)
            {
                ioctl_hdr = (PIOCTL_HEADER)pioctl_handle->pmgmt_buffer_h;
                dw0 = (ioctl_hdr->request.subsystem << 16 | ioctl_hdr->request.opcode);
            }

            ext_error[0] = dw0;
            ext_error[1] = pioctl_handle->payload_length;

            ap_log_osm_event(pcontroller,
                             UNI_CFG_LOG_LEVEL_CRITICAL,
                             UNI_SUBMODULE_ACIT_MGMT,
                             UNI_PHASE_IOCTL,
                             2,
                             ACIT_EC_MCC_SGE_EXCEEDED,
                             0, ACIT_NULL,
                             &ext_error[0]);
        }
    }
    else
    {
        OSM_ASSERT(!(pioctl_handle->flags & MGMT_IOCTLH_EXPLICIT_BUFFER));

        status = hwi_fill_embedded_wrb(pcontroller,
                                       pioctl_handle->pmcc_wrb,
                                       pioctl_handle->pmgmt_buffer,
                                       pioctl_handle->payload_length);
    }

    if(status != ACIT_STATUS_SUCCESS)
        return MGMT_STATUS_FAILED;

    return MGMT_STATUS_SUCCESS;
}

MGMT_STATUS
mgmt_submit_ioctl(
    PACIT_CONTROLLER_WS     pcontroller, 
    PMGMT_IOCTL_HANDLE      pioctl_handle)
{
    PBE_FUNCTION_OBJECT     pfo = ACIT_NULL;
    PMCC_WRB                pwrb = ACIT_NULL;
    BESTATUS                bestatus;
    MGMT_STATUS             mgmt_status = MGMT_STATUS_SUCCESS;
    MCC_WRB_CQE_CALLBACK    callback_ptr;
    ACIT_UEXACT32           poll_timeout = 0;

    OSM_ASSERT(pioctl_handle);

    pfo = hwi_get_be_pfo(pcontroller);
    pwrb = pioctl_handle->pmcc_wrb;

    OSM_ASSERT(pfo);
    OSM_ASSERT(pwrb);

    //
    // If there a UE, some ACIT internal module could be attempting to
    // issue IOCTLs. Block it here
    //
    if(pcontroller->eh_stats.ue_detected)
    {
        // Fake the submitted flag. Else the routine that cleans up the
        // internal resources will attempt to unlock mcc again
        ACIT_BIT_SET(pioctl_handle->flags, MGMT_IOCTLH_SUBMITTED);

        ACIT_BIT_SET(pioctl_handle->flags, MGMT_IOCTLH_COMPLETED);

        /* Release the MCC WRB lock for beclib */
        be_function_unlock_mcc(pfo);

        // If there was a MGMT error we could return for a UE, we could return it.
        // For now return MGMT_STATUS_FAILED
        ((PIOCTL_RESPONSE_HEADER)(pioctl_handle->pmgmt_buffer_h))->status = MGMT_STATUS_FAILED;
        goto call_completion;
    }

    // set polled mode in ioctl handle if interrupts are disabled
    if(pcontroller->state.interrupt_enabled == ACIT_FALSE)
    {
        ACIT_BIT_SET(pioctl_handle->flags, MGMT_IOCTLH_POLLED_MODE);
    }

    callback_ptr = mgmt_complete_beclib_ioctl;
    ACIT_BIT_SET(pioctl_handle->flags, MGMT_IOCTLH_SUBMITTED);
    pcontroller->stats.ioctls_submitted++;
    pcontroller->counters.pending_ioctls++;

#ifdef BE_EMULATION

    bestatus = emul_post_mcc_wrb (pcontroller, pioctl_handle, callback_ptr);

#else
    bestatus = be_function_post_mcc_wrb (pfo, 
                                         pwrb, 
                                         callback_ptr, 
                                         pioctl_handle,
                                         ACIT_NULL);
#endif
    /* Release the MCC WRB lock for beclib */
    be_function_unlock_mcc(pfo);

    if(pioctl_handle->flags & MGMT_IOCTLH_POLLED_MODE)
    {
        if(pcontroller->state.api_type != ACIT_OP_MODE_RESTRICTED)
        {
            ACIT_UEXACT32 hc_freq_usec, usec_delay = 200;

            hc_freq_usec = (pcontroller->params.health_check_freq * 1000 * 1000)/usec_delay;

            while((poll_timeout < pioctl_handle->timeout * 5 * 1000) &&
                    ACIT_IS_BIT_CLR(pioctl_handle->flags, MGMT_IOCTLH_COMPLETED))
            {
                OSM_DELAY_TIMER(pcontroller->pcontroller_key, usec_delay);
                poll_timeout ++;

                // If there is a UE, then the beclib will return the IOCTL back
                if((poll_timeout % hc_freq_usec) == 0)
                {
                    // Log an error from here if there is a UE
                    hwi_check_health(pcontroller, 1);

                    if(pcontroller->eh_stats.ue_detected)
                    {
                        // If there was a MGMT error we could return for a UE, we could return it.
                        // For now return MGMT_STATUS_FAILED
                        ((PIOCTL_RESPONSE_HEADER)(pioctl_handle->pmgmt_buffer_h))->status = MGMT_STATUS_FAILED;

                        break;
                    }
                }
                    
                hwi_process_mcc_cq(pcontroller);
            }
        }

        if(ACIT_IS_BIT_CLR(pioctl_handle->flags, MGMT_IOCTLH_COMPLETED))
        {
            OSM_ASSERT(0);
            ((PIOCTL_RESPONSE_HEADER)(pioctl_handle->pmgmt_buffer_h))->status = MGMT_STATUS_POLL_IOCTL_TIMEOUT;

        }

call_completion:

        if(pioctl_handle->pmgmt_post_routine)
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("Invoking post routine for IOCTL 0x%p\n", pioctl_handle));

            mgmt_status = ((*pioctl_handle->pmgmt_post_routine)(pioctl_handle->pcontroller,
                                                pioctl_handle,
                                                pioctl_handle->pmgmt_buffer_h,
                                                pioctl_handle->pmgmt_buffer));
        }
    }
   
    return mgmt_status;
}

ACIT_VOID
mgmt_complete_beclib_ioctl(
    PACIT_VOID              pcontext,
    BESTATUS                bestatus,
    PMCC_WRB                pwrb)
{
    PMGMT_IOCTL_HANDLE      pioctl_handle;
    pioctl_handle = (PMGMT_IOCTL_HANDLE)pcontext;

    OSM_ASSERT(pioctl_handle);
    OSM_ASSERT(ACIT_IS_BIT_CLR(pioctl_handle->flags,MGMT_IOCTLH_COMPLETED));

    ACIT_BIT_SET(pioctl_handle->flags, MGMT_IOCTLH_COMPLETED);

    OSM_ASSERT(pioctl_handle->pcontroller->counters.pending_ioctls);
    pioctl_handle->pcontroller->counters.pending_ioctls--;
    pioctl_handle->pcontroller->stats.ioctls_completed++;

    //
    // since there was a IOCTL completion, a MCC WRB was freed up which
    // can potentially be used to continue pending invalidations
    //
    if(pioctl_handle->pcontroller->state.defer_invalidations)
    {
        pioctl_handle->pcontroller->state.defer_invalidations = ACIT_FALSE;
    }

    //
    // In the completion routine, do not check for interrupt state to decide
    // whether to call the post routine. Only the ioctl handle flags should
    // be used. This is because some IOCTLs may have been submitted in 
    // interrupt mode, but gets completed with interrupts disabled. Under
    // this condition, call the post routine
    //
    if(pioctl_handle->flags & MGMT_IOCTLH_POLLED_MODE)
    {
        OSM_DEBUG_PRINT(ACIT_DBGMSG_TRACE,("IOCTL completed, not invoking post routine 0x%p\n", pioctl_handle));
    }
    else
    {
        if(pioctl_handle->pmgmt_post_routine)
        {
            (*pioctl_handle->pmgmt_post_routine)(pioctl_handle->pcontroller,
                                                pioctl_handle,
                                                pioctl_handle->pmgmt_buffer_h,
                                                pioctl_handle->pmgmt_buffer);

        }
    }
}

ACIT_VOID
mgmt_async_event_handler(
    PACIT_VOID              pcontext,
    ACIT_UEXACT32           event_code,
    PACIT_VOID              pevent)
{
    PACIT_CONTROLLER_WS     pcontroller;
    PASYNC_EVENT_TRAILER    pasync_event;

    if((pcontext == ACIT_NULL) || (pevent == ACIT_NULL))
    {
        OSM_ASSERT(0);
        return;
    }

    pasync_event = (PASYNC_EVENT_TRAILER)((ACIT_UEXACT8 *)pevent + sizeof(MCC_CQ_ENTRY) 
                                            - sizeof(ASYNC_EVENT_TRAILER));

    pcontroller = (PACIT_CONTROLLER_WS)pcontext;

    if(ACIT_IS_BIT_SET(pcontroller->busy, CONTROLLER_BUSY_SHUTDOWN_IN_PROGRESS))
    {
        if(event_code != ASYNC_EVENT_CODE_LINK_STATE)
        {
            pcontroller->stats.async_event_ignored++;
            // We're shutting down, do not process this event anymore..
            return;
        }
    }

    switch (event_code)
    {
    case ASYNC_EVENT_CODE_ISCSI:

        switch(pasync_event->event_type)
        {
        case ASYNC_ISCSI_EVENT_NEW_TARGET_DISCOVERED:

            pcontroller->stats.async_event_new_target++;
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("new target discovery, unhandled!!\n"));
            OSM_ASSERT(0);
            break;

        case ASYNC_ISCSI_EVENT_NEW_ISCSI_CONNECTION_ESTABLISHED:
        {
            PACIT_TARGET_WS ptarget;
            PASYNC_EVENT_NEW_ISCSI_CONNECTION_ESTABLISHED pnew_iscsi_cxn;
            pnew_iscsi_cxn = (PASYNC_EVENT_NEW_ISCSI_CONNECTION_ESTABLISHED)pevent;
            pcontroller->stats.async_event_new_iscsi_cxn++;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("new iSCSI connection established\n"));

            if(ap_check_duplicate_target(pcontroller, 
                                        pnew_iscsi_cxn->session_handle,
                                        ACIT_NULL) == ACIT_STATUS_DUPLICATE_ENTRY)
            {
                ptarget = ap_get_target_ws_from_session_handle(pcontroller,
                                                            pnew_iscsi_cxn->session_handle);

                if(ptarget)
                {
                    ptarget->eh_stats.armfw_async_reopens++;
                    iscsi_process_connection_event(pcontroller,
                                                   ptarget,
                                                   pnew_iscsi_cxn->connection_handle,
                                                   pnew_iscsi_cxn->session_handle);
                }
                else
                {
                    //
                    // This must only happen when we're still in the process of
                    // initializing the target and therefore ptarget was NULL
                    //
                    OSM_ASSERT(ptarget);
                }
            }
            else
            {
                OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("new target login - session handle %d\n", pnew_iscsi_cxn->session_handle));

                //
                // This is a new target that was notified. We do not have target
                // workspace setup yet and therefore cannot initialize async_reopens
                // stats
                //

                ap_add_target(pcontroller, 
                            pnew_iscsi_cxn->session_handle, 
                            0, // domain is 0 since its assumed to be a persistent target
                            MGMT_SESSION_STATUS_OPENED,
                            0 //is_login_pending
                            );

            }
        }

        break;

        case ASYNC_ISCSI_EVENT_NEW_TCP_CONNECTION_ESTABLISHED:
        {
            PACIT_TARGET_WS ptarget;
            PASYNC_EVENT_NEW_TCP_CONNECTION_ESTABLISHED pnew_tcp_cxn;
            pnew_tcp_cxn = (PASYNC_EVENT_NEW_TCP_CONNECTION_ESTABLISHED)pevent;

            ptarget = ap_get_target_ws_from_session_handle(pcontroller, 
                                                    pnew_tcp_cxn->session_handle);

            OSM_ASSERT(ptarget);
            pcontroller->stats.async_event_new_tcp_cxn++;
// No MCS support
#if 0
            // Mark target as busy and initiate a reopen a connection IOCTL
            ap_flag_busy_target(pcontroller, ptarget, TARGET_BUSY_NEW_CONNECTION_ADD);

            mgmt_reopen_cxn(pcontroller, 
                            ptarget, 
                            pnew_tcp_cxn->session_handle,
                            pnew_tcp_cxn->connection_handle,
                            ISCSI_GET_CMD_SN(ptarget));
#endif
            break;
        }

        case ASYNC_ISCSI_EVENT_INITIATOR_START_RECOVERY:
        {
            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("iSCSI CoS has been modified, initiating session recovery!!\n"));
            pcontroller->stats.async_event_iscsi_cos++;
            iscsi_recover_all_sessions(pcontroller, ACIT_NULL);
            break;
        }

        default:

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Unhandled event_type 0x%x\n", pasync_event->event_type));
            OSM_ASSERT(0);

        }   // switch end

    break;

    case ASYNC_EVENT_CODE_LINK_STATE:
    {
        PASYNC_EVENT_LINK_STATE plink_state;
        plink_state = (PASYNC_EVENT_LINK_STATE)pevent;

        if(plink_state->port_link_status == ASYNC_EVENT_LINK_DOWN)
        {
            pcontroller->stats.async_event_link_down++;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Port link is down!!\n"));
            ap_process_link_down(pcontroller, 0);
        }
        else if(plink_state->port_link_status == ASYNC_EVENT_LINK_UP)
        {
            pcontroller->stats.async_event_link_up++;

            OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Port link up!!\n"));

            //
            // report link up only if link was down earlier. A redundant
            // port having a link back up should not be considered link
            // up
            if(ACIT_IS_CONTROLLER_LINK_DOWN(pcontroller, 0))
            {
                ap_process_link_up(pcontroller, 0);
            }
        }
        break;
    }

    default:

        pcontroller->stats.async_event_unknown++;

        OSM_DEBUG_PRINT(ACIT_DBGMSG_WARNING,("Unhandled event_code 0x%x\n", event_code));
        OSM_ASSERT(0);
        break;
    }

    //
    // since either a link or a target came back up online, set this flag to mark for 
    // any pending flush
    //
    if(IS_LIST_NOT_EMPTY(&pcontroller->io_queued_list))
        pcontroller->state.flush_queued_ios = ACIT_TRUE;

    if(IS_LIST_NOT_EMPTY(&pcontroller->tmf_target_list))
        pcontroller->state.flush_queued_tmfs = ACIT_TRUE;
}
