/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#ifndef __sa_bitvector_h__
#define __sa_bitvector_h__

#include "system/sa_system.h"
#include "runtime/sa_int_math.h"
SA_CPP_HEADER


// Use this to declare both the bitvector object and the array of bits in a struct.
// The bits are actually stored in the Bits array.
// The "pads" array is declared to allow SA_BV_CREATE to know the exact number
// of bits that are valid in the bitvector.  This allows the size of the bitvector
// to be recorded in 1 location.  Obviously, it wastes around 32 bytes of memory.
#define SA_BV_DECLARE(_name_,_numBits_)          \
    SA_BIT_VECTOR _name_;                        \
    u32        _name_##Bits[(_numBits_)/32 + 1]; \
    u8         _name_##Pads[(_numBits_)%32 + 1]; \

// Use this to initialize the bitvector object with the statically
// declared array of bits.
#define SA_BV_CREATE(_name_)                                                               \
    sa_bv_create (&_name_,                                                                  \
                  _name_##Bits,                                                             \
                  ((sizeof (_name_##Bits)/sizeof(u32)-1)*32 + (sizeof (_name_##Pads)-1)));  \

// Internal magic value.
#define SA_BV_MAGIC 0x73706276 // 'spbv' - just to reduce warnings for less smart compilers

// Internal assert 
#define SA_BV_ASSERT(_bv_)                     \
        ASSERT (_bv_);                         \
        ASSERT ((_bv_)->magic == SA_BV_MAGIC)  \

// The bit vector structure.
typedef struct _SA_BIT_VECTOR {
    u32 magic;
    u32 numBits;
    u32 numDwords;
    u32 numSet;
    u32 *bits;
} SA_BIT_VECTOR, *PSA_BIT_VECTOR;


//
// Macro to run "code" for each bit set in given PSA_BIT_VECTOR.  The index param
// indicates the current index value.  This is optimized for a consecutive string of
// set bits at the beginning of the bit vector.
//
#define SA_FOR_EACH_SET_BIT(_bv_,_index_,_code_)                   \
    {                                                              \
        u32 numSet = (_bv_)->numSet;                               \
        for (_index_ = 0; _index_ < (_bv_)->numBits; _index_++) {  \
            if (numSet == 0) { break; }                            \
            if (sa_bv_check ((_bv_), _index_) == TRUE) {           \
                numSet--;                                          \
                _code_                                             \
            }                                                      \
        }                                                          \
    }                                                              \


PSA_BIT_VECTOR
sa_bv_allocate (u32 num_bits, u32 alloc_flags);

void 
sa_bv_free (PSA_BIT_VECTOR bv);


/*! 
@brief
    Clears all bits in bitvector.
@param
    bv - bit vector

@return
    None
*/
void sa_bv_clear_all (PSA_BIT_VECTOR bv);
void sa_bv_set_all (PSA_BIT_VECTOR bv);

/*! 
@brief
    Initializes bit vector to all zeros
@param
    bv - bit vector
@return
    none
*/
SA_STATUS sa_bv_create (PSA_BIT_VECTOR bv, u32 *bits, u32 numBits);


void sa_bv_destroy (PSA_BIT_VECTOR bv);


/*! 
@brief
    Checks value of given bit.
@param
    bv - bit vector
@param
    bit - index for the bit
@return
    value of bit
*/
boolean sa_bv_check (PSA_BIT_VECTOR bv, u32 bit);

/*! 
@brief
    Checks value of given bit.
@param
    bv - bit vector
@param
    bit - index for the bit
@return
    value of bit
*/
STATIC
INLINE u32 sa_bv_num_set (PSA_BIT_VECTOR bv) {
    SA_BV_ASSERT (bv);    
    return bv->numSet;
}


/*! 
@brief
    Sets the given index in the bit vector.
@param
    bv - bit vector
@param
    bit - index to set
@return
    Returns the original value of the bit.
*/
boolean sa_bv_set (PSA_BIT_VECTOR bv, u32 bit);


#define SA_BV_BAD_INDEX ((u32)0xFFFFFFFF)

/*! 
@brief
    Finds & sets the 1st clear bit
@param
    bv    - bit vector
@param
    Count - # of consecutive bits to find
@return
    Returns the 1st bit that was set or -1 if no clear suitable run was found
*/
u32 sa_bv_find_clear_bits_and_set( PSA_BIT_VECTOR bv, u32 Count );


// Searches from the start of bitvector each allocation.
// Returns the index of the bit that was found and set.
// Returns -1 on failure.
u32 sa_bv_find_clear_bit_from_start (PSA_BIT_VECTOR bv);
u32 sa_bv_find_clear_bit_from_hint (PSA_BIT_VECTOR bv, u32 hint);
u32 sa_bv_find_clear_bit_from_end (PSA_BIT_VECTOR bv);

/*! 
@brief
    Clears the given index in the bit vector.
@param
    bv - bit vector
@param
    bit - index to clear
@return
    Returns the original value of the bit.
*/
boolean sa_bv_clear (PSA_BIT_VECTOR bv, u32 bit);

void sa_bv_and (PSA_BIT_VECTOR result, PSA_BIT_VECTOR bv1, PSA_BIT_VECTOR bv2);
void sa_bv_or (PSA_BIT_VECTOR result, PSA_BIT_VECTOR bv1, PSA_BIT_VECTOR bv2);
void sa_bv_copy (PSA_BIT_VECTOR dest, PSA_BIT_VECTOR src);


SA_CPP_TRAILER
#endif // __sa_bitvector_h__


