/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   $RCSfile: sa_index_allocator.h,v $

@brief
    Implements generic index and object allocators that use a 
    free list to efficiently allocate/deallocate items.  Supports
    both multi-thread safe, and non-safe methods.

CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_index_allocator.h,v 1.7 2006/03/09 17:51:49 menz Exp $
*/
#ifndef __sa_index_allocator_h__
#define __sa_index_allocator_h__

#include "system/sa_system.h"
SA_CPP_HEADER

// ----------------------------------------------------------
// Index Allocator (Linked List Based)
// ----------------------------------------------------------

//
// Use this to declare the objects for a ring allocator. 
//
#define SA_INDEX_ALLOCATOR_DECLARE(_name_,_num_)          \
        SA_INDEX_ALLOCATOR _name_;                        \
        SA_DBG_ONLY (boolean _name_##DbgVector [_num_];)  \
        SA_INDEX_ALLOCATOR_ENTRY _name_##Entries [_num_]  \

//
// Use this to init ring allocator. 
//
#define SA_IA_CREATE SA_INDEX_ALLOCATOR_CREATE
#define SA_INDEX_ALLOCATOR_CREATE(_name_,_num_)                                                             \
    sa_indexalloc_create (&_name_, _name_##Entries, _num_, SA_DBG_PTR( _name_##DbgVector ), SA_DBG_PTR( #_name_ ))  \

// Entry used in linked list.
typedef struct _SA_INDEX_ALLOCATOR_ENTRY {

    struct _SA_INDEX_ALLOCATOR_ENTRY* next;

} SA_INDEX_ALLOCATOR_ENTRY, *PSA_INDEX_ALLOCATOR_ENTRY;

typedef struct _SA_INDEX_ALLOCATOR {

    u32 magic;                          // constant for detecting bad IA objects
    u32 num;                            // number of indices to allocate, from 0 to num-1.
    u32 num_free;                       // number in free list
    SA_INDEX_ALLOCATOR_ENTRY *entries;  // Array of "num" objects declared external to this object.        
    SA_INDEX_ALLOCATOR_ENTRY free_list; // Head of a singly-linked list of SA_INDEX_ALLOCATOR_ENTRYs    
    const c8* dbg_name;               // Name of allocator for debug prints
    boolean *dbg_vector;                // Array of BOOLEANs used for debug mode tracking, 
                                        // declared external to this structure.
} SA_INDEX_ALLOCATOR, *PSA_INDEX_ALLOCATOR;

//
// This is a "bad" index for the allocator.
//
#define SA_IA_BAD_INDEX (~ (u32)0UL)

// Magic value and assert.
#define SA_IA_MAGIC (0xCAFEDEAD)
#define SA_IA_ASSERT(_x_) ASSERT((_x_)->magic == SA_IA_MAGIC);


PSA_INDEX_ALLOCATOR
sa_indexalloc_allocate (u32 num, 
                        u32 alloc_flags, // typically SA_MALLOC_NORMAL
                        const c8* optional_name);

void 
sa_indexalloc_free (PSA_INDEX_ALLOCATOR ia);



SA_STATUS
sa_indexalloc_create (PSA_INDEX_ALLOCATOR ia, 
                      SA_INDEX_ALLOCATOR_ENTRY *entries, 
                      u32 num, 
                      boolean *optionalDbgVector, 
                      const c8* optionalName);

void
sa_indexalloc_destroy (PSA_INDEX_ALLOCATOR ia);

u32
sa_indexalloc_allocate_index (PSA_INDEX_ALLOCATOR ia);

u32
sa_indexalloc_index_allocate_exact (PSA_INDEX_ALLOCATOR ia, u32 exactIndex);

void
sa_indexalloc_free_index (PSA_INDEX_ALLOCATOR ia, u32 index);

u32
sa_indexalloc_allocate_index_interlocked (PSA_INDEX_ALLOCATOR ia);

void
sa_indexalloc_free_index_interlocked (PSA_INDEX_ALLOCATOR ia, u32 index);

STATIC
INLINE
u32 sa_indexalloc_num_free (PSA_INDEX_ALLOCATOR ia) {
    ASSERT (ia);
    ASSERT (ia->magic == SA_IA_MAGIC);
    return ia->num_free;
}

STATIC
INLINE
u32 sa_indexalloc_num_allocated (PSA_INDEX_ALLOCATOR ia) {
    ASSERT (ia);
    ASSERT (ia->magic == SA_IA_MAGIC);
    ASSERT (ia->num >= ia->num_free);
    return ia->num - ia->num_free;
}

//#if !defined(SA_OMIT_TEST_CODE)

//    void sa_indexalloc_test (void);

//#endif


SA_CPP_TRAILER
#endif
