/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_log.h

@brief
   
CVS Header Information:
   $Header: $
*/
#ifndef __sa_log_h__
#define __sa_log_h__

#include "system/sa_types.h"
#include "runtime/sa_ring.h"

SA_CPP_HEADER

// Options --------------------------------------

// No options
#define SA_LOG_OPTION_NONE             (0x0)

// Skip all spinlocks -- assume sa_log calls are single threaded
#define SA_LOG_OPTION_NO_LOCKING       (0x1)    

// Emit a SA_TRACE debug print with each sa_log_trace call using the
// same level and module name.
#define SA_LOG_OPTION_DEBUG_TRACE      (0x2)

//<no_inbox>

#define SA_LOG_MAGIC (0x12ab34dd)
#define SA_LOG_ASSERT(_pLog) {                \
    ASSERT (_pLog);                           \
    ASSERT ((_pLog)->magic == SA_LOG_MAGIC);  \
    }                                         \

// Trace log entry
typedef struct _SA_LOG_ENTRY {   
    u32 category;  // User provided category
    u32 level;     // User provided level
    u64 timestamp; // Timestamp of entry
    c8 module_name[9]; 
    c8 msg[4];
} SA_LOG_ENTRY, *PSA_LOG_ENTRY;

// Trace log object.
typedef struct _SA_LOG {
    u32 magic;
    
    // Keep this struct the same size between 32 & 64 bit builds
    // to allow the same debugger extension to work on both.
    union {
        PVOID buffer;    
        u64 buffer_u64;
    };
    u32 buffer_bytes;    
    u32 options;
    u32 msg_bytes;  // total length of each entry    
    u32 num_entries;  // total number of entry

    // Category and level names may be 11 characters (plus the NULL).
    c8 cat_names[32][9];   
    // c8 level_names[32][9]; // Use global trace level names
    
    SA_RING ring;
    SA_SPINLOCK lock;
    SA_IRQ irq;

    u32 store_category_mask;
    u32 store_level_mask;

} SA_LOG, *PSA_LOG;

PSA_LOG sa_log_allocate (u32 msg_bytes, u32 num_msg, u32 options);
void sa_log_free (PSA_LOG log);

SA_STATUS sa_log_create (PSA_LOG log, PVOID buffer, u32 buffer_bytes, u32 msg_bytes, u32 options);
void sa_log_destroy (PSA_LOG log);

const c8* sa_log_get_category_name (PSA_LOG log, u32 category);
// const c8* sa_log_get_level_name (PSA_LOG log, u32 level);

void sa_log_set_category_name (PSA_LOG log, u32 category, const c8*name);
//void sa_log_set_level_name (PSA_LOG log, u32 level, const c8*name);

void sa_log_clear (PSA_LOG log);
u32 sa_log_num_msg (PSA_LOG log);

u32 sa_log_set_category_mask (PSA_LOG log, u32 category_mask);
u32 sa_log_get_category_mask (PSA_LOG log);

u32 sa_log_set_level_mask (PSA_LOG log, u32 level_mask);
u32 sa_log_get_level_mask (PSA_LOG log);

STATIC INLINE void
sa_log_lock (PSA_LOG log) {
    SA_LOG_ASSERT (log);
    if (!(log->options & SA_LOG_OPTION_NO_LOCKING)) {
        sa_acquire_spinlock (&log->lock, &log->irq);
    }
}

STATIC INLINE void
sa_log_unlock (PSA_LOG log) {
    SA_LOG_ASSERT (log);
    if (!(log->options & SA_LOG_OPTION_NO_LOCKING)) {
        sa_release_spinlock (&log->lock, &log->irq);
    }
}

// Create an SA_OUTPUT object that writes to an SA_LOG.
void sa_output_salog_create (SA_OUTPUT *out, SA_LOG *log, u32 category, u32 level);

void sa_log_print (PSA_LOG log, SA_OUTPUT *out, u32 category_mask, u32 level_mask, u32 max_num);

STATIC INLINE void sa_log_dump_to_debugger (PSA_LOG log, u32 category_mask, u32 level_mask) {
    u32 tl;

    // Add this mask to the current trace mask.
    tl = sa_trace_set_level (level_mask | sa_trace_get_level ());    

    sa_log_print (log, sa_stddbg, category_mask, level_mask, 0xFFFFFFFF);

    // Restore trace mask
    sa_trace_set_level (tl);
}

STATIC INLINE void sa_log_dump_to_stdout (PSA_LOG log, u32 category_mask, u32 level_mask) {
    sa_log_print (log, sa_stdout, category_mask, level_mask, 0xFFFFFFFF);
}

void sa_log_trace_vargs            (PSA_LOG log, u32 category, u32 level, const c8* module_name, const c8* format, ... );
void sa_log_trace_vlist            (PSA_LOG log, u32 category, u32 level, const c8* module_name, const c8* format, va_list args);

#ifdef __VMKLNX__
#define module_name_stringify(x) #x
#define _SA_MODULE_NAME_STRING	module_name_stringify(_SA_MODULE_NAME)
#undef _SA_MODULE_NAME
#define _SA_MODULE_NAME	_SA_MODULE_NAME_STRING
#endif


void
sa_log_trace_buffer_internal (PSA_LOG log, 
                              u32 category, 
                              u32 level, 
                              const c8* module_name, 
                              u8* data,  // Array of data
                              u32 size,  // Length of data in bytes to print
                              u32 width, // Width of hex numbers, 1, 2 or 4 bytes
                              u32 flags);
STATIC INLINE void sa_log_trace_buffer_noop (PSA_LOG log, u32 category, u32 level, PVOID buffer, u32 bytes) {
}
// Simple dump buffer format.  For options, use the sa_trace_buffer_options.  e.g.
STATIC INLINE void sa_log_trace_buffer (PSA_LOG log, u32 category, u32 level, PVOID buffer, u32 bytes) {
    // Inline function is required to expand _SA_MODULE_NAME to the current module.
#ifndef __VMKERNEL__
    sa_log_trace_buffer_internal (log, category, level, _SA_MODULE_NAME, (u8*)buffer, bytes, 4, SA_TRACE_FLAG_DEFAULTS);
#endif
}

// More advanced control for dumping a byte buffer in hex & ascii.
STATIC INLINE void sa_log_trace_buffer_options (PSA_LOG log, 
                                                u32 category, 
                                                u32 level, 
                                                PVOID buffer, 
                                                u32 bytes, 
                                                u32 width, 
                                                u32 flags) {
    // Inline function is required to expand _SA_MODULE_NAME to the current module.
#ifndef __VMKERNEL__
    sa_log_trace_buffer_internal (log, category, level, _SA_MODULE_NAME, (u8*)buffer, bytes, width, flags);
#endif
}


#if defined(SA_LINUX)        

    // Linux supports variable number args for macros.
    #define sa_log_trace(_log_, _category_, _level_, _format_...)                             \
            sa_log_trace_vargs ((_log_), (_category_), (_level_), _SA_MODULE_NAME, _format_)  \

    // This is a NOP macro that can be used by an upper layer to conditionally remove
    // logging.  Just define a macro as either sa_log_trace or sa_log_trace_noop.
    #define sa_log_trace_noop(_log_, _category_, _level_, _format_...)

#else

    // Windows supports inline varg functions.
    STATIC INLINE void sa_log_trace (PSA_LOG log, u32 category, u32 level, c8* format, ...) {
        // Inline function is required to expand _SA_MODULE_NAME to the current module.
        va_list args;
        va_start (args, format) ;
        sa_log_trace_vlist (log, category, level, _SA_MODULE_NAME, format, args); 
        va_end (args);
    }

    // This is a NOP function that can be used by an upper layer to conditionally remove
    // logging.  Just define a macro as either sa_log_trace or sa_log_trace_noop.
    STATIC INLINE void sa_log_trace_noop (PSA_LOG log, u32 category, u32 level, c8* format, ...) {
    }

#endif


#ifdef SA_DEBUG
    #define SA_DEBUG_LOG_TRACE sa_log_trace
    #define SA_DEBUG_LOG_TRACE_BUFFER sa_log_trace_buffer
#else
    #define SA_DEBUG_LOG_TRACE sa_log_trace_noop
    #define SA_DEBUG_LOG_TRACE_BUFFER sa_log_trace_buffer_noop
#endif

void sa_log_test (void);

// This is never defined -- It is removed by the no_inbox and /no_inbox tags for the
// inbox driver submission.
#ifdef FAKE_INBOX 
// </no_inbox>

    typedef u32 SA_LOG, *PSA_LOG; // dummy definition

    // The INBOX driver definitions.
    #define sa_log_allocate(_m, _n, _o)   (NULL) // return NULL
    #define sa_log_set_category_mask(_l, _m)     // nop
    #define sa_log_set_category_name(_l, _m, _n) // nop
    #define sa_log_free(_log) // nop
    #define sa_log_trace __noop
    #define sa_log_trace_buffer __noop
    #define SA_DEBUG_LOG_TRACE __noop
    #define SA_DEBUG_LOG_TRACE_BUFFER __noop

// <no_inbox>
#endif
// </no_inbox>


SA_CPP_TRAILER
#endif // __sa_log_h__

