/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
// display, modification, or disclosure of this program is strictly
// prohibited.
//
#ifndef __sa_tiny_log_h__
#define __sa_tiny_log_h__

#include "system/sa_types.h"
#include "runtime/sa_ring.h"
#include "runtime/sa_log.h"
#include "runtime/sa_network_packet.h"

SA_CPP_HEADER

#define SA_TLOG_MAGIC (0x12ab34ee)
#define SA_TLOG_ASSERT(_pLog) {                \
    ASSERT (_pLog);                            \
    ASSERT ((_pLog)->magic == SA_TLOG_MAGIC);  \
    }                                          \

// Options --------------------------------------

// No options
#define SA_TLOG_OPTION_NONE             (0x0)

// Skip all spinlocks -- assume sa_tlog calls are single threaded
#define SA_TLOG_OPTION_NO_LOCKING       (0x1)    

// Emit a SA_TRACE debug print with each sa_tlog_trace call.
#define SA_TLOG_OPTION_DEBUG_TRACE      (0x2)

// Emit a sa_log_trace print to the associated debug SA_LOG.
#define SA_TLOG_OPTION_SALOG_TRACE      (0x4)

// Trace log entry
typedef struct _SA_TLOG_ENTRY {   
    u32 code;
    u32 sequence;
    u64 args[1];
} SA_TLOG_ENTRY, *PSA_TLOG_ENTRY;


// Trace log object.
typedef struct _SA_TLOG {    
    
    // Keep this struct the same size between 32 & 64 bit builds
    // to allow the same debugger extension to work on both.
    // This points to the log buffer.
    union {
        PVOID buffer;    
        u64 buffer_u64;
    };

    // Constant magic value.
    u32 magic;

    u32 buffer_bytes; // size of buffer in bytes
    u32 num_entries;  // total number of entry
    u32 options;      // option bits
    
    char prefix[32];  // Simple prefix used during printing to distinguish which tlog
                      // each message belongs to.  Good for printing to the debugger
                      // where multiple tlogs are interleaved.

    union {
        SA_RING ring;     // ring object to tracking the next log entry to write/read
        u8 pad[44];       // max size of SA_RING on a 64 bit machine with chk driver + 4 extra bytes of pad.
    };

    SA_SPINLOCK lock; // spinlock for accessing log
    SA_IRQ irq;

    const char** names;  // Pointer to a constant array of strings to decode the "codes".
    u32 max_names;       // Number of strings in the names array.

    SA_OUTPUT salog_output; // Output object for linking to a SA_LOG and printing each 
                            // message via the debugger.
    

} SA_TLOG, *PSA_TLOG;

// Allocates memory for the tlog and returns a pointer.
// For options, see SA_TLOG_OPTION_*
PSA_TLOG sa_tlog_allocate (u32 num_msg, u32 options);

// Frees memory for the tlog.
void sa_tlog_free (PSA_TLOG log);

// Creates/Destroys a tlog without allocating memory. 
// The allocate/free functions automatically call these versions, so these are 
// only necessary if the user wants to allocate the memory directly.
SA_STATUS sa_tlog_create (PSA_TLOG log, PVOID buffer, u32 buffer_bytes, u32 options);
void sa_tlog_destroy (PSA_TLOG log);

// Sets a <31 character prefix for printing tlog messages.  This is
// only used by the "print" functions below -- or when a debug sa_log
// is linked to the tlog.
void sa_tlog_set_prefix (PSA_TLOG log, __SA_PREFAST(__in) char *prefix);

// Clears contents of log.
void sa_tlog_clear (PSA_TLOG log);

// Prints the tlog to the specified output object.
// Typically this requires assigning names prior to printing
// using sa_tlog_assign_names.
void sa_tlog_print (PSA_TLOG log, SA_OUTPUT *out);

// Assigns an array to names that are indexed by the logged "code" values.
// This array of names must be assigned to print the tlog.
void sa_tlog_assign_names (PSA_TLOG log, const char* names[], u32 max_names);

// Unit test function that creates and prints a tlog.
void sa_tlog_test (void);

// Links a SA_LOG to the tlog so each tlog trace is also printed
// to the SA_LOG (& the debugger output).
void sa_tlog_assign_log (PSA_TLOG tlog, SA_LOG *log, u32 category, u32 level);

// Internal method for all traces written into the log.
void _sa_tlog_trace (PSA_TLOG log, u32 code, u64 arg0);

// Trace taking a 64-bit argument.
STATIC FORCEINLINE void sa_tlog_trace64 (PSA_TLOG log, u32 code, u64 arg0) {
    if (log) {
        _sa_tlog_trace (log, code, arg0);
    }
}

// Trace taking a 2 32-bit arguments.
STATIC FORCEINLINE void sa_tlog_trace2 (PSA_TLOG log, u32 code, u32 arg0, u32 arg1) {
    if (log) {
        _sa_tlog_trace (log, code, ((u64)arg1 << 32) | (u64)arg0);
    }
}

// Trace taking a 3 arguments.
STATIC FORCEINLINE void sa_tlog_trace3 (PSA_TLOG log, u32 code, u16 arg0, u16 arg1, u32 arg2) {
    if (log) {
        _sa_tlog_trace (log, code, ((u64)arg2 << 32) | ((u64)arg1 << 16) | (u64)arg0);
    }
}

// Trace taking a 3 arguments.
STATIC FORCEINLINE void sa_tlog_trace4 (PSA_TLOG log, u32 code, u16 arg0, u16 arg1, u16 arg2, u16 arg3) {
    if (log) {
        _sa_tlog_trace (log, code, ((u64)arg3 << 48) | ((u64)arg2 << 32) | ((u64)arg1 << 16) | (u64)arg0);
    }
}

// Trace taking a 2 arguments, 16-bit and MAC. Format string should contain 7 %d/%x directives for the 
// arg0 plus 6 bytes of MAC.
STATIC FORCEINLINE void sa_tlog_trace_mac (PSA_TLOG log, u32 code, u16 arg0, SA_MAC_ADDRESS *arg1) {
    if (log) {
        SA_MAC64 m64;

        sa_mac64_from_mac (&m64, arg1);

        _sa_tlog_trace (log, 
                        code, 
                        m64.quad | 
                        (((u64)arg0 & 0xFFFF) << 48)
                        );
    }
}

// Trace taking a pointer.
STATIC FORCEINLINE void sa_tlog_traceptr (PSA_TLOG log, u32 code, PVOID arg0) {
    if (log) {
        _sa_tlog_trace (log, code, SA_PTR_TO_U64 (arg0));
    }
}

// Trace taking no argument.
STATIC FORCEINLINE void sa_tlog_trace (PSA_TLOG log, u32 code) {
    if (log) {
        _sa_tlog_trace (log, code, 0);
    }
}

SA_CPP_TRAILER
#endif // __sa_tlog_h__

