/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

#ifndef __sa_assert_h__
#define __sa_assert_h__

#include "sa_trace.h"

SA_CPP_HEADER

// Platform files define sa_ fuctions to implement the asserts.
#ifdef SA_MICROSOFT
    #define sa_breakpoint __debugbreak
#elif defined (SA_DOS)
    #define sa_breakpoint SA_GETC
#elif ((defined (SA_XSCALE) || defined(SA_PPC) || defined(SA_IA64)) && defined (SA_LINUXK))
    #define sa_breakpoint() BUG()
#else
    #define sa_breakpoint() __asm__ volatile ("int $3")
#endif 

//<no_inbox>
/*lint -e413 */
/*lint -e831 */
STATIC INLINE
void sa_crash (void) {
    u32* null_ptr = NULL;
    TRACE (DL_ERR, "Fatal error. Crashing.");
    
    #ifdef SA_MICROSOFT
       
       #if defined(_PREFAST_)
       #pragma prefast (disable : 6011) //  warning 6011: (PFD)Dereferencing NULL pointer 'null_ptr'.
       #endif

       *null_ptr = 0xBAD;
       
       #if defined(_PREFAST_)
       #pragma prefast (default : 6011)
       #endif
  
    #else
       *null_ptr = 0xBAD;
    #endif

}
/*lint -restore */
//</no_inbox>

// --------------------------------------------------------
//  Compile Time Asserts (for struct sizes, primarily)
// --------------------------------------------------------    
// Various compile time asserts. Typedef style works in global scope. Switch type works 
// in functions.
#define SA_COMPILETIME_TYPEDEF_ASSERT(_expression_) typedef u8 __COMPILE_TIME_ASSERT_NAME__[(_expression_)?1:-1]

#define SA_COMPILETIME_SWITCH_ASSERT(_expression_)  \
    switch(0){case 0:case (_expression_):;}         \

#define SA_COMPILETIME_NAMED_TYPEDEF_ASSERT(_name_, _expression_)    \
    typedef u8 _name_##__COMPILE_TIME_ASSERT__[(_expression_)?1:-1]  \


#if defined(_PREFAST_)
    // Prefast cannot handle constant expression comparisons
    #define SA_C_ASSERT(_x)           
    #define SA_GLOBAL_C_ASSERT(_x,_y)
#else
    // Choose one as default. This may be compiler dependent in future.
    #define SA_C_ASSERT                 SA_COMPILETIME_SWITCH_ASSERT
    #define SA_GLOBAL_C_ASSERT          SA_COMPILETIME_NAMED_TYPEDEF_ASSERT
#endif


//<no_inbox>

    // Returns TRUE if breakpoint should occur.
    boolean 
    sa_assert_handler (const c8 *module_name, 
                       const c8 *expression, 
                       const c8 *file, 
                       const c8 *function, 
                       u32 line);

#if 0
//</no_inbox>

    // Windows inbox implementation:
    STATIC INLINE boolean 
    sa_assert_handler (const c8 *module_name, 
                       const c8 *expression, 
                       const c8 *file, 
                       const c8 *function, 
                       u32 line)
    {
        TRACE (DL_ERR, "ASSERT FAILURE: %s", expression);
        TRACE (DL_ERR, "Function %s() in %s line %d", function, file, line);
        return TRUE; // always break
    }

//<no_inbox>
#endif
//</no_inbox>

// --------------------------------------------------------
//  Standard runtime assert always defined (debug & release).
//  ONLY USE FOR RELEASE MODE ASSERTS.  Use SA_ASSERT for 
//  debug mode asserts.
// --------------------------------------------------------    
#define SA_RELEASE_ASSERT( _condition_)                                                             \
    if(!( _condition_ )) {                                                                          \
        if (sa_assert_handler (_SA_MODULE_NAME, #_condition_, __FILE__, __FUNCTION__, __LINE__)) {  \
            sa_breakpoint ();                                                                       \
        }                                                                                           \
    }                                                                                               \

#if defined(SA_DEBUG)

    // --------------------------------------------------------
    //  Debug build runtime asserts
    // --------------------------------------------------------    
    #define SA_ASSERT                   SA_RELEASE_ASSERT

    #define SA_BREAKPOINT()                                                                     \
        {                                                                                       \
            SA_TRACE ( DL_ERR, "BREAKPOINT at %s(%d) : %s", __FILE__, __LINE__, __FUNCTION__);  \
            /*lint -e26 */ sa_breakpoint(); /*lint -restore */                                  \
        }                                                                                       \

    #define SA_DBG_CSTR( _str_ )        (_str_)            /* "" in release builds      */
    #define SA_DBG_ONLY( _anything_ )   _anything_         /* nothing in release builds */
    #define SA_DBG_PTR( _ptr_ )         (_ptr_)            /* NULL in release builds    */

#else 
    // --------------------------------------------------------
    //  Free build (no asserts).
    // --------------------------------------------------------
    #define SA_ASSERT(_condition_)    
    #define SA_BREAKPOINT()    
    #define SA_DBG_CSTR( _str_ )        ""
    #define SA_DBG_ONLY( _anything_ )    
    #define SA_DBG_PTR( _ptr_ )         (NULL)
#endif


// --------------------------------------------------------
//  Additional fancy asserts
// --------------------------------------------------------
#define SA_ASSERT_PASSIVE()                    ASSERT (sa_is_passive ());
#define SA_ASSERT_DISPATCH()                   ASSERT (sa_is_dispatch ());
#define SA_ASSERT_BLOCKING()                   ASSERT (sa_can_block ());
#define SA_ASSERT_NOT_DIRQL()                  ASSERT (sa_is_not_dirql ());
#define SA_ASSERT_SPIN_LOCK_HELD(_pLock_)      ASSERT (sa_is_spinlock_held (_pLock_));
#define SA_ASSERT_SPIN_LOCK_NOT_HELD(_pLock_)  ASSERT (sa_is_spinlock_not_held (_pLock_));

#define SA_BREAKPOINT_MSG(_str_)               SA_DBG_ONLY ({ SA_TRACE ( DL_ERR, "BREAKPOINT: %s", _str_ );  BREAKPOINT (); })

// --------------------------------------------------------
//  Abbreviations and legacy names.
// --------------------------------------------------------
#undef ASSERT
#define ASSERT     SA_ASSERT
#define BREAKPOINT SA_BREAKPOINT


SA_CPP_TRAILER

#endif // __sa_assert_h__

