/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_ioctl.h

@brief
    Includes operating ioctl marshalling.

*/
#ifndef __sa_ioctl_h__
#define __sa_ioctl_h__

#include "sa_types.h"
#include "sa_assert.h"
#include "sa_std.h"

#if defined(SA_WIN32U) || defined(SA_WIN_USER)
#include "sa_ioctl_win32u.h"
#else 
SA_CPP_HEADER

typedef struct _SA_IOCTL_USER_CONTEXT {
    PVOID rsvd;
} SA_IOCTL_USER_CONTEXT, *PSA_IOCTL_USER_CONTEXT;

SA_CPP_TRAILER
#endif

SA_CPP_HEADER

/*
	Ioctl marshalling involves the ability to invoke a function
	in another module by means of a numemeric code and input and
	output buffers. This is a client server model. The SA
	abstracts the interface creation and the function demuxing and
	some generic error handling. 
*/

#define SA_IOCTL_CLIENT_CONTEXT_MAGIC   0xc11e3434
#define SA_IOCTL_CALL_CONTEXT_MAGIC     0xc11e3435

// The ioctl resource defines an ioctl interface. The implementation manages
// how this maps to the actual system resource. 
// Consumers must embed this or independently manage this structure. The 
// structure identifies a unique endpoint in the api
typedef struct _SA_IOCTL_CLIENT_CONTEXT{
	u32                   magic;
    u32                   ioctlType;
    SA_IOCTL_USER_CONTEXT saContext;
} SA_IOCTL_CLIENT_CONTEXT, *PSA_IOCTL_CLIENT_CONTEXT;

// These are standard error codes for the ioctls. This error code is private
// to the ioctl sa, ie any system error codes ("aka NT_STATUS, etc") need to
// be embedded in the ioctl payload and are managed by the implementation
typedef enum {
	SA_IOCTL_SUCCESS = 0,
	SA_IOCTL_ERROR_BAD_FUNCTION,
    SA_IOCTL_ERROR_BAD_TYPE,
	SA_IOCTL_ERROR_MAPPING_MEMORY,
    SA_IOCTL_PENDING,
    SA_IOCTL_ERROR_NOT_FOUND,
	SA_IOCTL_SERVER_ERROR = 0x80000000,
} SA_IOCTL_STATUS;

// This callback is the completion callback for any ioctl invocation. 
// Ioctls always complete asynchronously
typedef struct _SA_IOCTL_CALL_CONTEXT* PSA_IOCTL_CALL_CONTEXT;

typedef void (*SA_IOCTL_CLIENT_CB)(PSA_IOCTL_CALL_CONTEXT serverContext);

// Used to define and retrieve the context of an ioctl call. This structure
// defines an instance of a particular ioctl invocation.
typedef struct _SA_IOCTL_CALL_CONTEXT {
	u32                         magic;              // structure integrity
    u32                         function;           // function code to
    SA_IOCTL_CLIENT_CB          cb;                 // callback, unused in
                                                    // invoke
	PSA_IOCTL_CLIENT_CONTEXT    clientContext;      // end point identifier
	SA_IOCTL_STATUS             status;             // sa status
                                                    // server
    u32                         bufferLength;       // must be atleast
                                                    // sizeof(PSA_IOCTL_CALL_CONTEXT)
    u8                          opaque[64];         // Opaque data, this is a kind
                                                    // of a hack right now to plug
                                                    // into existing ioctl
                                                    // implementations
} SA_IOCTL_CALL_CONTEXT;

// This pointer is called in order the demux the ioctl at the server
// The implementation manages the call context
typedef void (*SA_IOCTL_SERVER_PROC)(PSA_IOCTL_CALL_CONTEXT clientContext);


// Client functions

// Called by the client to initiate an ioctl call
SA_IOCTL_STATUS sa_ioctl_invoke( PSA_IOCTL_CALL_CONTEXT serverContext );

// Called by client to initialize and obtain an ioctl resource
SA_IOCTL_STATUS sa_ioctl_initialize( PSA_IOCTL_CLIENT_CONTEXT resource );

// Called by client to close the ioctl resource
SA_IOCTL_STATUS sa_ioctl_destroy( PSA_IOCTL_CLIENT_CONTEXT resource );

// Server functions

// Called by server to indicate completion of the ioctl
void sa_ioctl_complete(PSA_IOCTL_CALL_CONTEXT clientContext, 
					   SA_IOCTL_STATUS serverStatus );


// This callback is called when a new client associates with the ioctl
typedef PSA_IOCTL_CLIENT_CONTEXT (*SA_IOCTL_CLIENT_CONTEXT_CREATE)(void);
typedef void (*SA_IOCTL_CLIENT_CONTEXT_REMOVE)(void);


// Installs an ioctl. Every implementation will have its own incarnation of this
// function. This call registers all the callbacks and calls them when the 
// appropriate functionality is invoked. Create is called when a new client binds
// to this callee. remove is called when the binding is removed. proc is called for
// every invocation of the ioctl.
void sa_ioctl_install(
    u32 ioctlType,                          // Implementation fixed ioctl type
    SA_IOCTL_CLIENT_CONTEXT_CREATE create,  // Callback for creating a client
                                            // association
    SA_IOCTL_CLIENT_CONTEXT_REMOVE remove,  // Callback for destroying a
                                            // association resoure
    SA_IOCTL_SERVER_PROC proc               // Callback for an invoked ioctl
);

// This is the ioctl registry for all the consumers of SA
enum SA_IOCTL_TYPE_REGISTRY {
    SA_IOCTL_TYPE_MIN = 0,
    SA_IOCTL_BENDIS_MINIPORT_ADAPTER0 = 0,
    SA_IOCTL_BENDIS_MINIPORT_ADAPTER_MAX = SA_IOCTL_BENDIS_MINIPORT_ADAPTER0 + 65535,
    SA_IOCTL_BERDMA_PROVIDER,
    SA_IOCTL_TYPE_MAX
};

SA_CPP_TRAILER

#endif // __sa_ioctl_h__

