/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#ifndef __sa_list_h__
#define __sa_list_h__

#include "sa_types.h"
#include "sa_assert.h"
#include "sa_sync.h"

SA_CPP_HEADER

#ifdef SA_MICROSOFT

    typedef LIST_ENTRY  SA_LIST_ENTRY;
    typedef PLIST_ENTRY PSA_LIST_ENTRY;

#else

    // Define the list entry.
    typedef struct _SA_LIST_ENTRY {
        struct _SA_LIST_ENTRY *Flink;
        struct _SA_LIST_ENTRY *Blink;
    } SA_LIST_ENTRY, *PSA_LIST_ENTRY;

#endif


// Macros to traverse a linked list, getting a pointer to the specified TYPE at each node.
// The pListElement is a pointer to the current item in the list.
// The ListHead is the head for a list to traverse.
// Type is the data type for the items in the list.
// ListFieldInType is the name of the SA_LIST_ENTRY within the Type that links the items.
// This statement works like a for loop to iterate thru each item, like this:
//
// typedef _TYPE { SA_LIST_ENTRY link; ...more items... } TYPE;
// TYPE *item;
// SA_LIST_ENTRY Head;
// SA_FOR_EACH_LIST_ENTRY (item, Head, TYPE, Link) {
//     do something with item
// }
//
#define SA_FOR_EACH_LIST_ENTRY(_pListElement, _ListHead, _Type, _ListFieldInType )                                            \
    for ( (_pListElement) =  SA_CONTAINING_RECORD( (_ListHead).Flink, _Type, _ListFieldInType.Flink );                        \
          &(_pListElement)->_ListFieldInType != &(_ListHead);                                                                 \
          (_pListElement) = SA_CONTAINING_RECORD( (_pListElement)->_ListFieldInType.Flink, _Type, _ListFieldInType.Flink ) )  \


#define SA_FOR_EACH_LIST_ENTRY_REVERSE(_pListElement, _ListHead, _Type, _ListFieldInType )                                    \
    for ( (_pListElement) =  SA_CONTAINING_RECORD( (_ListHead).Blink, _Type, _ListFieldInType.Flink );                        \
          &(_pListElement)->_ListFieldInType != &(_ListHead);                                                                 \
          (_pListElement) = SA_CONTAINING_RECORD( (_pListElement)->_ListFieldInType.Blink, _Type, _ListFieldInType.Flink ) )  \


STATIC
FORCEINLINE
void
sa_initialize_list_head(
    IN PSA_LIST_ENTRY head
    )
{
    // Empty list points to itself.
    head->Flink = head;
    head->Blink = head;
}

// Get the first entry in the list.
STATIC
FORCEINLINE
PSA_LIST_ENTRY sa_list_head (
    IN PSA_LIST_ENTRY head)
{
    ASSERT (head);
    return head->Flink;
}

// Get the last entry in the list.
STATIC
FORCEINLINE
PSA_LIST_ENTRY sa_list_tail (
    IN PSA_LIST_ENTRY head)
{
    ASSERT (head);
    return head->Blink;
}

//
// Returns TRUE if list is empty.
//
STATIC
FORCEINLINE
boolean
sa_is_list_empty(
    IN const SA_LIST_ENTRY * head
    )
{
    return (boolean)(head->Flink == head);
}

//
// Removes first entry in the list.
//
STATIC
FORCEINLINE
PSA_LIST_ENTRY
sa_remove_head_list(
    IN  PSA_LIST_ENTRY head)
{
    PSA_LIST_ENTRY new_first;
    PSA_LIST_ENTRY entry;

    // Entry is the element to remove.
    entry = head->Flink;

    // new_first is the new head of list
    new_first = entry->Flink;

    // The list head points forward to the new first item.
    head->Flink = new_first;

    // The new first item points back to the list head.
    new_first->Blink = head;

    return entry;
}

//
// Removes last entry in the list.
//
STATIC
FORCEINLINE
PSA_LIST_ENTRY
sa_remove_tail_list(
    IN PSA_LIST_ENTRY head
    )
{
    PSA_LIST_ENTRY new_last;
    PSA_LIST_ENTRY entry;

    // entry is item to remove
    entry = head->Blink;

    // new_last is the new last item in the list.
    // This may be the head itself.
    new_last = entry->Blink;

    // head points back to the new_last
    head->Blink = new_last;

    // new_last points forward to the head.
    new_last->Flink = head;

    return entry;
}

//
// Removes given entry from anywhere in the list.
// The list head is not required for this operation.
//
STATIC
boolean
FORCEINLINE
sa_remove_entry_list(
    IN PSA_LIST_ENTRY entry
    )
{
    PSA_LIST_ENTRY previous;
    PSA_LIST_ENTRY next;

    // Remember the previous and next pointers relative to 
    // the item to remove.
    next = entry->Flink;
    previous = entry->Blink;

    // Fix the previous forward link
    previous->Flink = next;

    // Fix the next back link
    next->Blink = previous;

    // Return TRUE if the list is empty. Both point to list head.
    return (boolean)(next == previous);
}

//
// Adds the entry to the end of the list.
//
STATIC
FORCEINLINE
void
sa_insert_tail_list(
    IN PSA_LIST_ENTRY head,
    IN PSA_LIST_ENTRY entry
    )
{
    PSA_LIST_ENTRY last;

    // last is the last item in the list.
    // last may point to the head if the list is empty.
    last = head->Blink;

    // The entry points forward to the head and backward to the
    // last item in the list.
    entry->Flink = head;
    entry->Blink = last;

    // The last item becomes the 2nd to last, pointing forward to the new last
    last->Flink = entry;

    // The head now points back to the new last item
    head->Blink = entry;
}

//
// Adds the entry to the beginning of the list.
//
STATIC
FORCEINLINE
void
sa_insert_head_list(
    IN PSA_LIST_ENTRY head,
    IN PSA_LIST_ENTRY entry
    )
{
    PSA_LIST_ENTRY first;

    // The first item in the list, may be the head itself.
    first = head->Flink;

    // The new first entry points forward to the old first item and
    // back to the list head.
    entry->Flink = first;
    entry->Blink = head;

    // The 2nd item points back to the new first entry.
    first->Blink = entry;

    // The head points forward to the new first entry.
    head->Flink = entry;
}

//
// Returns the number of elements in the list.
//
STATIC
FORCEINLINE
u32
sa_count_list (PSA_LIST_ENTRY head)
{
    PSA_LIST_ENTRY next = head;
    u32 count = 0;

    while (next->Flink != head) {
        count++;
        next = next->Flink;
    }

    return count;
}

//
// Moves the list pointed to by SrcListHead to the DestListHead.
//
STATIC
FORCEINLINE
void
sa_move_list (PSA_LIST_ENTRY DestListHead,
              PSA_LIST_ENTRY SrcListHead)
{    
    if (!sa_is_list_empty (SrcListHead)) {
        
        // Assign the dest list head element.
        DestListHead->Flink = SrcListHead->Flink;
        DestListHead->Blink = SrcListHead->Blink;

        // Fix all the pointers to the list head
        SrcListHead->Flink->Blink = DestListHead;
        SrcListHead->Blink->Flink = DestListHead;    

        // Empty the src list
        sa_initialize_list_head (SrcListHead);

    } else {
        // Empty list
        sa_initialize_list_head (DestListHead);
    }

}

//
// Appends the ListHeadToAppend to the end of the head list.
//
STATIC
FORCEINLINE
void
sa_append_list (PSA_LIST_ENTRY head,
                PSA_LIST_ENTRY ListHeadToAppend)
{
    if (!sa_is_list_empty (ListHeadToAppend)) {

        if (sa_is_list_empty (head)) {

            sa_move_list (head, ListHeadToAppend);

        } else {

            // Join lists
            head->Blink->Flink = ListHeadToAppend->Flink; 
            ListHeadToAppend->Flink->Blink = head->Blink;
            ListHeadToAppend->Blink->Flink = head;

            // Fix head blink
            head->Blink = ListHeadToAppend->Blink;

            // Empty appended list
            sa_initialize_list_head (ListHeadToAppend);
        }
    }
}

//
// Prepends the ListHeadToPrepend to the beginning of the head list.
//
STATIC
FORCEINLINE
void
sa_prepend_list (PSA_LIST_ENTRY head,
                 PSA_LIST_ENTRY ListHeadToPrepend)
{
    sa_append_list (ListHeadToPrepend, head);
    sa_move_list (head, ListHeadToPrepend);
    ASSERT (sa_is_list_empty (ListHeadToPrepend));
}

// Define logical names for the functions instead of the original random word order.
#define sa_list_initialize   sa_initialize_list_head
#define sa_list_is_empty     sa_is_list_empty
#define sa_list_remove_head  sa_remove_head_list
#define sa_list_remove_tail  sa_remove_tail_list
#define sa_list_remove_entry sa_remove_entry_list
#define sa_list_insert_tail  sa_insert_tail_list
#define sa_list_insert_head  sa_insert_head_list
#define sa_list_count        sa_count_list
#define sa_list_move         sa_move_list
#define sa_list_append       sa_append_list
#define sa_list_prepend      sa_prepend_list

// Define standard names as well.
#ifndef SA_MICROSOFT
    
    #define InitializeListHead sa_initialize_list_head
    #define IsListEmpty        sa_is_list_empty
    #define RemoveHeadList     sa_remove_head_list
    #define RemoveTailList     sa_remove_tail_list
    #define RemoveEntryList    sa_remove_entry_list
    #define InsertTailList     sa_insert_tail_list
    #define InsertHeadList     sa_insert_head_list
#endif


// ----------------------------------------------------------------------
//  Singly Linked List
// ----------------------------------------------------------------------

// Define the list entry.   
typedef struct _SA_SLIST_ENTRY {
    struct _SA_SLIST_ENTRY *link;
} SA_SLIST_ENTRY, *PSA_SLIST_ENTRY;

typedef struct _SA_SLIST_HEADER {
    struct _SA_SLIST_ENTRY *first;
    struct _SA_SLIST_ENTRY *last;
} SA_SLIST_HEADER, *PSA_SLIST_HEADER;


STATIC
FORCEINLINE
void
sa_slist_initialize (
    PSA_SLIST_HEADER head)
{
    ASSERT (head);
    head->first = NULL;
    head->last = NULL;
}

STATIC
FORCEINLINE
boolean
sa_slist_is_empty (
    PSA_SLIST_HEADER head)
{
    return (head->first == NULL);
}

STATIC
FORCEINLINE
void
sa_slist_insert_head (
    PSA_SLIST_HEADER head,
    PSA_SLIST_ENTRY  entry)
{
    ASSERT (head);
    ASSERT (entry);    

    entry->link = head->first;

    if (!head->last) {
         ASSERT (!head->first);         
         head->last = entry;
    } 

    head->first = entry;
}


STATIC
FORCEINLINE
void
sa_slist_insert_tail (
    PSA_SLIST_HEADER head,
    PSA_SLIST_ENTRY  entry)
{
    ASSERT (head);
    entry->link = NULL;

    if (head->last) {
        ASSERT (head->first);
        head->last->link = entry;
    } else {
        ASSERT (head->first == NULL);
        ASSERT (head->last == NULL);
        head->first = entry;
    }

    head->last = entry;       
}

STATIC
FORCEINLINE
PSA_SLIST_ENTRY
sa_slist_remove_head (
    PSA_SLIST_HEADER head)
{
    PSA_SLIST_ENTRY entry;

    ASSERT (head);
    entry = head->first;

    if (entry) {
        head->first = entry->link;

        if (head->first == NULL) {
            ASSERT (head->last == entry);
            head->last = NULL;
        }
    }
        
    return entry;
}

// Returns pointer to first element or 
// NULL if list is empty (without removing element).
STATIC
FORCEINLINE
PSA_SLIST_ENTRY
sa_slist_peek_head (
    PSA_SLIST_HEADER head)
{
    ASSERT (head);
    return head->first;
}

STATIC
FORCEINLINE
u32
sa_slist_count (
    PSA_SLIST_HEADER head)
{
    u32 count = 0;
    PSA_SLIST_ENTRY entry;
    
    ASSERT (head);
    
    entry = head->first;
    while (entry) {
        entry = entry->link;
        count++;
    }

    return count;
}

STATIC
FORCEINLINE
void
sa_slist_append (
    PSA_SLIST_HEADER head,
    PSA_SLIST_HEADER append)
{    
    ASSERT (head);
    ASSERT (append);

    if (!append->first) {
        ASSERT (!append->last);
        return;
    }

    if (!head->first) {
        ASSERT (!head->last);
        head->first = append->first;
    } else {
        ASSERT (head->first);
        ASSERT (head->last);
        head->last->link = append->first;
    }

    ASSERT (append->last);
    head->last = append->last;

    // empty appended list
    append->first = NULL;
    append->last = NULL;
}


STATIC
FORCEINLINE
void
sa_slist_move (
    PSA_SLIST_HEADER dest,
    PSA_SLIST_HEADER src)
{    
    ASSERT (dest);
    ASSERT (src);

    dest->first = src->first;
    dest->last = src->last;

    src->first = NULL;
    src->last = NULL;
}

// Prepends the "prepend" list at the beginning of the "head" list.
// At the end, the head list contains all the elements and "prepend" list is empty.
STATIC
FORCEINLINE
void
sa_slist_prepend (
    PSA_SLIST_HEADER head,
    PSA_SLIST_HEADER prepend)
{
    sa_slist_append (prepend, head);
    sa_slist_move (head, prepend);
    ASSERT (sa_slist_is_empty (prepend));
}


// ----------------------------------------------------------------------
//  Queue
// ----------------------------------------------------------------------

// Define the list entry.   
typedef SA_SLIST_ENTRY  SA_QUEUE_ENTRY,  *PSA_QUEUE_ENTRY;
typedef SA_SLIST_HEADER SA_QUEUE_HEADER, *PSA_QUEUE_HEADER;

STATIC FORCEINLINE void 
sa_queue_initialize (PSA_QUEUE_HEADER head) {
    sa_slist_initialize ((PSA_SLIST_HEADER)head);
}

STATIC FORCEINLINE boolean 
sa_queue_is_empty (PSA_QUEUE_HEADER head) {
    return sa_slist_is_empty ((PSA_SLIST_HEADER)head);
}

STATIC FORCEINLINE u32
sa_queue_count (PSA_QUEUE_HEADER head) {
    return sa_slist_count ((PSA_SLIST_HEADER)head);
}

STATIC FORCEINLINE void
sa_queue_push (PSA_QUEUE_HEADER head, PSA_QUEUE_ENTRY entry) {
    sa_slist_insert_tail((PSA_SLIST_HEADER)head, (PSA_QUEUE_ENTRY)entry);
}

STATIC FORCEINLINE void
sa_queue_unpop (PSA_QUEUE_HEADER head, PSA_QUEUE_ENTRY entry) {
    sa_slist_insert_head((PSA_SLIST_HEADER)head, (PSA_QUEUE_ENTRY)entry);
}

// Returns pointer to next element to pop without actually popping it.
STATIC FORCEINLINE PSA_QUEUE_ENTRY
sa_queue_peek (PSA_QUEUE_HEADER head) {
    return (PSA_QUEUE_ENTRY)sa_slist_peek_head ((PSA_SLIST_HEADER)head);
}

STATIC FORCEINLINE PSA_QUEUE_ENTRY
sa_queue_pop (PSA_QUEUE_HEADER head) {
    return (PSA_QUEUE_ENTRY)sa_slist_remove_head ((PSA_SLIST_HEADER)head);
}

STATIC FORCEINLINE void
sa_queue_append (PSA_QUEUE_HEADER head, PSA_QUEUE_HEADER append) {
    sa_slist_append ((PSA_SLIST_HEADER)head, (PSA_SLIST_HEADER)append);
}

STATIC FORCEINLINE void
sa_queue_move (PSA_QUEUE_HEADER dest, PSA_QUEUE_HEADER src) {
    sa_slist_move ((PSA_SLIST_HEADER)dest, (PSA_SLIST_HEADER)src);
}

STATIC FORCEINLINE void
sa_queue_prepend (PSA_QUEUE_HEADER head, PSA_QUEUE_HEADER prepend) {
    sa_slist_prepend ((PSA_SLIST_HEADER)head, (PSA_SLIST_HEADER)prepend);
}

// ----------------------------------------------------------------------
//  Stack
// ----------------------------------------------------------------------

// Define the list entry.   
typedef SA_SLIST_ENTRY  SA_STACK_ENTRY,  *PSA_STACK_ENTRY;
typedef struct _SA_STACK_HEADER {
    SA_STACK_ENTRY *link;
} SA_STACK_HEADER, *PSA_STACK_HEADER;

STATIC FORCEINLINE void 
sa_stack_initialize (PSA_STACK_HEADER head) {
    ASSERT (head);
    head->link = NULL;
}

STATIC FORCEINLINE boolean 
sa_stack_is_empty (PSA_STACK_HEADER head) {
    return (boolean)(head->link == NULL);
}

STATIC FORCEINLINE u32
sa_stack_count (PSA_STACK_HEADER head) {
    u32 cnt = 0;
    PSA_STACK_ENTRY entry;

    ASSERT (head);

    entry = head->link;
    while (entry) {
        entry = entry->link;
        cnt++;
    }

    return cnt;
}

STATIC FORCEINLINE void
sa_stack_push (PSA_STACK_HEADER head, PSA_STACK_ENTRY entry) {
    ASSERT (head);
    ASSERT (entry);

    entry->link = head->link;
    head->link = entry;
}

STATIC FORCEINLINE void
sa_stack_atomic_push (PSA_STACK_HEADER head, PSA_STACK_ENTRY entry) {    
    PSA_STACK_ENTRY old_head;

    ASSERT (head);
    ASSERT (entry);

    do {
        // Detect an entry linked to itself.  Not very likely this will happen.
        ASSERT (entry != head->link);

        old_head = entry->link = head->link;
    } while (old_head != (PSA_STACK_ENTRY)sa_atomic_compare_exchange_ptr ( (PVOID*)&head->link, 
                                                                             entry, 
                                                                             old_head));
}

STATIC FORCEINLINE PSA_STACK_ENTRY
sa_stack_pop (PSA_STACK_HEADER head) {
    PSA_STACK_ENTRY entry;
    ASSERT (head);

    if (!head->link)
        return NULL;
    
    entry = head->link;    
    head->link = entry->link;
    return entry;
}

STATIC FORCEINLINE PSA_STACK_ENTRY
sa_stack_atomic_pop (PSA_STACK_HEADER head) {    
    PSA_STACK_ENTRY entry;
    
    ASSERT (head);

    do {
        entry = head->link;

        // Detect an entry linked to itself.  Not very likely this will happen.
        ASSERT (entry != entry->link);

        if (!entry) {
            return NULL;
        }
            
    } while (entry != (PSA_STACK_ENTRY)sa_atomic_compare_exchange_ptr ((PVOID*)&head->link, 
                                                                       entry->link, 
                                                                       entry));

    return entry;
}


SA_CPP_TRAILER

#endif 

