/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_trace.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/sa_trace.h,v 1.24 2006/03/10 19:13:08 menz Exp $
*/
#ifndef __sa_trace_h__
#define __sa_trace_h__

#include "sa_types.h"
#include "sa_defines.h"
#include "sa_std.h"
#include "sa_status.h"
#include "sa_output.h"

SA_CPP_HEADER

// debug levels
typedef enum _SA_DEBUG_LEVELS {    
    DL_ALWAYS  = 0,   // cannot be masked

    DL_ERR     = 0x1, // errors that should never happen
    DL_WARN    = 0x2, // something is questionable.  recoverable errors
    DL_NOTE    = 0x4, // infrequent, important debug info
    DL_INFO    = 0x8, // debug information
    
    DL_VERBOSE = 0x10, // detailed information, such as buffer traces
    DL_ENTRY   = 0x20, // function entry
    DL_EXIT    = 0x40, // function exit
    DL_RSVD1   = 0x80,    
    
    SA_DL_MIN_VALUE = 0x1,  // this is the min value used
    SA_DL_MAX_VALUE = 0x80  // this is the higheset value used

} SA_DEBUG_LEVELS, *PSA_DEBUG_LEVELS;


// Flags for buffer tracing (SA_TRACE_FLAGS). Use only 1 of the endian flags (native, swap, little, or big).
#define SA_TRACE_FLAG_DEFAULTS      (0)
#define SA_TRACE_FLAG_NO_ASCII      (0x1)   // Don't print the ASCII representation (in byte order) at end of each line
#define SA_TRACE_FLAG_NO_OFFSET     (0x2)   // Don't print the hex byte offset at start of each line e.g. [000]
#define SA_TRACE_FLAG_NATIVE_ENDIAN (0x0)   // [Default] Assume byte packed in native endian order.
#define SA_TRACE_FLAG_ENDIAN_SWAP   (0x4)   // Endian swap integers before printing
#define SA_TRACE_FLAG_LITTLE_ENDIAN (0x8)   // Assume integers packed as little endian
#define SA_TRACE_FLAG_BIG_ENDIAN    (0x10)  // Assume integers packed as big endian

// <no_inbox>

// Function declarations.
u32 sa_trace_set_level (u32 level);
u32 sa_trace_get_level (void);

SA_STATUS sa_trace_start_file (const c8* filename);
void sa_trace_end_file (void);

const c8*   sa_trace_get_debug_level_string (u32 level);
void        sa_trace_set_debug_level_string (u32 level, const c8* str);

void sa_trace_vargs            (u32 level, const c8* module_name, const c8* format, ... );
void sa_trace_vargs_no_newline (u32 level, const c8* module_name, const c8* format, ... );
void sa_trace_vlist            (u32 level, const c8* module_name, const c8* format, va_list args );
void sa_trace_vlist_no_newline (u32 level, const c8* module_name, const c8* format, va_list args );

// Used in release mode to replace statements with another valid statement.
// This makes it safer in case someone writes code like this:
// if (something) SA_TRACE (...);
// This would be dangerous if SA_TRACE resolved to nothing during release builds.
STATIC INLINE void sa_noop (void) { }

#if defined(SA_LINUX)        

    // Linux supports variable number args for macros.
    #define sa_trace(_level_, _format_...)            \
            sa_trace_vargs ((_level_), _SA_MODULE_NAME, _format_)

    #define sa_trace_no_newline(_level_, _format_...) \
            sa_trace_vargs_no_newline ((_level_), _SA_MODULE_NAME, _format_)

    STATIC INLINE void sa_noop_trace_buffer (u32 l, PVOID b, u32 s) { }
    STATIC INLINE void sa_noop_trace_buffer_options (u32 l, PVOID m, u32 b, u32 w, u32 f) { }

#else
    // Windows supports inline varg functions.
    STATIC INLINE void sa_trace (u32 level, c8* format, ...) {
        // Inline function is required to expand _SA_MODULE_NAME to the current module.
        va_list args;
        va_start (args, format) ;
        sa_trace_vlist (level, _SA_MODULE_NAME, format, args); 
        va_end (args);
    }

    // Windows supports inline varg functions.
    STATIC INLINE void sa_trace_no_newline (u32 level, c8* format, ...) {
        // Inline function is required to expand _SA_MODULE_NAME to the current module.
        va_list args;
        va_start (args, format);
        sa_trace_vlist_no_newline (level, _SA_MODULE_NAME, format, args); 
        va_end (args);
    }

    // Dummy function for release builds.
    STATIC INLINE void sa_trace_noop (u32 x, ...) { SA_NOT_USED (x); }
    #define sa_noop_trace_buffer sa_trace_noop
    #define sa_noop_trace_buffer_options sa_trace_noop
#endif

void sa_buffer_print (
        SA_OUTPUT *out,
        u8* data,              // Array of data
        u32 size,              // Length of data in bytes to print
        u32 width,             // Width of hex numbers, 1, 2 or 4 bytes
        u32 flags);

// Do not call directly.  Use the SA_TRACE_BUFFER or SA_TRACE_BUFFER_OPTIONS for debug
// only tracing.  For release mode tracing using the sa_trace_buffer or sa_trace_buffer_options
// functions.
void sa_trace_buffer_internal (
    u32 level, 
    const c8* module_name, 
    u8* data,              // Array of data
    u32 size,              // Length of data in bytes to print
    u32 width,             // Width of hex numbers, 1, 2 or 4 bytes
    u32 flags);            // See SA_TRACE_FLAGS, use 0 for defaults.

// Simple dump buffer format.  For options, use the sa_trace_buffer_options.  e.g.
STATIC INLINE void sa_trace_buffer (u32 level, PVOID buffer, u32 bytes) {
    // Inline function is required to expand _SA_MODULE_NAME to the current module.
#ifndef __VMKERNEL__
    sa_trace_buffer_internal (level, _SA_MODULE_NAME, (u8*)buffer, bytes, 4, SA_TRACE_FLAG_DEFAULTS);
#endif
}

// More advanced control for dumping a byte buffer in hex & ascii.
STATIC INLINE void sa_trace_buffer_options (u32 level, 
                                            PVOID buffer, 
                                            u32 bytes, 
                                            u32 width, 
                                            u32 flags) { // see SA_TRACE_FLAGS
    // Inline function is required to expand _SA_MODULE_NAME to the current module.
#ifndef __VMKERNEL__
    sa_trace_buffer_internal (level, _SA_MODULE_NAME, (u8*)buffer, bytes, width, flags);
#endif
}


#if defined(SA_DEBUG)

    //  -------------------------------------------------------------------
    //   Debug Builds - Macros only defined for debug.
    //  -------------------------------------------------------------------    

    // void TRACE(u32 level, c8* format, ...);   
    #define SA_TRACE sa_trace
    #define TRACE    SA_TRACE

    // This version does not automatically append a newline. It does not flush either, since
    // fflush may insert a newline on some systems.  Don't use this version normally.
    // void SA_TRACE_NN(u32 level, c8* format, ...);        
    #define SA_TRACE_NN sa_trace_no_newline

    // This version avoids compiling the SA_TRACE function args during release builds
    // but it requires a double set of parentheses.  The other version will optimize away
    // the arguments, so it can be used as long as the arguments are defined in release builds.
    // void SA_TRACE_ARGS((u32 level, c8* format, ...));    
    #define SA_TRACE_ARGS(_x_)   sa_trace _x_
    
    // void SA_TRACE_BUFFER (u32 level, PVOID buffer, u32 bytes);
    #define SA_TRACE_BUFFER             sa_trace_buffer

    // void SA_TRACE_BUFFER_OPTIONS (u32 level, PVOID buffer, u32 bytes, u32 width, u32 flags); // see SA_TRACE_FLAGS
    #define SA_TRACE_BUFFER_OPTIONS     sa_trace_buffer_options

    #define SA_TRACE_ENTRY()          SA_TRACE( DL_ENTRY, "%s", __FUNCTION__ )
    #define SA_TRACE_IN(_dl_)         SA_TRACE( _dl_, "=>%s", __FUNCTION__)
    #define SA_TRACE_EXIT()           SA_TRACE( DL_EXIT, "%s", __FUNCTION__ )
    #define SA_TRACE_OUT(_dl_)        SA_TRACE( _dl_, "<=%s", __FUNCTION__)

    #define SA_TRACE_OUT_STATUS(_dl_, _status_) SA_TRACE( _dl_, "%s status: 0x%x\n", __FUNCTION__, (_status_))
    #define SA_TRACE_STATUS(_status_)           SA_TRACE( DL_EXIT, "%s status: 0x%x\n", __FUNCTION__, (_status_))

#else

    //  -------------------------------------------------------------------
    //   Release/Free Builds
    //  -------------------------------------------------------------------    

    // void TRACE(u32 level, c8* format, ...);
    // windows supports inline varg functions.
    #if defined(SA_LINUX)         
        // Use a noop function in case someone writes if (x) SA_TRACE (..); without 
    	// proper brackets.  The noop maintains the structure of the if statement.
        #define SA_TRACE(_level_, _format_...)              sa_noop()
        #define SA_TRACE_NN(_level_, _format_...)           sa_noop()
    #elif defined(SA_MICROSOFT)
        // This will be optimized to nothing.  However, the arguments to the function
        // will be parsed so they must exist.  (See SA_TRACE_ARGS if arguments don't exist)
        #define SA_TRACE                __noop
        #define SA_TRACE_NN             __noop
    #else
        // This will be optimized to nothing.  However, the arguments to the function
        // will be parsed so they must exist.  (See SA_TRACE_ARGS if arguments don't exist)
        #define SA_TRACE                sa_trace_noop
        #define SA_TRACE_NN             sa_trace_noop
    #endif  
    
    #define TRACE   SA_TRACE

    // void SA_TRACE_DBG((u32 level, c8* format, ...));    
    #define SA_TRACE_ARGS(_x_) sa_noop ()

    // void TRACE_BUFFER (u32 level, PVOID buffer, u32 bytes);
    #define SA_TRACE_BUFFER          sa_noop_trace_buffer
    #define SA_TRACE_BUFFER_OPTIONS  sa_noop_trace_buffer_options

    #define SA_TRACE_ENTRY()   sa_noop ()
    #define SA_TRACE_IN(_dl_)  sa_noop ()
    #define SA_TRACE_EXIT()    sa_noop ()
    #define SA_TRACE_OUT(_dl_) sa_noop ()

    #define SA_TRACE_OUT_STATUS(_dl_, _status_) sa_noop ()
    #define SA_TRACE_STATUS(_status_)           sa_noop ()

#endif



// This is never defined -- It is removed by the no_inbox tags for the
// inbox driver submission.
#ifdef FAKE_INBOX 
// </no_inbox>

    // Inbox definitions:
    #define sa_trace_set_debug_level_string(_l, _str)
    #define sa_trace_set_level(_a) (0)
    #define sa_trace_get_level()   (0)
    
    /*#if DBG
        STATIC INLINE SA_TRACE (u32 level, 
                                const char* format,
                                ...)
        {
            va_list args;
            va_start (args, format);
            vDbgPrintEx (DPFLTR_IHVNETWORK_ID, DPFLTR_ERROR_LEVEL, format, args);
            va_end (args);
        }
    #else*/
        #define SA_TRACE __noop
    //#endif

    #define SA_TRACE_NN             SA_TRACE
    #define TRACE                   SA_TRACE
    #define SA_TRACE_BUFFER         __noop
    #define SA_TRACE_BUFFER_OPTIONS __noop

    #define SA_TRACE_ENTRY()
    #define SA_TRACE_IN(_dl_)
    #define SA_TRACE_EXIT()
    #define SA_TRACE_OUT(_dl_)
    #define SA_TRACE_OUT_STATUS(_dl_, _status_) 
    #define SA_TRACE_STATUS(_status_)  

// <no_inbox>
#endif
// </no_inbox>

SA_CPP_TRAILER

#endif // __sa_trace_h__

