/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_bitvector.h

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_bitvector.c,v 1.4 2005/08/05 17:29:08 menz Exp $
*/
#include "sa_bitvector.h"

PSA_BIT_VECTOR
sa_bv_allocate (u32 num_bits, u32 alloc_flags) 
{
    SA_STATUS status = 0;
    PSA_BIT_VECTOR bv = sa_malloc (sizeof (SA_BIT_VECTOR),
                                    alloc_flags,
                                    SA_TAG("bv__"));

    u32 *dw = sa_malloc (sizeof (u32) * sa_ceiling (num_bits, 32),
                          alloc_flags,
                          SA_TAG ("bvdw"));

    if (!dw || !bv) {
        status = SA_STATUS_NOT_OK;
    } else {
        status = sa_bv_create (bv, dw, num_bits);
    }

    if (status != SA_STATUS_SUCCESS) {
        sa_free (dw);
        sa_free (bv);
        return NULL;
    }

    return bv;
}

void
sa_bv_free (PSA_BIT_VECTOR bv)
{
    if (!bv) {
        return;
    }

    sa_bv_destroy (bv);

    sa_free (bv->bits);
    sa_free (bv);
}


SA_STATUS sa_bv_create (PSA_BIT_VECTOR bv, u32 *bits, u32 numBits) {
    ASSERT (bv);
    
    bv->magic = SA_BV_MAGIC;
    bv->numSet = 0;
    bv->numBits = numBits;
    bv->numDwords = sa_ceiling (numBits, 32UL);
    bv->bits = bits;

    sa_bv_clear_all (bv);

    return SA_STATUS_SUCCESS;
}

void sa_bv_destroy (PSA_BIT_VECTOR bv) { 
    bv->magic = 0;
}

void sa_bv_clear_all (PSA_BIT_VECTOR bv) {
    SA_BV_ASSERT (bv);

    bv->numSet = 0;
    sa_memset (bv->bits, 0, bv->numDwords * sizeof (u32));    
}

void sa_bv_set_all (PSA_BIT_VECTOR bv) {
    SA_BV_ASSERT (bv);

    bv->numSet = bv->numBits;
    sa_memset (bv->bits, 0xFFFFFFFF, bv->numDwords * sizeof (u32));    
}


boolean sa_bv_check (PSA_BIT_VECTOR bv, u32 bit) {

    u32  index;
    u32  shift;
    u32  value;

    SA_BV_ASSERT (bv);
    ASSERT (bit < bv->numBits);

    index = (bit >> 5);
    shift = (bit & 0x1F);
    ASSERT (index < bv->numDwords);
    
    value = bv->bits[index];
    value >>= shift;
    value &= 1;

    return (boolean)value;
}


boolean sa_bv_set (PSA_BIT_VECTOR bv, u32 bit) {

    boolean isSet;

    SA_BV_ASSERT (bv);
    ASSERT (bit < bv->numBits);

    isSet = sa_bv_check (bv, bit);
    
    if (!isSet) {        

        u32 index = (bit >> 5);
        u32 shift = (bit & 0x1F);
        u32 value;

        ASSERT (index < bv->numDwords);

        bv->numSet++;
        
        value = 1 << shift;
        bv->bits[index] |= value;
    }

    return isSet;
}

u32 sa_bv_find_clear_bits_and_set( PSA_BIT_VECTOR bv, u32 Count ) {

    u32  _index_;
    u32  NumFound = 0;

    for (_index_ = 0; (_index_ < bv->numBits); _index_++) {  
        
        if (sa_bv_check ( bv, _index_) == FALSE) {               

            ++NumFound;
            if ( NumFound == Count ) {
                //
                // We've found enough bits
                //
                u32 j;
                for ( j=NumFound; j > 0; --j ) {
                    ASSERT( sa_bv_check( bv, _index_ ) == FALSE );
                    sa_bv_set( bv, _index_ );
                    --_index_;
                }

                return  (_index_+1);
            }

        } else {
            //
            // Our run of free bits has ended
            //
            NumFound    = 0;
        }
    }    


    return  (u32)-1;
}

u32 sa_bv_find_clear_bit_from_hint (PSA_BIT_VECTOR bv, u32 hint) {

    u32 i; // dword index
    u32 index; // bit index
    u32 num_dword = sa_ceiling (bv->numBits, 32);
    u32 hint_dword = hint / 32;
    u32 hint_offset = hint % 32;

    for (i = hint_dword; i < hint_dword + num_dword; i++) {  

        u32 dw_index = (i % num_dword);

        if (bv->bits[dw_index] == 0xFFFFFFFF) {
            continue;
        }

        for (index = dw_index * 32 + hint_offset; index < (dw_index+1) * 32; index++) {
        
            if (sa_bv_check (bv, index) == FALSE) {               

                sa_bv_set (bv, index);
                
                return index;
            } 
        }

        hint_offset = 0;
    }    

    return  (u32)-1; // not found
}

u32 sa_bv_find_clear_bit_from_end (PSA_BIT_VECTOR bv) {

    u32 i; // loop counter

    for (i = 0; i < bv->numBits; i++) { 

        u32 index = bv->numBits - 1 - i;

        if (sa_bv_check (bv, index) == FALSE) {               

            sa_bv_set (bv, index);
                
            return index;
        } 
    }    

    return  (u32)-1; // not found
}

u32 sa_bv_find_clear_bit_from_start (PSA_BIT_VECTOR bv) {

    u32 i; // dword index
    u32 index; // bit index
    u32 num_dword = sa_ceiling (bv->numBits, 32);

    for (i = 0; i < num_dword; i++) {  

        if (bv->bits[i] == 0xFFFFFFFF) {
            continue;
        }

        for (index = i * 32; index < (i+1) * 32; index++) {
        
            if (sa_bv_check (bv, index) == FALSE) {               

                sa_bv_set (bv, index);
                
                return index;
            } 
        }
    }    

    return  (u32)-1; // not found
}

boolean sa_bv_clear (PSA_BIT_VECTOR bv, u32 bit) {

    boolean isSet;

    SA_BV_ASSERT (bv);
    ASSERT (bit < bv->numBits);

    isSet = sa_bv_check (bv, bit);
    
    if (isSet) {
        u32 index;
        u32 shift;
        u32 value;

        bv->numSet--;
        ASSERT ( (int)bv->numSet >= 0);        
        
        index = (bit >> 5);
        shift = (bit & 0x1F);
        ASSERT (index < bv->numDwords);
        
        value = 1 << shift;
        bv->bits[index] &= ~value;
    }

    return isSet;
}

void sa_bv_and (PSA_BIT_VECTOR result, PSA_BIT_VECTOR bv1, PSA_BIT_VECTOR bv2)
{
    u32 i, min;
    ASSERT (result->numDwords >= MAX (bv1->numDwords, bv2->numDwords));

    sa_bv_clear_all (result);

    min = MIN (bv1->numDwords, bv2->numDwords);
    for (i=0; i<min; i++)
    {
        result->bits[i] = bv1->bits[i] & bv2->bits[i];
        result->numSet += sa_num_bits_set_u32 (result->bits[i]);
    }
}

void sa_bv_or (PSA_BIT_VECTOR result, PSA_BIT_VECTOR bv1, PSA_BIT_VECTOR bv2)
{
    u32 i, min;
    ASSERT (result->numDwords >= MAX (bv1->numDwords, bv2->numDwords));

    sa_bv_clear_all (result);

    min = MIN (bv1->numDwords, bv2->numDwords);
    for (i=0; i<min; i++)
    {
        result->bits[i] = bv1->bits[i] | bv2->bits[i];
        result->numSet += sa_num_bits_set_u32 (result->bits[i]);
    }

    for (; i<bv1->numDwords; i++)
    {
        result->bits[i] = bv1->bits[i];
        result->numSet += sa_num_bits_set_u32 (result->bits[i]);
    }

    for (; i<bv2->numDwords; i++)
    {
        result->bits[i] = bv2->bits[i];
        result->numSet += sa_num_bits_set_u32 (result->bits[i]);
    }
}

void sa_bv_copy (PSA_BIT_VECTOR dest, PSA_BIT_VECTOR src)
{
    u32 i;
    ASSERT (dest->numDwords >= src->numDwords);

    sa_bv_clear_all (dest);

    for (i=0; i<src->numDwords; i++)
    {
        dest->bits[i] = src->bits[i];
    }
    dest->numSet = src->numSet;
}
