/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/****************************************************************************
 * Filename    : checksum.cpp                                               *
 * Description :                                                            *
 *                                                                          *
 ***************************************************************************/
#include "sa_system.h"
#include "sa_checksum.h"


// This function performs a TCP frame checksum. This checksum includes
// the pseudo header, TCP header, and any TCP data.
u16 sa_ipv6_tcp_checksum (
    SA_IPV6_HEADER *ip,   // ptr to IPV6 header
    SA_TCP_HEADER *tcp)   // ptr to tcp header and data
{
    u32 sum = 0;
    u16 csum;
    u32 i;
    u32 tcplen = SA_SWAP16 (ip->payload_length);
    u16* next = (u16*)&ip->src;
    u8* tcpheader = (u8*)tcp;

    // Pseudo-header
    for (i=0; i<sizeof (SA_IPV6_ADDRESS)* 2 / sizeof (u16); i++) {
        sum += *next;
        next++;
    }
    sum += SA_SWAP16 (SA_PROTOCOL_TCP);
    sum += (u16)ip->payload_length;

    while (tcplen > 1) {
        sum += *(u16 *)tcpheader;
        tcpheader++;
        tcpheader++;
        tcplen -= 2;
    }
    if (tcplen) {
        sum += *tcpheader;
    }

    while (sum >> 16) 
        sum = (sum & 0xffff) + (sum >> 16);

    csum = (u16) ~sum;

    return csum;    
}


// This function performs a TCP frame checksum. This checksum includes
// the pseudo header, TCP header, and any TCP data.
u16 sa_ipv4_tcp_checksum (
    SA_IP_HEADER *ip,     
    SA_TCP_HEADER *tcp)
{
    u32 sum = 0;
    u16 csum;
    u32 i;
    u16 tcplen;
    u16 *pseudo = (u16*)(&ip->iph_src);
    u8* tcpheader = (u8*)tcp;

    ASSERT (SA_SWAP16 (ip->iph_length) >= sa_ip_header_length (ip));
    tcplen = (u16)(SA_SWAP16 (ip->iph_length) - sa_ip_header_length (ip));

    for (i=0; i<4; i++) {
        sum += *pseudo;
        pseudo++;
    }
    sum += SA_SWAP16 (SA_PROTOCOL_TCP);    
    sum += SA_SWAP16 (tcplen);

    while (tcplen > 1) {
        sum += *(u16 *)tcpheader;
        tcpheader++;
        tcpheader++;
        tcplen -= 2;
    }
    if (tcplen) {
        sum += *tcpheader;
    }

    while (sum >> 16) 
        sum = (sum & 0xffff) + (sum >> 16);

    csum = (u16) ~sum;

    return csum;    
}

// This function performs a TCP frame checksum. This checksum includes
// the pseudo header, TCP header, and any TCP data.
u16 sa_tcp_checksum (
    u8 *pheader,     // ptr to pseudo header (only src & dst IP addresses!)
    u8 *tcpheader,   // ptr to tcp header and data
    u32 tcplen)      // tcp frame length
{
    u32 sum = 0;
    u16 csum;
    u32 i;
    for (i=0; i<4; i++) {
        sum += *(u16 *)pheader;
        pheader++;
        pheader++;
    }
    sum += SA_SWAP16 (SA_PROTOCOL_TCP);    
    sum += SA_SWAP16 ((u16)tcplen);

    while (tcplen > 1) {
        sum += *(u16 *)tcpheader;
        tcpheader++;
        tcpheader++;
        tcplen -= 2;
    }
    if (tcplen) {
        sum += *tcpheader;
    }

    while (sum >> 16) 
        sum = (sum & 0xffff) + (sum >> 16);

    csum = (u16) ~sum;

    return csum;    
}


void sa_tcp_checksum_start (
    u8 *pheader,     // ptr to pseudo header
    u8 *tcpheader,   // ptr to tcp header and data
    u32 tcplen,      // tcp frame length
    SA_CHECKSUM_INFO *info)      
{
    u32 i;

    info->len = tcplen;
    info->sum = 0;

    for (i=0; i<4; i++) {
        info->sum += *(u16 *)pheader;
        pheader++;
        pheader++;
    }
    info->sum += SA_SWAP16 (6);    

    while (tcplen > 1) {
        info->sum += *(u16 *)tcpheader;
        tcpheader++;
        tcpheader++;
        tcplen -= 2;
    }
    if (tcplen) {
        info->sum += *tcpheader;
    }
}

void sa_tcp_checksum_continue (
    u8 *tcpdata,     // ptr to tcp header and data
    u32 tcplen,      // tcp frame length
    SA_CHECKSUM_INFO *info // opaque info
    ) 
{    
    ASSERT (info);    
    
    if (tcplen == 0) {
        return;
    }

    // If previous length was odd, fix correct the sum here.
    if (info->len & 1) {
        info->sum += *tcpdata << 8;
        tcpdata++;
        tcplen--;
        info->len++;
    }

    info->len += tcplen;

    while (tcplen > 1) {
        info->sum += *(u16 *)tcpdata;
        tcpdata++;
        tcpdata++;
        tcplen -= 2;
    }

    if (tcplen) {
        info->sum += *tcpdata;
    }    
}

u16 sa_tcp_checksum_end (SA_CHECKSUM_INFO *info) // opaque info
{
    u16 csum;

    // Add the tcp length to the checksum
    info->sum += SA_SWAP16 ((u16)info->len);

    while (info->sum >> 16) 
        info->sum = (info->sum & 0xffff) + (info->sum >> 16);

    csum = (u16) ~info->sum;

    return csum;
}


//  This function performs a UDP frame checksum. This checksum includes
//  the pseudo header, UDP header, and any UDP data.
u16 sa_ipv4_udp_checksum(
    SA_IP_HEADER *ip,   
    SA_UDP_HEADER *udp)
{
    u32 sum = 0;
    u16 csum;
    u32 i;
    u16 udplen;
    u16 *pseudo = (u16*)(&ip->iph_src);
    u8* udpheader = (u8*)udp;

    ASSERT (SA_SWAP16 (ip->iph_length) >= sa_ip_header_length (ip));
    udplen = (u16) (SA_SWAP16 (ip->iph_length) - sa_ip_header_length (ip));

    for (i=0; i<4; i++) {
        sum += *pseudo;
        pseudo++;
    }
    sum += SA_SWAP16 (SA_PROTOCOL_UDP);    
    sum += SA_SWAP16 (udplen);

    while (udplen > 1) {
        sum += *(u16 *)udpheader;
        udpheader++;
        udpheader++;
        udplen -= 2;
    }
    if (udplen) {
        sum += *udpheader;
    }

    while (sum >> 16) 
        sum = (sum & 0xffff) + (sum >> 16);

    csum = (u16) ~sum;

    return csum;
}


// This function performs a UDP frame checksum. This checksum includes
// the pseudo header, UDP header, and any UDP data.
u16 sa_ipv6_udp_checksum (
    SA_IPV6_HEADER *ip,   // ptr to IPV6 header
    SA_UDP_HEADER *udp)   // ptr to udp header and data
{
    u32 sum = 0;
    u16 csum;
    u32 i;
    u32 udplen = SA_SWAP16 (ip->payload_length);
    u16* next = (u16*)&ip->src;
    u8* udpheader = (u8*)udp;

    // Pseudo-header
    for (i=0; i<sizeof (SA_IPV6_ADDRESS)* 2 / sizeof (u16); i++) {
        sum += *next;
        next++;
    }
    sum += SA_SWAP16 (SA_PROTOCOL_UDP);
    sum += (u16)ip->payload_length;

    while (udplen > 1) {
        sum += *(u16 *)udpheader;
        udpheader++;
        udpheader++;
        udplen -= 2;
    }
    if (udplen) {
        sum += *udpheader;
    }

    while (sum >> 16) 
        sum = (sum & 0xffff) + (sum >> 16);

    csum = (u16) ~sum;

    return csum;    
}


//  This function performs a UDP frame checksum. This checksum includes
//  the pseudo header, UDP header, and any UDP data.
u16 sa_udp_checksum(
    u8 *pheader,     // ptr to pseudo header (only src & dst IP addresses!)
    u8 *udpheader,   // ptr to UDP header and data
    u32 udplen)      // UDP header + data length
{
    u32 sum = 0;
    u16 csum;
    u32 i;

    for (i=0; i<4; i++) {
        sum += *(u16 *)pheader;
        pheader++;
        pheader++;
    }
    sum += SA_SWAP16(SA_PROTOCOL_UDP);
    sum += SA_SWAP16((u16)udplen);

    while (udplen > 1) {
        sum += *(u16 *)udpheader;
        udpheader++;
        udpheader++;
        udplen -= 2;
    }
    if (udplen) {
        sum += *udpheader;
    }

    while (sum >> 16) 
        sum = (sum & 0xffff) + (sum >> 16);

    csum = (u16) ~sum;

    return csum;
}

// Performs the IPV4 header checksum.
u16 sa_ipv4_checksum(
    SA_IP_HEADER *ip)
{
    u32 sum = 0;
    u16 csum;
    u32 len = sa_ip_header_length (ip);
    u8* pdata = (u8*)ip;

    while (len > 1) {
        sum += *(u16 *)pdata;
        pdata++;
        pdata++;
        len -= 2;
    }
    if (len) {
        sum += *pdata;
    }

    while (sum >> 16) 
        sum = (sum & 0xffff) + (sum >> 16);

    csum = (u16) ~sum;

    return csum;

}

// This function performs an IP header checksum on a data block.
u16 sa_ip_checksum (
    u8 *pdata,   // pointer to data to checksum
    u32 len)     // length in bytes
{
    u32 sum = 0;
    u16 csum;
    while (len > 1) {
        sum += *(u16 *)pdata;
        pdata++;
        pdata++;
        len -= 2;
    }
    if (len) {
        sum += *pdata;
    }

    while (sum >> 16) 
        sum = (sum & 0xffff) + (sum >> 16);

    csum = (u16) ~sum;

    return csum;
}




