/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
#include "sa_dev.h"
#include "sa_trace.h"
#include "sa_assert.h"
#include "sa_mem.h"
#include <asm/io.h>
#include <linux/delay.h>

SA_DEV*
sa_dev_alloc (
    SA_DEV_BAR_LOCATIONS *bars,
    u32 num_bars,
    SA_DEV_OS_HANDLE     os_handle) 
{
    SA_NOT_USED (bars);
    SA_NOT_USED (num_bars);
    SA_NOT_USED (os_handle);

    SA_BREAKPOINT_MSG ("Not supported.");
    return NULL;
}

void
sa_dev_free (
    SA_DEV *dev)
{
    SA_NOT_USED (dev);
    SA_BREAKPOINT_MSG ("Not supported."); 
}

SA_STATUS
sa_dev_create (
    SA_DEV_BAR_LOCATIONS *bars,
    u32 num_bars,
    SA_DEV_OS_HANDLE os_handle,  // device extension pointer for storport/ndis
    SA_DEV           *dev)
{
    SA_NOT_USED (os_handle);
    ASSERT (dev);
    ASSERT (bars);

    // Zero the struct
    SA_ZERO_MEM (dev);

    // Copy the bar info
    dev->magic = SA_DEV_MAGIC;
    dev->num_bars = num_bars;   
    sa_memcpy (dev->bars, bars, sizeof (SA_DEV_BAR_LOCATIONS) * num_bars);

    // Figure out the csr/pd/pci bar numbers.
    _sa_cache_bar_nums (dev);
        
    return SA_SUCCESS;
}


void
sa_dev_destroy (
    SA_DEV      *dev)
{    
    SA_DEV_ASSERT (dev);
    SA_ZERO_MEM (dev);
}

u8 sa_dev_read_u8  (SA_DEV *dev, u32 bar, u32 offset) 
{
    u8* real_offset = 0;
    u8 value = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset < dev->bars[bar].length);

    real_offset = (u8*)dev->bars[bar].base_va + offset;
    value = readb ((void*)real_offset);
    return value;
}

u16 sa_dev_read_u16 (SA_DEV *dev, u32 bar, u32 offset) 
{
    u8* real_offset = 0;
    u16 value = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset < dev->bars[bar].length);

    real_offset = (u8*)dev->bars[bar].base_va + offset;
    value = readw ((void*)real_offset);
    return value;}


u32 sa_dev_read_u32 (SA_DEV *dev, u32 bar, u32 offset) 
{
    u8* real_offset = 0;
    u32 value = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset < dev->bars[bar].length);

    real_offset = (u8*)dev->bars[bar].base_va + offset;
    value = readl ((void*)real_offset);
    return value;
}

void sa_dev_write_u8  (SA_DEV *dev, u32 bar, u32 offset, u8  value) 
{
    u8* real_offset = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset < dev->bars[bar].length);

    real_offset = (u8*)dev->bars[bar].base_va + offset;    
    writeb (value, (void*)real_offset);
}

void sa_dev_write_u16 (SA_DEV *dev, u32 bar, u32 offset, u16 value) 
{
    u8* real_offset = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset < dev->bars[bar].length);

    real_offset = (u8*)dev->bars[bar].base_va + offset;    
    writew (value, (void*)real_offset);
}

void sa_dev_write_u32 (SA_DEV *dev, u32 bar, u32 offset, u32 value) 
{
    u8* real_offset = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset < dev->bars[bar].length);

    real_offset = (u8*)dev->bars[bar].base_va + offset;    
    writel (value, (void*)real_offset);
}


SA_STATUS
sa_dev_read_buffer_u8 ( 
    SA_DEV      *dev,
    u32         bar,
    u32         offset,
    u8*         buffer, 
    u32         count) 
{
    u8* real_offset = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset + count < dev->bars[bar].length);
    ASSERT (count > 0);
    ASSERT (buffer);
    
    real_offset = (u8*)dev->bars[bar].base_va + offset;  
    memcpy_fromio (buffer, (void*)real_offset, count);

    return SA_SUCCESS;
}

SA_STATUS
sa_dev_read_buffer_u16 ( 
    SA_DEV      *dev,
    u32         bar,
    u32         offset,
    u16*        buffer, 
    u32         count) 
{
    u8* real_offset = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset + count*sizeof(u16) < dev->bars[bar].length);
    ASSERT (count > 0);
    ASSERT (buffer);
    
    real_offset = (u8*)dev->bars[bar].base_va + offset;  
    memcpy_fromio (buffer, (void*)real_offset, count*sizeof(u16));

    return SA_SUCCESS;
}

SA_STATUS
dev_read_buffer_u32 ( 
    SA_DEV      *dev,
    u32         bar,
    u32         offset,
    u32*        buffer, 
    u32         count) 
{
    u8* real_offset = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset + count*sizeof(u32) < dev->bars[bar].length);
    ASSERT (count > 0);
    ASSERT (buffer);
    
    real_offset = (u8*)dev->bars[bar].base_va + offset;  
    memcpy_fromio (buffer, (void*)real_offset, count*sizeof(u32));

    return SA_SUCCESS;
}


SA_STATUS
sa_dev_write_buffer_u8 ( 
    SA_DEV      *dev,
    u32         bar,
    u32         offset,
    u8*         buffer, 
    u32         count) 
{
    u8* real_offset = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset + count < dev->bars[bar].length);
    ASSERT (count > 0);
    ASSERT (buffer);
    
    real_offset = (u8*)dev->bars[bar].base_va + offset;  
    memcpy_toio ((void*)real_offset, buffer, count);

    return SA_SUCCESS;
}

SA_STATUS
sa_dev_write_buffer_u16 ( 
    SA_DEV      *dev,
    u32         bar,
    u32         offset,
    u16*        buffer, 
    u32         count) 
{
    u8* real_offset = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset + count*sizeof(u16) < dev->bars[bar].length);
    ASSERT (count > 0);
    ASSERT (buffer);
    
    real_offset = (u8*)dev->bars[bar].base_va + offset;  
    memcpy_toio ((void*)real_offset, buffer, count*sizeof(u16));

    return SA_SUCCESS;
}


SA_STATUS
sa_dev_write_buffer_u32 ( 
    SA_DEV      *dev,
    u32         bar,
    u32         offset,
    u32*        buffer, 
    u32         count) 
{
    u8* real_offset = 0;

    SA_DEV_ASSERT (dev);
    ASSERT (bar < dev->num_bars);
    ASSERT (offset + count*sizeof(u32) < dev->bars[bar].length);
    ASSERT (count > 0);
    ASSERT (buffer);
    
    real_offset = (u8*)dev->bars[bar].base_va + offset;  
    memcpy_toio ((void*)real_offset, buffer, count*sizeof(u32));

    return SA_SUCCESS;
}


SA_DEV_BUS_ADDRESS
sa_dev_get_bus_address (
    SA_DEV     *dev,
    PSA_DEV_BUS_ADDRESS_CONTEXT context,
    void*      va,
    u32        minimum_length) 
{   
    SA_DEV_ASSERT (dev);
    SA_NOT_USED (context);
#ifdef __VMKERNEL_MODULE__
    return virt_to_phys (va);    
#else
    return virt_to_bus (va);    
#endif
}

void
sa_dev_stall (
    SA_DEV     *dev,
    u32        us_to_stall) 
{
    udelay (us_to_stall);
}

void
sa_dev_log_error (
    SA_DEV                      *dev,
    PSA_DEV_LOG_ERROR_CONTEXT   context,    // pointer to SRB for storport
    u32                         error_code,
    u32                         unique_code) 
{
    SA_BREAKPOINT_MSG ("Not supported."); 
}


