/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_log.c

@brief
   
CVS Header Information:
   $Header: $
*/
#include "runtime/sa_log.h"
#include "system/sa_output.h"

PSA_LOG sa_log_allocate (u32 msg_bytes, u32 num_msg, u32 options)
{
    SA_STATUS status;
    PVOID buffer = NULL;
    PSA_LOG log = NULL;
    u32 bytes = (num_msg+1) * (msg_bytes + sizeof (SA_LOG_ENTRY));

    ASSERT (num_msg > 0);

    log = sa_malloc (
        sizeof (SA_LOG), 
        SA_MALLOC_NORMAL, 
        SA_TAG ("log-"));
    if (!log) {
        status = SA_STATUS_NO_SYSTEM_RESOURCES;
        goto Error;
    }
   
    buffer = sa_malloc (bytes, SA_MALLOC_NORMAL, SA_TAG ("logb"));
    if (!buffer) {
        status = SA_STATUS_NO_SYSTEM_RESOURCES;
        goto Error;
    }

    status = sa_log_create (log, buffer, bytes, msg_bytes, options);
    
Error:

    if (status != SA_SUCCESS) {
        if (buffer)
            sa_free (buffer);

        if (log)
            sa_free (log);

        log = NULL;
    }

    return log;
}

void sa_log_free (PSA_LOG log)
{    
    PVOID buffer = log->buffer;

    SA_LOG_ASSERT (log);
    sa_log_destroy (log);

    sa_free (buffer);
    sa_free (log);
}

SA_STATUS sa_log_create (PSA_LOG log, PVOID buffer, u32 buffer_bytes, u32 msg_bytes, u32 options)
{
    ASSERT (buffer);
    ASSERT (log);
    ASSERT (buffer_bytes > 0);

    sa_zero_mem (log, sizeof (SA_LOG));
    log->magic = SA_LOG_MAGIC;
    log->buffer = buffer;
    log->options = options;
    log->buffer_bytes = buffer_bytes;
    log->msg_bytes = msg_bytes;
    log->num_entries = buffer_bytes / (msg_bytes + sizeof (SA_LOG_ENTRY));

    ASSERT (log->num_entries > 0);
    sa_ring_create (&log->ring, log->num_entries, msg_bytes + sizeof (SA_LOG_ENTRY), buffer);

    log->store_category_mask = 0xFFFFFFFF;
    log->store_level_mask = 0xFFFFFFFF;

    if (!(options & SA_LOG_OPTION_NO_LOCKING)) {
        sa_init_spinlock (&log->lock);
    }

    return SA_SUCCESS;
}

void sa_log_destroy (PSA_LOG log) {
    sa_zero_mem (log, sizeof (SA_LOG));
}

u32 sa_log_set_category_mask (PSA_LOG log, u32 category_mask)
{
    u32 x;
    SA_LOG_ASSERT (log);
    
    sa_log_lock (log);
    x = log->store_category_mask;
    log->store_category_mask = category_mask;
    sa_log_unlock (log);

    return x;
}

u32 sa_log_get_category_mask (PSA_LOG log) {
    SA_LOG_ASSERT (log);
    return log->store_category_mask;
}

u32 sa_log_set_level_mask (PSA_LOG log, u32 level_mask)
{
    u32 x;
    SA_LOG_ASSERT (log);

    sa_log_lock (log);
    x = log->store_level_mask;
    log->store_level_mask = level_mask;
    sa_log_unlock (log);

    return x;
}

u32 sa_log_get_level_mask (PSA_LOG log) {
    SA_LOG_ASSERT (log);
    return log->store_level_mask;
}

void sa_log_clear (PSA_LOG log) {
    SA_LOG_ASSERT (log);

    sa_log_lock (log);
    sa_ring_clear (&log->ring);
    sa_log_unlock (log);
}

// Number of pending messages.
u32 sa_log_num_msg (PSA_LOG log) {
    SA_LOG_ASSERT (log);
    return sa_ring_num_pending (&log->ring);
}

extern const c8* sa_trace_get_debug_level_string (u32 level);

void sa_log_print (PSA_LOG log, SA_OUTPUT *out, u32 category_mask, u32 level_mask, u32 max_num)
{
    u32 index;
    u32 printed = 0;
    //u32 base_timestamp = 0;

    SA_LOG_ASSERT (log);
    sa_log_lock (log);

    if (sa_ring_num_pending (&log->ring) > 0) {

        if (sa_ring_num_pending (&log->ring) > max_num) {
            index = sa_subc (sa_ring_producer_index (&log->ring), max_num, sa_ring_num (&log->ring));
        } else {
            index = sa_ring_consumer_index (&log->ring);
        }

        for (; 
             index != sa_ring_producer_index (&log->ring) && printed < max_num;
             index = sa_addc (index, 1, sa_ring_num (&log->ring)))
        {
            PSA_LOG_ENTRY e = sa_ring_item (&log->ring, index);

            //if (!printed) {
            //    base_timestamp = e->timestamp;   
            //}

            if ((e->category & category_mask) &&
                (e->level & level_mask))
            {
                printed++;
                sa_output_printf (out,
                    "%-8s 0x%08x%08x %-8s %-8s %s", 
                    e->module_name,
                    sa_hi (e->timestamp),
                    sa_lo (e->timestamp),
                    //(e->timestamp - base_timestamp) / 1000,
                    //((e->timestamp - base_timestamp) % 1000)/10,
                    sa_log_get_category_name (log, e->category),
                    sa_trace_get_debug_level_string (e->level),
                    e->msg);
            }
        }
    }

    sa_output_printf (out, 
                      "TraceLog contains %d messages.  Print %d messages.\n", 
                      sa_ring_num_pending (&log->ring), 
                      printed);

    sa_log_unlock (log);
}

void sa_log_trace_vargs (PSA_LOG log, u32 category, u32 level, const c8* module_name, const c8* format, ... )
{    
    va_list args;
    va_start (args, format) ;
    sa_log_trace_vlist (log, category, level, module_name, format, args); 
    va_end (args);
}

void sa_log_trace_vlist (PSA_LOG log, u32 category, u32 level, const c8* module_name, const c8* format, va_list args )
{
    UNALIGNED PSA_LOG_ENTRY e = NULL;
    u64 timestamp;
    u32 len;
    u32 i;
    u32 max_len;
    c8 *buffer = NULL;  // temporary buffer to allow sprintf'ing without holding the log's lock

    if (!log) {
        // Treat this as a normal debug TRACE.  
        // This allows functions to have an optional SA_LOG* parameter to use for tracing to either
        // debug console or a user supplied log.
        sa_trace_vlist (level, module_name, format, args);
        return;
    }

    SA_LOG_ASSERT (log);

    // Check if category is masked.
    if ((sa_log_get_category_mask (log) & category) == 0) {
        return;
    }

    // Check if current level is masked.
    if ((sa_log_get_level_mask (log) & level) == 0) {
        return;
    }

    max_len = log->msg_bytes;
    
    // for the additional '\n' & NULL
    ASSERT (max_len > 2);    
    if (max_len <= 2) {
        return;
    }
    max_len -= 2;  

    buffer = sa_malloc (log->msg_bytes, SA_MALLOC_NORMAL, SA_TAG ("tmpb"));
    if (!buffer) {
        return;
    }

    len = SA_VSNPRINTF_S (buffer, log->msg_bytes, max_len, format, args);
    len = MIN (len, max_len);
    buffer[len] = 0; // terminate

    if (buffer[len - 1] != '\n') {        
        buffer[len] = '\n';  // Append a newline, if one doesn't exist already.
        buffer[len + 1] = 0;
        len++;
    }

    sa_log_lock (log);

    // Delete the old entry if necessary.
    if (sa_ring_num_empty (&log->ring) == 0) {
        sa_ring_consume (&log->ring);
    }

    e = sa_ring_producer_ptr (&log->ring);
    ASSERT (e);
    ASSERT (SA_PTR_TO_U64 (e) - SA_PTR_TO_U64 (log->buffer) < log->buffer_bytes);
    ASSERT (SA_PTR_TO_U64 (e) + sizeof (SA_LOG_ENTRY) + log->msg_bytes - SA_PTR_TO_U64 (log->buffer) <= log->buffer_bytes);
        
    i = 0;
    ASSERT (module_name);
    for (; i<sizeof (e->module_name) - 1 && module_name[i]; i++) {
        e->module_name[i] = module_name[i];
    }
    e->module_name[i] = 0; // terminate

    e->category = category;
    e->level = level;
    timestamp = sa_get_system_time(); 
    memcpy( (UNALIGNED PVOID)&(e->timestamp), (UNALIGNED PVOID)&timestamp, sizeof(u64));
    memcpy (e->msg, buffer, len+1);

    sa_ring_produce (&log->ring);

    sa_log_unlock (log);

    if (log->options & SA_LOG_OPTION_DEBUG_TRACE) {
        sa_trace_vargs (level, module_name, "%s", buffer);
    }    

    sa_free (buffer);
}

u32 sa_log_index (u32 value) {
    u32 i = sa_log2 (value);
    return MIN (SA_NUMBER_OF_FIELD(SA_LOG, cat_names) - 1, i);
}

void sa_log_set_category_name (PSA_LOG log, u32 category, const c8*name) {
    u32 i;
    u32 length;

    SA_LOG_ASSERT (log);
    ASSERT (name);
    ASSERT (sa_is_power2 (category));    

    i = sa_log2 (category);
    ASSERT (i < SA_NUMBER_OF(log->cat_names));

    length = (u32)strlen (name);
    length = MIN (sizeof (log->cat_names[0]) - 1, length);
    sa_memcpy (log->cat_names[i], name, length);
    log->cat_names[i][length] = 0; // terminate
}

//void sa_log_set_level_name (PSA_LOG log, u32 level, const c8*name) {
//    u32 i;
//    u32 length;
//    
//    SA_LOG_ASSERT (log);
//    ASSERT (name);
//    ASSERT (sa_is_power2 (level));    
//
//    i = sa_log2 (level);
//    ASSERT (i < SA_NUMBER_OF(log->level_names));
//
//    length = (u32)strlen (name);
//    length = MIN (sizeof (log->level_names[0]) - 1, length);
//    sa_memcpy (log->level_names[i], name, length);
//    log->level_names[i][length] = 0; // terminate
//}

const c8* sa_log_get_category_name (PSA_LOG log, u32 category)
{
    SA_LOG_ASSERT (log);
    return log->cat_names[sa_log_index (category)];
}

//const c8* sa_log_get_level_name (PSA_LOG log, u32 level) {
//    SA_LOG_ASSERT (log);
//    return log->level_names[sa_log_index (level)];
//}

#define SA_LOG_OUTPUT_MAGIC (0x11223344)

typedef struct _SA_LOG_OUTPUT {
    SA_OUTPUT base; // must be first in struct
    u32 magic;
    SA_LOG *log;
    u32 level;
    u32 category;
    const c8* module_name;
} SA_LOG_OUTPUT, *PSA_LOG_OUTPUT;

void SA_STDCALL sa_log_output_print (SA_OUTPUT *output, const c8* str) {
    SA_LOG_OUTPUT *out = (SA_LOG_OUTPUT*)output;
    ASSERT (out->magic == SA_LOG_OUTPUT_MAGIC);
    sa_log_trace_vargs (out->log, out->category, out->level, out->module_name, "%s", str);
}

void
sa_log_output_create (PSA_LOG log,
                      u32 category,
                      u32 level, 
                      const c8* module_name, 
                      SA_LOG_OUTPUT *out)
{
    out->base.print = sa_log_output_print;
    out->base.flush = NULL;
    out->base.vprintf = NULL;
    out->magic = SA_LOG_OUTPUT_MAGIC;
    out->log = log;
    out->category = category;
    out->level = level;
    out->module_name = module_name;
}

void
sa_log_trace_buffer_internal ( PSA_LOG log,
                      u32 category,
                      u32 level, 
                      const c8* module_name, 
                      u8* data,              // Array of data
                      u32 size,              // Length of data in bytes to print
                      u32 width,             // Width of hex numbers, 1, 2 or 4 bytes
                      u32 flags)
{
    SA_LOG_OUTPUT out;
    sa_log_output_create (log, category, level, module_name, &out);
    sa_buffer_print ((SA_OUTPUT*)&out, data, size, width, flags);
}

void
sa_log_test (void)
{
    PSA_LOG log;

    const u32 err = 1;
    const u32 warn = 2;

    const u32 good = 1;
    const u32 bad = 2;

    log = sa_log_allocate (64, 4, SA_LOG_OPTION_NONE);
    if (log) {

        sa_log_set_category_name (log, good, "good");
        sa_log_set_category_name (log, bad, "bad_is_really_too_long"); // test a long name
#ifndef __VMKERNEL__
        sa_log_trace (log, good, err, "This is a log trace number: %d Verify that the buffer is truncated to a few characters.", 1);
        sa_spin (10); // let timestamp increment
        sa_log_trace (log, good, warn, "This is a log trace number: %d", 2);
        sa_spin (1000); // let timestamp increment
        sa_log_trace (log, bad, err, "This is a log trace number: %d\n", 3);

        sa_log_dump_to_stdout (log, 0xFFFFFFFF, 0xFFFFFFFF); // 3 entries will print
        sa_log_dump_to_stdout (log, good, err); // 1 entry will print

        sa_log_trace (log, bad, err, "This is a log trace number: %d\n", 4);
        sa_log_trace (log, bad, err, "This is a log trace number: %d\n", 5);
#endif
        sa_log_dump_to_stdout (log, 0xFFFFFFFF, 0xFFFFFFFF);  // only last 4 entries will print

        sa_log_clear (log);  

        sa_log_dump_to_stdout (log, 0xFFFFFFFF, 0xFFFFFFFF);  // zero entries will print

        sa_log_trace_buffer (log, good, err, log->buffer, log->buffer_bytes/5);
        sa_log_dump_to_stdout (log, 0xFFFFFFFF, 0xFFFFFFFF);  // prints the formatted buffer.

        sa_log_free (log);
    }
}

//
// SA_LOG output object.
//
void SA_STDCALL sa_output_salog_print (SA_OUTPUT *out, const c8* outstr) {
    sa_log_trace ((SA_LOG*)out->rsvd[0],
                  SA_PTR_TO_U32 (out->rsvd[1]),
                  SA_PTR_TO_U32 (out->sa_rsvd[0]),
                  "%s",
                  outstr);
}  

void SA_STDCALL sa_output_salog_flush (SA_OUTPUT *out) {    
    // nop
}

void sa_output_salog_create (SA_OUTPUT *out, SA_LOG *log, u32 category, u32 level) 
{
    SA_LOG_ASSERT (log);

    out->print = sa_output_salog_print;
    out->flush = sa_output_salog_flush;
    out->vprintf = NULL;
    out->rsvd[0] = log;
    out->rsvd[1] = SA_U32_TO_PTR (category);        
    out->sa_rsvd[0] = SA_U32_TO_PTR (level);        
}

