/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_mem_storport.c

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/system/linuxk/sa_mem_linuxk.c,v 1.14 2006/02/28 01:13:44 menz Exp $
*/

#include "sa_types.h"
#include "sa_trace.h"
#include "sa_assert.h"
#include "sa_mem.h"
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <asm/io.h>

//    Allocates virtual memory.  This function is generally used to allocate
//    software owned (control) data structures and not for memory shared with
//    the controller.
//
//    Size        - Size in bytes to allocate
//    Flag        - SA_MALLOC_NORMAL 
//                  SA_MALLOC_PAGEABLE to allocate pageable memory
//    Tag         - Tracking tag for this memory allocation
//    context     - Opaque tracking struct to use for this allocation.
//                  Optional, but for portable code it should
//                  be supplied.
//                                                                                                      
//    Returns a pointer to newly allocated virtual memory
PVOID 
_sa_malloc_release (
    u32                   size,
    u32                   flag,
    u32                   tag,
    SA_MEM_ALLOC_CONTEXT* context)
{    
    PVOID ptr = NULL;
#ifdef __VMKERNEL_MODULE__
    int order = 0;
#endif

    SA_NOT_USED (tag);
    
    if (flag == SA_MALLOC_PAGEABLE)
    {
        ptr = vmalloc(size);

        ASSERT (context); // context required for vmalloc calls
    }
    else
    {
        ASSERT (size <= (128 * 1024));
#ifdef __VMKERNEL_MODULE__
        ASSERT(context);
        while (size > (PAGE_SIZE * (1 << order)))
            order++;

        ptr = (PVOID)__get_free_pages(GFP_ATOMIC | GFP_DMA, order);
#else
        ptr = kmalloc(size, GFP_ATOMIC | GFP_DMA); 
#endif

        // context is optional for kmalloc calls
    }

    if (context) {
        context->flags = flag;
        context->ptr = ptr;
	context->len = size;
    }

    return ptr;
}

// Frees virtual memory allocated by sa_malloc_release.  
// ptr - Pointer to be freed.  Null pointers are ignored.
void            
_sa_free_release (
    PVOID                 ptr,
    SA_MEM_ALLOC_CONTEXT* context)
{
#ifdef __VMKERNEL_MODULE__
    int order = 0;
#endif
    if (!ptr)
        return;

    // For pageable memory, the context must be provided.
    if (context) {
        ASSERT (ptr == context->ptr);

        if (context->flags == SA_MALLOC_PAGEABLE) {
            vfree (ptr);
            return;
        }
    } 
      
#ifdef __VMKERNEL_MODULE__
    ASSERT(context);
    while (context->len > (PAGE_SIZE * (1 << order)))
        order++;
    free_pages((unsigned long)ptr, order);
#else
    // Either no context, or not pageable memory. SA_MALLOC_NORMAL
    kfree (ptr);    
#endif
}


/*!
@brief
    Allocates physical memory.  This function is generally used to allocate
     memory shared with the controller.
@param    
    Size                    - Size in bytes to allocate
@param    
    HighestAcceptableAddress- The highest physical address that returned in the SGL
@return   
    Returns a pointer to newly allocated virtual memory
@note
*/
PSA_SGL      
sa_physalloc( 
    u32                  size,
    SA_PHYSICAL_ADDRESS  highest_acceptable_address )
{
    // Go with a simple implementation for now.  No need to use
    // SA for allocated huge arrays of pages at this point.
    PVOID va = sa_malloc (size, SA_MALLOC_NORMAL, SA_TAG ("phys"));
    
    SA_NOT_USED (highest_acceptable_address);
    
    if (va) {
        return sa_sgl_allocate_for_va (va, size);
    }

    return NULL;
}


/*!
@brief
    Frees physical memory allocated by sa_PhysAlloc.    
@param    
    Sgl             - Size in bytes to allocate
*/
void            
sa_physfree(  
    PSA_SGL sgl)
{
    PVOID va = sa_sgl_get_base_va (sgl);
    
    sa_sgl_free (sgl); // free the SGL struct

    ASSERT (va);    
    sa_free (va);  // free the buffer
}


// Allocates a SA_SGL, correctly sizing the PageFrame array based in the VA and
// Length parameters.
PSA_SGL
sa_sgl_allocate_for_va (
    PVOID                virtual_address,
    u32                  length    )
{
    SA_SGL *sgl = NULL;

    ASSERT (virtual_address);

    sgl = sa_malloc (sizeof (SA_SGL), SA_MALLOC_NORMAL, SA_TAG("_sgl"));

    if (sgl) {
        SA_ZERO_MEM (sgl);

        sgl->va = virtual_address;
#ifdef __VMKERNEL_MODULE__
        sgl->pa = (virtual_address == NULL) ? 0 : virt_to_phys (virtual_address);
#else
        sgl->pa = virt_to_bus (virtual_address);
#endif
        sgl->length = length;
        TRACE (DL_VERBOSE, 
               "Allocated SGL for va:0x%x pa:0x%08x_%08x length:%d", 
               virtual_address, 
               sa_hi (sgl->pa), 
               sa_lo (sgl->pa), 
               length); 
    } else {
        TRACE (DL_WARN, "Failed to allocate SGL for va:0x%x length:%d", virtual_address, length);
    }
    
    return sgl;
}

//
// Frees a SA_SGL data structure allocated by sa_SglAllocate()
//
// This guy does not free or touch page frames described by the SGL
//
// Replaces - IoFreeMdl()
//
void
sa_sgl_free(
    PSA_SGL sgl  )
{
    sa_free (sgl);
}


// This creates a virtual address mapping of the physical pages described
// by Sgl.  The target address can be either UserSpace or KernelSpace. Only
// Kernel space need be implemented on non-windows platforms for now.
//
// If a mapping already exists, it should be returned w/o creating a new mapping.
// It is up to the caller to track mappings, should nested mapping attempts occur.
PVOID
sa_sgl_map(
     PSA_SGL              sgl,
     boolean              user_space)
{
    return sgl->va;
}

// Undoes the VA mapping done by sa_SglBuild()
void
sa_sgl_unmap(
    PSA_SGL              sgl )
{
    // NOP
}


// Translates a virtual address to physical - virt_to_phys
SA_PHYSICAL_ADDRESS
sa_get_physical_address (
    PVOID virtual_address )
{
    ASSERT (virtual_address);

#ifdef __VMKERNEL_MODULE__
    return (SA_PHYSICAL_ADDRESS)(virtual_address == NULL) ? 0 : virt_to_phys (virtual_address);
#else
    return (SA_PHYSICAL_ADDRESS)virt_to_bus (virtual_address);
#endif
}

// Used to generate a mapping from a physical address to a virtual
// address.  
// This is different from sa_sgl_map in that it maps
// contiguous memory (usually only 1 page in klib) at a time and does not require an SGL
PVOID
sa_map_page(
    SA_PHYSICAL_ADDRESS  physical_address,
    SA_SIZE_T            number_of_bytes )
{
    ASSERT (physical_address);
    SA_NOT_USED (number_of_bytes);
#ifdef __VMKERNEL_MODULE__
    return phys_to_virt (physical_address);
#else
    return bus_to_virt (physical_address);
#endif
}


// Undoes the mapping done by sa_MapPage()
void
sa_unmap_page(
    PVOID                virtual_address,
    SA_SIZE_T            number_of_bytes )
{
    SA_NOT_USED (virtual_address);
    SA_NOT_USED (number_of_bytes);
    // NOP
}

// Creates a SGL for a physically contiguous chunk of memory.
// The sgl struct must be previously allocated.
SA_STATUS
sa_sgl_create_contiguous (
    PVOID                virtual_address,
    SA_PHYSICAL_ADDRESS  physical_address,
    u32                  length,
    PSA_SGL              sgl)
{
    ASSERT (virtual_address);
    ASSERT (physical_address);
    ASSERT (length > 0);
    ASSERT (sgl);

    sgl->va     = virtual_address;
    sgl->pa     = physical_address;
    sgl->length = length;

    return SA_SUCCESS;
}

void
sa_sgl_destroy_contiguous (PSA_SGL sgl) {
    ASSERT (sgl);
    SA_ZERO_MEM (sgl);
}

