/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */
/*!
@file
   sa_network_packet.c

@brief
   
CVS Header Information:
   $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/sa/runtime/sa_network_packet.c,v 1.3 2005/08/05 18:08:17 menz Exp $
*/

#include "sa_system.h"
#include "sa_network_packet.h"
#include "sa_checksum.h"
#include "sa_text.h"
#include "sa_crc.h"

void
sa_network_packet_validate(
    u8 *Data,                   // ptr to contiguous data
    u32 Size,                   // size of packet
    u16 *IPChecksum,            // place to put ip checksum
    u16 *TCPChecksum,           // place to put tcp checksum
    u16 *UDPChecksum,           // place to put udp checksum
    u32 *bcast,                 // place to put detection of bcast frame
    u32 *mcast,                 // multicast?
    u32 *ipv4type,              // ipv4 frame?
    u32 *ipv6type,              // ipv6 frame?
    u32 *tcptype,               // tcp frame?
    u32 *udptype,               // udp frame?
    u32 *tcpcksm,               // tcp checksum done?
    u32 *udpcksm,               // udp checksum done?
    u32 *isipFrag)              // is a single ip frag? 
{
    SA_ENET_HEADER *EHdr = (SA_ENET_HEADER *) Data;
    SA_SNAP_HEADER *SNAPHdr = (SA_SNAP_HEADER *) Data;
    SA_IP_HEADER *IPH = NULL;
    SA_IPV6_HEADER *IPHV6 = NULL;
    SA_UDP_HEADER *UDPH = NULL;
    SA_TCP_HEADER *TCPH = NULL;

    u16 type;
    u32 ProtOffset = 0;
    u16 udpCksm;
    u16 tcpCksm;
    u16 ipCksm;
    u32 HeaderLength;
    u32 IPDataLength;
    SA_IPV4 DAddr;
    SA_IPV4 SAddr;
    u16 Protocol;
    u16 fragOffset;

    ASSERT(ipv4type);
    ASSERT(ipv6type);
    ASSERT(tcptype);
    ASSERT(udptype);
    ASSERT(mcast);
    ASSERT(bcast);
    ASSERT(IPChecksum);
    ASSERT(TCPChecksum);
    ASSERT(UDPChecksum);
    ASSERT(Data);

    *ipv4type = 0;
    *ipv6type = 0;
    *tcptype = 0;
    *udptype = 0;
    *tcpcksm = 0;
    *udpcksm = 0;
    ipCksm = 0;
    tcpCksm = 0;
    udpCksm = 0;
    *bcast = 0;
    *mcast = 0;

    type = SA_BE16_TO_HOST (EHdr->eh_type);
    ProtOffset = sizeof(SA_ENET_HEADER);
    
    if (type >= SA_MIN_ETYPE) {
        ProtOffset += 0;
    } else {
        SNAPHdr = (SA_SNAP_HEADER *)Data;

        if (Size >= sizeof(SA_SNAP_HEADER) &&
             SNAPHdr->sh_dsap == SA_SNAP_SAP &&
             SNAPHdr->sh_ssap == SA_SNAP_SAP &&
             SNAPHdr->sh_ctl == SA_SNAP_UI) {

            type = SA_BE16_TO_HOST (SNAPHdr->sh_etype);
            ProtOffset += sizeof(SA_SNAP_HEADER);
        } else {
            // return with nothing checked - we do not care about SNAP packets
            // for now.
            return;
        }
    }

    if ((Data[0] == 0xff) && (Data[1] == 0xff) && (Data[2] == 0xff) &&
        (Data[3] == 0xff) && (Data[4] == 0xff) && (Data[5] == 0xff)) {
        *bcast = 1;
    } else if (Data[0] & 1) {
        *mcast = 1;
    }
    IPH = (SA_IP_HEADER *) ((u8 *)Data+ProtOffset);
    //Incoming packet may have a valid ethernet type 0x800 but ipheader ver_len may not be valid
    if (type == SA_ETHTYPE_IP && ((IPH->iph_verlen & SA_IP_VER_FLAG) == SA_IP_VERSION_4)){
        *ipv4type = 1;
        HeaderLength = ((IPH->iph_verlen & (u8)~SA_IP_VER_FLAG) << 2);
        
        fragOffset = SA_BE16_TO_HOST (IPH->iph_offset);

        if( (sa_ip_header_length(IPH) < sizeof(SA_IP_HEADER)) || 
            ((SA_SWAP16 (IPH->iph_length) + sizeof(SA_ENET_HEADER)) > Size)  ||
            SA_SWAP16 (IPH->iph_length) < sa_ip_header_length(IPH)){            
            return;
        }

        ipCksm = sa_ip_checksum (Data+ProtOffset, HeaderLength);

        if (HeaderLength >= sizeof(SA_IP_HEADER) && HeaderLength <= Size) {
            IPDataLength = (u32) SA_BE16_TO_HOST (IPH->iph_length);

            if ((IPH->iph_verlen & SA_IP_VER_FLAG) == SA_IP_VERSION_4 && 
                IPDataLength > sizeof(SA_IP_HEADER)) {

                IPDataLength -= HeaderLength;
                Data = (u8 *) Data + HeaderLength + ProtOffset; 

                DAddr = IPH->iph_dest;
                SAddr = IPH->iph_src;
                Protocol = IPH->iph_protocol;
                if ((sa_ip_header_flags (IPH) & SA_IPV4_FLAG_MORE_FRAGMENTS) ||
                        sa_ip_header_offset (IPH) > 0) { 
                            // No checksum computation on IP fragments.
                            *isipFrag = 1;
                }
                if (Protocol == SA_PROTOCOL_TCP) {
                    *tcptype = 1;                                 
                    //changed the mask from bfff to 3fff (1 bit more fragments + 13 bit fragoffset) 
                    if ((fragOffset & 0x3fff) == 0) {
                        TCPH = (SA_TCP_HEADER *) Data;
                        if(((u32)(TCPH->tcp_offs >> 4)* 4) <= ((u32) SA_BE16_TO_HOST (IPH->iph_length) - HeaderLength)) { 
                            tcpCksm = sa_tcp_checksum ((u8*)&IPH->iph_src, Data, IPDataLength);
                            *tcpcksm = 1;
                        }
                    }                    
                } else if (Protocol == SA_PROTOCOL_UDP) {
                    *udptype = 1;
                    if ((fragOffset & 0x3fff) == 0) {
                        *udpcksm = 1;
                        UDPH = (SA_UDP_HEADER *) Data;
                        if (UDPH->xsum != 0) {
                            udpCksm = sa_udp_checksum ((u8*)&IPH->iph_src, Data, IPDataLength);
                        } else {
                            udpCksm = 0;
                        }
                    }

                } else {
                    switch (Protocol) {
                        case SA_PROTOCOL_ICMP:
                        case SA_PROTOCOL_IGMP:
                        case SA_PROTOCOL_CUSTOM:
                            break;
                        default:
                            break;
                    }
                }
            } else {
                TRACE(DL_INFO, "Bad case - check packet");
                //We can expect ip packet with wrong TL in which case err should be set in RXentry.
                //BREAKPOINT ();
            }
        } else {
            TRACE(DL_INFO, "Bad case - check packet");
            //If pass_err bit set these packets are expected.
            BREAKPOINT ();
        }
    } else if (type == SA_ETHTYPE_IPV6) {
        *ipv6type = 1;
        IPHV6 = (SA_IPV6_HEADER *) ((u8 *)Data+ProtOffset);

        ASSERT( ( sa_swap32(IPHV6->version_class_flow) & 0xf0000000) == 0x60000000 );

        Protocol = IPHV6->next_header;
        
        if (Protocol == SA_PROTOCOL_TCP) {
            *tcptype = 1;
            *tcpcksm = 1;
            TCPH = (SA_TCP_HEADER *) (((u8 *)IPHV6)+(sizeof(SA_IPV6_HEADER)));

            if ((SA_SWAP16 (IPHV6->payload_length) + sizeof (SA_ENET_HEADER) + sizeof (SA_IPV6_HEADER) > 
                Size))
            {             
                return;
            }

            if(sa_tcp_header_length(TCPH) < sizeof(SA_TCP_HEADER))
            {          
                return;
            }

            tcpCksm = sa_ipv6_tcp_checksum(IPHV6, TCPH);
        } else if (Protocol == SA_PROTOCOL_UDP) {
            *udptype = 1;
            *udpcksm = 1;
            UDPH = (SA_UDP_HEADER *) (((u8 *)IPHV6)+(sizeof(SA_IPV6_HEADER)));
            if (UDPH->xsum != 0) {
                udpCksm = sa_ipv6_udp_checksum(IPHV6, UDPH);
            } else {
                udpCksm = 0;
            }
        } else {
            //BREAKPOINT();
        }
    }
    *IPChecksum = ipCksm;
    *UDPChecksum = udpCksm;
    *TCPChecksum = tcpCksm;
}


void sa_network_packet_debug (PVOID packet, u32 length) 
{
    u16 ipCksm, udpCksm, tcpCksm;
    u32 mcast, bcast;
    u32 ipv4type, ipv6type;
    u32 tcptype;
    u32 udptype;
    u32 tcpcksm;
    u32 udpcksm;
    u32 isipFrag;


    sa_network_packet_dump ((u8*)packet);

    sa_network_packet_validate ((u8*)packet, 
                                length, 
                                &ipCksm, 
                                &tcpCksm, 
                                &udpCksm, 
                                &bcast, 
                                &mcast,
                                &ipv4type,
                                &ipv6type,
                                &tcptype,
                                &udptype,
                                &tcpcksm,
                                &udpcksm,
                                &isipFrag);

    TRACE (DL_INFO, 
           "Normal Recv: Len %d, ipCksm %02X, udpCksm %02X, tcpCksm %02X, bcast %d, mcast %d", 
           length, 
           ipCksm, 
           udpCksm, 
           tcpCksm, 
           bcast, 
           mcast);
}

void sa_network_packet_dump (u8 *pkt)
{
    TRACE (DL_INFO, "Pkt => %02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x:%02x", 
        pkt[0], 
        pkt[1], 
        pkt[2], 
        pkt[3], 
        pkt[4], 
        pkt[5], 
        pkt[6], 
        pkt[7], 
        pkt[8], 
        pkt[9], 
        pkt[10], 
        pkt[11], 
        pkt[12], 
        pkt[13], 
        pkt[14], 
        pkt[15]);
}


void
sa_packet_parse_contiguous (
    u8* va, // pointer to the raw packet
    u32 length,
    SA_PARSED_PACKET *parsed)
{
    SA_STATUS status = 0;
    u32 offset = sizeof (SA_ENET_HEADER);
    u32 ip_length;

    parsed->bytes = length;
    parsed->eth = (SA_ENET_HEADER*)va;
    parsed->type = sa_swap16 (parsed->eth->eh_type);

    parsed->flags.broadcast = sa_mac_is_broadcast ((SA_MAC_ADDRESS*)parsed->eth->eh_daddr);
    parsed->flags.multicast = sa_mac_is_multicast ((SA_MAC_ADDRESS*)parsed->eth->eh_daddr);

    if (parsed->type == SA_ETHTYPE_VLAN) {
        
        // Offset for a VLAN tag plus the additional type field. (4 bytes)
        if (length - offset < sizeof (SA_VLAN_TYPE_TAG)) {
            status = -1;
            goto Error;
        }
        
        parsed->flags.vlan = 1;
        parsed->vlan = (SA_VLAN_TAG_TYPE*)(va + offset);
        parsed->type = sa_swap16 (parsed->vlan->type);

        offset += sizeof (SA_VLAN_TYPE_TAG);
    }

    if (parsed->type < SA_MIN_ETYPE) {

        // Require contiguous SNAP header.
        if (length - offset < sizeof (SA_SNAP_HEADER)) {
            status = -1;
            goto Error;
        }

        parsed->flags.snap = 1;
        parsed->snap = (SA_SNAP_HEADER*)(va + offset);
        offset += sizeof (SA_SNAP_HEADER);

        if  (!(parsed->snap->sh_dsap == SA_SNAP_SAP &&
               parsed->snap->sh_ssap == SA_SNAP_SAP &&
               parsed->snap->sh_ctl  == SA_SNAP_UI))
        {
            status = 0; // Okay, but stop parsing at this point.
            goto Error;
        }

        parsed->type = sa_swap16 (parsed->snap->sh_etype);
    } 
    

    if (parsed->type == SA_ETHTYPE_IP) {        

        if (offset + sizeof (SA_IP_HEADER) > length) 
        {
            status = -1;
            goto Error;
        }

        parsed->flags.ipv4 = 1;
        parsed->ipv4 = (SA_IP_HEADER*)(va + offset);        

        if (offset + sa_ip_header_length (parsed->ipv4) > length) 
        {
            status = -1;
            goto Error;
        }

        ip_length = sa_swap16 (parsed->ipv4->iph_length);
        if (ip_length + offset > parsed->bytes) {
            // bogus IP length
            parsed->flags.bad_ip_length = 1;
        }

        offset += sa_ip_header_length (parsed->ipv4);

    } else if (parsed->type == SA_ETHTYPE_IPV6) {

        if (length - offset < sizeof (SA_IPV6_HEADER)) 
        {
            status = -1;
            goto Error;
        }

        parsed->flags.ipv6 = 1;
        parsed->ipv6 = (SA_IPV6_HEADER*)(va + offset);

        ip_length = sa_swap16 (parsed->ipv6->payload_length);
        if (ip_length + offset > parsed->bytes) {
            // bogus IP length
            parsed->flags.bad_ip_length = 1;
        }

        offset += sizeof (SA_IPV6_HEADER);
    
    } else if (parsed->type == SA_ETHTYPE_ARP) {

        if (length - offset < sizeof (SA_ARP_HEADER)) 
        {
            status = -1;
            goto Error;
        }

        parsed->flags.arp = 1;
        parsed->arp = (SA_ARP_HEADER*)(va + offset);
        offset += sizeof (SA_ARP_HEADER);
    }


    if ((parsed->flags.ipv4 && parsed->ipv4->iph_protocol == SA_PROTOCOL_TCP) ||
        (parsed->flags.ipv6 && parsed->ipv6->next_header == SA_PROTOCOL_TCP)) {

        // Require contiguous TCP header
        if (length - offset < sizeof (SA_TCP_HEADER)) {
            status = -1;
            goto Error;
        }

        parsed->flags.tcp = 1;
        parsed->tcp = (SA_TCP_HEADER*)(va + offset);

        if (offset + sa_tcp_header_length (parsed->tcp) > length) 
        {
            status = -1;
            goto Error;
        }

        offset += sa_tcp_header_length (parsed->tcp);

    } else if ((parsed->flags.ipv4 && parsed->ipv4->iph_protocol == SA_PROTOCOL_UDP) ||
               (parsed->flags.ipv6 && parsed->ipv6->next_header == SA_PROTOCOL_UDP)) {

        // Require contiguous UDP header
        if (length - offset < sizeof (SA_UDP_HEADER)) {
            status = -1;
            goto Error;
        }

        parsed->flags.udp = 1;
        parsed->udp = (SA_UDP_HEADER*)(va + offset);

        offset += sizeof (SA_UDP_HEADER);
    } else if ((parsed->flags.ipv4 && parsed->ipv4->iph_protocol == SA_PROTOCOL_ICMP) ||
               (parsed->flags.ipv6 && parsed->ipv6->next_header == SA_PROTOCOL_IPV6_ICMP)) 
    {
        if (length - offset < sizeof (SA_ICMP_HEADER)) {
            status = -1;
            goto Error;
        }

        parsed->flags.icmp = 1;
        parsed->icmp = (SA_ICMP_HEADER*)(va + offset);

        switch (sa_swap16 (parsed->type)) {
            case SA_ICMP_TYPE_ECHO_REQUEST:
            case SA_ICMP_TYPE_ECHO_REPLY:

                if (length - offset < sizeof (SA_ICMP_ECHO)) {
                    status = -1;
                    goto Error;
                }

                parsed->flags.icmp_echo = 1;
                parsed->icmp_echo = (SA_ICMP_ECHO*)(va + offset);
                break;

            default:
                break;
        }
    }

    parsed->ulp = (PVOID)(va + offset);

Error:

    if (status) {
        parsed->flags.truncated = 1;
    }

}

const char*
sa_icmp_type_string (u16 type)
{
    switch (type)
    {
        case SA_ICMP_TYPE_DEST_UNREACHABLE                         : return "dest unreachable";
        case SA_ICMP_TYPE_PACKET_TOO_BIG                           : return "packet too big";
        case SA_ICMP_TYPE_TIME_EXCEEDED                            : return "time exceeded";
        case SA_ICMP_TYPE_PARAM_PROBLEM                            : return "parameter problem";
        case SA_ICMP_TYPE_PRIVATE                                  : return "private";
        case SA_ICMP_TYPE_ERROR_MSG                                : return "error msg";
        case SA_ICMP_TYPE_ECHO_REQUEST                             : return "echo request";
        case SA_ICMP_TYPE_ECHO_REPLY                               : return "echo reply";
        case SA_ICMP_TYPE_GROUP_MEMBERSHIP_QUERY                   : return "group membership query";
        case SA_ICMP_TYPE_GROUP_MEMBERSHIP_REPORT                  : return "group membership report";
        case SA_ICMP_TYPE_GROUP_MEMBERSHIP_REDUCTION               : return "group membership reduction";
        case SA_ICMP_TYPE_ROUTER_SOLICITATION                      : return "router solicitation";
        case SA_ICMP_TYPE_ROUTER_ADVERTISEMENT                     : return "router advertisement";
        case SA_ICMP_TYPE_NEIGHBOR_SOLICITATION                    : return "neighbor solicitation";
        case SA_ICMP_TYPE_NEIGHBOR_ADVERTISEMENT                   : return "neighbor advertisement";
        case SA_ICMP_TYPE_REDIRECT                                 : return "redirect";
        case SA_ICMP_TYPE_ROUTER_RENUMBERING                       : return "router renumbering";
        case SA_ICMP_TYPE_NODE_INFO_QUERY                          : return "node info query";
        case SA_ICMP_TYPE_NODE_INFO_RESPONSE                       : return "node info response";
        case SA_ICMP_TYPE_INVERSE_NEIGHBOR_DISCOVERY_SOLICITATION  : return "inverse neighbor discovery solicitation";
        case SA_ICMP_TYPE_INVERSE_NEIGHBOR_DISCOVERY_ADVERTISEMENT : return "inverse neighbor discovery advertisement";
        case SA_ICMP_TYPE_MLDV2_MULTICAST_LISTENER_REPORT          : return "MLDv2 Multicast Listener Report";
        case SA_ICMP_TYPE_HAAD_REQUEST                             : return "Home Agent Address Discovery Request";
        case SA_ICMP_TYPE_HAAD_REPLY                               : return "Home Agent Address Discovery Reply";
        case SA_ICMP_TYPE_MOBILE_PREFIX_SOLICITATION               : return "Mobile Prefix Solicitation";
        case SA_ICMP_TYPE_MOBILE_PREFIX_ADVERTISEMENT              : return "Mobile Prefix Advertisement";
        case SA_ICMP_TYPE_CERTIFICATION_PATH_SOLICITATION          : return "Certification Path Solicitation";
        case SA_ICMP_TYPE_CERTIFICATION_PATH_ADVERTISEMENT         : return "Certification Path Advertisement";
        case SA_ICMP_TYPE_EXPERIMENTAL_MOBILITY                    : return "Experimental Mobility";
        case SA_ICMP_TYPE_MRD_ADVERTISEMENT                        : return "MRD Advertisement";
        case SA_ICMP_TYPE_MRD_SOLICITATION                         : return "MRD Solicitation";
        case SA_ICMP_TYPE_MRD_TERMINATION                          : return "MRD Termination";
        case SA_ICMP_TYPE_FMIPV6                                   : return "FMIPV6";
        case SA_ICMP_TYPE_PRIVATE2                                 : return "Private";
        default:
            return "unknown";
    }
}

void
sa_print_icmp (
    SA_ICMP_HEADER *icmp,
    u32 length,
    SA_OUTPUT *out)
{
    u16 type = sa_swap16 (icmp->type);
    sa_output_printf (out, "ICMP Header\n");
    sa_output_printf (out, "\tType:0x%x (%s)\n", type, sa_icmp_type_string (type));
    sa_output_printf (out, "\tCode:0x%x\n", sa_swap16 (icmp->code));
    sa_output_printf (out, "\tChecksum:0x%x\n", sa_swap16 (icmp->checksum));

    if (length >= sizeof (SA_ICMP_ECHO) &&
        (type == SA_ICMP_TYPE_ECHO_REQUEST || type == SA_ICMP_TYPE_ECHO_REPLY)) 
    {
        SA_ICMP_ECHO *echo = (SA_ICMP_ECHO *)icmp;
        sa_output_printf (out, "\tEcho Id:0x%x\n", sa_swap16 (echo->id));
        sa_output_printf (out, "\tEcho Sequence:0x%x\n", sa_swap16 (echo->sequence));
    }
}

const char*
sa_ethtype_string (u16 type)
{
    switch (type) {
        case SA_ETHTYPE_VLAN: return "vlan";
        case SA_ETHTYPE_IP:   return "ipv4";
        case SA_ETHTYPE_IPV6: return "ipv6";
        case SA_ETHTYPE_ARP:  return "arp";
        default:
            return "unknown";
    }
}

const char*
sa_ethaddress_string (SA_MAC_ADDRESS* mac) {

    if (sa_mac_is_broadcast (mac)) {
        return "broadcast";
    } else if (sa_mac_is_multicast (mac)) {
        return "multicast";
    } else {
        return "unicast";
    }
}

void
sa_print_eth (
    SA_ENET_HEADER *eth,
    SA_OUTPUT *out)
{
    sa_output_printf (out, "Ethernet Header\n");
    sa_output_printf (out, "\tDst: " SA_MAC_FORMAT " (%s)\n", 
                      SA_MAC_ARGS (*(SA_MAC_ADDRESS*)eth->eh_daddr), 
                      sa_ethaddress_string ((SA_MAC_ADDRESS*)eth->eh_daddr));
    sa_output_printf (out, "\tSrc: " SA_MAC_FORMAT " (%s)\n", 
                      SA_MAC_ARGS (*(SA_MAC_ADDRESS*)eth->eh_saddr),
                      sa_ethaddress_string ((SA_MAC_ADDRESS*)eth->eh_saddr));
    sa_output_printf (out, "\tType: 0x%x (%s)\n", 
                      (u32)(sa_swap16 (eth->eh_type)), 
                      sa_ethtype_string (sa_swap16 (eth->eh_type)));
}

void
sa_print_vlan (
    SA_VLAN_TAG_TYPE *vlan_type,
    SA_OUTPUT *out)
{
    SA_VLAN_TAG tag;
    u16 type;

    tag.tag = sa_swap16 (vlan_type->tag.tag);
    type = sa_swap16 (vlan_type->type);

    sa_output_printf (out, "\tVLAN id:0x%x (%d)\n", tag.id, tag.id);
    sa_output_printf (out, "\tVLAN priority:0x%x\n", tag.priority);
    if (tag.canonical_format) {
        sa_output_printf (out, "\tVLAN canonical format:%d [ERROR!]\n", tag.canonical_format);
    }
    sa_output_printf (out, "\tType: 0x%x (%s)\n", (u32)(type), sa_ethtype_string (type));
    
}

void
sa_print_snap (
    SA_SNAP_HEADER *snap,
    SA_OUTPUT *out)
{
    sa_output_printf (out, "Snap Header\n");
    sa_output_printf (out, "\tDSAP/SSAP/CTL: 0x%x 0x%x 0x%x\n", 
        snap->sh_dsap, snap->sh_ssap, snap->sh_ctl);
    sa_output_printf (out, "\tProtocol: 0x%x 0x%x 0x%x\n", 
        snap->sh_protid[0], snap->sh_protid[1], snap->sh_protid[2]);    
    sa_output_printf (out, "\tType: 0x%x (%s)\n", (u32)(sa_swap16 (snap->sh_etype)), sa_ethtype_string (sa_swap16 (snap->sh_etype)));
}

const char*
sa_ip_protocol_string (u16 protocol)
{
    switch (protocol) {
        case SA_PROTOCOL_HOPBYHOP      : return "Hop-by-Hop";
        case SA_PROTOCOL_ICMP          : return "icmp";
        case SA_PROTOCOL_IGMP          : return "igmp";
        case SA_PROTOCOL_CUSTOM        : return "SE Custom";
        case SA_PROTOCOL_TCP           : return "tcp";
        case SA_PROTOCOL_UDP           : return "udp";
        case SA_PROTOCOL_IPV6          : return "ipv6";
        case SA_PROTOCOL_IPV6_ROUTING  : return "ipv6-routing";
        case SA_PROTOCOL_IPV6_FRAGMENT : return "ipv6-frag";
        case SA_PROTOCOL_ESP           : return "esp";
        case SA_PROTOCOL_AH            : return "ah";
        case SA_PROTOCOL_IPV6_ICMP     : return "ipv6-icmp";
        case SA_PROTOCOL_IPV6_NO_NEXT  : return "ipv6-NoNxt";
        case SA_PROTOCOL_IPV6_DESTOPTS : return "IPv6-Opts";
        default:
            return "unknown";
    }
}

void
sa_print_ipv4 (
    SA_IP_HEADER *ip,
    SA_OUTPUT *out)
{
    sa_output_printf (out, "IPv4 Header\n");
    sa_output_printf (out, "\tVersion: 0x%x\n", (u32)(ip->iph_verlen >> 4));
    sa_output_printf (out, "\tHeader Length: 0x%x (%d)\n", (u32)sa_ip_header_length (ip), (u32)sa_ip_header_length (ip));
    sa_output_printf (out, "\tTotal Length: 0x%x (%d)\n", (u32)sa_swap16 (ip->iph_length), (u32)sa_swap16 (ip->iph_length));
    sa_output_printf (out, "\tID: 0x%x\n", (u32)sa_swap16 (ip->iph_id));
    sa_output_printf (out, "\tOffset: 0x%x\n", (u32)sa_ip_header_offset (ip));
    sa_output_printf (out, "\tFlags: 0x%x %s %s\n", 
        (u32)sa_ip_header_flags (ip),
        (sa_ip_header_flags (ip) & SA_IPV4_FLAG_MORE_FRAGMENTS ? "[More Fragments]" : ""),
        (sa_ip_header_flags (ip) & SA_IPV4_FLAG_DONT_FRAGMENT ? "[Don't Fragment]" : ""));

    sa_output_printf (out, "\tTime to Live: 0x%x\n", (u32)ip->iph_ttl);
    sa_output_printf (out, "\tProtocol: 0x%x (%s)\n", (u32)ip->iph_protocol, sa_ip_protocol_string (ip->iph_protocol));
    sa_output_printf (out, "\tChecksum: 0x%x (%s)\n", 
                      (u32)sa_swap16 (ip->iph_xsum),
                      sa_ipv4_checksum (ip) == 0 ? "Correct" : "ERROR");
    sa_output_printf (out, "\tSrc: " SA_IP_FORMAT "\n", SA_IP_ARGS (sa_swap32 (ip->iph_src)));
    sa_output_printf (out, "\tDst: " SA_IP_FORMAT "\n", SA_IP_ARGS (sa_swap32 (ip->iph_dest)));
}

void
sa_print_ipv6 (
    SA_IPV6_HEADER *ip,
    SA_OUTPUT *out)
{
    sa_output_printf (out, "IPv6 Header\n");
    sa_output_printf (out, "\tVersion: 0x%x\n", sa_ipv6_header_version (ip));
    sa_output_printf (out, "\tClass: 0x%x\n", sa_ipv6_header_class (ip));
    sa_output_printf (out, "\tFlow: 0x%x\n", sa_ipv6_header_flow (ip));    
    sa_output_printf (out, "\tPayload Length: 0x%x (%d)\n", (u32)sa_swap16 (ip->payload_length), (u32)sa_swap16 (ip->payload_length));
    sa_output_printf (out, "\tNext Header: 0x%x (%s)\n", (u32)ip->next_header, sa_ip_protocol_string (ip->next_header));
    sa_output_printf (out, "\tHop Limit: 0x%x\n", (u32)ip->hop_limit);
    sa_output_printf (out, "\tSrc: " SA_IPV6_FORMAT "\n", SA_IPV6_ARGS (ip->src));
    sa_output_printf (out, "\tDst: " SA_IPV6_FORMAT "\n", SA_IPV6_ARGS (ip->dst));
}


const char*
sa_arp_opcode_string (u16 op)
{
    switch (op) {
        case SA_ARP_OPCODE_REQUEST: return "Request";
        case SA_ARP_OPCODE_REPLY: return "Reply";
        case SA_RARP_OPCODE_REQUEST: return "Reverse ARP";
        case SA_RARP_OPCODE_REPLY: return "Reverse Reply";
        default:
            return "unknown";
    }
}

void
sa_print_arp (
    SA_ARP_HEADER *arp,
    SA_OUTPUT *out)
{
    sa_output_printf (out, "ARP Header\n");
    sa_output_printf (out, "\tOpcode: 0x%x (%s)\n", sa_swap16 (arp->op), sa_arp_opcode_string (sa_swap16 (arp->op)));
    if (arp->hwlen != 6 || arp->prlen != 4) {
        sa_output_printf (out, "\tUnknown ARP address types.\n");
    } else {
        sa_output_printf (out, "\tSender MAC: " SA_MAC_FORMAT "\n", SA_MAC_ARGS (arp->ethernet.send_mac));
        sa_output_printf (out, "\tSender IP:  " SA_IP_FORMAT "\n", SA_IP_ARGS (sa_swap32 (arp->ethernet.send_ip)));
        sa_output_printf (out, "\tTarget MAC: " SA_MAC_FORMAT "\n", SA_MAC_ARGS (arp->ethernet.target_mac));
        sa_output_printf (out, "\tTarget IP:  " SA_IP_FORMAT "\n", SA_IP_ARGS (sa_swap32 (arp->ethernet.target_ip)));
    }
}

void
sa_print_udp (
    SA_UDP_HEADER *udp,
    SA_IP_HEADER *ipv4, // Optional, used for checksum verification
    SA_IPV6_HEADER *ipv6, // Optional, used for checksum verification
    SA_OUTPUT *out)
{
    sa_output_printf (out, "UDP Header\n");    
    sa_output_printf (out, "\tSrc Port: 0x%x\n", sa_swap16 (udp->src_port));
    sa_output_printf (out, "\tDst Port: 0x%x\n", sa_swap16 (udp->dst_port));
    sa_output_printf (out, "\tLen: 0x%x (%d)\n", sa_swap16 (udp->len), sa_swap16 (udp->len));
    sa_output_printf (out, "\tChecksum: 0x%x (%s)\n", 
                      sa_swap16 (udp->xsum),
                      (!ipv4 && !ipv6 ? "unknown - bad ip length?" :
                          (ipv4 && sa_ipv4_udp_checksum (ipv4, udp) == 0 ? "correct" : 
                          (ipv6 && sa_ipv6_udp_checksum (ipv6, udp) == 0 ? "correct" : "ERROR"))));
}

void
sa_print_tcp (
    SA_TCP_HEADER *tcp,
    SA_IP_HEADER *ipv4, // Optional, used for checksum verification
    SA_IPV6_HEADER *ipv6, // Optional, used for checksum verification
    SA_OUTPUT *out)
{
    sa_output_printf (out, "TCP Header\n");
    sa_output_printf (out, "\tSrc Port: 0x%x (%d)\n", sa_swap16 (tcp->tcp_src), sa_swap16 (tcp->tcp_src));
    sa_output_printf (out, "\tDst Port: 0x%x (%d)\n", sa_swap16 (tcp->tcp_dest), sa_swap16 (tcp->tcp_dest));
    sa_output_printf (out, "\tHeader Length: 0x%x (%d)\n", (u32)sa_tcp_header_length (tcp), (u32)sa_tcp_header_length (tcp));    
    sa_output_printf (out, "\tFlags: 0x%x [%s%s%s%s%s%s]\n", (u32)tcp->tcp_flags,
                           tcp->tcp_flags & SA_TCP_FLAG_FIN ? "FIN " : "",
                           tcp->tcp_flags & SA_TCP_FLAG_SYN ? "SYN " : "",
                           tcp->tcp_flags & SA_TCP_FLAG_RST ? "RST " : "",
                           tcp->tcp_flags & SA_TCP_FLAG_PUSH ? "PSH " : "",
                           tcp->tcp_flags & SA_TCP_FLAG_ACK ? "ACK " : "",
                           tcp->tcp_flags & SA_TCP_FLAG_URG ? "URG " : ""
                           );
    sa_output_printf (out, "\tSequence: 0x%x\n", sa_swap32 (tcp->tcp_seq));
    sa_output_printf (out, "\tACK     : 0x%x\n", sa_swap32 (tcp->tcp_ack));
    sa_output_printf (out, "\tWindow  : 0x%x\n", sa_swap16 (tcp->tcp_window));
    sa_output_printf (out, "\tUrgent  : 0x%x\n", sa_swap16 (tcp->tcp_urgent));
    sa_output_printf (out, "\tChecksum: 0x%x (%s)\n", 
                      sa_swap16 (tcp->tcp_xsum),
                      (!ipv4 && !ipv6 ? "unknown - bad ip length?" :
                          (ipv4 && sa_ipv4_tcp_checksum (ipv4, tcp) == 0 ? "correct" : 
                          (ipv6 && sa_ipv6_tcp_checksum (ipv6, tcp) == 0 ? "correct" : "ERROR"))));
    
}

void
sa_print_crc (
    u8* packet,
    u32 len,
    SA_OUTPUT *out)
{
    u32 crc_offset = len - 4;
    u32 crc = sa_crc_ethernet (packet, crc_offset);
    u32 *packet_crc = (u32*) &packet[crc_offset];

    sa_output_printf (out, "Ethernet CRC at offset 0x%x: 0x%x [Computed CRC: 0x%x]\n", crc_offset, *packet_crc, crc);
}


void
sa_print_parsed_packet (
    SA_PARSED_PACKET *parsed,
    SA_OUTPUT *out)
{
    if (parsed->eth) {
        sa_print_eth (parsed->eth, out);
    }
    if (parsed->flags.vlan) {
        sa_print_vlan (parsed->vlan, out);
    }
    if (parsed->flags.snap) {
        sa_print_snap (parsed->snap, out);
    }
    if (parsed->flags.arp) {
        sa_print_arp (parsed->arp, out);
    }
    if (parsed->flags.ipv4) {
        sa_print_ipv4 (parsed->ipv4, out);
    }
    if (parsed->flags.ipv6) {
        sa_print_ipv6 (parsed->ipv6, out);
    }
    if (parsed->flags.tcp) {
        if (parsed->flags.ipv4) {
            sa_print_tcp (parsed->tcp, parsed->flags.bad_ip_length ? NULL : parsed->ipv4, NULL, out);
        } else {
            sa_print_tcp (parsed->tcp, NULL, parsed->flags.bad_ip_length ? NULL : parsed->ipv6, out);
        }
    }
    if (parsed->flags.udp) {
        if (parsed->flags.ipv4) {
            sa_print_udp (parsed->udp, parsed->flags.bad_ip_length ? NULL : parsed->ipv4, NULL, out);
        } else {
            sa_print_udp (parsed->udp, NULL, parsed->flags.bad_ip_length ? NULL : parsed->ipv6, out);
        }
    }   
    if (parsed->flags.icmp) {
        sa_print_icmp (parsed->icmp, 
                       (parsed->flags.icmp_echo ? sizeof (SA_ICMP_ECHO) : sizeof (SA_ICMP_HEADER)), 
                       out);
    }

    if (parsed->flags.truncated) {
        sa_output_printf (out, "Warning: Packet truncated during parsing.\n");
    }

    if (parsed->eth) {
        sa_print_crc ((u8*)parsed->eth, parsed->bytes, out);
    }
}

void
sa_print_packet (
    u8* packet,
    u32 length,
    SA_OUTPUT *out)
{
    SA_PARSED_PACKET parsed;

    SA_ZERO_MEM (&parsed);

    sa_packet_parse_contiguous (packet, length, &parsed);

    sa_print_parsed_packet (&parsed, out);
}

