/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!

@file
    main.c

@brief
    Shared code entry points for TcpTx Send queue setup and takedown functions. 

@author vbridgers

@note
    vbridgers  10/24/2004    initial
    $Header: /opt/be_cvsrepo/dsn.sw.host.drivers/dsn/sw/host/drivers/beclib/tcptx_ll.c,v 1.16 2006/04/09 23:32:54 menz Exp $

*/

#include "pch.h"

void be_toe_sq_create_internal_callback (PVOID context, BESTATUS status, MCC_WRB* wrb)
{
    PBE_TOE_SQ_OBJECT tcp_sq = (PBE_TOE_SQ_OBJECT)context;
    IOCTL_TOE_WRBQ_CREATE *ioctl = NULL;

    TOE_SQ_ASSERT (tcp_sq);
    ASSERT (wrb); // must be embedded WRB
    ASSERT (wrb->embedded);

    ioctl = (IOCTL_TOE_WRBQ_CREATE*)&wrb->payload;
    ASSERT (ioctl->header.response.opcode == OPCODE_TOE_WRBQ_CREATE);
    ASSERT (ioctl->header.response.subsystem == SUBSYSTEM_TOE_WRBQ_CREATE);

    if (status == BE_SUCCESS) {
        
        // Remember the allocated cid.
        tcp_sq->cid = ioctl->params.response.cid;

        // Increment reference count for associated CQ.
        be_cq_object_reference (tcp_sq->cq_object);

        // Add to the function object for later cleanup.
        _be_function_add_toe_sq (tcp_sq->parent_function, tcp_sq);

    } else {
        
        TRACE (DL_ERR, "toe sq ioctl failed. ioctl_status:%s(0x%x) additional_status:%s(0x%x)",
               mgmt_error_codes_get_name (ioctl->header.response.status),
               ioctl->header.response.status,
               mgmt_addl_status_get_name (ioctl->header.response.additional_status),
               ioctl->header.response.additional_status);

        // Request failed.
        SA_ZERO_MEM (tcp_sq);
    }
}


/*!

@brief
    This routine creates a tcp send queue 

@param
    function_object      - handle to a function object                           
@param
    ring_size_power_of_two  - number of entries in the ring (power of two). 32K = 0, 1 = rsvd, 2 = 1, 3 = 4 ....
@param
    connection_object    - TOE connection object that this SQ will be used for.    
@param
    cq_object            - internal CQ handle returned.
@param
    pp_tcp_sq             - internal tcp sq handle returned.                          

@return
    BE_SUCCESS if successfull, , otherwise a useful error code is returned.

@note
    IRQL: < DISPATCH_LEVEL
    this function allocates a tcp_sq object

*/
BESTATUS 
be_toe_sq_create (
    IN  PBE_FUNCTION_OBJECT             function_object,
    IN  PSA_SGL                         sgl,
    IN  u32                             length,
    IN  PBE_CQ_OBJECT                   cq_object,
    OUT PBE_TOE_SQ_OBJECT               tcp_sq,
    IN  MCC_WRB_CQE_CALLBACK            callback,
    IN  PVOID                           callback_context,
    IN  PBE_TOE_SQ_CREATE_QUEUE_CONTEXT queue_context)
{
    BESTATUS status = 0;
    MCC_WRB *wrb = NULL;
    IOCTL_TOE_WRBQ_CREATE *ioctl = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    SA_TRACE_ENTRY ();

    ASSERT (sgl);
    ASSERT (tcp_sq);
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    sa_zero_mem (tcp_sq, sizeof(*tcp_sq));
    
    tcp_sq->magic             = BE_TOE_SQ_MAGIC;
    tcp_sq->parent_connection = NULL;            // don't know this yet
    tcp_sq->parent_function   = function_object;
    tcp_sq->cq_object         = cq_object;
    tcp_sq->sgl               = sgl;
    tcp_sq->size_encoding     = be_ring_length_to_encoding (length, sizeof (BE2_TOE_WRB));

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, TOE_WRBQ_CREATE);

    if (function_object->fw_cfg.ulp[0].ulp_mode & TOE_MODE) {
        ioctl->params.request.ulp_num = 0;
    } else {
        ASSERT (function_object->fw_cfg.ulp[1].ulp_mode & TOE_MODE);
        ioctl->params.request.ulp_num = 1;
    }

    ioctl->params.request.num_pages = (UEXACT16) SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);
    
    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (function_object, 
                                                              wrb, 
                                                              generic_context,
                                                              callback, 
                                                              callback_context, 
                                                              be_toe_sq_create_internal_callback, 
                                                              tcp_sq, // internal callback
                                                              ioctl);
    if (status < BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to create toe wrbq failed.");
        goto Error;
    }

Error:
    if (status < BE_SUCCESS) {
        SA_ZERO_MEM (tcp_sq);
    }

    be_unlock_wrb_post (function_object);
    return status;
}


/*!

@brief
    This routine destroys a TCP send queue     

@param    
    tcp_sq - tcp_sq handle returned from tcp_sq_create

@return   
    this function always return BE_SUCCESS.

@note
    this function frees memory allocated by tcp_sq_create for the tcp_sq object.

*/
BESTATUS 
be_toe_sq_destroy (
    IN      PBE_TOE_SQ_OBJECT        tcp_sq)
{
    PBE_FUNCTION_OBJECT function_object;
    IOCTL_TOE_WRBQ_DESTROY *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    u32 id = tcp_sq->cid;
    
    TOE_SQ_ASSERT (tcp_sq);
    
    function_object = tcp_sq->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    TRACE (DL_INFO, "Destroy TOE SQ ring id:%d", id);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in destroy TOE SQ ring.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, TOE_WRBQ_DESTROY);

    ioctl->params.request.id = id;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (
                 function_object,
                 wrb,
                 NULL,
                 NULL,
                 NULL,
                 NULL,
                 NULL,
                 ioctl);
    
    ASSERT (status != BE_PENDING);
    if (status != BE_PENDING) 
    {
        // Dereference CQ object
        be_cq_object_dereference (tcp_sq->cq_object);

        // Remove from function object
        _be_function_remove_toe_sq (function_object, tcp_sq);

        // Zero the tracking object
        SA_ZERO_MEM (tcp_sq);
    }

Error:
    
    be_unlock_wrb_post (function_object);
    return status;

}


u32
be_toe_sq_get_ring_size_encoding (IN PBE_TOE_SQ_OBJECT sq)
{
    TOE_SQ_ASSERT (sq);
    return sq->size_encoding;
}

u32
be_toe_sq_get_cid (
    IN PBE_TOE_SQ_OBJECT sq)
{
    TOE_SQ_ASSERT (sq);
    return sq->cid;
}

