/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*!

@file
    toeconn.cpp

@brief
    Shared code entry points for creating and managing TOE per-connection 
    resources such as CIDs and Send and (preposted) Receive rings.

@author jflores

@note
    jflores  10/24/2004    initial
    $Header: /proj/be/archive/dsn/sw/host/drivers/beclib/toeconn_ll.c,v 1.32 2006/04/20 23:18:54 menz Exp $

*/

#include "pch.h"


void
be_toe_rq_create_callback (PVOID context, BESTATUS status, MCC_WRB* optional_wrb)
{
    PBE_TOE_RQ_OBJECT toe_rq = (PBE_TOE_RQ_OBJECT)context;
    TOE_RQ_ASSERT (toe_rq);

    TRACE (DL_INFO, "toe rq internal callback.  status:0x%x", status);
    if (status == BE_SUCCESS) {
        
        // Increment reference count
        be_cq_object_reference (toe_rq->cq_object);

        //
        // note we don't set the ring context here since we don't have
        // all the information we need. (rq/dq/cid, etc)
        //
        _be_function_add_toe_rq (toe_rq->parent_function, toe_rq);
    }
}


/*!
@brief
    This routine creates a TOE recv queue 

@param
    FunctionObject      - Handle to a function object                           
@param
    RecvQueueVa         - Base VA for a the TOE receive ring
@param
    RingSizePowerOfTwo  - Number of entries in the ring (power of two). 64K = 0
@param
    ConnectionObject    - TOE Connection object this RQ will be used for.
@param
    CqObject            - Internal CQ Handle returned.
@param
    ppToeRq             - Internal TOE Rq Handle returned.                          

@return
    BE_SUCCESS if successfull,, otherwise a useful error code is returned.

@note
    IRQL: < DISPATCH_LEVEL
    This function allocates a ToeRq object

*/
BESTATUS 
be_toe_rq_create (
    IN  PBE_FUNCTION_OBJECT  function_object,
    IN  PSA_SGL              sgl,
    IN  u32                  length,
    IN  u32                  cid,
    IN  PBE_CQ_OBJECT        cq_object,
    OUT PBE_TOE_RQ_OBJECT    toe_rq,
    IN  MCC_WRB_CQE_CALLBACK callback,
    IN  PVOID                callback_context,
    IN  PBE_TOE_RQ_CREATE_QUEUE_CONTEXT    queue_context)
{
    BESTATUS status = 0;
    IOCTL_TOE_RQ_CREATE *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    SA_TRACE_ENTRY ();
    
    ASSERT (sgl);
    ASSERT (toe_rq);
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    SA_ZERO_MEM (toe_rq);

    toe_rq->magic             = BE_TOE_RQ_MAGIC;
    toe_rq->parent_connection = NULL; // don't know this yet
    toe_rq->parent_function   = function_object;
    toe_rq->cq_object         = cq_object;
    toe_rq->sgl               = sgl;
    toe_rq->size_encoding     = be_ring_length_to_encoding (length, sizeof(TOE_PREBUFFER));
    toe_rq->cid               = cid;

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, TOE_RQ_CREATE);

    ioctl->params.request.cid = cid;
    ioctl->params.request.num_pages = (UEXACT16) SA_PAGES_SPANNED (sa_sgl_get_offset (sgl), length);
    
    // Create a page list for the IOCTL.
    be_sgl_to_pa_list (sgl, 
                       ioctl->params.request.pages, 
                       SA_NUMBER_OF (ioctl->params.request.pages));

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (function_object, 
                                                              wrb, 
                                                              generic_context,
                                                              callback, 
                                                              callback_context, 
                                                              be_toe_rq_create_callback, 
                                                              toe_rq, // internal callback context
                                                              ioctl);
    if (status < BE_SUCCESS) {
        TRACE (DL_ERR, "ioctl to create toe rq failed.");
        goto Error;
    }

Error:

    if (status < BE_SUCCESS) {
        SA_ZERO_MEM (toe_rq);
    }

    be_unlock_wrb_post (function_object);
    return status;
}


/*!

@brief
    This routine destroys a TOE recv queue     

@param    
    toe_rq - toe_rq handle returned from toe_rq_create

@return   
    this function always return BE_SUCCESS.

@note
    this function frees memory allocated by toe_rq_create for the toe_rq object.

*/
BESTATUS 
be_toe_rq_destroy (
    IN      PBE_TOE_RQ_OBJECT        toe_rq)
{
    PBE_FUNCTION_OBJECT function_object;
    IOCTL_TOE_RQ_DESTROY *ioctl = NULL;
    MCC_WRB *wrb = NULL;
    BESTATUS status = 0;
    u32 id = toe_rq->cid;

    TOE_RQ_ASSERT (toe_rq);
    ASSERT (toe_rq->parent_connection == NULL); // destroy connection first
    
    function_object = toe_rq->parent_function;
    FUNCTION_ASSERT (function_object);

    be_lock_wrb_post (function_object);
    
    TRACE (DL_INFO, "Destroy TOE RQ ring id:%d", id);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        ASSERT (wrb);
        TRACE (DL_ERR, "No free MCC WRBs in destroy TOE RQ ring.");
        status = BE_STATUS_NO_MCC_WRB;
        goto Error;
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, TOE_RQ_DESTROY);

    ioctl->params.request.id = id;

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_internal_callback (
                 function_object,
                 wrb,
                 NULL,
                 NULL,
                 NULL,
                 NULL,
                 NULL,
                 ioctl);

    ASSERT (status != BE_PENDING);
    if (status != BE_PENDING) {
    
        be_cq_object_dereference (toe_rq->cq_object);

        _be_function_remove_toe_rq (function_object, toe_rq);

        // zero the tracking object
        SA_ZERO_MEM (toe_rq);
    }

Error:
    ASSERT (0 == status || function_object->ue);

    be_unlock_wrb_post (function_object);
    return status;
}

u32
be_toe_rq_get_ring_size_encoding (IN PBE_TOE_RQ_OBJECT rq)
{
    ASSERT (rq);
    return rq->size_encoding;
}

#ifdef SA_MICROSOFT

#pragma optimize("",off)
#pragma warning (disable:4748)

#endif
u32
be_toe_fpga_crc_calculation (u32 *dst, u32 *src , u32 da_id, u32 ifId, u16 srcPort, u16 dstPort, u32 isIpv6)
{
    u32 srcAddr[4], dstAddr, addrLength;
    u32 mask[8] = {0x1, 0x80, 0x40, 0x20, 0x10, 0x8, 0x4, 0x2}, crc = 0, seed = 0xFFFFFFFF;
    char dataStream[27] = {(char)0, (char)0, (char)0, (char)0, (char)0, (char)0, (char)0, (char)0, (char)0, (char)0,
                           (char)0, (char)0, (char)0, (char)0, (char)0, (char)0, (char)0, (char)0, (char)0, (char)0,
                           (char)0, (char)0, (char)0, (char)0, (char)0, (char)0, (char)0};
    if(isIpv6) {
        addrLength = sizeof(SA_IPV6_ADDRESS);
    } else {
        addrLength = sizeof(SA_IPV4);
    }
    sa_memcpy(&srcAddr[0],src,addrLength);

    if(!isIpv6) {

        srcAddr[0] = (u32) *src;
        dstAddr = (u32) *dst;
        // Setup the IPv4 five-tuple that is used by the CAM table lookup.
        dataStream[8]  = (char)(dstAddr >> 30);
        dataStream[9]  = (char)(dstAddr >> 22);
        dataStream[10] = (char)(dstAddr >> 14);
        dataStream[11] = (char)(dstAddr >> 6);
        dataStream[12] = (char)((dstAddr << 2) | (srcAddr[0] >> 30));
        dataStream[13] = (char)(srcAddr[0] >> 22);
        dataStream[14] = (char)(srcAddr[0] >> 14);
        dataStream[15] = (char)(srcAddr[0] >> 6);
        dataStream[16] = (char)((srcAddr[0] << 2) | (dstPort >> 14));
        dataStream[17] = (char)(dstPort >> 6);
        dataStream[18] = (char)((dstPort << 2) | (srcPort >> 14));
        dataStream[19] = (char)(srcPort >> 6);
        dataStream[20] = (char)((srcPort << 2) | (ifId >> 4));
        dataStream[21] = (char)(ifId << 4);
    } else {
        // Setup the IPv6 five-tuple that is used by the CAM table lookup.
        dataStream[0] =  (char)((da_id << 2) | (srcAddr[3] >> 30));
        dataStream[1] =  (char)(srcAddr[3] >> 22);
        dataStream[2] =  (char)(srcAddr[3] >> 14);
        dataStream[3] =  (char)(srcAddr[3] >> 6);
        dataStream[4] = (char)((srcAddr[3] << 2) | (srcAddr[2] >> 30));
        dataStream[5] =  (char)(srcAddr[2] >> 22);
        dataStream[6] =  (char)(srcAddr[2] >> 14);
        dataStream[7] =  (char)(srcAddr[2] >> 6);
        dataStream[8] = (char)((srcAddr[2] << 2) | (srcAddr[1] >> 30));
        dataStream[9]  = (char)(srcAddr[1] >> 22);
        dataStream[10] = (char)(srcAddr[1] >> 14);
        dataStream[11] = (char)(srcAddr[1] >> 6);
        dataStream[12] = (char)((srcAddr[1] << 2) | (srcAddr[0] >> 30));
        dataStream[13] = (char)(srcAddr[0] >> 22);
        dataStream[14] = (char)(srcAddr[0] >> 14);
        dataStream[15] = (char)(srcAddr[0] >> 6);
        dataStream[16] = (char)((srcAddr[0] << 2) | (dstPort >> 14));
        dataStream[17] = (char)(dstPort >> 6);
        dataStream[18] = (char)((dstPort << 2) | (srcPort >> 14));
        dataStream[19] = (char)(srcPort >> 6);
        dataStream[20] = (char)((srcPort << 2) | (ifId >> 4));
        dataStream[21] = (char)(ifId << 4);
    }
    // Compute the lower 7 bits of the CRC, since that's all we really need, using trivial Vedic math :-)
    // (CRC polynomial being used here is the Ethernet polynomial 0x04c11db7.  Look at rtl/include/crc_functions.v
    // for more details.)
#define D(x) ((dataStream[(((172-(x))/8)-(((172-(x))%8)?0:1))] & mask[((172-(x))%8)]) ? 1 : 0)
#define C(x) ((seed & (1<<(x))) ? 1 : 0)
    crc |= (D(171) ^ D(170) ^ D(168) ^ D(163) ^ D(162) ^ D(160) ^ 
            D(157) ^ D(156) ^ D(154) ^ D(153) ^ D(152) ^ D(151) ^ 
            D(149) ^ D(148) ^ D(147) ^ D(145) ^ D(144) ^ D(139) ^ 
            D(138) ^ D(137) ^ D(136) ^ D(134) ^ D(132) ^ D(131) ^ 
            D(130) ^ D(129) ^ D(127) ^ D(126) ^ D(125) ^ D(124) ^ 
            D(120) ^ D(118) ^ D(117) ^ D(114) ^ D(113) ^ D(110) ^ 
            D(108) ^ D(106) ^ D(104) ^ D(101) ^ D(100) ^ D(93) ^ 
            D(92) ^ D(88) ^ D(87) ^ D(84) ^ D(82) ^ D(80) ^ D(79) ^ 
            D(78) ^ D(75) ^ D(71) ^ D(70) ^ D(69) ^ D(68) ^ D(66) ^ 
            D(65) ^ D(64) ^ D(62) ^ D(59) ^ D(54) ^ D(53) ^ D(51) ^ 
            D(48) ^ D(47) ^ D(46) ^ D(45) ^ D(43) ^ D(42) ^ D(37) ^ 
            D(36) ^ D(35) ^ D(31) ^ D(30) ^ D(28) ^ D(27) ^ D(26) ^ 
            D(25) ^ D(23) ^ D(21) ^ D(18) ^ D(17) ^ D(16) ^ D(12) ^ 
            D(9) ^ D(8) ^ D(7) ^ D(6) ^ D(5) ^ D(4) ^ D(0) ^ C(4) ^ 
            C(5) ^ C(7) ^ C(8) ^ C(9) ^ C(11) ^ C(12) ^ C(13) ^ 
            C(14) ^ C(16) ^ C(17) ^ C(20) ^ C(22) ^ C(23) ^ C(28) ^ 
            C(30) ^ C(31));
    crc |= (D(171) ^ D(169) ^ D(164) ^ D(163) ^ D(161) ^ D(158) ^ 
            D(157) ^ D(155) ^ D(154) ^ D(153) ^ D(152) ^ D(150) ^ 
            D(149) ^ D(148) ^ D(146) ^ D(145) ^ D(140) ^ D(139) ^ 
            D(138) ^ D(137) ^ D(135) ^ D(133) ^ D(132) ^ D(131) ^ 
            D(130) ^ D(128) ^ D(127) ^ D(126) ^ D(125) ^ D(121) ^ 
            D(119) ^ D(118) ^ D(115) ^ D(114) ^ D(111) ^ D(109) ^ 
            D(107) ^ D(105) ^ D(102) ^ D(101) ^ D(94) ^ D(93) ^ 
            D(89) ^ D(88) ^ D(85) ^ D(83) ^ D(81) ^ D(80) ^ D(79) ^ 
            D(76) ^ D(72) ^ D(71) ^ D(70) ^ D(69) ^ D(67) ^ D(66) ^ 
            D(65) ^ D(63) ^ D(60) ^ D(55) ^ D(54) ^ D(52) ^ D(49) ^ 
            D(48) ^ D(47) ^ D(46) ^ D(44) ^ D(43) ^ D(38) ^ D(37) ^ 
            D(36) ^ D(32) ^ D(31) ^ D(29) ^ D(28) ^ D(27) ^ D(26) ^ 
            D(24) ^ D(22) ^ D(19) ^ D(18) ^ D(17) ^ D(13) ^ D(10) ^ 
            D(9) ^ D(8) ^ D(7) ^ D(6) ^ D(5) ^ D(1) ^ C(0) ^ C(5) ^ 
            C(6) ^ C(8) ^ C(9) ^ C(10) ^ C(12) ^ C(13) ^ C(14) ^ 
            C(15) ^ C(17) ^ C(18) ^ C(21) ^ C(23) ^ C(24) ^ C(29) ^ 
            C(31)) << 1;
    crc |= (D(170) ^ D(165) ^ D(164) ^ D(162) ^ D(159) ^ D(158) ^ 
            D(156) ^ D(155) ^ D(154) ^ D(153) ^ D(151) ^ D(150) ^ 
            D(149) ^ D(147) ^ D(146) ^ D(141) ^ D(140) ^ D(139) ^ 
            D(138) ^ D(136) ^ D(134) ^ D(133) ^ D(132) ^ D(131) ^ 
            D(129) ^ D(128) ^ D(127) ^ D(126) ^ D(122) ^ D(120) ^ 
            D(119) ^ D(116) ^ D(115) ^ D(112) ^ D(110) ^ D(108) ^ 
            D(106) ^ D(103) ^ D(102) ^ D(95) ^ D(94) ^ D(90) ^ 
            D(89) ^ D(86) ^ D(84) ^ D(82) ^ D(81) ^ D(80) ^ D(77) ^ 
            D(73) ^ D(72) ^ D(71) ^ D(70) ^ D(68) ^ D(67) ^ D(66) ^ 
            D(64) ^ D(61) ^ D(56) ^ D(55) ^ D(53) ^ D(50) ^ D(49) ^ 
            D(48) ^ D(47) ^ D(45) ^ D(44) ^ D(39) ^ D(38) ^ D(37) ^ 
            D(33) ^ D(32) ^ D(30) ^ D(29) ^ D(28) ^ D(27) ^ D(25) ^ 
            D(23) ^ D(20) ^ D(19) ^ D(18) ^ D(14) ^ D(11) ^ D(10) ^ 
            D(9) ^ D(8) ^ D(7) ^ D(6) ^ D(2) ^ C(0) ^ C(1) ^ C(6) ^ 
            C(7) ^ C(9) ^ C(10) ^ C(11) ^ C(13) ^ C(14) ^ C(15) ^ 
            C(16) ^ C(18) ^ C(19) ^ C(22) ^ C(24) ^ C(25) ^ C(30)) << 2;
    crc |= (D(171) ^ D(166) ^ D(165) ^ D(163) ^ D(160) ^ D(159) ^ 
            D(157) ^ D(156) ^ D(155) ^ D(154) ^ D(152) ^ D(151) ^ 
            D(150) ^ D(148) ^ D(147) ^ D(142) ^ D(141) ^ D(140) ^ 
            D(139) ^ D(137) ^ D(135) ^ D(134) ^ D(133) ^ D(132) ^ 
            D(130) ^ D(129) ^ D(128) ^ D(127) ^ D(123) ^ D(121) ^ 
            D(120) ^ D(117) ^ D(116) ^ D(113) ^ D(111) ^ D(109) ^ 
            D(107) ^ D(104) ^ D(103) ^ D(96) ^ D(95) ^ D(91) ^ 
            D(90) ^ D(87) ^ D(85) ^ D(83) ^ D(82) ^ D(81) ^ D(78) ^ 
            D(74) ^ D(73) ^ D(72) ^ D(71) ^ D(69) ^ D(68) ^ D(67) ^ 
            D(65) ^ D(62) ^ D(57) ^ D(56) ^ D(54) ^ D(51) ^ D(50) ^ 
            D(49) ^ D(48) ^ D(46) ^ D(45) ^ D(40) ^ D(39) ^ D(38) ^ 
            D(34) ^ D(33) ^ D(31) ^ D(30) ^ D(29) ^ D(28) ^ D(26) ^ 
            D(24) ^ D(21) ^ D(20) ^ D(19) ^ D(15) ^ D(12) ^ D(11) ^ 
            D(10) ^ D(9) ^ D(8) ^ D(7) ^ D(3) ^ C(0) ^ C(1) ^ C(2) ^ 
            C(7) ^ C(8) ^ C(10) ^ C(11) ^ C(12) ^ C(14) ^ C(15) ^ 
            C(16) ^ C(17) ^ C(19) ^ C(20) ^ C(23) ^ C(25) ^ C(26) ^ 
            C(31)) << 3;
    crc |= (D(167) ^ D(166) ^ D(164) ^ D(161) ^ D(160) ^ D(158) ^ 
            D(157) ^ D(156) ^ D(155) ^ D(153) ^ D(152) ^ D(151) ^ 
            D(149) ^ D(148) ^ D(143) ^ D(142) ^ D(141) ^ D(140) ^ 
            D(138) ^ D(136) ^ D(135) ^ D(134) ^ D(133) ^ D(131) ^ 
            D(130) ^ D(129) ^ D(128) ^ D(124) ^ D(122) ^ D(121) ^ 
            D(118) ^ D(117) ^ D(114) ^ D(112) ^ D(110) ^ D(108) ^ 
            D(105) ^ D(104) ^ D(97) ^ D(96) ^ D(92) ^ D(91) ^ D(88) ^ 
            D(86) ^ D(84) ^ D(83) ^ D(82) ^ D(79) ^ D(75) ^ D(74) ^ 
            D(73) ^ D(72) ^ D(70) ^ D(69) ^ D(68) ^ D(66) ^ D(63) ^ 
            D(58) ^ D(57) ^ D(55) ^ D(52) ^ D(51) ^ D(50) ^ D(49) ^ 
            D(47) ^ D(46) ^ D(41) ^ D(40) ^ D(39) ^ D(35) ^ D(34) ^ 
            D(32) ^ D(31) ^ D(30) ^ D(29) ^ D(27) ^ D(25) ^ D(22) ^ 
            D(21) ^ D(20) ^ D(16) ^ D(13) ^ D(12) ^ D(11) ^ D(10) ^ 
            D(9) ^ D(8) ^ D(4) ^ C(0) ^ C(1) ^ C(2) ^ C(3) ^ C(8) ^ 
            C(9) ^ C(11) ^ C(12) ^ C(13) ^ C(15) ^ C(16) ^ C(17) ^ 
            C(18) ^ C(20) ^ C(21) ^ C(24) ^ C(26) ^ C(27)) << 4;
    crc |= (D(168) ^ D(167) ^ D(165) ^ D(162) ^ D(161) ^ D(159) ^ 
            D(158) ^ D(157) ^ D(156) ^ D(154) ^ D(153) ^ D(152) ^ 
            D(150) ^ D(149) ^ D(144) ^ D(143) ^ D(142) ^ D(141) ^ 
            D(139) ^ D(137) ^ D(136) ^ D(135) ^ D(134) ^ D(132) ^ 
            D(131) ^ D(130) ^ D(129) ^ D(125) ^ D(123) ^ D(122) ^ 
            D(119) ^ D(118) ^ D(115) ^ D(113) ^ D(111) ^ D(109) ^ 
            D(106) ^ D(105) ^ D(98) ^ D(97) ^ D(93) ^ D(92) ^ D(89) ^ 
            D(87) ^ D(85) ^ D(84) ^ D(83) ^ D(80) ^ D(76) ^ D(75) ^ 
            D(74) ^ D(73) ^ D(71) ^ D(70) ^ D(69) ^ D(67) ^ D(64) ^ 
            D(59) ^ D(58) ^ D(56) ^ D(53) ^ D(52) ^ D(51) ^ D(50) ^ 
            D(48) ^ D(47) ^ D(42) ^ D(41) ^ D(40) ^ D(36) ^ D(35) ^ 
            D(33) ^ D(32) ^ D(31) ^ D(30) ^ D(28) ^ D(26) ^ D(23) ^ 
            D(22) ^ D(21) ^ D(17) ^ D(14) ^ D(13) ^ D(12) ^ D(11) ^ 
            D(10) ^ D(9) ^ D(5) ^ C(1) ^ C(2) ^ C(3) ^ C(4) ^ C(9) ^ 
            C(10) ^ C(12) ^ C(13) ^ C(14) ^ C(16) ^ C(17) ^ C(18) ^ 
            C(19) ^ C(21) ^ C(22) ^ C(25) ^ C(27) ^ C(28)) << 5;
    crc |= (D(171) ^ D(170) ^ D(169) ^ D(166) ^ D(159) ^ D(158) ^ 
            D(156) ^ D(155) ^ D(152) ^ D(150) ^ D(149) ^ D(148) ^ 
            D(147) ^ D(143) ^ D(142) ^ D(140) ^ D(139) ^ D(135) ^ 
            D(134) ^ D(133) ^ D(129) ^ D(127) ^ D(125) ^ D(123) ^ 
            D(119) ^ D(118) ^ D(117) ^ D(116) ^ D(113) ^ D(112) ^ 
            D(108) ^ D(107) ^ D(104) ^ D(101) ^ D(100) ^ D(99) ^ 
            D(98) ^ D(94) ^ D(92) ^ D(90) ^ D(87) ^ D(86) ^ D(85) ^ 
            D(82) ^ D(81) ^ D(80) ^ D(79) ^ D(78) ^ D(77) ^ D(76) ^ 
            D(74) ^ D(72) ^ D(69) ^ D(66) ^ D(64) ^ D(62) ^ D(60) ^ 
            D(57) ^ D(52) ^ D(49) ^ D(47) ^ D(46) ^ D(45) ^ D(41) ^ 
            D(35) ^ D(34) ^ D(33) ^ D(32) ^ D(30) ^ D(29) ^ D(28) ^ 
            D(26) ^ D(25) ^ D(24) ^ D(22) ^ D(21) ^ D(17) ^ D(16) ^ 
            D(15) ^ D(14) ^ D(13) ^ D(11) ^ D(10) ^ D(9) ^ D(8) ^ 
            D(7) ^ D(5) ^ D(4) ^ D(0) ^ C(0) ^ C(2) ^ C(3) ^ C(7) ^ 
            C(8) ^ C(9) ^ C(10) ^ C(12) ^ C(15) ^ C(16) ^ C(18) ^ 
            C(19) ^ C(26) ^ C(29) ^ C(30) ^ C(31)) << 6;    
#undef D
#undef C

    return crc;
}
#ifdef SA_MICROSOFT

#pragma optimize("",on) // added so as so build free drivers.
#pragma warning (default:4748) 

#endif

BESTATUS 
be_toe_cxn_create (
    IN  PBE_FUNCTION_OBJECT       function_object,
    IN  PBE_TOE_DQ_OBJECT         dq,
    IN  PBE_TOE_SQ_OBJECT         sq,
    IN  PBE_TOE_RQ_OBJECT         rq,
    OUT PBE_TOE_CONNECTION_OBJECT cxn)
{
    FUNCTION_ASSERT (function_object);
    TOE_DQ_ASSERT (dq);
    TOE_SQ_ASSERT (sq);
    if (rq) { TOE_RQ_ASSERT (rq); }
    ASSERT (cxn);

    // Zero the cxn
    SA_ZERO_MEM (cxn);

    // Create the cxn object
    cxn->magic = BE_TOE_CONNECTION_MAGIC;
    cxn->dq = dq;
    cxn->rq = rq;
    cxn->sq = sq;
    cxn->parent_function = function_object;

    // Update offload counts.
    if (rq) {
        ASSERT (rq->offloaded == FALSE);
        rq->offloaded = TRUE;
    }    
    ASSERT (sq->offloaded == FALSE);    
    sq->offloaded = TRUE;
    sa_atomic_increment (&dq->num_offloads);

    // Insert into the function
    _be_function_add_toe_connection (function_object, cxn);

    return BE_SUCCESS;
}

u32 
be_toe_cxn_get_cid (
    IN      PBE_TOE_CONNECTION_OBJECT connection_object)
{
    TOE_CXN_ASSERT (connection_object);
    return be_toe_sq_get_cid (connection_object->sq);
}

BESTATUS 
be_toe_cxn_destroy (
    IN      PBE_TOE_CONNECTION_OBJECT     cxn)
{
    TOE_CXN_ASSERT (cxn);

    // Update offload counts.
    if (cxn->rq) {
        ASSERT (cxn->rq->offloaded == TRUE);
        cxn->rq->offloaded = FALSE;
    }

    ASSERT (cxn->sq->offloaded == TRUE);    
    cxn->sq->offloaded = FALSE;

    ASSERT (cxn->dq->num_offloads > 0);
    sa_atomic_decrement (&cxn->dq->num_offloads);

    // Remove from  the function object
    _be_function_remove_toe_connection (cxn->parent_function, cxn);

    SA_ZERO_MEM (cxn);

    return BE_SUCCESS;
}

BESTATUS BECALL
be_toe_cxn_offload_ex (
    IN  PBE_TOE_SQ_OBJECT             sq,
    IN  PBE_TOE_RQ_OBJECT             rq,           // optional
    IN  PBE_TOE_DQ_OBJECT             dq,
    IN  PBE_CQ_OBJECT                 dq_cq,
    IN  PBE_RXIF_OBJECT               rxif_object,
    IN  IOCTL_TOE_CONNECTION_OFFLOAD* va_for_ioctl, // TEMPLATE_HEADER & TCP_CONTEXT must be filled in
    IN  u64                           pa_for_ioctl,
    IN  u32                           rx_delegated_data_bytes,
    IN  u32                           tx_delegated_data_bytes,
    IN  u32                           global_flow_id,
    IN  u32                           local_flow_id,
    IN  u32                           default_mode,
    IN  u32                           disable_repeat_wrb,
    OUT PBE_TOE_CONNECTION_OBJECT     cxn,
    IN  MCC_WRB_CQE_CALLBACK          callback,
    IN  PVOID                         callback_context,
    IN  PBE_TOE_OFFLOAD_QUEUE_CONTEXT queue_context,
    IN  u32                           ka_probe_count,
    IN  u32                           ka_timeout_delta)
{
    BESTATUS status = 0;
    MCC_WRB* wrb = NULL;
    TOE_CONTEXT *context = NULL;
    BE_FUNCTION_OBJECT *function_object = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    TOE_SQ_ASSERT (sq);
    if (rq) { TOE_RQ_ASSERT (rq); }
    TOE_DQ_ASSERT (dq);
    CQ_ASSERT (dq_cq);
    ASSERT (va_for_ioctl);
    ASSERT (pa_for_ioctl);
    ASSERT (cxn);

    function_object = sq->parent_function;
    be_lock_wrb_post (function_object);

    // Zero the cxn    
    status = be_toe_cxn_create (function_object, dq, sq, rq, cxn);
    if (status != BE_SUCCESS) {
        goto Error;
    }

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    TRACE (DL_INFO, "IOCTL_TOE_CONNECTION_OFFLOAD:");
    SA_TRACE_BUFFER (DL_INFO, va_for_ioctl, sizeof (*va_for_ioctl));

    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object,
                                           wrb, 
                                           va_for_ioctl, 
                                           pa_for_ioctl, 
                                           TOE_CONNECTION_OFFLOAD,
                                           0);        
    /* Program the KA timeout and probe count */
    va_for_ioctl->params.request.ka_probe_count = ka_probe_count;
    va_for_ioctl->params.request.ka_timeout_delta = ka_timeout_delta;
    va_for_ioctl->params.request.dq_id = be_toe_dq_get_id (dq);
    // Get the cid from the previously allocate SQ.
    va_for_ioctl->params.request.cid = be_toe_sq_get_cid (sq);

    // The ulp id must be a valid TOE chute.  The ulp_id is 1 based, i.e. 1,2,3.
    ASSERT (va_for_ioctl->params.request.tcp.ulp_id-1 < SA_NUMBER_OF (function_object->fw_cfg.ulp));
    ASSERT (function_object->fw_cfg.ulp[va_for_ioctl->params.request.tcp.ulp_id-1].ulp_mode & TOE_MODE);

    // We fill in portions of TOE context
    context = &va_for_ioctl->params.request.context;
    SA_ZERO_MEM (context);    

    context->fetch_wrb         = TRUE; // Consel should fetch WRB on a driver request 
    context->pci_func_id       = be_function_get_function_number(function_object);
    context->rx_pci_func       = be_function_get_function_number(function_object);
    context->ctx_valid         = 1;
    context->pd_id             = va_for_ioctl->params.request.tcp.pd_id;
    context->tx_ring_size      = be_toe_sq_get_ring_size_encoding (sq);
    context->local_flow_id     = local_flow_id;
    context->global_flow_id    = global_flow_id;
    context->tx_delegated      = tx_delegated_data_bytes > 0 ? 1 : 0;

    context->cq_id_def         = be_cq_get_id (dq_cq);                             // Default receive CQ
    context->cq_id_send        = be_cq_get_id (sq->cq_object);                     // Send CQ
    context->rx_preposted_sequence = va_for_ioctl->params.request.tcp.rcv_nxt - rx_delegated_data_bytes;
    context->previous_state    = TOERX_STATE_DEFAULT_BUFFERS;
    context->current_state     = TOERX_STATE_DEFAULT_BUFFERS;
	context->default_ring_rid  = be_toe_dq_get_id (dq); //TODO: Need to make sure this is default ring rid (relative) not cid.
	context->default_buff_size  = sa_log2 (be_toe_dq_get_buffer_length (dq)) - 11; // previously == 11

    context->stake_sequence    = va_for_ioctl->params.request.tcp.snd_nxt;
    context->rx_ring_size      = (rq ? be_toe_rq_get_ring_size_encoding (rq) : 0);
    context->cq_id_recv        = (rq ? be_cq_get_id (rq->cq_object) : 0xFFFFFFFF); // Preposted receive CQ
    
    context->rx_pdid           = va_for_ioctl->params.request.tcp.pd_id;
    context->rx_pdid_valid     = 1; 
    
    context->cur_seq           = va_for_ioctl->params.request.tcp.snd_nxt;
    context->data_bytes_posted = rx_delegated_data_bytes;
    context->stake_valid       = 0;
	context->t_maxseg          = va_for_ioctl->params.request.tcp.t_maxseg;
    context->default_mode      = default_mode;
	context->disable_repeat_wrb = disable_repeat_wrb;
	
    // The length of the template header in TCP context does not include the beginning 2 pad bytes
    // in the struct.  Therefore, I add 2 before comparing the size.
    switch (va_for_ioctl->params.request.tcp.hdr_len+2)
    {
        case sizeof (TEMPLATE_HEADER):
        case sizeof (TEMPLATE_HEADER_TS):
            va_for_ioctl->params.request.filter.ipv4.src_addr = sa_swap32(va_for_ioctl->params.request.header.ethernetII.Ip.dst);
            va_for_ioctl->params.request.filter.ipv4.dst_addr = sa_swap32(va_for_ioctl->params.request.header.ethernetII.Ip.src);
            va_for_ioctl->params.request.filter.ipv4.src_port = sa_swap16((u16)va_for_ioctl->params.request.header.ethernetII.Tcp.srcPort);
            va_for_ioctl->params.request.filter.ipv4.dst_port = sa_swap16((u16)va_for_ioctl->params.request.header.ethernetII.Tcp.dstPort);
            break;
        case sizeof (TEMPLATE_VLAN_HEADER):
        case sizeof (TEMPLATE_VLAN_HEADER_TS):
            va_for_ioctl->params.request.filter.ipv4.src_addr = sa_swap32(va_for_ioctl->params.request.header.vlan_ethernetII.Ip.dst);
            va_for_ioctl->params.request.filter.ipv4.dst_addr = sa_swap32(va_for_ioctl->params.request.header.vlan_ethernetII.Ip.src);
            va_for_ioctl->params.request.filter.ipv4.src_port = sa_swap16((u16)va_for_ioctl->params.request.header.vlan_ethernetII.Tcp.srcPort);
            va_for_ioctl->params.request.filter.ipv4.dst_port = sa_swap16((u16)va_for_ioctl->params.request.header.vlan_ethernetII.Tcp.dstPort);
            break;
        case sizeof (TEMPLATE_VLAN_IPV6_HEADER):
        case sizeof (TEMPLATE_VLAN_IPV6_HEADER_TS):
            va_for_ioctl->params.request.filter.ipv6.src_addr[0] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.dst[3]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[1] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.dst[2]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[2] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.dst[1]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[3] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.dst[0]);

            va_for_ioctl->params.request.filter.ipv6.dst_addr[0] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.src[3]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[1] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.src[2]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[2] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.src[1]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[3] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.src[0]);

            va_for_ioctl->params.request.filter.ipv6.src_port = sa_swap16((u16)va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Tcp.srcPort);
            va_for_ioctl->params.request.filter.ipv6.dst_port = sa_swap16((u16)va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Tcp.dstPort);
            break;
        case sizeof (TEMPLATE_IPV6_HEADER):
        case sizeof (TEMPLATE_IPV6_HEADER_TS):
            va_for_ioctl->params.request.filter.ipv6.src_addr[0] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.dst[3]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[1] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.dst[2]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[2] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.dst[1]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[3] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.dst[0]);

            va_for_ioctl->params.request.filter.ipv6.dst_addr[0] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.src[3]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[1] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.src[2]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[2] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.src[1]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[3] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.src[0]);

            va_for_ioctl->params.request.filter.ipv6.src_port = sa_swap16((u16)va_for_ioctl->params.request.header.ipv6_ethernetII.Tcp.srcPort);
            va_for_ioctl->params.request.filter.ipv6.dst_port = sa_swap16((u16)va_for_ioctl->params.request.header.ipv6_ethernetII.Tcp.dstPort);            
            break;

        // SNAP headers are not supported!!!!!!!
        default:
            TRACE (DL_ERR, "Illegal template header length: %d", va_for_ioctl->params.request.tcp.hdr_len);
            BREAKPOINT ();
            break;
    }

    // This is the interface id, not the MAC id.  The interface may have multiple MACs.
    // All MACs for an interface will match, which is a rare, yet potential, problem (e.g. 2 VMs
    // with different MACs and same IP address)
    va_for_ioctl->params.request.filter.ipv4.mac_id  = rxif_object->handle;

    // Both port 0 & 1 packets are set to the same RXULP.  
    // If the same MAC is programmed on both ports, this allows teaming our 2 ports with TOE.
    // Most likely different MACs are programmed on the 2 ports, so packets will only arrive on
    // the port for this PCI function -- the opposite port's packets are dropped due to MAC.
    va_for_ioctl->params.request.filter.ipv4.ct0     = va_for_ioctl->params.request.tcp.ulp_id;
    va_for_ioctl->params.request.filter.ipv4.ct1     = va_for_ioctl->params.request.tcp.ulp_id;
    va_for_ioctl->params.request.filter.ipv4.valid   = 1;   

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (function_object, wrb, generic_context, callback, callback_context, va_for_ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "toe offload ioctl failed.");
        goto Error;
    }
    
Error:

    be_unlock_wrb_post (function_object);
    return status;
}

BESTATUS BECALL
be_toe_cxn_offload (
    IN  PBE_TOE_SQ_OBJECT             sq,
    IN  PBE_TOE_RQ_OBJECT             rq,           // optional
    IN  PBE_TOE_DQ_OBJECT             dq,
    IN  PBE_CQ_OBJECT                 dq_cq,
    IN  PBE_RXIF_OBJECT               rxif_object,
    IN  IOCTL_TOE_CONNECTION_OFFLOAD* va_for_ioctl, // TEMPLATE_HEADER & TCP_CONTEXT must be filled in
    IN  u64                           pa_for_ioctl,
    IN  u32                           rx_delegated_data_bytes,
    OUT PBE_TOE_CONNECTION_OBJECT     cxn,
    IN  MCC_WRB_CQE_CALLBACK          callback,
    IN  PVOID                         callback_context,
    IN  PBE_TOE_OFFLOAD_QUEUE_CONTEXT queue_context)
{
    BESTATUS status = 0;
    MCC_WRB* wrb = NULL;
    TOE_CONTEXT *context = NULL;
    BE_FUNCTION_OBJECT *function_object = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    TOE_SQ_ASSERT (sq);
    if (rq) { TOE_RQ_ASSERT (rq); }
    TOE_DQ_ASSERT (dq);
    CQ_ASSERT (dq_cq);
    ASSERT (va_for_ioctl);
    ASSERT (pa_for_ioctl);
    ASSERT (cxn);

    function_object = sq->parent_function;
    be_lock_wrb_post (function_object);

    // Zero the cxn    
    status = be_toe_cxn_create (function_object, dq, sq, rq, cxn);
    if (status != BE_SUCCESS) {
        goto Error;
    }

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    TRACE (DL_INFO, "IOCTL_TOE_CONNECTION_OFFLOAD:");
    SA_TRACE_BUFFER (DL_INFO, va_for_ioctl, sizeof (*va_for_ioctl));

    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, 
                                           wrb, 
                                           va_for_ioctl, 
                                           pa_for_ioctl, 
                                           TOE_CONNECTION_OFFLOAD,
                                           0);        
    
    // Get the cid from the previously allocate SQ.
    va_for_ioctl->params.request.cid = be_toe_sq_get_cid (sq);

    // The ulp id must be a valid TOE chute.  The ulp_id is 1 based, i.e. 1,2,3.
    ASSERT (va_for_ioctl->params.request.tcp.ulp_id-1 < SA_NUMBER_OF (function_object->fw_cfg.ulp));
    ASSERT (function_object->fw_cfg.ulp[va_for_ioctl->params.request.tcp.ulp_id-1].ulp_mode & TOE_MODE);

    // We fill in portions of TOE context
    context = &va_for_ioctl->params.request.context;
    SA_ZERO_MEM (context);    

    context->fetch_wrb         = TRUE; // Consel should fetch WRB on a driver request 
    context->pci_func_id       = be_function_get_function_number (function_object);
    context->ctx_valid         = 1;
    context->pd_id             = va_for_ioctl->params.request.tcp.pd_id;
    context->tx_ring_size      = be_toe_sq_get_ring_size_encoding (sq);
    // HACK for testing QOS
    context->local_flow_id     = 0;//flow_id & 0x7;
    context->global_flow_id     = 0;//flow_id & 0x7;
    //flow_id++;
    

    context->cq_id_def         = be_cq_get_id (dq_cq);                             // Default receive CQ
    context->cq_id_send        = be_cq_get_id (sq->cq_object);                     // Send CQ
    context->rx_preposted_sequence = va_for_ioctl->params.request.tcp.rcv_nxt - rx_delegated_data_bytes;
    context->previous_state    = TOERX_STATE_DEFAULT_BUFFERS;
    context->current_state     = TOERX_STATE_DEFAULT_BUFFERS;
    context->default_ring_rid  = be_toe_dq_get_id (dq); //TODO: Need to make sure this is default ring rid (relative) not cid.
    context->default_buff_size  = sa_log2 (be_toe_dq_get_buffer_length (dq)) - 11; // previously == 11

    context->stake_sequence    = va_for_ioctl->params.request.tcp.snd_nxt;
    context->rx_ring_size      = (rq ? be_toe_rq_get_ring_size_encoding (rq) : 0);
    context->cq_id_recv        = (rq ? be_cq_get_id (rq->cq_object) : 0xFFFFFFFF); // Preposted receive CQ
    
    context->rx_pci_func       = be_function_get_function_number (function_object);
    context->rx_pdid           = va_for_ioctl->params.request.tcp.pd_id;
    context->rx_pdid_valid     = 1; 
    
    context->cur_seq           = va_for_ioctl->params.request.tcp.snd_nxt;
    context->data_bytes_posted = rx_delegated_data_bytes;
    context->t_maxseg          = va_for_ioctl->params.request.tcp.t_maxseg;

    // The length of the template header in TCP context does not include the beginning 2 pad bytes
    // in the struct.  Therefore, I add 2 before comparing the size.
    switch (va_for_ioctl->params.request.tcp.hdr_len+2)
    {
        case sizeof (TEMPLATE_HEADER):
        case sizeof (TEMPLATE_HEADER_TS):
            va_for_ioctl->params.request.filter.ipv4.src_addr = sa_swap32(va_for_ioctl->params.request.header.ethernetII.Ip.dst);
            va_for_ioctl->params.request.filter.ipv4.dst_addr = sa_swap32(va_for_ioctl->params.request.header.ethernetII.Ip.src);
            va_for_ioctl->params.request.filter.ipv4.src_port = sa_swap16((u16)va_for_ioctl->params.request.header.ethernetII.Tcp.srcPort);
            va_for_ioctl->params.request.filter.ipv4.dst_port = sa_swap16((u16)va_for_ioctl->params.request.header.ethernetII.Tcp.dstPort);
            break;
        case sizeof (TEMPLATE_VLAN_HEADER):
        case sizeof (TEMPLATE_VLAN_HEADER_TS):
            va_for_ioctl->params.request.filter.ipv4.src_addr = sa_swap32(va_for_ioctl->params.request.header.vlan_ethernetII.Ip.dst);
            va_for_ioctl->params.request.filter.ipv4.dst_addr = sa_swap32(va_for_ioctl->params.request.header.vlan_ethernetII.Ip.src);
            va_for_ioctl->params.request.filter.ipv4.src_port = sa_swap16((u16)va_for_ioctl->params.request.header.vlan_ethernetII.Tcp.srcPort);
            va_for_ioctl->params.request.filter.ipv4.dst_port = sa_swap16((u16)va_for_ioctl->params.request.header.vlan_ethernetII.Tcp.dstPort);
            break;
        case sizeof (TEMPLATE_VLAN_IPV6_HEADER):
        case sizeof (TEMPLATE_VLAN_IPV6_HEADER_TS):
            va_for_ioctl->params.request.filter.ipv6.src_addr[0] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.dst[3]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[1] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.dst[2]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[2] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.dst[1]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[3] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.dst[0]);

            va_for_ioctl->params.request.filter.ipv6.dst_addr[0] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.src[3]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[1] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.src[2]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[2] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.src[1]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[3] = sa_swap32(va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Ipv6.src[0]);

            va_for_ioctl->params.request.filter.ipv6.src_port = sa_swap16((u16)va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Tcp.srcPort);
            va_for_ioctl->params.request.filter.ipv6.dst_port = sa_swap16((u16)va_for_ioctl->params.request.header.vlan_ipv6_ethernetII.Tcp.dstPort);
            break;
        case sizeof (TEMPLATE_IPV6_HEADER):
        case sizeof (TEMPLATE_IPV6_HEADER_TS):
            /*TODO: We may not need this swapping by 32 bit for real ipv6 offloaded packets */
            ///*Temporarily hardcoding to ipv4*/
//            va_for_ioctl->params.request.ip_version = 6;

            va_for_ioctl->params.request.filter.ipv6.src_addr[0] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.dst[3]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[1] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.dst[2]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[2] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.dst[1]);
            va_for_ioctl->params.request.filter.ipv6.src_addr[3] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.dst[0]);

            va_for_ioctl->params.request.filter.ipv6.dst_addr[0] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.src[3]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[1] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.src[2]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[2] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.src[1]);
            va_for_ioctl->params.request.filter.ipv6.dst_addr[3] = sa_swap32(va_for_ioctl->params.request.header.ipv6_ethernetII.Ipv6.src[0]);

            va_for_ioctl->params.request.filter.ipv6.src_port = sa_swap16((u16)va_for_ioctl->params.request.header.ipv6_ethernetII.Tcp.srcPort);
            va_for_ioctl->params.request.filter.ipv6.dst_port = sa_swap16((u16)va_for_ioctl->params.request.header.ipv6_ethernetII.Tcp.dstPort);
            //va_for_ioctl->params.request.filter.ipv6.ct0     = va_for_ioctl->params.request.tcp.ulp_id; // ENUM_BE_TCP_OFFLOAD;
            //va_for_ioctl->params.request.filter.ipv6.ct1     = va_for_ioctl->params.request.tcp.ulp_id; // ENUM_BE_TCP_OFFLOAD;
            //va_for_ioctl->params.request.filter.ipv6.valid   = 1;   

            break;

            // SNAP headers are not supported
        //case sizeof (TEMPLATE_SNAP_HEADER):
        //case sizeof (TEMPLATE_VLAN_SNAP_HEADER):
        default:
            TRACE (DL_ERR, "Illegal template header length: %d", va_for_ioctl->params.request.tcp.hdr_len);
            BREAKPOINT ();
            break;
    }
    va_for_ioctl->params.request.filter.ipv4.mac_id  = rxif_object->handle;
    va_for_ioctl->params.request.filter.ipv4.ct0     = va_for_ioctl->params.request.tcp.ulp_id;
    va_for_ioctl->params.request.filter.ipv4.ct1     = va_for_ioctl->params.request.tcp.ulp_id;
    va_for_ioctl->params.request.filter.ipv4.valid   = 1;   

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (function_object, wrb, generic_context, callback, callback_context, va_for_ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "toe offload ioctl failed.");
        goto Error;
    }
    
Error:

    be_unlock_wrb_post (function_object);
    return status;
}

void be_toe_cxn_upload_internal_callback (PVOID context, BESTATUS status, MCC_WRB* wrb)
{
    PBE_TOE_CONNECTION_OBJECT cxn = (PBE_TOE_CONNECTION_OBJECT)context;
    TOE_CXN_ASSERT (cxn);

    TRACE (DL_NOTE, "Upload cid:%d(0x%x) complete.", be_toe_sq_get_cid (cxn->sq), be_toe_sq_get_cid (cxn->sq));
    
    //
    // Destroy our cxn tracking object.
    //
    be_toe_cxn_destroy (cxn);
}


BESTATUS BECALL
be_toe_cxn_upload (
    IN  PBE_TOE_CONNECTION_OBJECT    cxn,
    IN  PBE_RXIF_OBJECT              rxif_object,
    IN  u32                          flags,
    IN  u32                          rst_sequence,
    OUT PTCP_CONTEXT                 tcp_state,
    IN  MCC_WRB_CQE_CALLBACK         callback,
    IN  PVOID                        callback_context,
    IN  PBE_TOE_UPLOAD_QUEUE_CONTEXT queue_context)
{
    BESTATUS status                     = 0;
    MCC_WRB* wrb                        = NULL;
    IOCTL_COMMON_TCP_UPLOAD *ioctl      = NULL;
    PBE_FUNCTION_OBJECT function_object = NULL;
    BE_TOE_SQ_OBJECT *sq                = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    TOE_CXN_ASSERT (cxn);
    function_object = cxn->parent_function;
    sq = cxn->sq;

    be_lock_wrb_post (function_object);      
    
    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;              
            goto Error;
        }
    }

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_TCP_UPLOAD);    
    ioctl->params.request.id = be_toe_sq_get_cid (sq);
    ioctl->params.request.reset_seq = 0;

    if (SA_IS_BIT_SET (flags, BE_TOE_UPLOAD_FLAG_RESET)) {
        ioctl->params.request.upload_type = CONNECTION_UPLOAD_ABORT_RESET;
        TRACE (DL_NOTE, "Upload cid:%d(0x%x) flags:abort reset", ioctl->params.request.id, ioctl->params.request.id);
    } else if (SA_IS_BIT_SET (flags, BE_TOE_UPLOAD_FLAG_ABORT)) {
        ioctl->params.request.upload_type = CONNECTION_UPLOAD_ABORT;
        TRACE (DL_NOTE, "Upload cid:%d(0x%x) flags:abort", ioctl->params.request.id, ioctl->params.request.id);
    } else if (SA_IS_BIT_SET (flags, BE_TOE_UPLOAD_FLAG_RST_WITH_SEQUENCE)) {
        ioctl->params.request.upload_type = CONNECTION_UPLOAD_ABORT_WITH_SEQ;
        ioctl->params.request.reset_seq = rst_sequence;
    } else {
        ioctl->params.request.upload_type = CONNECTION_UPLOAD_GRACEFUL;
        TRACE (DL_NOTE, "Upload cid:%d(0x%x) flags:graceful", ioctl->params.request.id, ioctl->params.request.id);
    }

    // Post the Ioctl
    if (tcp_state) {
        status = be_function_post_mcc_wrb_complete (
                    function_object,
                    wrb,
                    generic_context,
                    callback,
                    callback_context,
                    be_toe_cxn_upload_internal_callback,
                    cxn, // internal callback context
                    ioctl,
                    BE_CREATE_MCC_RESPONSE_COPY (IOCTL_COMMON_TCP_UPLOAD, params.response.tcp, tcp_state));
    } else {
        status = be_function_post_mcc_wrb_with_internal_callback (
                    function_object,
                    wrb,
                    generic_context,
                    callback,
                    callback_context,
                    be_toe_cxn_upload_internal_callback,
                    cxn, // internal callback context
                    ioctl);
    }    

Error:
    be_unlock_wrb_post (function_object);
    return status;
}


// This updates the template header and tcp_context for the connection.
// If the connection is currently enabled, it is disabled during the update,
// and reenabled.  If the connection is currently disabled, it remains disabled.
BESTATUS BECALL
be_toe_cxn_update (
    IN PBE_TOE_CONNECTION_OBJECT    cxn,
    IN IOCTL_TOE_CONNECTION_UPDATE* va_for_ioctl, // TEMPLATE_HEADER & TCP_CONTEXT must be filled in
    IN u64                          pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK         callback,
    IN PVOID                        callback_context,
    IN PBE_TOE_UPDATE_QUEUE_CONTEXT queue_context)
{
    BESTATUS status = 0;
    MCC_WRB* wrb = NULL;
    BE_FUNCTION_OBJECT *function_object = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    TOE_CXN_ASSERT (cxn);
    function_object = cxn->parent_function;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, 
                                           wrb, 
                                           va_for_ioctl, 
                                           pa_for_ioctl, 
                                           TOE_CONNECTION_UPDATE,
                                           0);
    
    // Get the cid from the previously allocate SQ.
    va_for_ioctl->params.request.cid = be_toe_sq_get_cid (cxn->sq);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 va_for_ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "toe update ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}



BESTATUS BECALL
be_toe_cxn_query (
    IN PBE_TOE_CONNECTION_OBJECT    cxn,
    IN IOCTL_TOE_CONNECTION_QUERY*  va_for_ioctl, // TEMPLATE_HEADER & TCP_CONTEXT must be filled in
    IN u64                          pa_for_ioctl,
    IN MCC_WRB_CQE_CALLBACK         callback,
    IN PVOID                        callback_context,
    IN PBE_TOE_QUERY_QUEUE_CONTEXT queue_context)
{
    BESTATUS status = 0;
    MCC_WRB* wrb = NULL;
    BE_FUNCTION_OBJECT *function_object = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;
    
    TOE_CXN_ASSERT (cxn);
    function_object = cxn->parent_function;

    be_lock_wrb_post (function_object);

    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }

    va_for_ioctl = BE_FUNCTION_PREPARE_NONEMBEDDED_IOCTL (function_object, 
                                           wrb, 
                                           va_for_ioctl, 
                                           pa_for_ioctl, 
                                           TOE_CONNECTION_QUERY,
                                           0);
    
    // Get the cid from the previously allocate SQ.
    va_for_ioctl->params.request.cid = be_toe_sq_get_cid (cxn->sq);

    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 va_for_ioctl);
    if (status < 0) {
        TRACE (DL_ERR, "toe update ioctl failed.");
        goto Error;
    }

Error:
    be_unlock_wrb_post (function_object);
    return status;
}

// Determine if we need this function!

//
//BESTATUS BECALL
//be_toe_offload_enable  (
//    IN  PBE_TOE_CONNECTION_OBJECT cxn,
//    IN  boolean                   enable,
//    IN  MCC_WRB_CQE_CALLBACK      callback,
//    IN  PVOID                     callback_context)
//{
//    BESTATUS status = 0;
//    MCC_WRB* wrb = NULL;
//    IOCTL_NTWK_CONN_ENABLE *ioctl = NULL;
//    PBE_FUNCTION_OBJECT function_object = NULL;
//
//    TOE_CXN_ASSERT (cxn);
//    function_object = cxn->parent_function;
//
//    wrb = be_function_peek_mcc_wrb (function_object);
//    if (!wrb) {
//        TRACE (DL_ERR, "MCC wrb peek failed.");
//        return BE_STATUS_NO_MCC_WRB;
//    }
//
//    // Prepares an embedded ioctl, including request/response sizes.
//    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, IOCTL_NTWK_CONN_ENABLE);
//
//    ioctl->header.request.subsystem = IOCTL_SUBSYSTEM_TOE;
//    ioctl->header.request.opcode    = OPCODE_TOE_NTWK_CONN_ENABLE;
//
//    ioctl->params.request.cid    = be_toe_sq_get_cid (cxn->sq);
//    ioctl->params.request.enable = enable;
//
//    // Post the Ioctl
//    status = be_function_post_mcc_wrb (function_object, wrb, callback, callback_context, ioctl);
//    if (status < 0) {
//        TRACE (DL_ERR, "toe enable ioctl failed.");
//        return status;
//    }
//
//    return status;
//}



//---------------------------------------------------------------------------------------
// Function: be_function_set_tcp_parameters
//
//   This function updates the global tcp parameters per Upper Layer Protocol (ULP) chute.
//   Only the iSCSI PCI function may modify iSCSI protocol chutes. Only the networking PCI
//   function may modify TOE and RDMA chutes. 
//
// function_object    - 
// ulp_mask           - The ULPs to which the parameters will be applied. The storage PCI
//                      function may only change iSCSI ULPs. The networking PCI function
//                      may change TOE or RDMA ULPs. The mask may contain the bitwise OR
//                      of the valid values. 0x1 = ULP 0, 0x2 = ULP 1, etc. 
// parameters         - The new TCP parameters used for all connections on the given ULP
//                      chutes. 
// callback           - 
// callback_context   - 
// queue_context      - 
// return pend_status - BE_SUCCESS (0) on success. BE_PENDING (postive value) if the IOCTL
//                      completion is pending. Negative error code on failure.
//---------------------------------------------------------------------------------------
BESTATUS
be_function_set_tcp_parameters (
   IN PBE_FUNCTION_OBJECT                  function_object,
   IN u32                                  ulp_mask,
   IN PBE_DRIVER_TCP_PARAMETERS            parameters,
   IN MCC_WRB_CQE_CALLBACK                 callback,
   IN PVOID                                callback_context,
   IN PBE_SET_TCP_PARAMETERS_QUEUE_CONTEXT queue_context
)
{
    BESTATUS status = 0;
    MCC_WRB *wrb = NULL;
    IOCTL_COMMON_SET_TCP_PARAMETERS *ioctl = NULL;
    PBE_GENERIC_QUEUE_CONTEXT generic_context = NULL;

    FUNCTION_ASSERT (function_object);
    ASSERT (parameters);
    ASSERT (ulp_mask != 0);

    be_lock_wrb_post (function_object);
    
    wrb = be_function_peek_mcc_wrb (function_object);
    if (!wrb) {
        if (queue_context && callback) {    
            wrb = (MCC_WRB*)&queue_context->wrb_header;    // Location to write the WRB.
            generic_context = (PBE_GENERIC_QUEUE_CONTEXT)queue_context; // Indicate to queue
            generic_context->context.bytes = sizeof (*queue_context);
        } else {
            status = BE_STATUS_NO_MCC_WRB;  
            goto Error;
        }
    }   

    // Prepares an embedded ioctl, including request/response sizes.
    ioctl = BE_FUNCTION_PREPARE_EMBEDDED_IOCTL (function_object, wrb, COMMON_SET_TCP_PARAMETERS);    

    ioctl->params.request.ulp_mask = ulp_mask;
    sa_memcpy (&ioctl->params.request.tcp_params,
               parameters,
               sizeof (*parameters));
    
    // Post the Ioctl
    status = be_function_post_mcc_wrb_with_queue_context (
                 function_object,
                 wrb,
                 generic_context,
                 callback,
                 callback_context,
                 ioctl);

    if (status < 0) {
        TRACE (DL_ERR, "ioctl to set tcp parameters failed.");
        goto Error;
    }

    TRACE (DL_INFO, 
           "set tcp params succeeded. ulp_mask:0x%x function:%d ack_freq:%d max_rxmt_count:%d max_warn_count:%d dup_ack:%d sws_ms:%d idle_ms:%d delay_ms:%d push_ms:%d fin_wait2_ms:%d", 
           ulp_mask, 
           be_function_get_function_number (function_object), 
           parameters->ack_frequency, 
           parameters->max_rxmt_count, 
           parameters->max_warning_count, 
           parameters->dup_ack_threshold, 
           parameters->sws_prevention_ms, 
           parameters->idle_ms, 
           parameters->delayed_ack_ms, 
           parameters->push_timer_ms, 
           parameters->fin_wait2_ms);

Error:

    be_unlock_wrb_post (function_object);
    return status;
}






