/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

        vase.c

    Purpose:

        Header file for defining VASE functions.
        

--*/

#include "acitpch.h"

#if ACIT_CFG_VASE_ENABLED




/*
 * Complete File is enclosed in the ACIT_CFG_VASE_ENABLED definition.
 */



/*+

Variable:
    DbgMasterBlkPtr

Purpose:
    This varaible stores the pointer to the Master block structure. This is
    used by the WinDbg extension DLL to find out the location of the Master
    Block in the driver image in memory. Currently it only stores pointer
    to one Master Block. If multiple adapters are present in the system it 
    will point to the last initialized adapter.
-*/
volatile ACIT_PTR DbgMasterBlkPtr;



/*+

Function:
    DbgLogPreInit

Purpose:
    This function is called by the miniport driver to estimate the memory 
    required for debug logging and to pre-initialize debug logging structures.
    The calculated memory is returned in the pMemNeeded pointer.
    This function expects the registry interface to be available and calls the
    registry interface functions.
-*/

ACIT_UEXACT32
DbgLogPreInit(
    PDBG_MASTER_BLOCK   pMasterBlock,
    PACIT_UEXACT32      pMemNeeded,
    PVASE_PROFILE       pVaseProfile)
{
    ACIT_UEXACT32       CompiledEnabledBlocks;
    ACIT_UEXACT32       RegistryEnabledBlocks;
    ACIT_UEXACT32       ThisSessionBlocks;
    ACIT_UEXACT32       MemNeeded;

    /* DBG_ASSERT(pMasterBlock); */
    /* DBG_ASSERT(pMemNeeded); */

    /*
     * Initialize master block
     */
    memset(pMasterBlock, 0, sizeof(DBG_MASTER_BLOCK));

    /*
     * Store the device extension pointer locally for later use
     */
    pMasterBlock->Context = pVaseProfile->Context;

    /*
     * Store the version number in the master block
     */
    pMasterBlock->Version = DBG_BLK_MASTER_VERSION;

    /*
     * Look at what blocks this compiled binary supports
     */
    CompiledEnabledBlocks = DBG_BLK_COMPILED_IN();

    /*
     * Check if the registry enables or disables any block support.
     */
    RegistryEnabledBlocks = pVaseProfile->ConfigFlags;

    /*
     * AND both to see what we need to support in this session
     */
    ThisSessionBlocks = (CompiledEnabledBlocks & RegistryEnabledBlocks);

    /*
     * Anything happens, enable the master block this cannot be disabled
     */
    ThisSessionBlocks |= (1U << DBG_BLK_MASTER);

    /*
     * Save all these values in the master block
     */
    pMasterBlock->CompileEnabledBlocks = CompiledEnabledBlocks;
    pMasterBlock->RegistryEnabledBlocks = RegistryEnabledBlocks;
    pMasterBlock->ThisSessionBlocks = ThisSessionBlocks;

    /*
     * Go thru all the blocks we support and calculate the memory we need.
     * Start with master block. Also fillup other details about each block
     */

    
    /*
     * Master block is stored in device extension so no other memory 
     * needed to store it but initialize other fields. Initialize the 
     * first Block structure in master block to itself.
     */

    /* Maybe master block does not have to be here
    pMasterBlock->BlockHdrs[DBG_BLK_MASTER].ID = DBG_BLK_MASTER;
    strcpy(pMasterBlock->BlockHdrs[DBG_BLK_MASTER].Name, (PACIT_UEXACT8)"MasterBlock");
    pMasterBlock->BlockHdrs[DBG_BLK_MASTER].CurrentEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_MASTER].MaxEntries = 1;
    pMasterBlock->BlockHdrs[DBG_BLK_MASTER].EntrySize =sizeof(DBG_MASTER_BLOCK);
    pMasterBlock->BlockHdrs[DBG_BLK_MASTER].FirstEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_MASTER].TotalSize =sizeof(DBG_MASTER_BLOCK);
    pMasterBlock->pBlocks[DBG_BLK_MASTER] = (PDBG_BLOCK_HEADER) pMasterBlock;
    */
    /*
     * Increment number of blocks stored 
     */
    pMasterBlock->NumberOfBlocksAvailable ++;
    pMasterBlock->NumberOfBlocksEnabled ++;


    /*
     * Now do processing for the function block.
     *
     * Initialize the entry for function block in master block.
     * We may not alloc memory for this block if it has been disabled
     * in registry but we still need to initialize its entry in master block.
     */
    pMasterBlock->BlockHdrs[DBG_BLK_FUNC].ID = DBG_BLK_FUNC;
    strcpy(pMasterBlock->BlockHdrs[DBG_BLK_FUNC].Name, (PACIT_UEXACT8)"FunctionBlock");
    pMasterBlock->BlockHdrs[DBG_BLK_FUNC].HeadEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_FUNC].MaxEntries = 
                                        pVaseProfile->FuncBlockSize;
    pMasterBlock->BlockHdrs[DBG_BLK_FUNC].EntrySize = 
                                        sizeof(DBG_FUNC_BLOCK_ENTRY);
    pMasterBlock->BlockHdrs[DBG_BLK_FUNC].TailEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_FUNC].Version = DBG_BLK_FUNC_VERSION;

    /*
     * The size of this log may come from the registry too but for now
     * we assume the size to be a default num of entries
     */
    pMasterBlock->BlockHdrs[DBG_BLK_FUNC].TotalSize =
            sizeof(DBG_FUNC_BLOCK_ENTRY) * (pVaseProfile->FuncBlockSize);

    /*
     * We need to alloc mem for this block if it is enabled now
     */
    if(DBG_IS_BLOCK_ENABLED(ThisSessionBlocks, DBG_BLK_FUNC))
    {
        pMasterBlock->TotalDebugLogSize += 
            pMasterBlock->BlockHdrs[DBG_BLK_FUNC].TotalSize;

        pMasterBlock->NumberOfBlocksEnabled ++;
    }

    /*
     * Increment number of blocks stored 
     */
    pMasterBlock->NumberOfBlocksAvailable ++;


    /*
     * Now do processing for the text message block.
     *
     * Initialize the entry for text message block in master block.
     * We may not alloc memory for this block if it has been disabled
     * in registry but we still need to initialize its entry in master block.
     */
    pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].ID = DBG_BLK_TEXT_MESSAGE;
    strcpy(pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].Name, 
                                        (PACIT_UEXACT8)"TextMessageBlock");
    pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].HeadEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].MaxEntries = 1;
    pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].EntrySize = 
                                        pVaseProfile->TextBlockSize;
    pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].TailEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].Version = 
                                        DBG_BLK_TEXT_MESSAGE_VERSION;
    /*
     * The size of this log may come from the registry too but for now
     * we assume the size to be a default num of entries.
     */
    pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].TotalSize = 
            pVaseProfile->TextBlockSize;

    /*
     * We need to alloc mem for this block if it is enabled now
     */
    if(DBG_IS_BLOCK_ENABLED(ThisSessionBlocks, DBG_BLK_TEXT_MESSAGE))
    {
        pMasterBlock->TotalDebugLogSize += 
            pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].TotalSize;

        pMasterBlock->NumberOfBlocksEnabled ++;
    }
    /*
     * Increment number of blocks stored
     */
    pMasterBlock->NumberOfBlocksAvailable ++;


    /*
     * Now do processing for the operating parameter block.
     *
     * Initialize the entry for operating parameter block in master block.
     * We may not alloc memory for this block if it has been disabled
     * in registry but we still need to initialize its entry in master block.
     */
    pMasterBlock->BlockHdrs[DBG_BLK_OP_PARAMETER].ID = DBG_BLK_OP_PARAMETER;
    strcpy(pMasterBlock->BlockHdrs[DBG_BLK_OP_PARAMETER].Name, 
                                        (PACIT_UEXACT8)"OpParameterBlock");
    pMasterBlock->BlockHdrs[DBG_BLK_OP_PARAMETER].HeadEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_OP_PARAMETER].MaxEntries = 
                                        DBG_OP_PARAMETER_BLOCK_ENTRIES;
    pMasterBlock->BlockHdrs[DBG_BLK_OP_PARAMETER].EntrySize = 
                                        sizeof(DBG_OP_PARAMETER_BLOCK_ENTRY);
    pMasterBlock->BlockHdrs[DBG_BLK_OP_PARAMETER].TailEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_OP_PARAMETER].Version = 
                                        DBG_BLK_OP_PARAMETER_VERSION;
    /*
     * The size of this log may come from the registry too but for now
     * we assume the size to be a default num of entries
     */
    pMasterBlock->BlockHdrs[DBG_BLK_OP_PARAMETER].TotalSize =
            sizeof(DBG_OP_PARAMETER_BLOCK_ENTRY) * 
                    (DBG_OP_PARAMETER_BLOCK_ENTRIES);

    /*
     * We need to alloc mem for this block if it is enabled now
     */
    if(DBG_IS_BLOCK_ENABLED(ThisSessionBlocks, DBG_BLK_OP_PARAMETER))
    {
        pMasterBlock->TotalDebugLogSize += 
            pMasterBlock->BlockHdrs[DBG_BLK_OP_PARAMETER].TotalSize;

        pMasterBlock->NumberOfBlocksEnabled ++;
    }
    /*
     * Increment number of blocks stored 
     */
    pMasterBlock->NumberOfBlocksAvailable ++;


    /*
     * Now do processing for the statistics block.
     *
     * Initialize the entry for statistics block in master block.
     * We may not alloc memory for this block if it has been disabled
     * in registry but we still need to initialize its entry in master block.
     */
    pMasterBlock->BlockHdrs[DBG_BLK_STATISTICS].ID = DBG_BLK_STATISTICS;
    strcpy(pMasterBlock->BlockHdrs[DBG_BLK_STATISTICS].Name, 
                                        (PACIT_UEXACT8)"StatisticsBlock");
    pMasterBlock->BlockHdrs[DBG_BLK_STATISTICS].HeadEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_STATISTICS].MaxEntries = 1;
    pMasterBlock->BlockHdrs[DBG_BLK_STATISTICS].EntrySize = 
                                            sizeof(DBG_STATISTICS_BLOCK);
    pMasterBlock->BlockHdrs[DBG_BLK_STATISTICS].TailEntry = 0;
    pMasterBlock->BlockHdrs[DBG_BLK_STATISTICS].Version = 
                                        DBG_BLK_STATISTICS_VERSION;
    /*
     * The size of this log may come from the registry too but for now
     * we assume the size to be a default num of entries
     */
    pMasterBlock->BlockHdrs[DBG_BLK_STATISTICS].TotalSize =
                    sizeof(DBG_STATISTICS_BLOCK);

    /*
     * We need to alloc mem for this block if it is enabled now
     */
    if(DBG_IS_BLOCK_ENABLED(ThisSessionBlocks, DBG_BLK_STATISTICS))
    {
        pMasterBlock->TotalDebugLogSize += 
            pMasterBlock->BlockHdrs[DBG_BLK_STATISTICS].TotalSize;

        pMasterBlock->NumberOfBlocksEnabled ++;
    }
    /*
     * Increment number of blocks stored 
     */
    pMasterBlock->NumberOfBlocksAvailable ++;





    /*
     * Done allocating and initializing all blocks
     * Make sure we are 8 bytes alligned in memory we are asking for.
     */
    /* DBG_ASSERT(!(pMasterBlock->TotalDebugLogSize & 0x00000007)); */


    /*
     * We have gathered our memory requirements now return it.
     */
    *pMemNeeded = pMasterBlock->TotalDebugLogSize;

    return (0);
}


/*+

Function:
    DbgLogInitialize

Purpose:
    This function is called by the miniport driver to initialize debug logging
    All debug blocks are initialized completely and memory assigned from the 
    debug log pool. Logging can begin after returning from this function.

-*/

ACIT_UEXACT32
DbgLogInitialize(
    PDBG_MASTER_BLOCK   pMasterBlock,
    PACIT_VOID          pMemBuffer,
    ACIT_UEXACT32       FreqFor10ms)
{
    ACIT_UEXACT32       BlockNum;
    PACIT_UEXACT8       pAvailableMemory;

/*     DBG_ASSERT(pMasterBlock); */
/*     DBG_ASSERT(pMemBuffer); */

    pAvailableMemory = (PACIT_UEXACT8) pMemBuffer;

    DbgMasterBlkPtr = (ACIT_PTR) pMasterBlock;

    pMasterBlock->BlockHdrs[DBG_BLK_FUNC].BlockSpecific1 = FreqFor10ms;

    /*
     * Start with block 1 and go till max blocks. We don't start with 0
     * as 0 is master block that has already been initialized and does not
     * need memory as it is stored in device extension.
     */
    for(BlockNum = 1; BlockNum < DBG_MAX_BLOCKS; BlockNum ++)
    {

        if(DBG_IS_BLOCK_ENABLED(pMasterBlock->ThisSessionBlocks, BlockNum))
        {
            /*
             * This block has been enabled and we need memory for it.
             * Assign memory to this block if it asked for any
             */
            if (pMasterBlock->BlockHdrs[BlockNum].TotalSize)
            {
                /*
                 * Make sure blocks memory requirement is 4 bytes alligned
                 */

                 /*DBG_ASSERT(!(pMasterBlock->BlockHdrs[BlockNum].TotalSize & 
                                                    DBG_BLOCK_ALIGN_MASK));*/

                pMasterBlock->BlockHdrs[BlockNum].Buffer = 
                                        (ACIT_UEXACT64) pAvailableMemory;

                /*
                 * So now move our available memory by the amount we 
                 * gave away to the above block
                 */
                pAvailableMemory += pMasterBlock->BlockHdrs[BlockNum].TotalSize;

            }
        }
    }

    /*
     * By now all the block that were enabled have the memory they asked for.
     */
    return (0);
}


/*+

Function:
    DbgLogIoctl

Purpose:
    This function is called by the miniport driver to service the Debug IOCTL
    received by the driver.

-*/

ACIT_UEXACT32
DbgLogIoctl(
    PACIT_VOID              pMasterBuffer,
    PACIT_VOID              pCmdBuffer)
{
    PDBG_COMMAND            pDbgCommand;
    PACIT_VOID              pBlock;
    PDBG_MASTER_BLOCK       pMasterBlock;
    ACIT_UEXACT32           Ctr;
    ACIT_UEXACT32           CopySize;

    /* DBG_ASSERT(pMasterBuffer); */
    /* DBG_ASSERT(pCmdBuffer); */


//CR: Add assert to make sure proper buffer length and other assumptions are fullfilled.

    //
    // Initialize our locals
    //
    pDbgCommand = (PDBG_COMMAND) pCmdBuffer;
    pMasterBlock = (PDBG_MASTER_BLOCK) pMasterBuffer;

    //
    // Calculate the buffer pointer for the block. This is below 
    // the DBG_COMMAND structure in memory
    //
    pBlock = ((PACIT_UEXACT8) pCmdBuffer) + pDbgCommand->BufferOffset;
//CR: FAIL cmd is buffer offset does not matches assumptions

    //
    // OK, we received a Debug IOCTL command from our debug application.
    // Setup to fail the command by default
    //
    pDbgCommand->Status = DBG_IOCTLCMD_STATUS_FAIL;

    //
    // Find out what action is required.
    //
    switch(pDbgCommand->CommandCode)
    {
        case DBG_IOCTLCMD_NOP:
            {
                //
                // The pDbgCommand->Version field contains the applications version
                // while coming in and should contain the debug implementations version
                // on return. This is also a good place to return a generic error string
                // to the applications.
                //
                
                // Overwrite the applications version with ours
                pDbgCommand->Version = DBG_API_VERSION;
                pDbgCommand->Status = DBG_IOCTLCMD_STATUS_SUCCESS;

                //
                // Just for test purpose for now. Report alpha state.
                //
                strcpy((PACIT_UEXACT8)pDbgCommand->StatusStr, "IOCTL implementation in alpha stage");
                pDbgCommand->StatusFlag |= DBG_IOCTLCMD_FLAG_CHECKSTR;
            }
            break;
        case DBG_IOCTLCMD_GET_MASTER_BLOCK:
            {
                if(DBG_IS_BLOCK_ENABLED(
                    pMasterBlock->ThisSessionBlocks,
                    DBG_BLK_MASTER))
                {
                    if(pDbgCommand->BufferLength < sizeof(DBG_MASTER_BLOCK))
                    {
                        //
                        // Not enough buffer provided to us
                        //
                        pDbgCommand->Status = DBG_IOCTLCMD_STATUS_INSUFFICIENT_BUFFER;
                        break;
                    }

                    //
                    // First update all the fields in the master block with data from
                    // individual blocks. When entries are made in individual blocks
                    // The counts are not updated in master block. So we have to update
                    // them here before we report it back to aplication.
                    //
                    /* Old code
                    for (Ctr = 1; Ctr < DBG_MAX_BLOCKS; Ctr ++)
                    {
                        if(pMasterBlock->pBlocks[Ctr])
                        {
                            pMasterBlock->BlockHdrs[Ctr].HeadEntry = 
                                pMasterBlock->pBlocks[Ctr]->HeadEntry;
                            pMasterBlock->BlockHdrs[Ctr].TailEntry = 
                                pMasterBlock->pBlocks[Ctr]->TailEntry;
                        }
                    }
                    */

                    //
                    // Copy the master block to the buffer and return success
                    //
                    memcpy(pBlock, pMasterBuffer, sizeof(DBG_MASTER_BLOCK));
                    pDbgCommand->Status = DBG_IOCTLCMD_STATUS_SUCCESS;
                }
            }
            break;
        case DBG_IOCTLCMD_GET_DEBUG_BLOCK:
            {
                ACIT_UEXACT32   index;
                /* DBG_ASSERT(pDbgCommand->Parameter1 < DBG_MAX_BLOCKS); */

                //
                // First find out the ID of the block thats needed.
                //

                for(index = 0; index < DBG_MAX_BLOCKS; index ++)
                {
                    if(pMasterBlock->BlockHdrs[index].ID == pDbgCommand->Parameter1)
                        break;
                }
                if(index == DBG_MAX_BLOCKS)
                {
                    //
                    // We did not find the block we were looking for
                    //
                    pDbgCommand->Status = DBG_IOCTLCMD_STATUS_UNKNOWN_BLOCKID;
                    break;
                }

                //
                // We have found the block that we were looking for.
                //
                if(DBG_IS_BLOCK_ENABLED(
                        pMasterBlock->ThisSessionBlocks,
                        pDbgCommand->Parameter1))
                {
                    //
                    // Find out the size of copy operation we need to do
                    // we need to provide the header as well as the block
                    //
                    CopySize = pMasterBlock->BlockHdrs[index].TotalSize;
                    CopySize += sizeof(DBG_BLOCK_HEADER);

                    if(pDbgCommand->BufferLength < CopySize)
                    {
                        //
                        // Not enough buffer provided to us
                        //
                        pDbgCommand->Status = DBG_IOCTLCMD_STATUS_INSUFFICIENT_BUFFER;
                        break;
                    }

                    //
                    // Copy the block header from master block
                    //
                    memcpy(pBlock,
                        &(pMasterBlock->BlockHdrs[index]),
                        sizeof(DBG_BLOCK_HEADER));

                    //
                    // move to the offset and copy the block itself
                    //
                    pBlock = ((PACIT_UEXACT8)pBlock) + sizeof(DBG_BLOCK_HEADER);

                    memcpy(pBlock,
                        (PACIT_VOID)pMasterBlock->BlockHdrs[index].Buffer,
                        pMasterBlock->BlockHdrs[index].TotalSize);

                    //
                    // Everything ok, return success.
                    //
                    pDbgCommand->Status = DBG_IOCTLCMD_STATUS_SUCCESS;
                }
                else
                {
                    pDbgCommand->Status = DBG_IOCTLCMD_STATUS_BLOCK_UNAVAILABLE;
                }
            }
            break;

        default:
            {
                //
                // Command code is unrecognized. so fail it.
                //
                pDbgCommand->Status = DBG_IOCTLCMD_STATUS_FAIL;
                //
                // Fill up the failure string and set flag so app can print it.
                //
                strcpy(pDbgCommand->StatusStr, 
                    "DBG_COMMAND.CommandCode is unknown, Command failed");
                pDbgCommand->StatusFlag |= DBG_IOCTLCMD_FLAG_CHECKSTR;
            }
            break;
    }

    return FALSE;
}


/*+

Function:
    DbgLogFuncBlock

Purpose:
    This function is called to log an entry into the Function Block.

-*/

ACIT_VOID 
DbgLogFuncBlock(
    PDBG_MASTER_BLOCK       pMasterBlock, 
    USHORT                  FunctionID, 
    UCHAR                   FuncState,
    UCHAR                   Module,
    ACIT_UEXACT64           Param1,
    ACIT_UEXACT64           Param2,
    ACIT_UEXACT64           Param3)
{
    PDBG_FUNC_BLOCK         pFuncBlock;
    ACIT_UEXACT32           CurrentEntry;
    ACIT_UEXACT32           TimeStampHi, TimeStampLo;
    static ACIT_UEXACT32    LastStampLo;

    /* DBG_ASSERT(pMasterBlock); */

    /*
     * Get pointer to the function block
     */
    pFuncBlock = (PDBG_FUNC_BLOCK) pMasterBlock->BlockHdrs[DBG_BLK_FUNC].Buffer;

    if(pFuncBlock)
    {
        /*
         * Increment our head/tail entry index in the function block and 
         * also increment the sequence number in master block
         */
        INCREMENT_ENTRY_INDEX(DBG_BLK_FUNC, CurrentEntry);

        /*
         * Store this entry at the index of current entry
         */
        pFuncBlock->pLog[CurrentEntry].TimeStamp = OSM_GETCLOCK();
        pFuncBlock->pLog[CurrentEntry].Module = (OSM_GETCPUNUM() << 4);
        pFuncBlock->pLog[CurrentEntry].FunctionID = FunctionID;
        pFuncBlock->pLog[CurrentEntry].State = FuncState;
        pFuncBlock->pLog[CurrentEntry].SequenceNumber = 
                                        pMasterBlock->CurrentSequenceNumber;
        pFuncBlock->pLog[CurrentEntry].Parameter1 = Param1;
        pFuncBlock->pLog[CurrentEntry].Parameter2 = Param2;
        pFuncBlock->pLog[CurrentEntry].Parameter3 = Param3;
    }
}


/*+

Function:
    DbgLogTextBlock

Purpose:
    This function is called to log an entry into the TextMessage Block.

-*/

ACIT_VOID
DbgLogTextBlock(
    PDBG_MASTER_BLOCK    pMasterBlock,
    ACIT_UEXACT32        messageSize,
    UCHAR                dbgMessage[])
{
    PDBG_TEXT_BLOCK_ENTRY    pBlockEntry;
    ACIT_UEXACT32            EntryLocation; 

    /*
     * Please note that since this is a free flowing text block, we cannot use 
     * HeadEntry and Tail entry. The whole log has only one entry, the entire 
     * buffer. So how do we make multiple text entries to this block. We use 
     * the BlockSpecific1 as our head and BlockSpecific2 as the tail.
     */

    /* offset to our entry to fill out
     * Check for the wrap around first, since we don't know in advance the size
     * of the debug string we get.
     * NOTE: if is assume that the message will be shorter than the total 
     * buffer size.
     */
    EntryLocation = 
        pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].BlockSpecific1;

    if ( (EntryLocation + messageSize + sizeof(DBG_TEXT_BLOCK_ENTRY)) >= 
         (pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].TotalSize) )

    {
        /* Ignore the reminder of the buffer, just wrap around.*/
        EntryLocation = 0;
        pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].BlockSpecific1 = 0;
        /*
         * Also if we are wrapping around, now we have to carry the tail 
         * pointer forward as we have completed going thru the entire log buffer
         * Just change the value to non zero, that way later we can calculate
         * its right position and set the value.
         */
        pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].BlockSpecific2 = 1;
    }

    /*
     * Get pointer to the text block entry we are about to use.
     */
    pBlockEntry = (PDBG_TEXT_BLOCK_ENTRY) 
        (((PACIT_UEXACT8)pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].Buffer) +
        EntryLocation);

    if(pBlockEntry)
    {
        /*
         * increment the sequence number after saving in the entry
         */
        pBlockEntry->SequenceNumber = pMasterBlock->CurrentSequenceNumber++;

        /*
         * Set the markers so that the decoders could search for this entry.
         */
        pBlockEntry->Marker1 = DBG_TEXT_BLOCK_MARKER;
        pBlockEntry->Marker2 = DBG_TEXT_BLOCK_MARKER;

        /*
         * copy the dbg string to our log buffer immediately after the entry 
         * structure
         */
        strcpy((PACIT_UEXACT8) (pBlockEntry+1), dbgMessage);

        /*
         * and update the BlockSpecific1 offset.
         * add the size of the type of SequenceNumber
         */
        pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].BlockSpecific1 += 
            messageSize + 
            sizeof(DBG_TEXT_BLOCK_ENTRY) + 1; /* Add 1 for the null char.*/

        if(pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].BlockSpecific2)
        {
            /*
             * Tail pointer has moved, so we have wrapped around atleast once
             * So from now on we have to maintain the tail pointer too
             */
            pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].BlockSpecific2 =
                pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].BlockSpecific1;
        }

        /*
         * Increment the number of entries made in this block
         */
        pMasterBlock->BlockHdrs[DBG_BLK_TEXT_MESSAGE].EntriesMade ++;

    }
}



/*+

Function:
    DbgLogOpParameter

Purpose:
    This function is called to log an entry into the Operating Parameter Block.

-*/

ACIT_VOID
DbgLogOpParameter(
    PDBG_MASTER_BLOCK           pMasterBlock,
    PACIT_UEXACT8               pParameterString,
    ACIT_UEXACT64               Value)
{
    PDBG_OP_PARAMETER_BLOCK     pBlock;
    ACIT_UEXACT32               CurrentEntry;


    /* DBG_ASSERT(pMasterBlock); */

    /*
     * Get pointer to the operating parameter block
     */
    pBlock = (PDBG_OP_PARAMETER_BLOCK) pMasterBlock->
                        BlockHdrs[DBG_BLK_OP_PARAMETER].Buffer;

    if(pBlock)
    {
        /*
         * Increment our head/tail entry index in the function block and 
         * also increment the sequence number in master block
         */
        INCREMENT_ENTRY_INDEX(DBG_BLK_OP_PARAMETER, CurrentEntry);

        /*
         * Store this entry at the index of current entry
         */
        pBlock->Param[CurrentEntry].Value = Value;
        OSM_MEMCPY(pBlock->Param[CurrentEntry].ParameterString, 
                    pParameterString, 
                    OP_PARAMETER_STRING_SIZE);
        pBlock->Param[CurrentEntry].SequenceNumber = 
                    pMasterBlock->CurrentSequenceNumber;
    }
}


#else

//
// Compilation note
// C4206: Translation unit cannot be empty
//
int vase_dummy = 1;

#endif /* for  #if ACIT_CFG_VASE_ENABLED */
