/*
 * This file is part of the Emulex Linux Device Driver for iSCSI Host Bus
 * Adapters.
 * Copyright (C) 2003 Adaptec, Inc.
 * Copyright (C) 2005-2011 Emulex.
 * All rights reserved.
 * EMULEX and SLI are trademarks of Emulex.
 * www.emulex.com
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as published
 * by the Free Software Foundation. This program is distributed in the hope that
 * it will be useful. ALL EXPRESS OR IMPLIED CONDITIONS, REPRESENTATIONS AND
 * WARRANTIES, INCLUDING ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT, ARE DISCLAIMED, EXCEPT TO THE EXTENT
 * THAT SUCH DISCLAIMERS ARE HELD TO BE LEGALLY INVALID. See the GNU General
 * Public License for more details, a copy of which can be found in the file
 * COPYING included with this package.
 */

/*++

    Filename:

	be2iscsiosm.c

    Purpose:

	This file implements the different entry points that are needed
	of a SCSI Lower Layer Driver (LLD) under Linux. This follows the
	ACIT APIs for achieving different functionalities needed.

	TODO: 
	- implement sending NULL if WS size is 0
	- streamline the process of calling post routines
--*/

#include <linux/version.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/blkdev.h>
#include <linux/pci.h>
#if LINUX_VERSION_CODE != KERNEL_VERSION(2,6,5)
#include "scsi/scsi.h"
#else
#include "../drivers/scsi/scsi.h"
#endif
#include "scsi/scsi_cmnd.h"
#include "scsi/scsi_device.h"
#include "acitcfg.h"
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/reboot.h>
#include "acitdefs.h"
#include "acitscsi.h"
#include "uerr.h"
#include "ioctl_top_bmap.h"
#include "ioctl_top_enum.h"
#include "acitapi.h"
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/list.h>
#include <asm/unistd.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
#include <asm/semaphore.h>
#else
#include <linux/semaphore.h>
#endif
#include <asm/delay.h>
#include <linux/genhd.h>
#include <linux/spinlock.h>
#include <scsi/scsi_host.h>
#include "be2iscsi_common.h"
#include "version.h"
#include "be2iscsi_osm.h"
#include "be2iscsi_debug.h"
#include "acit_iface.h"

#ifdef __VMKLNX__
#include <scsi/scsi_transport.h>
#ifdef REGISTER_ISCSI
#include <scsi/scsi_transport_iscsi.h>
#endif
#if defined(ESX_50)
#include <vmklinux_scsi.h>   // For Transport type
#include <asm/uaccess.h>
#else
#include <vmklinux26_scsi.h>   // For Transport type
#endif
#endif

#define BUILD_STR XSTR(BE_MAJOR) "." XSTR(BE_MINOR) "." XSTR(BE_BUILD) "." XSTR(BE_BRANCH)
#define XSTR(x) MSTR(x)
#define MSTR(x) #x
#define MPU_EP_SEMAPHORE 0xac

#ifdef MODULE
#include <linux/module.h>
#define MOD_DESCRIPTION "Emulex OneConnect iSCSI Driver version " BUILD_STR
MODULE_AUTHOR("Emulex Corporation");
MODULE_VERSION(BUILD_STR);
MODULE_INFO(supported, "external");
MODULE_DESCRIPTION(MOD_DESCRIPTION);
#if defined(__VMKERNEL_MODULE__)
#else
MODULE_LICENSE("GPL");
#endif
#endif

struct be_default_msix_entry def_msix_entries[BE_NUM_MSIX_ENTRIES]= {
	{ 0, 0,
		"be2iscsi_zero", se_isr_msix},
};

/* Mutex for locking while we walk the hba_list */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,0,0)
DEFINE_SEMAPHORE(se_configuration_lock);
#else
DECLARE_MUTEX(se_configuration_lock);
#endif

unsigned be2iscsi_dump_in_progress = 0;

/*------------------- PCI Driver operations and data ----------------- */
struct pci_device_id        be2iscsi_pci_id_table[] = 
{
	{0x19A2, 0x0212, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0x19A2, 0x0222, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0x19A2, 0x0702, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0x19A2, 0x0703, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0x19A2, 0x0712, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 0},
	{0}
};

MODULE_DEVICE_TABLE(pci, be2iscsi_pci_id_table);

struct pci_driver be2iscsi_pci_driver = 
{
	name:                "be2iscsi",
	probe:                se_pci_dev_probe,
#ifdef POWER_MANAGEMENT
	suspend:              se_pci_dev_suspend,
	resume:               se_pci_dev_resume,
#endif
	remove:               se_pci_dev_remove,
	id_table:             be2iscsi_pci_id_table
};

#if ACIT_CFG_DEBUG_BUILD
int acit_debug_level  = 
		ACIT_DBGMSG_WARNING |
		ACIT_DBGMSG_ERROR | ACIT_DBGMSG_FATAL |
		ACIT_DBGMSG_UCODE | ACIT_DBGMSG_MGMT;
#else
int acit_debug_level  = 0;
#endif

/* global variables */
ACIT_API_TABLE acitapi_table;
ACIT_VERSION acit_api_version;
struct se_param_struct se_param[PARAMETERS];
unsigned long gshtdwn = 0;
unsigned int  gcrashmode = 0;

char *be2iscsi_driver_version = BUILD_STR;
char *driver_name  = "be2iscsi";
unsigned int be2iscsi_major = 0;
static  struct class       *vnic_class = NULL;
LIST_HEAD(hba_chain);                  /* Chain of se_hba structs     */
ACIT_SEXACT8     hba_count = 0;            /* Total count of hbas we have   */

unsigned int se_hba_uid[BE2ISCSI_MAX_HBA];
#define SE_HBA_REMOVED	0x80000000

extern struct file_operations se_fops;

/* Extern variables Kernel specific code */
#ifdef CONFIG_SYSFS
extern struct class_simple *be2iscsi_class;
#endif

/* Configurable parameters */
/* NOTE: If a new param is configured here, the value of PARAMETERS
   is to be incremented. Also an entry in se_init_params() is to be 
   made */

#define MAX_SG_IO_SIZE		((SG_LIST_ELEMENTS * 4096) / 1024)
#define MAX_CHIP_IO_SIZE	(DEFAULT_MAX_BURST_LENGTH/1024)

#if MAX_SG_IO_SIZE > MAX_CHIP_IO_SIZE
#define MAX_IO_SIZE	MAX_CHIP_IO_SIZE
#else
#define MAX_IO_SIZE	MAX_SG_IO_SIZE
#endif

unsigned max_io_size = MAX_IO_SIZE;
unsigned dic = 1;

DECLARE_CONF_PARAM_NUM(eto,30);
DECLARE_CONF_PARAM_NUM(ldto,20);
DECLARE_CONF_PARAM_NUM(im_policy,2);
DECLARE_CONF_PARAM_NUM(large_io,128);
DECLARE_CONF_PARAM_NUM(log_level,0);


int tmf_reset = 1;
int max_luns = 8;
int ddm = 0;
int enable_msix = 1;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,6)
MODULE_PARM_DESC(eto, " Extended TimeOut, Default=30, Range => 0-3600");
MODULE_PARM_DESC(ldto, " Link Down TimeOut, Default=20, Range => 0-3600");
MODULE_PARM_DESC(im_policy, " Interrupt Moderation Policy, Default=2, Range => 0-4");
MODULE_PARM_DESC(large_io, "Maximum transfer size in a single I/O request in KiloBytes, Default=128, Range => 128-512");
MODULE_PARM_DESC(log_level,"Enable Logging for specific/all paths in driver [Default : Disabled]");
#endif

#if !defined(__VMKERNEL_MODULE__)
DECLARE_CONF_PARAM_STR(drv_name, "be2iscsi");
DECLARE_CONF_PARAM_STR(be2iscsi_version,BUILD_STR);
#endif

unsigned int log_init_unload = 0;
unsigned int log_config = 0;
unsigned int log_io = 0;
unsigned int log_eh = 0;
unsigned int log_ioctl = 0;


// These values have been removed from MODULE_PARM
unsigned int ios_per_lun = 256;
unsigned int cxn_cmds = 256;
unsigned int cxns = 128;
unsigned int tgts= 128;
unsigned int logouts = 64;
unsigned int tmfs = 16;
unsigned int nops = 16;
unsigned int ioctls= 16;
unsigned int eqs = 1024;
unsigned int cqs = 1024;
unsigned int defpdu_sz = 128;
unsigned int sge = 32;
unsigned int l_to_t = 10;  //logout_timeout_ticks
unsigned int mcs = 1;
unsigned int ticks_per_sec = 2;
unsigned int dbg = 0x8000000;
unsigned int cq_wm = 16;
unsigned int eq_wm = 16;               
unsigned int eqt_wm = 48;
unsigned int cxn_tmfs = 1;               
unsigned int inv_ticks = 2;
unsigned int tgt_def_tick = 8;
unsigned int to_factor = 2;
unsigned int io_retries_b4_erl0 = 3;
unsigned int hc_freq = 5;
unsigned int min_wm = 8;
unsigned int max_wm = 240;
unsigned int wm_incs = 16;
unsigned int max_ir = 10000;
unsigned int min_ir = 3500;
unsigned int iot4_dic = 16;
unsigned int ior4_dic = 8;
unsigned int intr_res = 500;
unsigned int intr_steps = 2200;
unsigned int sync_cd_ioc = 0;


/* OSM's function pointer table exposed to ACIT */
ACIT_OSM_API_TABLE      osm_api_table = BE2ISCSI_OSM_API_TABLE;

extern struct file_operations se_fops;

struct notifier_block se_shutdown_notifier =
{
	se_shutdown_event,
	NULL,
	0
};

#ifdef BE2ISCSI_VARYIO_SHT_EXT
struct scsi_host_template_Ext driver_template = BE2ISCSI_SHT;
#else
struct scsi_host_template driver_template = BE2ISCSI_SHT;
#endif


PACIT_VOID
se_pci_alloc_consistent( struct	pci_dev *pcidev, 
			int	size_in_bytes,
			dma_addr_t	*bus_add)
{
	PACIT_VOID	vaddr = NULL;
#ifdef VIRT_TO_BUS_NOTSUPPORTED
	se_dma_mem_profile	*pprofile;
	se_hba	*phba;
	phba = (se_hba *) pci_get_drvdata(pcidev);
#endif
	vaddr = (PACIT_VOID) pci_alloc_consistent(pcidev, size_in_bytes, bus_add);
#ifdef VIRT_TO_BUS_NOTSUPPORTED
	if (vaddr != NULL)
		pprofile = se_add_dma_mem_profile( phba, (unsigned long) vaddr,
					*bus_add, size_in_bytes);
#endif
	return vaddr;
}

void
se_pci_free_consistent( struct	pci_dev *pcidev, 
			int	size_in_bytes,
			void	*va,
			dma_addr_t	bus_add)
{
#ifdef VIRT_TO_BUS_NOTSUPPORTED
	se_hba	*phba;
	phba = (se_hba *) pci_get_drvdata(pcidev);
	se_remove_dma_mem_profile( phba, (unsigned long) va);
#endif
	pci_free_consistent(pcidev, size_in_bytes, va, bus_add);
}

ACIT_STATUS
se_alloc_mem_desc(se_hba *phba)
{
	ACIT_STATUS acit_status, acit_status1;
	int index = 0;
#ifndef __VMKERNEL_MODULE__
	int retry_count = 0;
#endif
	BEGIN;

	/* Iterate over all memory descriptors asked by ACIT */
	do {

		dma_addr_t                 bus_add;
		ACIT_MEM_DESCRIPTOR        acit_mem_desc;

		memset(&acit_mem_desc, 0, sizeof(acit_mem_desc));

		acit_status = acitapi_get_mem_descriptor(phba,
					&phba->controller_key, 
					&phba->acit_device, index, 
					&acit_mem_desc);

		if ((acit_status == ACIT_STATUS_ERROR) ||
			(acit_status == SHTDWN_EVENT))
				break; 

		if (!acit_mem_desc.size_in_bytes) {
			index++;
			continue; 
		}
 
#ifndef __VMKERNEL_MODULE__
/*
 * Xen dynamic memory control sometimes causes
 * allocation to fail.  Retry mem alloc provides
 * OS a chance to up its available mem requirements.
 * Also, this will avoid module initialization failure.
 */
retry_alloc:
#endif
		acit_mem_desc.virtual_address = (PACIT_VOID)
			se_pci_alloc_consistent(phba->pcidev, 
				acit_mem_desc.size_in_bytes,
				&bus_add);

		if (!acit_mem_desc.virtual_address) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"mem alloc failed for desc "
				"index %d, size %d\n", index,
				acit_mem_desc.size_in_bytes);
#ifndef __VMKERNEL_MODULE__
			if (retry_count < 10) {
				retry_count++;
				goto retry_alloc;
			}
#endif
			goto free_mem;
		}
#ifndef __VMKERNEL_MODULE__
		retry_count = 0;
#endif

		acit_mem_desc.bus_address.u.a64.address = (__u64)bus_add;

		OSM_LOG_INIT_UNLOAD(phba,KERN_INFO,"acitapi_set_mem_descriptor : index %d size %d "
			"bus_add %lx VA 0x%lx\n",
			index, acit_mem_desc.size_in_bytes,
			(unsigned long)bus_add, 
			(unsigned long)(acit_mem_desc.virtual_address));

		acit_status1 = acitapi_set_mem_descriptor(phba,
			&phba->controller_key, index, &acit_mem_desc);

		if (acit_status1 == SHTDWN_EVENT)
		      break; 

		index++;
	} while (acit_status == ACIT_STATUS_MORE_MEMORY);

	return acit_status;

free_mem:
	se_free_mem_desc(phba);
	OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"se_alloc_mem_desc: error\n");
	LnxLogOSMError(UNI_CFG_LOG_LEVEL_CRITICAL,
		UNI_PHASE_INIT,
		UNI_SUBMODULE_LNXOSM_GENERIC,
		OSM_EC_RESOURCE_ALLOCATION_FAILURE);
	return acit_status;
}


/*++
Function:
	       se_free_mem_desc 

Parameter:
	phba:           [in]      pointer to se_hba structure

Return:
	nothing 

Description:
	This routine releases the memory allocated as acit memory descriptors

 --*/

void se_free_mem_desc(se_hba  *phba)
{
	ACIT_U32 index = 0;
	ACIT_U32 more_mem_needed;
	BEGIN;
	do {
		ACIT_MEM_DESCRIPTOR        acit_mem_desc;

		more_mem_needed = acitapi_get_mem_descriptor(phba,
					&phba->controller_key, 
					&phba->acit_device, index, 
					&acit_mem_desc);

		if (acit_mem_desc.virtual_address) {
			se_pci_free_consistent(phba->pcidev, 
				acit_mem_desc.size_in_bytes,
				(void *)acit_mem_desc.virtual_address,
				acit_mem_desc.bus_address.u.a64.address);
		}
		index++;
	} while (more_mem_needed == ACIT_STATUS_MORE_MEMORY);
}

#ifdef VIRT_TO_BUS_NOTSUPPORTED
se_dma_mem_profile *
se_add_dma_mem_profile( se_hba *phba, unsigned long va, unsigned long pa, unsigned long size)
{
	se_dma_mem_profile *pnew_profile=NULL;
	ACIT_U32_64     flags = 0;

	pnew_profile = (se_dma_mem_profile *)kmalloc(sizeof(se_dma_mem_profile),
				GFP_ATOMIC);

	if (!pnew_profile) {
		printk("%s: failed dma mem profile alloc\n",__func__);
		return NULL;
	}

	pnew_profile->va = va;
	pnew_profile->pa = pa;
	pnew_profile->size = size;

	spin_lock_irqsave(&phba->dma_profile_lock,flags);
	list_add(&pnew_profile->queue_link, &phba->dma_profile);
	spin_unlock_irqrestore(&phba->dma_profile_lock,flags);

	return pnew_profile;
}

void
se_remove_dma_mem_profile( se_hba *phba, unsigned long va)
{
	se_dma_mem_profile *pprofile;
	struct list_head     *lh, *lh_sf = NULL;
	ACIT_U32_64     flags = 0;

	spin_lock_irqsave(&phba->dma_profile_lock,flags);
	list_for_each_safe(lh, lh_sf, &phba->dma_profile) {
		pprofile = list_entry(lh, se_dma_mem_profile, queue_link);
		if (pprofile->va == va) {
			list_del_init(lh);
			kfree( pprofile);
			break;
		}
	}
	spin_unlock_irqrestore(&phba->dma_profile_lock,flags);
}

se_dma_mem_profile *
se_get_dma_mem_profile( se_hba *phba, unsigned long va)
{
	se_dma_mem_profile *pprofile = NULL;
	struct list_head     *lh, *lh_sf = NULL;
	ACIT_U32_64     flags = 0;

	spin_lock_irqsave(&phba->dma_profile_lock,flags);
	list_for_each_safe(lh, lh_sf, &phba->dma_profile) {
		pprofile = list_entry(lh, se_dma_mem_profile, queue_link);
		if ((pprofile->va <= va) && ((pprofile->va + pprofile->size) >= va))
			break;
	}

	spin_unlock_irqrestore(&phba->dma_profile_lock,flags);
	return pprofile;
}

inline
unsigned long
se_virt_to_bus( se_hba *phba, void * posm_ref, unsigned long va)
{
	se_dma_mem_profile *pprofile;
	struct list_head     *lh, *lh_sf = NULL;
	unsigned long offset=0;
	ACIT_U32_64     flags = 0;

	if ((posm_ref) && (posm_ref != (void *) phba)) {
		pprofile = (se_dma_mem_profile *) posm_ref;
		offset = (unsigned long)va - (unsigned long)pprofile->va;
		return (pprofile->pa + offset);
	}

	spin_lock_irqsave(&phba->dma_profile_lock,flags);
	list_for_each_safe(lh, lh_sf, &phba->dma_profile) {
		pprofile = list_entry(lh, se_dma_mem_profile, queue_link);
		if ((va >= pprofile->va) && ( va <= pprofile->va + pprofile->size)) {
			offset = (unsigned long)va - (unsigned long)pprofile->va;
			spin_unlock_irqrestore(&phba->dma_profile_lock,flags);
			return (pprofile->pa + offset);
		}
	}
	spin_unlock_irqrestore(&phba->dma_profile_lock,flags);

	printk(KERN_ERR "se_virt_to_bus: error: addr not in local table\n");
	return virt_to_bus((void *)va);
}
#endif

#define GET_CACHE_NAME(ptr, name, count)  \
	snprintf(ptr, 20, "%s%d", name, count);
ACIT_STATUS
se_alloc_osm_res(se_hba *phba)
{
	unsigned int osm_lun_mem_size;
	unsigned int req_ws_size;
	BEGIN;
	/*
	 * Our memory allocation scheme in the init is as follows.
	 * There are 3 main structures, channel, target and LUN.
	 * In init we set up a skeleton structures with pointers
	 * for later use. Actual memory for target and LUN is 
	 * allocated on the fly using kernel's memory cache functions.
	 * We initialize the area for cache here.
	 * A channel structure contains memory for a channel including
	 * pointers for all possible targets it may have. Similarly,
	 * a target structure contains memory for a target including
	 * pointers for all possible LUNs it may have. ACIT workspace
	 * memory are not part of these structures, seperate cache
	 * areas are created here.
	 */

	/* TODO: hardcoded for now, get from acit/some defines */

	phba->max_luns_per_target = MAX_LUN + 1;

	phba->osm_target_mem_size = sizeof(se_target) + 
					sizeof(se_device *) *
					phba->max_luns_per_target;

	GET_CACHE_NAME(phba->osmtarget_name, OSM_TARGET_CACHE_NAME,
		phba->hba_unique_id);

	phba->posm_target_cache = kmem_cache_create(
			phba->osmtarget_name, phba->osm_target_mem_size,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
			0, SLAB_HWCACHE_ALIGN, NULL,NULL);
#else
			0, SLAB_HWCACHE_ALIGN, NULL);
#endif
	if (!phba->posm_target_cache) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"cache create failed for osm target "
			"alloc size %d\n", phba->osm_target_mem_size);
		goto error;
	}

	if (phba->acit_target_ws_size) {

		GET_CACHE_NAME(phba->acittarget_name, ACIT_TARGET_WS_CACHE_NAME,
			 phba->hba_unique_id);

		phba->pacit_target_ws_cache = kmem_cache_create(
			phba->acittarget_name, phba->acit_target_ws_size, 0,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	  		SLAB_HWCACHE_ALIGN| SLAB_MUST_HWCACHE_ALIGN, NULL,
#else
			SLAB_HWCACHE_ALIGN , 
#endif
			NULL);

		if (!phba->pacit_target_ws_cache) {
			OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"cache create failed for acit "
					"target ws alloc size %d\n", 
					phba->acit_target_ws_size);
			goto error;
		}
	}

	osm_lun_mem_size = sizeof(se_device);

	GET_CACHE_NAME(phba->osmlun_name, OSM_LUN_CACHE_NAME, phba->hba_unique_id);
	phba->posm_lun_cache = kmem_cache_create(phba->osmlun_name,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
			osm_lun_mem_size, 0, SLAB_HWCACHE_ALIGN, NULL,NULL);
#else
			osm_lun_mem_size, 0, SLAB_HWCACHE_ALIGN, NULL);
#endif
	if (!phba->posm_lun_cache) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"cache create failed for osm lun "
			"alloc size %d\n", osm_lun_mem_size);
		goto error;
	}

	if (phba->acit_lun_ws_size) {

		GET_CACHE_NAME(phba->acitlun_name, ACIT_LUN_WS_CACHE_NAME, phba->hba_unique_id);
		phba->pacit_lun_ws_cache = kmem_cache_create(phba->acitlun_name,
			     phba->acit_lun_ws_size, 0,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	  		SLAB_HWCACHE_ALIGN| SLAB_MUST_HWCACHE_ALIGN, NULL,
#else
			SLAB_HWCACHE_ALIGN,
#endif
			NULL);

		if (!phba->pacit_lun_ws_cache) {
			OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"cache create failed for acit "
					"lun ws alloc size %d\n", 
					phba->acit_lun_ws_size);
			goto error;
		}
	}

	/* Initialize the free list of IO*/
	/* In each cxn for sending TMF/Logout reserve an ICD. 
	 * The IOS_PER_CTRL w.r.t to h/w gets to (1024 - cxns) and w.r.t SML it's 1024.
	 */
	phba->max_ios_per_controller = BE2ISCSI_MAX_Q_DEPTH - cxns;
	req_ws_size = phba->acit_io_ws_size < phba->acit_tmf_ws_size ? 
			phba->acit_tmf_ws_size : phba->acit_io_ws_size;   
	if (se_init_free_req_list(phba, req_ws_size) != ACIT_STATUS_SUCCESS) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"Not a single IO cmd was allocated\n");
		goto error;

	}

	return ACIT_STATUS_SUCCESS;

error:
	se_free_osm_res(phba);
	LnxLogOSMError(UNI_CFG_LOG_LEVEL_CRITICAL,
		UNI_PHASE_INIT,
		UNI_SUBMODULE_LNXOSM_GENERIC,
		OSM_EC_RESOURCE_ALLOCATION_FAILURE);

	return ACIT_STATUS_INSUFFICIENT_RESOURCE;
}
void
se_free_osm_res(se_hba *phba)
{
	BEGIN;
	if (phba->pacit_lun_ws_cache)
		kmem_cache_destroy(phba->pacit_lun_ws_cache);
	if (phba->posm_lun_cache) 
		kmem_cache_destroy(phba->posm_lun_cache);
	if (phba->pacit_target_ws_cache)
		kmem_cache_destroy(phba->pacit_target_ws_cache);
	if (phba->posm_target_cache) 
		kmem_cache_destroy(phba->posm_target_cache);

	phba->pacit_lun_ws_cache = NULL; 
	phba->posm_lun_cache     = NULL;
	phba->posm_target_cache  = NULL;
	return;
}

/*++
Function:
	aim_alloc_mem_params

Parameter:
	phba:           [in]      pointer to se_hba structure
	target_ws_size: [in]      ACIT WS size for target
	lun_ws_size:    [in]      ACIT WS size for LUN
	io_ws_size:     [in]      ACIT WS size for IO
	tmf_ws_size:    [in]      ACIT WS size for TMF

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_INSUFFICIENT_RESOURCE - memory allocation failure

Context:
	Init context

Description:
	This routine sets up memory parameters for the OSM. 
	Negotiation with ACIT for mem_level is also done in this
	routine.
 --*/
ACIT_STATUS 
aim_alloc_mem_params(se_hba  *phba)
{
	int         mem_level;
	ACIT_STATUS acit_status;
	BEGIN;

	mem_level = ACIT_MEM_LEVEL_DEFAULT;

#ifdef VIRT_TO_BUS_NOTSUPPORTED
	INIT_LIST_HEAD(&phba->dma_profile);
#endif

	while (mem_level >= 0) {
		acit_status = acitapi_set_mem_params(phba, 
				&phba->controller_key, mem_level);
		if (acit_status != ACIT_STATUS_SUCCESS) {
			return ACIT_STATUS_INSUFFICIENT_RESOURCE;
		}
		acit_status = se_alloc_mem_desc(phba);
		if (acit_status != ACIT_STATUS_SUCCESS) {
			return ACIT_STATUS_INSUFFICIENT_RESOURCE;
		}
		acit_status = se_alloc_osm_res(phba);
		if (acit_status != ACIT_STATUS_SUCCESS) {
			se_free_mem_desc(phba);
			return ACIT_STATUS_INSUFFICIENT_RESOURCE;
		}
		return ACIT_STATUS_SUCCESS;
	}
	return ACIT_STATUS_INSUFFICIENT_RESOURCE;
}

/*++
Function:
	aim_free_mem_params

Parameter:
	phba: [in] pointer to se_hba structure

Return:
	ACIT_STATUS_SUCCESS - no errors

Context:
	Init context

Description:
	This routine cleans memory allocated by aim_alloc_mem_params
 --*/

ACIT_STATUS 
aim_free_mem_params(se_hba *phba)
{
	se_free_osm_res(phba);
	se_free_mem_desc(phba);
	se_delete_free_req_list(phba);
	
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	se_initialize_lun

Parameter:
	ptarget:        [in]    pointer to target
	pdevice:        [in]    pointer to device
	channel:        [in]    bus number
	target:         [in]    target number
	lun:            [in]    lun
	pacit_lun:      [in]    pointer to 8 byte LUN

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_INSUFFICIENT_RESOURCE - memory allocation failure
	errors if any from acit_init_io_ws

Context:
	Any

Description:
	This routine sets up a LUN at the given BTL address. The
	workspace allocation and initialization is also done in
	this routine.

--*/

inline
ACIT_STATUS
se_initialize_lun(
		se_target     *ptarget,
		se_device     **ppdevice,
		ACIT_UEXACT32   channel,
		ACIT_UEXACT32   target,
		ACIT_UEXACT32   lun,
		PACIT_LUN       pacit_lun)
{
	ACIT_STATUS acit_status = ACIT_STATUS_SUCCESS;
	se_hba *phba = ptarget->pparent_hba;
	se_device *pdevice = NULL;

	BEGIN;
	pdevice = kmem_cache_alloc( phba->posm_lun_cache,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
			SLAB_ATOMIC);
#else
			GFP_KERNEL);
#endif
	*ppdevice = pdevice;

	if (!pdevice) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Mem alloc for new lun failed\n");
		return ACIT_STATUS_INSUFFICIENT_RESOURCE;
	}


	memset(pdevice,0, sizeof(se_device));
	INIT_LIST_HEAD(&pdevice->sibling_list);

	pdevice->pparent_target = ptarget;
	pdevice->scsi_channel = channel;
	pdevice->scsi_id = target;
	pdevice->scsi_lun = lun;
	pdevice->lun.u.lun_complete = pacit_lun->u.lun_complete;
	pdevice->pacit_lun_ws = NULL;

	if (phba->acit_lun_ws_size) {

		pdevice->pacit_lun_ws = kmem_cache_alloc(
						phba->pacit_lun_ws_cache,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
						SLAB_ATOMIC);
#else
						GFP_KERNEL);
#endif
		if (!pdevice->pacit_lun_ws) {

			OSM_LOG_CONFIG(phba, KERN_ERR,"__se_initialize_lun failed "
				"allocating workspace \n");

			return ACIT_STATUS_INSUFFICIENT_RESOURCE;

		}
	}

	pdevice->lun_key.osm_key = pdevice;

	acit_status = acitapi_init_lun_ws( phba,
					   &phba->controller_key,
					   ptarget->posm_target_key,
					   &pdevice->lun_key,
					   pdevice->pacit_lun_ws);

	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"acit_init_lun_ws failed\n");
		if (pdevice->pacit_lun_ws)
			kmem_cache_free(phba->pacit_lun_ws_cache, 
					pdevice->pacit_lun_ws);
	}

	if (lun > phba->top_scsi_lun) {
		phba->top_scsi_lun = lun;
	}

	ptarget->num_luns++;

	list_add(&pdevice->sibling_list, &ptarget->lun_list);

	return acit_status;
}

void
se_destroy_lun(se_hba *phba, se_device *pdevice)
{
	se_target *ptarget = pdevice->pparent_target;

	OSM_LOG_INIT_UNLOAD(phba,KERN_INFO,"destroying pdevice=%p BTL:%d:%d:%d\n",
		pdevice,
		pdevice->scsi_channel,
		pdevice->scsi_id,
		pdevice->scsi_lun);

	OSM_ASSERT(ptarget->num_luns != 0);

	pdevice->state |= SE_DEV_OFFLINE;

	if (pdevice->pacit_lun_ws) {
		kmem_cache_free(phba->pacit_lun_ws_cache, 
				pdevice->pacit_lun_ws);
	}

	list_del_init(&pdevice->sibling_list);

	if (pdevice->scsi_lun == phba->top_scsi_lun) {
		/* TODO: This should point to next smaller LUN number */
		phba->top_scsi_lun--;
	}

	kmem_cache_free(phba->posm_lun_cache, pdevice);

	ptarget->num_luns--;
}

/*++
Function:
	se_verify_parameters

Parameter:
	 None
Return:
	None

Context:
	Expected to be called from init context
Description:
	This routine is used to check whether the parameters provided by the user are valid. If any value is out of range, it picks up the default value.

--*/


void se_verify_parameters(void)
{
	if (eto < 0 || eto > ACIT_CFG_MAX_ETO) {
		printk("be2iscsi: WARNING: eto value %d out of range.  Valid range is 0 - %d.  \
				Using default value of %d.\n", eto,ACIT_CFG_MAX_ETO,ACIT_CFG_DEFAULT_ETO);

		eto = ACIT_CFG_DEFAULT_ETO;
	}
	if (ldto < 0 || ldto > ACIT_CFG_MAX_LDTO) {
		printk("be2iscsi: WARNING: ldto value %d out of range.  Valid range is 0 - %d. \
				Using default value of %d.\n", ldto,ACIT_CFG_MAX_LDTO,ACIT_CFG_DEFAULT_LDTO);

		ldto = ACIT_CFG_DEFAULT_LDTO;
	}

	if (large_io < ACIT_CFG_DEFAULT_LARGE_IO || large_io > ACIT_CFG_MAX_LARGE_IO) {
		printk("be2iscsi: WARNING: large_io value %d out of range.  Valid range is %d - %d. \
				Using default value of %d.\n", large_io,ACIT_CFG_DEFAULT_LARGE_IO,
				ACIT_CFG_MAX_LARGE_IO,ACIT_CFG_DEFAULT_LARGE_IO);
		
		large_io = ACIT_CFG_DEFAULT_LARGE_IO;
	}

	if (im_policy < 0 || im_policy > ACIT_CFG_MAX_IM_POLICY ) {
		printk("be2iscsi : WARNING: im_policy value = %d out of range . Valid Range is  0 - %d. \
				Using Default Value = %d \n",im_policy,ACIT_CFG_MAX_IM_POLICY,ACIT_CFG_DEFAULT_IM_POLICY);
		
		im_policy= ACIT_CFG_DEFAULT_IM_POLICY;
	}
}

int
se_setup_msix(
se_hba		*phba,
struct pci_dev	*pcidev, 
PACIT_DEVICE	pacit_device
)
{
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,5)
	unsigned		i;
	unsigned		n;
	ACIT_UEXACT32		retval; 
	struct msix_entry	entries[BE_NUM_MSIX_ENTRIES];
	struct be_msix_entry  	*be_entry;

	for (i = 0 ; i < BE_NUM_MSIX_ENTRIES ; i++) {
		entries[i].entry = def_msix_entries[i].entry;
	}

	if (pci_find_capability(pcidev, PCI_CAP_ID_MSIX) == 0) {

		return 0;
	}

	retval = pci_enable_msix(pcidev, entries,
		ARRAY_SIZE(entries));

	if (retval) {
		printk(KERN_ERR "pci_enable_msix failed\n");

		return 0;
	}

	for (i = 0 ; i < BE_NUM_MSIX_ENTRIES ; i++) {

		be_entry = &phba->msix_entries[def_msix_entries[i].index];

		be_entry->msix_vector = entries[i].vector;
		be_entry->msix_entry = entries[i].entry;
		be_entry->msix_enabled = 0;

		if (request_irq(be_entry->msix_vector,
		    def_msix_entries[i].handler, 0,
		    def_msix_entries[i].name, phba) != 0) {

			printk(KERN_ERR "msix request_irq failed \n");
			break;

		}

		be_entry->msix_enabled = 1;

	}

	if (i != BE_NUM_MSIX_ENTRIES) {

		for (n = 0 ; n < BE_NUM_MSIX_ENTRIES ; n++) {

			be_entry = &phba->msix_entries[
				def_msix_entries[n].index];

			if (be_entry->msix_enabled) {
				free_irq(be_entry->msix_vector, phba);
			}
		}

		pci_disable_msix(pcidev);

		return 0;
	}

	return 1;
#else
	return 0;
#endif
}

/*++
Function:
	se_initialize_hba

Parameter:
	pcidev:         [in]    pointer to pci_dev structure
	pacit_device:   [in]    pointer to ACIT_DEVICE
	phba:           [out]   pointer to pointer to HBA struct

Return:
	0 - no errors
	-ENODEV - device not supported or could not be enabled
	-ENOMEM - memory allocation failure

Context:
	Expected to be called from init context once per HBA.

Description:
	This routine performs ACIT specific initializations that are
	needed by an OSM. This function is to be called on a per-HBA
	basis, with pacit_device parameter populated with PCI details.

	If successful, this routine allocates a structure for se_hba
	and returns the pointer in phba parameter.

--*/

ACIT_SEXACT32 
se_initialize_hba(
struct pci_dev	*pcidev, 
PACIT_DEVICE	pacit_device, 
se_hba		*phba
)
{
	ACIT_STATUS             acit_status = 0;
	ACIT_UEXACT32           controller_ws_size = 0;  
	PACIT_CONTROLLER_KEY    pacit_controller_key = NULL;
	unsigned		msix_initialized;
	unsigned 		error_code = OSM_EC_HW_INIT_FAILED;

	BEGIN;

	OSM_ASSERT(pacit_device);
	OSM_ASSERT(phba);

	/* check if acit support this device. */
	acit_status = acitapi_check_supported_device(NULL, pacit_device);
	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"acit does not support device\n");
		error_code = OSM_EC_UNSUPPORTED_HARDWARE;
		goto no_device;
	}
	/* 
	 * Detemine ACIT ws sizes and allocate them.
	 */
	acit_status = acitapi_get_ws_sizes(&acit_api_version,
			&controller_ws_size,
			&phba->acit_target_ws_size, &phba->acit_lun_ws_size,
			&phba->acit_io_ws_size, &phba->acit_tmf_ws_size);

	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"acit_get_ws_sizes failed\n");
		error_code = OSM_EC_ACIT_API_FAILED;
		goto agws_failed;        
	}

	phba->pcontroller_ws = (PACIT_VOID)
			       kmalloc(controller_ws_size, GFP_ATOMIC);

	if (phba->pcontroller_ws == NULL) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"memory alloc failed for Controller_WS size : %d\n",controller_ws_size);

		error_code = OSM_EC_RESOURCE_ALLOCATION_FAILURE;
		goto agws_failed;        
	}


	/* Tell the ACIT about the ws allocated */
	pacit_controller_key            = &(phba->controller_key);
	pacit_controller_key->osm_key   = phba;

	acit_status = acitapi_init_controller_ws(phba,
			&phba->controller_key, phba->pcontroller_ws);

	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"acit_init_controller_ws failed\n");
		error_code = OSM_EC_ACIT_API_FAILED;
		goto aicws_failed; 
	}

	acit_status = acitapi_set_osm_api_table(phba, &phba->controller_key,
				&osm_api_table, sizeof(osm_api_table));

	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"acit_set_osm_api_table failed \n");
		error_code = OSM_EC_OSM_API_TABLE_INITIALIZE_FAILED;
		goto aicws_failed;
	}

	acit_status = acitapi_init_configurable_params(phba,
					&phba->controller_key);

	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"acit_init_configurable_params\n");
		error_code = OSM_EC_ACIT_API_FAILED;
		goto aicws_failed;
	}

	/* Check se_init_params(), for the index values */
	phba->max_iscsi_targets = se_param[2].val;
	phba->max_luns_per_target = se_param[3].val;

	sprintf(phba->name, "be2iscsi");
	
	phba->hba_unique_id = se_get_new_hba_unique_id(pcidev);

	acit_status = aim_alloc_mem_params(phba);

	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"aim_alloc_mem_params failed\n");
		error_code = OSM_EC_RESOURCE_ALLOCATION_FAILURE;
		goto aicws_failed;
	}


	/* All set! Ask ACIT to initialize hardware */

	if (gcrashmode) {
		acit_status = acitapi_init_controller(phba,
			&phba->controller_key, ACIT_INIT_TYPE_CRASHDUMP);
	} else {
		acit_status = acitapi_init_controller(phba,
			&phba->controller_key, ACIT_INIT_TYPE_NORMAL);
	}
	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"acit_init_controller failed\n");
		error_code = OSM_EC_ACIT_API_FAILED;
		goto aic_failed;                
	}

	init_completion(&phba->worker_thread.work_queued);

	/*
	 * Looks like its a good time to configure and enable IRQ
	 */
	msix_initialized = 0;

	if (enable_msix) {
		/*
		 * Try MSI-X first ...
		 */
		if (se_setup_msix(phba, pcidev, pacit_device) != 0) {
			/*
			 * It worked.
			 */
			msix_initialized = 1;
		} else {
			enable_msix = 0;
		}
	}

	if (msix_initialized == 0) {
		/*
		 * If we couldn't or didn't want to use MSI-X, then
		 * try regular interrupts.
		 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
		if (request_irq(pcidev->irq, se_isr, SA_INTERRUPT | SA_SHIRQ,
#else 
		if (request_irq(pcidev->irq, se_isr,IRQF_DISABLED | IRQF_SHARED, 
#endif
			"be2iscsi", phba) != 0) {

			OSM_LOG_INIT_UNLOAD(phba,KERN_CRIT,"Cannot register irq\n");
			goto rirq_failed;
		}
	}

	acitapi_enable_irq(phba, &phba->controller_key);

	spin_lock(&phba->acitapi_lock);
	phba->state |= SE_STATE_INT_ENABLED;
	spin_unlock(&phba->acitapi_lock);

	/* 
	 * Successfully completed initialize, chain it to HBA
	 * structure
	 */

	down(&se_configuration_lock);
	list_add(&phba->hba_queue, &hba_chain);

	hba_count++;

	up(&se_configuration_lock);

	return 0;

rirq_failed:
aic_failed:
	aim_free_mem_params(phba);

aicws_failed:
	kfree(phba->pcontroller_ws);

agws_failed:
	pci_set_drvdata(pcidev, NULL);

no_device:
	LnxLogOSMError(UNI_CFG_LOG_LEVEL_CRITICAL,
		UNI_PHASE_INIT,
		UNI_SUBMODULE_LNXOSM_GENERIC,
		error_code);
	return -ENODEV;
}

int se_get_new_hba_unique_id(struct pci_dev	*pcidev)
{
	int i;
	unsigned int pciid;

	pciid = SE_HBA_REMOVED | (pcidev->devfn & 0xffffff) | (pcidev->bus->number << 24);

	/* look for previous pos */
	for (i = 0; i < BE2ISCSI_MAX_HBA; i++) {
		if (se_hba_uid[i] == pciid) {
			se_hba_uid[i] = pciid & (~SE_HBA_REMOVED);
			return i;
		}
	}

	/* look for new pos */
	for (i = 0; i < BE2ISCSI_MAX_HBA; i++) {
		if (se_hba_uid[i] == 0) {
			se_hba_uid[i] = pciid & (~SE_HBA_REMOVED);
			return i;
		}
	}

	/* look for first abandoned pos */
	for (i = 0; i < BE2ISCSI_MAX_HBA; i++) {
		if (se_hba_uid[i] & SE_HBA_REMOVED) {
			se_hba_uid[i] = pciid & (~SE_HBA_REMOVED);
			return i;
		}
	}

	/* all else failed, give it the hba_count */
	return hba_count;
}

#ifdef __VMKLNX__
void
se_vmware_state_change(
se_hba		*phba
)
{
	struct vmk_ScsiAdapter *vmk_adapter;
	struct vmklnx_ScsiAdapter *vmklnx_adapter;
	se_target     			*ptarget;
	se_device			*pdevice;
	
	if(phba != NULL){
		vmklnx_adapter = (struct vmklnx_ScsiAdapter *)phba->host->adapter;
		if(vmklnx_adapter)
			vmk_adapter = (struct vmk_ScsiAdapter *)vmklnx_adapter->vmkAdapter;
	}else{
		printk("PHBA is NULL \n");
		return;
	}

	list_for_each_entry(ptarget, &phba->target_list, sibling_list) {

		list_for_each_entry(pdevice, &ptarget->lun_list,
				sibling_list) {

			if(vmk_adapter){
				if(VMK_OK != vmk_ScsiNotifyPathStateChangeAsync(vmk_adapter,pdevice->scsi_channel,pdevice->scsi_id,pdevice->scsi_lun)){

					printk("vmk_ScsiNotifyPathStateChangeAsync Failed for C%d:T%d:L%d\n",pdevice->scsi_channel,pdevice->scsi_id,pdevice->scsi_lun);
						}
			}

		}
	}
}


#ifdef REGISTER_ISCSI

extern struct iscsi_transport be2iscsi_iscsi_transport;

ACIT_STATUS
se_vmware_add_target(
se_hba			*phba,
se_target		*ptarget,
PACIT_TARGET_ENTRY	ptarget_entry
)
{
	unsigned		portal_group_tag;
	int status = 0 ,len =0;

	ptarget->sess = iscsi_alloc_session(phba->host,
		&be2iscsi_iscsi_transport);

	if (iscsi_add_session(ptarget->sess, ptarget->target, ptarget->bus)) {
		return ACIT_STATUS_ERROR;
	}

	ptarget->conn = iscsi_create_conn(ptarget->sess, 0);

	if (ptarget->conn == NULL) {
		iscsi_remove_session(ptarget->sess);
		iscsi_free_session(ptarget->sess);
		ptarget->sess = NULL;
		return ACIT_STATUS_ERROR;
	}

	/*
	 * Getting all the info required for iscsi host parameters
	 */

	/*
	 * Getting Initiator IP
	 */
	ptarget->sess->dd_data = ptarget;

	PGET_IF_INFO_PARAMS pattr_buf;

	pattr_buf = (PGET_IF_INFO_PARAMS) kmalloc(sizeof(GET_IF_INFO_PARAMS),
		GFP_KERNEL);

	status = se_get_nw_info(phba, pattr_buf);

	if (status) {
		kfree(pattr_buf);
		return ACIT_STATUS_ERROR;
	}

	len = sprintf(phba->buf_ip, "%d.%d.%d.%d",
		pattr_buf->response.ip_address[0].IPAddress[0],
		pattr_buf->response.ip_address[0].IPAddress[1],
		pattr_buf->response.ip_address[0].IPAddress[2],
		pattr_buf->response.ip_address[0].IPAddress[3]);

	OSM_LOG_CONFIG(phba, KERN_ERR,"IP : %s \n", phba->buf_ip);
	kfree(pattr_buf);

	/*
	 * Getting Initiator Name
	 */
	PISCSI_HBA_NAME_PARAMS pattr_buf1;

	pattr_buf1 = (PISCSI_HBA_NAME_PARAMS)kmalloc(
		sizeof(ISCSI_HBA_NAME_PARAMS), GFP_KERNEL);

	status = se_get_iqn_info(phba, pattr_buf1);

	if (status) {
		kfree(pattr_buf1);
		return ACIT_STATUS_ERROR;
	}

	len = sprintf(phba->buf_name,"%s",pattr_buf1->iscsiname);

	OSM_LOG_CONFIG(phba, KERN_ERR,"InitiatorName is %s\n",pattr_buf1->iscsiname);

	kfree(pattr_buf1);

	/*
	 * Getting Target Properties
	 */
	PGET_A_SESSION_PARAMS psess_buf;

	psess_buf = (PGET_A_SESSION_PARAMS)kmalloc(
		sizeof(GET_A_SESSION_PARAMS), GFP_KERNEL);

        memset(psess_buf, 0x0, sizeof(GET_A_SESSION_PARAMS));

        status = se_get_asession_info(phba,
		psess_buf, ptarget_entry->session_handle);

        if (status) {
                kfree(psess_buf);
		return ACIT_STATUS_ERROR;
        }

	len = sprintf(ptarget->buf_name, "%s",
		psess_buf->response.session_info.target_name);

	len = sprintf(ptarget->buf_isid, "%02x%02x%02x%02x%02x%02x",
		psess_buf->response.session_info.isid[0],
		psess_buf->response.session_info.isid[1],
		psess_buf->response.session_info.isid[2],
		psess_buf->response.session_info.isid[3],
		psess_buf->response.session_info.isid[4],
		psess_buf->response.session_info.isid[5]);

	len = sprintf(ptarget->buf_port, "%u",
		psess_buf->response.session_info.conn_list[0].dest_port);

	len = sprintf(ptarget->buf_ip, "%d.%d.%d.%d",
		psess_buf->response.session_info.conn_list[0].
			dest_ipaddr.IPAddress[0],
		psess_buf->response.session_info.conn_list[0].
			dest_ipaddr.IPAddress[1],
		psess_buf->response.session_info.conn_list[0].
			dest_ipaddr.IPAddress[2],
		psess_buf->response.session_info.conn_list[0].
			dest_ipaddr.IPAddress[3]);	
	
	OSM_LOG_CONFIG(phba, KERN_ERR,"Target props IP: %s Name = %s port =%s isid = %s\n",
		ptarget->buf_ip,
		ptarget->buf_name,
		ptarget->buf_port,
		ptarget->buf_isid);	

	kfree(psess_buf);	

	status = se_get_portal_group_tag(phba, ptarget, &portal_group_tag);

	if (status) {
		return ACIT_STATUS_ERROR;
	}

	ptarget->portal_group_tag = portal_group_tag;

	return ACIT_STATUS_SUCCESS;
}
#endif /* REGISTER_ISCSI */
#endif /* __VMKLNX__ */

/*++
Function:
	se_initialize_target

Parameter:
	phba:           [in]    pointer to hba struct
	entry_data:     [in]    pointer to target entry job data

Return:
	pointer to target entry initialized, upon success
	NULL on errors

Context:
	Called when a ACIT calls osm_async_event with target added event code.

Description:
	This routine allocates memory for a target and sets up the target
	for later use. Setting up involves initializing workspace for the
	target, notifying the ACIT that the target is in use, etc.

--*/
void 
se_initialize_target(
se_hba		*phba,
void		*entry_data
)
{
	se_target     *ptarget;
	PACIT_VOID      pacit_ws = NULL;
	ACIT_UEXACT32   target;
	ACIT_UEXACT32   channel;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)
	ACIT_UEXACT32   lun;
#endif
	ACIT_STATUS     acit_status;
	ACIT_U32_64     flags = 0;
	PACIT_TARGET_ENTRY   ptarget_entry = NULL; 
	PMGMT_BTL_MAPPING    pbtl_mapping = NULL;
	struct se_target_job_data *pdata = 
				(struct se_target_job_data *)entry_data;
	BEGIN;

	ptarget_entry = pdata->ptarget_entry;

	pbtl_mapping = (PMGMT_BTL_MAPPING) kmalloc(sizeof(MGMT_BTL_MAPPING) + 
		(phba->max_luns_per_target - 1) * sizeof(MGMT_LUN_MAPPING),
		GFP_ATOMIC);

	if (pbtl_mapping == NULL) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"pbtl_mapping alloc failed\n");
		goto no_target;
	}

	/* Following function will send GET_SESSION_CONFIGURATION ioctl */
	acit_status = se_get_btl_info(phba, ptarget_entry,
			pbtl_mapping);

	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"Max limit for targets\n");
		goto no_target;
	}

#if defined(__VMKERNEL_MODULE__)
	channel = 0;
#else
	channel = pbtl_mapping->os_bus;
#endif
	target = pbtl_mapping->os_target;       

	ptarget = (se_target *)kmem_cache_alloc(phba->posm_target_cache,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
					SLAB_ATOMIC);
#else
					GFP_KERNEL);
#endif
	if (!ptarget) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"kmem_cache_alloc failed for ptarget\n");
		goto no_target;
	}

	memset(ptarget, 0, phba->osm_target_mem_size);
	ptarget->bus = channel;
	ptarget->target = target;
	ptarget->num_luns = 0;

	OSM_LOG_CONFIG(phba, KERN_ERR,"Channel %d Target %d\n",
			ptarget->bus, ptarget->target);

	phba->host->max_id = 256;

	phba->host->max_channel = phba->top_scsi_channel;

	if (phba->acit_target_ws_size) {
		pacit_ws = (PACIT_VOID)kmem_cache_alloc(
				phba->pacit_target_ws_cache,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
				SLAB_ATOMIC);
#else
				GFP_KERNEL);
#endif
		if (!pacit_ws) {
			OSM_LOG_CONFIG(phba, KERN_ERR,"kmem_cache_alloc failed for ACIT "
				"target WS\n");
			goto no_target_ws;
		}
		memset(pacit_ws, 0, phba->acit_target_ws_size);
	}

	/* Initialize target structure */
	ptarget->pparent_hba = phba;
	ptarget->posm_target_key = ptarget_entry->ptarget_key;
	ptarget->posm_target_key->osm_key = ptarget;
	ptarget->pacit_target_ws = pacit_ws;
	ptarget->pbtl_mapping = pbtl_mapping;

	/*
	 * We initialize the ACIT WS for the target even before
	 * the slot is nailed down. The target can move to a
	 * different slot if a custom map decides so. Even in
	 * that case its OK because we will then move the
	 * entire target entry, not just the contents
	 */
	INIT_LIST_HEAD(&ptarget->lun_list);
	INIT_LIST_HEAD(&ptarget->sibling_list);
	list_add(&ptarget->sibling_list, &phba->target_list);
	acit_status = acitapi_init_target_ws(phba,
				&phba->controller_key,
				ptarget->posm_target_key,
				pacit_ws);

	if (acit_status != ACIT_STATUS_SUCCESS) {
		//PDEBUG(DBG_ERR_HNDL, "acit_init_target_ws failed for target "
		//                "at Bus %d Target %d CKey 0x%lx TKey 0x%lx\n",
		//                channel, target, 
		//                (ACIT_U32_64)&phba->controller_key,
		//                (ACIT_U32_64)ptarget->posm_target_key);
		//goto no_init_target_ws;
		OSM_LOG_CONFIG(phba, KERN_ERR,"acitapi_init_target_ws failed for Target :%d Bus : %d\n",
				 ptarget->target,ptarget->bus);

		goto no_btl_info;

	}
	OSM_LOG_CONFIG(phba, KERN_ERR,"Initializing target workspace .... Done\n");

	ptarget_entry->osm_tflags |= ACIT_OSM_TFLAG_IN_USE;

#ifdef REGISTER_ISCSI
#ifdef __VMKLNX__
	if (se_vmware_add_target(phba, ptarget, ptarget_entry) != 
		ACIT_STATUS_SUCCESS)
	{
		OSM_LOG_CONFIG(phba, KERN_ERR,"se_vmware_add-target failed for Target :%d Bus : %d\n",
				ptarget->bus, ptarget->target);
		goto no_btl_info;
	}
#endif
#endif

	/*Getting ETO and storing in target*/
	se_get_session_eto(phba, ptarget_entry,ptarget);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)
	for(lun = 0 ; lun < max_luns ; lun++)  
		scsi_add_device(phba->host,ptarget->bus,ptarget->target,lun);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	down(&phba->host->scan_mutex);
	scsi_scan_target(phba->host,ptarget->bus,ptarget->target,-1,1);
	up(&phba->host->scan_mutex);
#else
#if defined(__VMKERNEL_MODULE__)
	scsi_scan_target(&phba->host->shost_gendev,ptarget->bus,ptarget->target,-1,1);
#else
	scsi_scan_target(&phba->host->shost_gendev,ptarget->bus,ptarget->target,~0,1);
#endif
#endif
#endif

	OSM_LOG_CONFIG(phba, KERN_ERR,"se_scan_target_for_luns ....Done\n");

	/*
	 * send SET_SESSION_CONF ioctl to save btl info in flash
	 * In theory it could fail. ignore if it fails
	 */
	se_set_session_conf(phba, ptarget, ptarget_entry);

	/*
	 * The First TGT is the boot target and we do not need to check
	 */
	phba->boot_tgt_added = 1;

	goto success;

no_target_ws:
	kmem_cache_free(phba->posm_target_cache, ptarget);

no_target:
	if (pbtl_mapping != NULL) {
		kfree(pbtl_mapping); 
	}

no_btl_info:
	LnxLogOSMError(UNI_CFG_LOG_LEVEL_ERROR,
		UNI_PHASE_DONT_CARE,
		UNI_SUBMODULE_LNXOSM_GENERIC,
		OSM_EC_HW_INIT_FAILED);

	/* if login was pending .. complete it */
	if (phba->state & SE_STATE_LOGIN_PENDING) {
		spin_lock(&phba->acitapi_lock);
		phba->state &= ~SE_STATE_LOGIN_PENDING;
		phba->login_status = ACIT_STATUS_ERROR;
		spin_unlock(&phba->acitapi_lock);
		complete(&phba->login_completion);
	}
	return;

success:
	/* if login was pending .. complete it */
	if (phba->state & SE_STATE_LOGIN_PENDING) {
		spin_lock_irqsave(&phba->acitapi_lock,flags);
		phba->state &= ~SE_STATE_LOGIN_PENDING;
		phba->login_status = ACIT_STATUS_SUCCESS;
		spin_unlock_irqrestore(&phba->acitapi_lock,flags);
		complete(&phba->login_completion);
	}
	return;
}

void
se_handle_controller_event(
	se_hba        *phba,
	ACIT_UEXACT32   event_code,
	ACIT_UEXACT32   info)
{
	BEGIN;

	switch (event_code) {
	case ACIT_EVENT_CONTROLLER_LINK_DOWN:
		if (phba->host != NULL) {
			printk("scsi%d: ", phba->host->host_no);
		}

		printk("be2iscsi: OneConnect iSCSI Driver: Link Down\n");

		break;

	case ACIT_EVENT_CONTROLLER_LINK_UP:
		if (phba->host != NULL) {
			printk("scsi%d: ", phba->host->host_no);
		}

		printk("be2iscsi: OneConnect iSCSI Driver: Link Up\n");

		break;

	case ACIT_EVENT_CONTROLLER_LINKDOWN_TIMEOUT:
#if defined(__VMKERNEL_MODULE__)
		se_vmware_state_change(phba);
#endif
		break;

	case ACIT_EVENT_CONTROLLER_BUSY:
	case ACIT_EVENT_CONTROLLER_READY:
	case ACIT_EVENT_CONTROLLER_DISCOVERY_COMPLETE:
		/*
		 * Nothing to do for these events for now.
		 */
		break;
	}
}


void
se_handle_target_event(
	se_hba        *phba,
	ACIT_UEXACT32   event_code,
	ACIT_UEXACT32   info)
{
	struct se_timer_queue	*job;
	PACIT_TARGET_TABLE      pacit_target_table;
	PACIT_TARGET_ENTRY      ptarget_entry;
	se_target		*ptarget;

	BEGIN;

	pacit_target_table = phba->pacit_target_table;

	OSM_LOG_CONFIG(phba, KERN_ERR,"phba %p: ACIT_EVENT %d : %d\n",
					phba, event_code, info);
	switch (event_code) {

	case ACIT_EVENT_TARGET_READY:
		OSM_LOG_CONFIG(phba, KERN_ERR,"ACIT_EVENT_TARGET_READY:");

		ptarget_entry = &pacit_target_table->target_entry[info];
		ptarget = ptarget_entry->ptarget_key->osm_key;
		if (ptarget != NULL) {
			ptarget->state &= ~SE_TARGET_BUSY;
		}
		break;

	case ACIT_EVENT_TARGET_BUSY:
		OSM_LOG_CONFIG(phba, KERN_ERR,"ACIT_EVENT_TARGET_BUSY:");

		ptarget_entry = &pacit_target_table->target_entry[info];
		ptarget = ptarget_entry->ptarget_key->osm_key;
		if (ptarget != NULL) {
			ptarget->state |= SE_TARGET_BUSY;
		}
		break;

	case ACIT_EVENT_TARGET_REMOVE_PENDING:
		OSM_LOG_CONFIG(phba, KERN_ERR,"ACIT_EVENT_TARGET_REMOVE_PENDING:");
		/*
		 * Nothing to do for these events for now.
		 */
		break;

	case ACIT_EVENT_TARGET_MISSING: {
		se_device *pdevice;
		struct list_head        *lh, *lh_sf = NULL;
#if defined(__VMKERNEL_MODULE__)
		struct vmk_ScsiAdapter          *vmk_adapter = NULL;
		struct vmklnx_ScsiAdapter       *vmklnx_adapter = NULL;

		vmklnx_adapter = (struct vmklnx_ScsiAdapter *)phba->host->adapter;

		if(vmklnx_adapter != NULL){
			vmk_adapter = (struct vmk_ScsiAdapter *)vmklnx_adapter->vmkAdapter;
			if(vmk_adapter == NULL){
				OSM_LOG_CONFIG(phba, KERN_ERR,"vmk_adapter is NULL \n");
				break;
			}
		}else{
			OSM_LOG_CONFIG(phba, KERN_ERR,"vmklnx_adapter is NULL \n");
			break;
		}
#endif
		OSM_LOG_CONFIG(phba, KERN_ERR,"ACIT_EVENT_TARGET_MISSING:");

		ptarget_entry = &pacit_target_table->target_entry[info];

		ptarget = ptarget_entry->ptarget_key->osm_key;

		if (ptarget != NULL) {
			ptarget->state |= SE_TARGET_OFFLINE;
		}

		if (ddm) {
			break;
		}

		if (ptarget == NULL) {
			break;
		}

		list_for_each_safe(lh, lh_sf, &ptarget->lun_list) {

			pdevice = list_entry(lh, se_device, sibling_list);  

			if (pdevice == NULL) {
				continue;
			}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6))
			scsi_device_set_state(pdevice->pScsi_dev, SDEV_OFFLINE);
#else
			if (pdevice->pScsi_dev != NULL)
				pdevice->pScsi_dev->online = FALSE;
#endif
#if defined(__VMKERNEL_MODULE__)
			if(VMK_OK != vmk_ScsiNotifyPathStateChangeAsync(vmk_adapter,
						pdevice->scsi_channel,pdevice->scsi_id,pdevice->scsi_lun)){

				printk("vmk_ScsiNotifyPathStateChangeAsync Failed for C%d:T%d:L%d\n",
						pdevice->scsi_channel,pdevice->scsi_id,pdevice->scsi_lun);
			}
#endif

		}

		break;
	}

	case ACIT_EVENT_TARGET_RECONNECTED: {
		se_device *pdevice;
		struct list_head        *lh, *lh_sf = NULL;
#if defined(__VMKERNEL_MODULE__)
		struct vmk_ScsiAdapter          *vmk_adapter = NULL;
		struct vmklnx_ScsiAdapter       *vmklnx_adapter = NULL;

		vmklnx_adapter = (struct vmklnx_ScsiAdapter *)phba->host->adapter;

		if(vmklnx_adapter != NULL){
			vmk_adapter = (struct vmk_ScsiAdapter *)vmklnx_adapter->vmkAdapter;
			if(vmk_adapter == NULL){
				OSM_LOG_CONFIG(phba, KERN_ERR,"vmk_adapter is NULL \n");
				break;
			}
		}else{
			OSM_LOG_CONFIG(phba, KERN_ERR,"vmklnx_adapter is NULL \n");
			break;
		}
#endif

		OSM_LOG_CONFIG(phba, KERN_ERR,"ACIT_EVENT_TARGET_RECONNECTED:\n");

		ptarget_entry = &pacit_target_table->target_entry[info];
		ptarget = ptarget_entry->ptarget_key->osm_key;

		if (ptarget == NULL) {
			break;
		}

		list_for_each_safe(lh, lh_sf, &ptarget->lun_list) {
			pdevice = list_entry(lh, se_device, sibling_list);  

			if (pdevice == NULL) {
				continue;
			}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,6))
			/* 
			 * Need to make sure that it is Not in SDEV_RUNNING 
			 * state as that will be illegal
			 */
			if (pdevice->pScsi_dev->sdev_state != SDEV_RUNNING) 
				scsi_device_set_state(pdevice->pScsi_dev,
					SDEV_RUNNING);
#else
			if (pdevice->pScsi_dev != NULL)
				pdevice->pScsi_dev->online = TRUE;
#endif
#if defined(__VMKERNEL_MODULE__)
			if(VMK_OK != vmk_ScsiNotifyPathStateChangeAsync(vmk_adapter,
						pdevice->scsi_channel,pdevice->scsi_id,pdevice->scsi_lun)){

				printk("vmk_ScsiNotifyPathStateChangeAsync Failed for C%d:T%d:L%d\n",
						pdevice->scsi_channel,pdevice->scsi_id,pdevice->scsi_lun);
			}
#endif
		}

		ptarget->state &= ~SE_TARGET_OFFLINE;

		break;
	}

	case ACIT_EVENT_TARGET_RESCAN_REQUIRED: {
		OSM_LOG_CONFIG(phba, KERN_ERR,"ACIT_EVENT_TARGET_RESCAN_REQUIRED:\n");

		ptarget_entry = &pacit_target_table->target_entry[info];
		ptarget = ptarget_entry->ptarget_key->osm_key;
		job = kmalloc(sizeof(*job), GFP_ATOMIC);
		if (job) {
			job->fn = se_change_luns;
			job->data.tj_data.ptarget = ptarget;
			job->data.tj_data.command = event_code;
			se_schedule_job(phba, job);
			OSM_LOG_CONFIG(phba, KERN_ERR,"queued the job \n");
		}

		break;
	}

	case ACIT_EVENT_TARGET_REMOVE: {

		OSM_LOG_CONFIG(phba, KERN_ERR,"ACIT_EVENT_TARGET_REMOVE:\n");
		/*
		 * don't want to remove target in suspend mode
		 * since resume will need target to read/write
		 * and one IO error will fail hibernation
		 */
		if (phba->state & SE_STATE_SUSPEND)
			break;

		ptarget_entry = &pacit_target_table->target_entry[info];

		ptarget = ptarget_entry->ptarget_key->osm_key;

		if (ptarget == NULL) {
			break;
		}

		ptarget_entry->osm_tflags &= ~ACIT_OSM_TFLAG_IN_USE;
		ptarget_entry->ptarget_key->osm_key = NULL;
		ptarget->state |= SE_TARGET_OFFLINE;

		job = kmalloc(sizeof(*job), GFP_ATOMIC);
		if (job) {
			job->fn = se_delete_target_job;
			job->data.tj_data.ptarget = ptarget;
			job->data.tj_data.command = event_code;
			se_schedule_job(phba, job);
			OSM_LOG_CONFIG(phba, KERN_ERR,"queued the job \n");
		}

		break;
	}

	/* these events can be scheduled for later */ 
	case ACIT_EVENT_TARGET_ADD:
		OSM_LOG_CONFIG(phba, KERN_ERR,"ACIT_EVENT_TARGET_ADD:\n");

		if (phba->host != NULL)
			OSM_LOG_CONFIG(phba, KERN_ERR,"scsi%d: ", phba->host->host_no);
		
		job = kmalloc(sizeof(*job), GFP_ATOMIC);
		ptarget_entry = &pacit_target_table->target_entry[info];

		if (job) {
			job->fn = se_initialize_target;
			job->data.tj_data.ptarget_entry = ptarget_entry;
			job->data.tj_data.command = event_code;
			se_schedule_job(phba, job);
		}

		break;
	}
}



/*++
Function:
	se_process_deferred_jobs

Parameter:
	phba : pointer to se_hba structure

Return:
	None 

Context:
	Timer or initialization thread

Description:
	This routine will process the queue of deferred job.
	Except for the initialization it will be called by timer.

--*/
void se_process_deferred_jobs(se_hba *phba)
{
	struct se_timer_queue  *job = NULL;
	ACIT_U32_64              flags = 0;
redo:
	spin_lock_irqsave(&phba->job_queue_lock, flags);
	if (!list_empty(&phba->job_queue)) {
		job = list_entry(phba->job_queue.next, struct 
					se_timer_queue, job_queue); 
		list_del_init(phba->job_queue.next);
	} 
	spin_unlock_irqrestore(&phba->job_queue_lock, flags);
	if (job) {
		job->fn(phba, &job->data);
		kfree(job);
	} else {
		return;
	}
	job = NULL;
	goto redo;
}

int se_worker_thread(void *data)
{
	se_hba *phba = (se_hba *)data;
	int status;

	daemonize("be2iscsi_wt_%d", phba->hba_unique_id);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)) || defined(__VMKLNX__)
#else
	current->flags |= PF_NOFREEZE;        
#endif

	while (!phba->worker_thread.terminate)
	{
#if !defined(__VMKERNEL_MODULE__)
		status = down_interruptible(&phba->kthread_sema);
#else
		init_completion(&phba->worker_thread.work_queued);
		wait_for_completion(&phba->worker_thread.work_queued);
#endif
		se_process_deferred_jobs(phba);
	}

	complete(&phba->worker_thread.thread_completion);

	return 0;
}

/*++
Function:
	se_alloc_io_handles

Parameter:
	phba: [in]       pointer to allocated memory for HBA

Return:
	ACIT_STATUS_SUCCESS on success, ACIT_STATUS_ERROR on error

Context:
	Initialization

Description:
	This routine allocates all pci relelated resources.

--*/
int se_alloc_io_handles(se_hba *phba)
{
	int reg, range_index = 0;
	struct pci_dev *dev = phba->pcidev;
	// struct resource *pregion;

	for (reg = 0; reg < MAX_PCI_RESOURCES; reg++) {
		se_pci_resource *presource = &phba->resourcemap[range_index];
		unsigned long base, end, flags, length;

		base = pci_resource_start(dev,reg);
		end = pci_resource_end(dev,reg);
		flags = pci_resource_flags(dev,reg);
		length = pci_resource_len(dev,reg); 
		if (!end)
			continue;
		presource->pci_bar_index = reg;

	       if (!(flags & PCI_BASE_ADDRESS_SPACE_IO)) {
			if (pci_request_region(dev, reg, "be2iscsi") != 0) {
				OSM_LOG_INIT_UNLOAD(phba,KERN_INFO,"Mapping conflict\n");
				continue;
			}

			presource->mmap = 1;
			presource->address = (ACIT_U32_64)
				ioremap_nocache(base, length);
			presource->paddr = base;
		} else {
			presource->mmap = 0;

			if (pci_request_region(dev, reg, "be2iscsi") != 0) {
				OSM_LOG_INIT_UNLOAD(phba,KERN_INFO,"Mapping conflict\n");
				continue;
			}

			presource->address = base;
			presource->paddr = base;
		}
		
		   if (!presource->address) {
			return ACIT_STATUS_ERROR;
		}
		range_index++;
	}
	return ACIT_STATUS_SUCCESS;
}
/*++
Function:
	se_release_io_handles

Parameter:
	phba: [in]       pointer to allocated memory for HBA

Return:
	None

Context:
	Unload

Description:
	This routine frees all pci relelated resources.

--*/
void se_release_io_handles(se_hba *phba)
{
	int range_index = 0;

	struct pci_dev *dev = phba->pcidev;

		 se_pci_resource *presource = &phba->resourcemap[range_index];

	while (presource->address) {
		if (presource->mmap) {
			iounmap((void *)presource->address);
		}
		range_index++;
		presource = &phba->resourcemap[range_index];
	}
	pci_release_regions(dev);
}

/*++
Function:
	se_init_params

Parameter:
	None

Return:
	None

Context:
	Init

Description:
	This routine initializes the module parameters for the OSM.
	User set values can override these default settings.

--*/

void se_init_params(void)
{
	int i = 0;
	se_hba *phba = NULL;

	INIT_NUM_CONF_PARAM(i,ios_per_lun,  ios_per_lun);   i++;
	INIT_NUM_CONF_PARAM(i,maxiosize,    max_io_size);     i++;
	INIT_NUM_CONF_PARAM(i,tmfreset,     tmf_reset);      i++;

	INIT_NUM_CONF_PARAM(i,cxn_cmds,     cxn_cmds);      i++;
	INIT_NUM_CONF_PARAM(i,cxns,         cxns);          i++;
	INIT_NUM_CONF_PARAM(i,tgts,   	    tgts);    i++;
	INIT_NUM_CONF_PARAM(i,ios_per_ctrl, BE2ISCSI_MAX_Q_DEPTH - cxns);  i++;
	INIT_NUM_CONF_PARAM(i,logouts,      logouts);       i++;

	INIT_NUM_CONF_PARAM(i,tmfs,         tmfs);          i++;
	INIT_NUM_CONF_PARAM(i,nops,         nops);          i++;
	INIT_NUM_CONF_PARAM(i,ioctls,       ioctls);        i++;
	INIT_NUM_CONF_PARAM(i,eqs,          eqs);           i++;
	INIT_NUM_CONF_PARAM(i,cqs,          cqs);           i++;

	INIT_NUM_CONF_PARAM(i,defpdu_sz,    defpdu_sz);     i++;
	INIT_NUM_CONF_PARAM(i,sge,          sge);           i++;

	INIT_NUM_CONF_PARAM(i,l_to_t,       l_to_t);        i++;
	INIT_NUM_CONF_PARAM(i,mcs,          mcs);           i++;
	INIT_NUM_CONF_PARAM(i,ticks_per_sec,ticks_per_sec); i++;
	INIT_NUM_CONF_PARAM(i,eto,          eto);           i++;
	INIT_NUM_CONF_PARAM(i,ldto,         ldto);          i++;

	INIT_NUM_CONF_PARAM(i,dbg,          dbg);           i++;
	INIT_NUM_CONF_PARAM(i,cq_wm,        cq_wm);         i++;
	INIT_NUM_CONF_PARAM(i,eq_wm,        eq_wm);         i++;
	INIT_NUM_CONF_PARAM(i,eqt_wm,       eqt_wm);        i++;
	INIT_NUM_CONF_PARAM(i,cxn_tmfs,     cxn_tmfs);      i++;

	INIT_NUM_CONF_PARAM(i,inv_ticks,    inv_ticks);     i++;
	INIT_NUM_CONF_PARAM(i,tgt_def_tick, tgt_def_tick);  i++;
	INIT_NUM_CONF_PARAM(i,to_factor,    to_factor);     i++;
	INIT_NUM_CONF_PARAM(i,io_retries_b4_erl0,       io_retries_b4_erl0);        i++;
	INIT_NUM_CONF_PARAM(i,hc_freq,      hc_freq);       i++;

	INIT_NUM_CONF_PARAM(i,dic,          dic);           i++;
	INIT_NUM_CONF_PARAM(i,min_wm,       min_wm);        i++;
	INIT_NUM_CONF_PARAM(i,max_wm,       max_wm);        i++;
	INIT_NUM_CONF_PARAM(i,wm_incs,      wm_incs);       i++;
	INIT_NUM_CONF_PARAM(i,max_ir,       max_ir);        i++;

	INIT_NUM_CONF_PARAM(i,min_ir,       min_ir);        i++;
	INIT_NUM_CONF_PARAM(i,iot4_dic,     iot4_dic);      i++;
	INIT_NUM_CONF_PARAM(i,ior4_dic,     ior4_dic);      i++;
	INIT_NUM_CONF_PARAM(i,intr_res,     intr_res);      i++;
	INIT_NUM_CONF_PARAM(i,intr_steps,   intr_steps);    i++;

	INIT_NUM_CONF_PARAM(i,sync_cd_ioc,  sync_cd_ioc)   i++;
	INIT_NUM_CONF_PARAM(i,im_policy,  im_policy)   i++;

	//INIT_STR_CONF_PARAM(i, drv_name);        
	for(i = 0; i < PARAMETERS; i++) {
	    OSM_LOG_INIT_UNLOAD(phba,KERN_INFO,"%d) %s = %d %s\n",i+1, 
		    se_param[i].osm_config_param, se_param[i].val,
		    se_param[i].str_val);
	}
}


/*++
Function:
	se_isr

Parameter:
	irq:    [in]    irq number
	dev_id: [in]    context information
	regs:   [in]    pointer to structure containing CPU registers

Return:
	None

Context:
	Interrupt 

Description:
	This routine is the ISR of the driver. It calls acit_process_irq to
	handle the interrupts.

--*/
irqreturn_t
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18)
se_isr(
	ACIT_SEXACT32 irq, 
	void *dev_id, 
	struct pt_regs *regs)
#else
se_isr(
	ACIT_SEXACT32 irq, 
	void *dev_id)
#endif
{
	se_hba        *phba = NULL;
	phba = (se_hba *) dev_id;

	if (!(phba->state & SE_STATE_INT_ENABLED)) {

		OSM_LOG_CONFIG(phba, KERN_ERR,"Not BE interrupt\n");

		return IRQ_NONE;
	}

	if (acitapi_process_irq(phba, &phba->controller_key) == 
		ACIT_STATUS_SUCCESS) {

		return IRQ_HANDLED;
	}


	return IRQ_NONE;
}


/*++
Function:
	se_isr_msix

Parameter:
	irq:    [in]    irq number
	dev_id: [in]    context information
	regs:   [in]    pointer to structure containing CPU registers

Return:
	None

Context:
	Interrupt 

Description:
	This routine is the ISR of the driver. It calls acit_process_irq to
	handle the interrupts.

--*/
irqreturn_t
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18)
se_isr_msix(
	ACIT_SEXACT32 irq, 
	void *dev_id, 
	struct pt_regs *regs)
#else
se_isr_msix(
	ACIT_SEXACT32 irq, 
	void *dev_id)
#endif
{
	se_hba        *phba = NULL;

	phba = (se_hba *) dev_id;
	if (!(phba->state & SE_STATE_INT_ENABLED)) {

		OSM_LOG_CONFIG(phba, KERN_ERR,"Not BE interrupt\n");
		return IRQ_NONE;
	}
	if (acitapi_process_irq_msix(phba, &phba->controller_key) == ACIT_STATUS_SUCCESS) {
		return IRQ_HANDLED;
	}

	return IRQ_NONE;
}





ACIT_UEXACT32
se_init_tmf(  se_device     *pdevice,
		se_req        *preq,
		se_io         *pio,
		ACIT_UEXACT8    tmf_function,
		ACIT_UEXACT32   timeout)
{
	se_tmf          *ptmf = &preq->u.tmf;
	se_target       *ptarget = pdevice->pparent_target;
	se_hba          *phba = ptarget->pparent_hba;


	PACIT_TMF_REQUEST ptmf_req = NULL;
	PACIT_IO_REQUEST  pio_req = NULL;
	ACIT_UEXACT32 status;

	ptmf_req = &ptmf->tmf_request;
	ptmf_req->tmf_function = tmf_function;

	/* Parameters depending on the TMF function code */
	if (tmf_function == ACIT_TMF_ABORT_TASK)
	{
		pio_req = &pio->io_request;
		ptmf_req->pcontroller_key = pio_req->pcontroller_key;
		ptmf_req->ptarget_key = pio_req->ptarget_key;
		ptmf_req->plun_key = pio_req->plun_key;
		ptmf_req->pio_request = pio_req;

	}
	else
	{
		ptmf_req->pcontroller_key = &phba->controller_key;
		ptmf_req->ptarget_key = ptarget->posm_target_key;
		ptmf_req->plun_key = &pdevice->lun_key;
		ptmf_req->pio_request = NULL;

	}
	ptmf_req->ptmf_key = &ptmf->acit_tmf_key; 
	ptmf_req->ptmf_key->osm_key = preq; 
	ptmf_req->ptmf_key->acit_key = 0;
	ptmf_req->ppost_routine = NULL;
	ptmf_req->timeout = timeout;
	ptmf_req->lun.u.lun_complete = pdevice->lun.u.lun_complete;

	status = acitapi_init_tmf_ws(phba, 
			ptmf_req->pcontroller_key, ptmf_req, 
			ptmf_req->ptmf_key, &ptmf->tmf_ws);


	return status;
}


ACIT_SEXACT32
se_complete_tmf_request(se_hba *phba, 
			se_tmf *ptmf)
{
	complete(&ptmf->complete_tmf);
	return ACIT_STATUS_SUCCESS;
}

ACIT_SEXACT32 
se_abort(struct scsi_cmnd *pscsi_cmnd)
{
	se_req		*prequest = NULL;
	se_tmf		*ptmf = NULL;
	se_hba		*phba = NULL;
	se_device	*pdevice = NULL;
	se_io		*pio_request = NULL;

	PACIT_CONTROLLER_KEY pcontroller_key;
	ACIT_UEXACT32 status = FAILED;        

	struct Scsi_Host *phost;
	unsigned int channel;
	unsigned int target;
	unsigned int lun;


	OSM_LOG_EH(phba, KERN_INFO,"Abort Function Invoked\n");
	if (gcrashmode) 
		return SUCCESS;
	
	pio_request = (se_io *)pscsi_cmnd->host_scribble;
	if(pio_request	== NULL){
		OSM_LOG_EH(phba, KERN_ERR,"pio_request is NULL\n");
		return FAILED;
	}

	phost = pscsi_cmnd->device->host;
	channel = pscsi_cmnd->device->channel;
	target = pscsi_cmnd->device->id;
	lun = pscsi_cmnd->device->lun;

	phba = (se_hba *) pscsi_cmnd->device->host->hostdata[0];
	if(phba == NULL){
		OSM_LOG_EH(phba, KERN_ERR,"phba is NULL\n");
		return FAILED;
	}

	OSM_LOG_EH(phba, KERN_ERR,"scsi%d: OneConnect iSCSI Driver: B%02d:T%02d:L%02d: "
		"Abort scsi_cmnd %p\n",
		phba->host->host_no, channel, target, lun, pscsi_cmnd);

	if (phba->state & SE_STATE_SHUTDOWN)
		return SUCCESS;

	if (pscsi_cmnd->serial_number != pio_request->curr_ser_num) {
		if (pscsi_cmnd->serial_number == pio_request->old_ser_num) {
			printk(KERN_ERR "Bad  reuse in se_abort of se_io "
				"that belongs to another cmd !!!!!  \n");
		} else {
			printk(KERN_ERR "Mismatch in ser num "
				"pscsi_cmnd->serial_number "
				"= %lx pio_request->curr_ser_num=%lx "
				"pio_request->old_ser_num=%lx\n",
				pscsi_cmnd->serial_number,
				pio_request->curr_ser_num,
				pio_request->old_ser_num);
		}
		return FAILED;
	}

	pcontroller_key = pio_request->io_request.pcontroller_key;
	if(pcontroller_key == NULL){
		OSM_LOG_EH(phba, KERN_ERR,"pcontroller_key is NULL\n");
		return FAILED;
	}

	pdevice = (se_device *)pscsi_cmnd->device->hostdata;
	if (pdevice == NULL) {
		OSM_LOG_EH(phba, KERN_ERR,"Device is NOT CONNECTED!\n");
		goto abort_failed;
	}
	pdevice->pScsi_dev = pscsi_cmnd->device;

	/* Get a free io request from the queue */
	prequest = se_get_free_cmd(phba);
	if (!prequest) {
		OSM_LOG_EH(phba, KERN_ERR,"Did not get a free command\n");
		goto abort_failed;
	}
	ptmf = &prequest->u.tmf;
	ptmf->command = pscsi_cmnd;
	init_completion(&ptmf->complete_tmf);
	status = se_init_tmf( pdevice,
				prequest, 
				pio_request,
				ACIT_TMF_ABORT_TASK,
				100);
	if (status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_EH(phba, KERN_ERR,"Cannot initialize tmf request\n");
		se_return_cmd(phba, prequest);
		goto abort_failed;
	}

	status = acitapi_start_tmf(phba, pcontroller_key, &ptmf->tmf_request); 

	if (status == SHTDWN_EVENT) {
		OSM_LOG_EH(phba, KERN_ERR," acitapi_start_tmf failed due to "
			"shutdown\n");
		se_return_cmd(phba, prequest);
		goto abort_failed;
	}

#if !( LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)) && !defined(__VMKERNEL_MODULE__)
	spin_unlock_irq(pscsi_cmnd->device->host->host_lock);
#endif
	wait_for_completion(&ptmf->complete_tmf);
#if !( LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)) && !defined(__VMKERNEL_MODULE__)
	spin_lock_irq(pscsi_cmnd->device->host->host_lock);
#endif
	if (ptmf->tmf_request.tmf_status != ACIT_TMF_STATUS_SUCCESS) {
		OSM_LOG_EH(phba, KERN_ERR,"tmf request failed\n");
		goto abort_failed;
	}
	return SUCCESS;
abort_failed:
	OSM_LOG_EH(phba, KERN_ERR,"tmf failed\n");
	return FAILED;
}


/*++
Function:
	se_device_reset

Parameter:
	cmd:    [in]    pointer to scsi command that was stuck in the
			device queue.

Return:
	DRV_SUCCESS - equal to SUCCESS defined in scsi.h
	DRV_FAILED - equal to FAILED defined in scsi.h

Context:
	Error handler entry point 

Description:
	This routine is the error handler routine called by the SCSI Midlayer
	when it wants to reset the SCSI Device. The command that is sent as
	the argument is the command that is in the device queue of the SCSI
	Midlayer that did not complete (successfully or otherwise).

--*/

ACIT_SEXACT32 
se_device_reset(struct scsi_cmnd *pscsi_cmnd)
{
	se_req *prequest = NULL;
	se_tmf *ptmf = NULL;
	se_hba *phba = NULL;
	se_device     *pdevice = NULL; 
	se_io         *pio_request = NULL;
	PACIT_CONTROLLER_KEY pcontroller_key = NULL;
	ACIT_UEXACT32 status = FAILED;        

	struct Scsi_Host *phost;
	unsigned int channel;
	unsigned int target;
	unsigned int lun;


	if (gcrashmode) {
		return SUCCESS;
	}
	pio_request = (se_io *)pscsi_cmnd->host_scribble;
	if(pio_request	== NULL){
		OSM_LOG_EH(phba, KERN_ERR,"pio_request is NULL\n");
		return FAILED;
	}

	phost = pscsi_cmnd->device->host;
	channel = pscsi_cmnd->device->channel;
	target = pscsi_cmnd->device->id;
	lun = pscsi_cmnd->device->lun;

 
	phba = (se_hba *) phost->hostdata[0];
	if(phba	== NULL){
		OSM_LOG_EH(phba, KERN_ERR,"phba is NULL\n");
		return FAILED;
	}

	OSM_LOG_EH(phba, KERN_ERR,"scsi%d: OneConnect iSCSI Driver: B%02d:T%02d:L%02d: "
		"Device Reset\n",
		phba->host->host_no, channel, target, lun);



	pdevice = (se_device *)pscsi_cmnd->device->hostdata;
	if (pdevice == NULL) {
		OSM_LOG_EH(phba, KERN_ERR,"Device is NOT CONNECTED!\n");
		goto dev_rst_failed;
	}
	pdevice->pScsi_dev = pscsi_cmnd->device;

	pcontroller_key = &phba->controller_key;
	if(pcontroller_key == NULL){
		OSM_LOG_EH(phba, KERN_ERR,"pcontroller_key is NULL\n");
		return FAILED;
	}


	if (pdevice->state & SE_DEV_OFFLINE) {
		OSM_LOG_EH(phba, KERN_ERR,"Device is currently offline !\n");
		goto dev_rst_failed;

	}

	OSM_LOG_EH(phba, KERN_ERR,"Reset device for pdevice OS-CTL(%d,%d,%d)!\n", 
			pdevice->scsi_channel,
			pdevice->scsi_id,
			pdevice->scsi_lun);

	/* Get a free io request from the queue */
	prequest = se_get_free_cmd(phba);

	if (!prequest) {
		OSM_LOG_EH(phba, KERN_ERR,"Did not get a free command\n");
		goto dev_rst_failed;
	}
	ptmf = &prequest->u.tmf;
	ptmf->command = pscsi_cmnd;
	init_completion(&ptmf->complete_tmf);
	status = se_init_tmf( pdevice,
				prequest, 
				pio_request,
				ACIT_TMF_LUN_RESET,
				100);

	if (status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_EH(phba, KERN_ERR,"Cannot initialize tmf request\n");
		se_return_cmd(phba, prequest);
		goto dev_rst_failed;
	}

	status = acitapi_start_tmf(phba, pcontroller_key, &ptmf->tmf_request); 
	if (status == SHTDWN_EVENT) {
		OSM_LOG_EH(phba, KERN_ERR," acitapi_start_tmf failed due to shutdown\n");
		se_return_cmd(phba, prequest);
		goto dev_rst_failed;
	}

#if !( LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)) && !defined(__VMKERNEL_MODULE__)
	spin_unlock_irq(pscsi_cmnd->device->host->host_lock);
#endif
	wait_for_completion(&ptmf->complete_tmf);
#if !( LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)) && !defined(__VMKERNEL_MODULE__)
	spin_lock_irq(pscsi_cmnd->device->host->host_lock);
#endif
	if (ptmf->tmf_request.tmf_status != ACIT_TMF_STATUS_SUCCESS) {
		OSM_LOG_EH(phba, KERN_ERR,"tmf request failed\n");
		goto dev_rst_failed;
	}

	return DRV_SUCCESS;

dev_rst_failed:
	OSM_LOG_EH(phba, KERN_ERR,"tmf failed\n");
	return DRV_FAILED;

}

/*++
Function:
	se_bus_reset

Parameter:
	cmd:    [in]    pointer to scsi command that was stuck in the
			device queue.

Return:
	DRV_SUCCESS - equal to SUCCESS defined in scsi.h
	DRV_FAILED - equal to FAILED defined in scsi.h

Context:
	Error handler entry point 

Description:
	This routine is the error handler routine called by the SCSI Midlayer
	when it wants to reset the SCSI Bus. The command that is sent as
	the argument is the command that is in the device queue of the SCSI
	Midlayer that did not complete (successfully or otherwise).

--*/

ACIT_SEXACT32 
se_bus_reset(struct scsi_cmnd *pscsi_cmnd)
{
	se_req *prequest = NULL;
	se_tmf *ptmf = NULL;
	se_hba *phba = NULL;
	se_target *ptarget = NULL, *ptemp = NULL;
	se_device     *pdevice = NULL;
	ACIT_U32_64     flags = 0;
	se_io         *pio_request = NULL;
	struct list_head         *lh = NULL, *lh_sf = NULL;
	PACIT_CONTROLLER_KEY pcontroller_key = NULL;
	ACIT_UEXACT32 status = FAILED;        
	struct Scsi_Host *phost;
	unsigned int channel;
	unsigned int target;
	unsigned int lun;


	if ((gcrashmode) || (be2iscsi_dump_in_progress))
		return SUCCESS;

	pio_request = (se_io *)pscsi_cmnd->host_scribble;
	if(pio_request	== NULL){
		OSM_LOG_EH(phba, KERN_ERR,"pio_request is NULL\n");
		return FAILED;
	}

	phost = pscsi_cmnd->device->host;
	channel = pscsi_cmnd->device->channel;
	target = pscsi_cmnd->device->id;
	lun = pscsi_cmnd->device->lun;

	phba = (se_hba *) phost->hostdata[0];
	if(phba	== NULL){
		OSM_LOG_EH(phba, KERN_ERR,"phba is NULL\n");
		return FAILED;
	}

	OSM_LOG_EH(phba, KERN_ERR,"scsi%d: OneConnect iSCSI Driver: B%02d:T%02d:L%02d: "
		"Bus Reset\n",
		phba->host->host_no, channel, target, lun);


	pdevice = (se_device *)pscsi_cmnd->device->hostdata;
	if (pdevice == NULL) {
		OSM_LOG_EH(phba, KERN_ERR,"Device is NOT CONNECTED!\n");
		return DRV_FAILED;
	}
	pdevice->pScsi_dev = pscsi_cmnd->device;

	ptarget = pdevice->pparent_target;
	phba = ptarget->pparent_hba;
	pcontroller_key    = &phba->controller_key;
	if(pcontroller_key	== NULL){
		OSM_LOG_EH(phba, KERN_ERR,"pcontroller_key is NULL\n");
		return FAILED;
	}

	if (pdevice->state & SE_DEV_OFFLINE) {
		OSM_LOG_EH(phba, KERN_ERR,"Device is currently offline !\n");
		return DRV_FAILED;
	}

	/* Initiate a bus reset == Target Warm?Cold?SeriesOfLuns Reset */
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	phba->state |= SE_STATE_BUS_RESET;
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);



	list_for_each_entry(ptemp, &phba->target_list, sibling_list) { 
		if (ptemp->bus == channel) {
			ptarget = ptemp;
			OSM_LOG_EH(phba, KERN_INFO,"ptarget is on same BUS (%d) as the cmd's target\n",ptemp->bus);
		}
		else
		{
			/* CONTINUE with the LIST as the target is on some other BUS */
			ptarget = ptemp;
			OSM_LOG_EH(phba, KERN_INFO,"ptarget is on some other BUS (%d) as the cmd's target\n",ptemp->bus);
			goto next_target;
		}

		list_for_each_safe(lh, lh_sf, &ptarget->lun_list) {
			pdevice = list_entry(lh, se_device, sibling_list);  
			if (pdevice) {

				OSM_LOG_EH(phba, KERN_INFO,"Reset device for pdev = 0x%lx!\n", 
							(unsigned long)pdevice);

				/* Get a free io request from the queue */
				prequest = se_get_free_cmd(phba);

				if (!prequest) {
					OSM_LOG_EH(phba, KERN_INFO,"Did not get a free command\n");
					goto bus_rst_failed;
				}
				ptmf = &prequest->u.tmf;
				ptmf->command = pscsi_cmnd;
				init_completion(&ptmf->complete_tmf);

				pdevice->state |= SE_DEV_RESET;

				if (tmf_reset == 3)
				{
					OSM_LOG_EH(phba, KERN_INFO,"BUS RESET <==> TGT COLD RST\n"); 
					/* TARGET Cold RSTs for all tgts on BUS */
					status = se_init_tmf(pdevice,
							 prequest, 
							 pio_request,
						  	 ACIT_TMF_TARGET_COLD_RESET,
						  	 100);
				} else if (tmf_reset == 2)
				{
					OSM_LOG_EH(phba, KERN_INFO,"BUS RESET <==> TGT WARM RST\n"); 
					/* TARGET Warm RSTs for all tgts on BUS */
					status = se_init_tmf(pdevice,
							 prequest, 
							 pio_request,
						  	 ACIT_TMF_TARGET_WARM_RESET,
						  	 100);
				} else 
				{
					if (tmf_reset != 1)
					{
						OSM_LOG_EH(phba, KERN_INFO," TMFReset =%d, Allowed Range = 1 - 3 \n",tmf_reset);
						tmf_reset =1;
					}

					OSM_LOG_EH(phba, KERN_INFO,"BUS RESET <==> LUN RST\n"); 
					/* LUN Resets for all LUNs on the BUS */
					status = se_init_tmf(pdevice,
						prequest, 
						pio_request,
						ACIT_TMF_LUN_RESET,
						100);
				} 

				if (status != ACIT_STATUS_SUCCESS) {
					OSM_LOG_EH(phba, KERN_ERR,"Cannot initialize tmf request\n");
					se_return_cmd(phba, prequest);
					goto bus_rst_failed;
				}

				status = acitapi_start_tmf(phba, pcontroller_key, &ptmf->tmf_request); 
				if (status == SHTDWN_EVENT) {
					OSM_LOG_EH(phba, KERN_ERR," acitapi_start_tmf failed due to shutdown\n");
					se_return_cmd(phba, prequest);
					goto bus_rst_failed;
				}

#if !( LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)) && !defined(__VMKERNEL_MODULE__)
				spin_unlock_irq(pscsi_cmnd->device->host->host_lock);
#endif
				wait_for_completion(&ptmf->complete_tmf);


#if !( LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)) && !defined(__VMKERNEL_MODULE__)
				spin_lock_irq(pscsi_cmnd->device->host->host_lock);
#endif
				pdevice->state &= ~SE_DEV_RESET;

				if (ptmf->tmf_request.tmf_status != ACIT_TMF_STATUS_SUCCESS) {
					OSM_LOG_EH(phba, KERN_ERR,"tmf request failed\n");
					goto bus_rst_failed;
				}
			}

			if (tmf_reset > 1 )
			{
				/* Target RST is sent on the 1st device for the Tgt */
				goto next_target;
			}
		}
next_target:
	OSM_LOG_EH(phba, KERN_INFO,"Sending reset on the next target\n");
	}


	spin_lock_irqsave(&phba->acitapi_lock,flags);
	phba->state &= ~SE_STATE_BUS_RESET;
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);
	return DRV_SUCCESS;

bus_rst_failed:
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	phba->state &= ~SE_STATE_BUS_RESET;
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);
	OSM_LOG_EH(phba, KERN_ERR,"tmf failed\n");
	return DRV_FAILED;

}

void se_dma_unmap(se_hba *phba, se_io *posm_request, struct scsi_cmnd *pscsi_cmnd)
{
	ACIT_U32        dma_dir;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)
	dma_dir = scsi_to_pci_dma_dir(pscsi_cmnd->sc_data_direction);
#else  /* ChangeLog-2.6.6:2404:  - Stop using scsi_to_pci_dma_dir() */
	dma_dir = pscsi_cmnd->sc_data_direction;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	if (pscsi_cmnd->use_sg) {
			      
		struct scatterlist *sg;
		sg = (struct scatterlist *)pscsi_cmnd->request_buffer;
		pci_unmap_sg(phba->pcidev, sg, pscsi_cmnd->use_sg, dma_dir);

	} else if (pscsi_cmnd->request_bufflen != 0) {

	       dma_addr_t bus_addr = posm_request->sg_list.sg[0].address.u.a64.address;

		pci_unmap_single(phba->pcidev, 
			bus_addr, 
			pscsi_cmnd->request_bufflen, 
			dma_dir);
	}
#else
	scsi_dma_unmap(pscsi_cmnd); 
#endif

}

/*++
Function:
	se_complete_io_request

Parameter:
	phba:            [in]   pointer to hba
	ptarget:         [in]   pointer to target
	pdevice:         [in]   pointer to device
	posm_request:    [in]   pointer to se_io
	pacit_io_request:[in]   pointer to acit_io_request

Return:
	ACIT_STATUS_SUCCESS

Context:
	osm_complete_io

Description:
	This routine does the reverse translation from acit_io_request
	to struct scsi_cmnd. This also sets the struct scsi_cmnd result field from
	SCSI status returned by ACIT. This is also the place where you
	can snoop into the scsi data e.g. REPORT_LUN etc.

--*/

ACIT_STATUS
se_complete_io_request(
		se_hba                *phba,
		se_target             *ptarget,
		se_device             *pdevice,
		se_io                 *posm_request,
		PACIT_IO_REQUEST        pacit_io_request
		)
{
	struct scsi_cmnd       *pscsi_cmnd = posm_request->command;


	se_dma_unmap(phba, posm_request, pscsi_cmnd);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	pscsi_cmnd->resid = pacit_io_request->underrun_length;
#else
	scsi_set_resid(pscsi_cmnd,pacit_io_request->underrun_length); 
#endif
	/* SAM2 defines the following status codes */

	switch (pacit_io_request->io_status)
	{
	case ACIT_IO_STATUS_SUCCESS:
		switch (pacit_io_request->scsi_status) {

			case SCSI_STATUS_GOOD:
			case SCSI_STATUS_CONDITION_MET:
			case SCSI_STATUS_INTERMEDIATE:
			case SCSI_STATUS_INTERMEDIATE_CONDITION_MET:
			case SCSI_STATUS_RESERVATION_CONFLICT:

				pscsi_cmnd->result = (DID_OK << 16);
				break;

			case SCSI_STATUS_BUSY:
			case SCSI_STATUS_TASK_SET_FULL:
			case SCSI_STATUS_ACA_ACTIVE:
				/*
				 * If status is SCSI_STATUS_ACA_ACTIVE, then it
				 * means target port has a ACA condition 
				 * currently that caused it to return failure. 
				 * As per SAM2 specs, initiator port may 
				 * reissue the command after the ACA condition 
				 * has been cleared. So lets ask for a retry 
				 * to the SCSI Midlayer.
				 */
				pscsi_cmnd->result = (DID_BUS_BUSY << 16);
				break;

			case SCSI_STATUS_TASK_ABORTED:

				if (pdevice->state & SE_DEV_RESET) {
					pscsi_cmnd->result = (DID_RESET << 16);
				} else {
					pscsi_cmnd->result = (DID_ABORT << 16);
				}
				break;

			case SCSI_STATUS_CHECK_CONDITION:
#ifdef __VMKLNX__
				memset(pscsi_cmnd->sense_buffer, 0, sizeof(pscsi_cmnd->sense_buffer));
				memcpy(pscsi_cmnd->sense_buffer, &posm_request->sense_data,SCSI_SENSE_BUFFERSIZE);
#endif
				pscsi_cmnd->result =(DID_OK << 16) | (CHECK_CONDITION << 1);
				break;

			default:

				pscsi_cmnd->result = (DID_ERROR << 16);
				break;
		}
		break;

	case ACIT_IO_STATUS_IO_TIMEOUT:
		pscsi_cmnd->result = (DID_BUS_BUSY << 16);
		break;

	case ACIT_IO_STATUS_TARGET_TIMEOUT:
	case ACIT_IO_STATUS_LINK_DOWN_TIMEOUT:
		pscsi_cmnd->result = (DID_NO_CONNECT << 16);
		break;

	case ACIT_IO_STATUS_ABORTED:
		if (pdevice->state & SE_DEV_RESET) {
			pscsi_cmnd->result = (DID_RESET << 16);
		} else {
			pscsi_cmnd->result = (DID_ABORT << 16);
		}
		break;

	case ACIT_IO_STATUS_ERROR:
		pscsi_cmnd->result = (DID_ERROR << 16);
		break;

	case ACIT_IO_STATUS_CRC_ERROR:
	case ACIT_IO_STATUS_TRANSPORT_ERROR:
	default:
		pscsi_cmnd->result = (DID_ERROR << 16);
		break;
	}

	pscsi_cmnd->result |= pacit_io_request->scsi_status;

	if (pscsi_cmnd->scsi_done != NULL) {
			OSM_LOG_IO(phba, KERN_ERR,"calling done routine with result = %x\n",
					(unsigned int)pscsi_cmnd->result);

#ifdef BE2ISCSI_DEBUG
			if (pdevice != NULL) {
				pdevice->num_ios_outstanding--;
			}
#endif
			pscsi_cmnd->scsi_done(pscsi_cmnd);
	}                
#ifdef BE2ISCSI_DEBUG
	if ((pdevice != NULL) && 
		((host_byte(pscsi_cmnd->result) != DID_OK) ||
		 (status_byte(pscsi_cmnd->result) != 0))) {

		pdevice->num_errors++;
	}
#endif
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	se_init_freeio_list

Parameter:
	phba:       [in]    pointer to allocated memory for HBA
	ws_size:    [in]    max(size of acit io ws, size of acit tmf ws)

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_INSUFFICIENT_RESOURCE - no memory available

Context:
	Init

Description:
	This routine initializes the freeio list maintained by the OSM.
	Each list entry holds memory for an IO at runtime, like 
	acit_io_request, SGs, acit io workspace, etc.
--*/

ACIT_STATUS
se_init_free_req_list(
		se_hba        *phba, 
		ACIT_UEXACT32   ws_size)
{
	se_req *tmp_req;
	int index;
	dma_addr_t pa;

	INIT_LIST_HEAD(&phba->free_req);
	spin_lock_init(&phba->free_list_lock);

	/* we allocate 1024 number of ios_per_controller. This number
	 * includes 128 tmf requests.
	 */
	for(index = 0; index < phba->max_ios_per_controller + 128; index++) {
		tmp_req = (se_req *) se_pci_alloc_consistent(phba->pcidev, 
				sizeof(se_req) + ws_size, &pa);
		if (!tmp_req) {
			se_delete_free_req_list(phba);
			return ACIT_STATUS_INSUFFICIENT_RESOURCE;
		}

		tmp_req->pa = pa;
		list_add(&tmp_req->queue_link, &phba->free_req);
	}
	
	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	se_delete_free_req_list

Parameter:
	phba:       [in]    pointer to allocated memory for HBA

Return:
	None

Context:
	Init/Unload

Description:
	This routine deletes the freeio list. It unlinks each element from
	the list and frees memory allocated for the request.

--*/

ACIT_VOID
se_delete_free_req_list(se_hba *phba)
{
	se_req             *posm_request;
	struct list_head     *lh, *lh_sf = NULL;
	unsigned int	req_ws_size;

	req_ws_size = phba->acit_io_ws_size < phba->acit_tmf_ws_size ? 
			phba->acit_tmf_ws_size : phba->acit_io_ws_size;   

	list_for_each_safe(lh, lh_sf, &phba->free_req) {
		posm_request = list_entry(lh, se_req, queue_link);
		list_del_init(lh);
		se_pci_free_consistent(phba->pcidev, sizeof(se_req) + req_ws_size,
				(void *) posm_request, posm_request->pa);
	}
}


/*++
Function:
	se_return_cmd

Parameter:
	phba:      [in] pointer to se_hba structure
	prequest   [in] pointer to command to be returned

Return:
	void

Context:
	IO, ISR

--*/
void
se_return_cmd(
		se_hba        *phba,
		se_req        *prequest)
{
	ACIT_U32_64 flags = 0;
	spin_lock_irqsave(&phba->free_list_lock, flags);
	list_add_tail(&prequest->queue_link, &phba->free_req);
	spin_unlock_irqrestore(&phba->free_list_lock, flags);

}

/*++
Function:
	se_get_free_cmd

Parameter:
	phba:      [in] pointer to se_hba structure
	prequest   [in] pointer to pointer to free command

Return:
	ACIT_STATUS_SUCCESS - no errors
	ACIT_STATUS_INSUFFICIENT_RESOURCE - no more free command

Context:
	IO

Description:
	This routine gets a free command from freeio_list

--*/

se_req *
se_get_free_cmd(
		se_hba        *phba)
{
	ACIT_U32_64     flags = 0;
	se_req        *prequest;

	prequest = NULL;

	spin_lock_irqsave(&phba->free_list_lock, flags);

	if (!list_empty(&phba->free_req)) {
		prequest = list_entry(phba->free_req.next, se_req, queue_link); 
		list_del_init(&prequest->queue_link);
	} else {
		OSM_ASSERT(0);
	}
	spin_unlock_irqrestore(&phba->free_list_lock, flags);
	
	return prequest;
}



/*++
Function:
	se_build_io_request

Parameter:
	phba:         [in] pointer to se_hba structure
	ptarget:      [in] pointer to target
	pdevice:      [in] pointer to lun
	pio_request:  [in] pointer to se_io
	pscsi_cmnd:   [in] pointer to scsi_cmnd
	scsi_done:    [in] completion routine
	timeout:      [in] timeout for the command

Return:
	ACIT_STATUS_SUCCESS - no errors
	errors from acit_init_io_ws if any

Context:
	IO

Description:
	This routine builds a acit_io_request for the SCSI command sent
	by the OS.

--*/

ACIT_UEXACT32
se_build_io_request(
		se_hba        *phba,
		se_target     *ptarget,
		se_device     *pdevice,
		se_req        *prequest,
		struct scsi_cmnd       *pscsi_cmnd,
		ACIT_UEXACT32   timeout)
{
	ACIT_STATUS             acit_status;
	PACIT_IO_REQUEST        pacit_io_request;
	ACIT_SG_ELEMENT         *psg;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	ACIT_S32                dma_dir;
#endif
	se_io                 *pio_request = &prequest->u.io;

	pio_request->command = pscsi_cmnd;
	pacit_io_request = &pio_request->io_request;

	memset(pacit_io_request, 0, sizeof(ACIT_IO_REQUEST));
	pacit_io_request->pio_key = &pio_request->acit_io_key;

	/* Command - CDB */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
	memcpy(&pacit_io_request->cdb, pscsi_cmnd->cmnd, 
		ACIT_SCSI_MAX_CDB_LENGTH);
#else
	memcpy(&pacit_io_request->cdb, &pscsi_cmnd->cmnd, 
		ACIT_SCSI_MAX_CDB_LENGTH);
#endif
	pacit_io_request->cdb_length = pscsi_cmnd->cmd_len;

	/* Data and SG */
	/*TODO: There is buffer pointer in command, is this what ACIT
	 * is looking for?
	 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	pacit_io_request->pdata_buffer = pscsi_cmnd->request_buffer;
#endif    

	/* construct SG list */
	pacit_io_request->psg_list = (PACIT_SG_LIST)&pio_request->sg_list;
	psg = &pio_request->sg_list.sg[0];

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)

	if (pscsi_cmnd->use_sg) {
		struct scatterlist *sg;
		ACIT_UEXACT32   num_sg;
		ACIT_S32        index;

		pacit_io_request->data_transfer_length = 0;

		sg = (struct scatterlist *) pscsi_cmnd->request_buffer;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)
		dma_dir = scsi_to_pci_dma_dir(pscsi_cmnd->sc_data_direction);
#else   /* ChangeLog-2.6.6:2404:  - Stop using scsi_to_pci_dma_dir() */
		dma_dir = pscsi_cmnd->sc_data_direction;
#endif
		num_sg = pci_map_sg(phba->pcidev, sg, 
					pscsi_cmnd->use_sg, dma_dir);
	      
		OSM_ASSERT(!(num_sg > SG_ALL));
#ifndef __VMKLNX__
		if (pscsi_cmnd->cmnd[0] == 0x12) {
			pacit_io_request->pdata_buffer = 
				page_address(sg->page) + sg->offset;
		}
#endif

#if defined ESX_50
		for (index = 0 ; index < num_sg; index++, sg_next(sg)) {
#else
		for (index = 0 ; index < num_sg; index++, sg++) {
#endif
			psg[index].address.u.a64.address = sg_dma_address(sg);
			psg[index].length = sg_dma_len(sg);

			pacit_io_request->data_transfer_length += 
							psg[index].length;
		}

		/* update total number of scatter gather elements */
		pio_request->sg_list.number_elements = index;
	}
	else
	{
		dma_addr_t        bus_addr;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)
		dma_dir = scsi_to_pci_dma_dir(pscsi_cmnd->sc_data_direction);
#else   /* ChangeLog-2.6.6:2404:  - Stop using scsi_to_pci_dma_dir() */
		dma_dir = pscsi_cmnd->sc_data_direction;
#endif

		pacit_io_request->data_transfer_length = pscsi_cmnd->request_bufflen;

		if (pscsi_cmnd->request_bufflen == 0) {
			pio_request->sg_list.number_elements = 0;
		} else {
			bus_addr = pci_map_single(phba->pcidev,
						pscsi_cmnd->request_buffer,
						pscsi_cmnd->request_bufflen,
						dma_dir);
			psg[0].address.u.a64.address = bus_addr;
			psg[0].length = pscsi_cmnd->request_bufflen;
			pio_request->sg_list.number_elements = 1;
		}
	}
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
	pacit_io_request->pdata_buffer = NULL;
#endif
	{
		struct scatterlist *sg;
		ACIT_UEXACT32   num_sg;
		ACIT_S32        index;

		pacit_io_request->data_transfer_length = 0;

		num_sg = scsi_dma_map(pscsi_cmnd);
	      
		OSM_ASSERT(!(num_sg > SG_ALL));

		scsi_for_each_sg(pscsi_cmnd, sg, num_sg, index) {        
			psg[index].address.u.a64.address = sg_dma_address(sg);
			psg[index].length = sg_dma_len(sg);

			pacit_io_request->data_transfer_length += 
				       	psg[index].length;
		}

		/* update total number of scatter gather elements */
		pio_request->sg_list.number_elements = num_sg;
	}
#endif //<2.6.22
	/* Sense */
	/*In ESX, the cache line size is 64 bytes. Because of this limit, the SCSI sense buffer (which is 96 bytes in size)
	  cannot be guaranteed to be physically contiguous when allocated*/
#ifdef __VMKLNX__

	pacit_io_request->psense_buffer_physical.u.a64.address = virt_to_phys(&pio_request->sense_data);

	pacit_io_request->psense_buffer_virtual = &pio_request->sense_data; 
#else


	pacit_io_request->psense_buffer_physical.u.a64.address = virt_to_phys(
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27)
			(void *)&pscsi_cmnd->sense_buffer);
#else
			(void *)pscsi_cmnd->sense_buffer);
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27)
	pacit_io_request->psense_buffer_virtual = pscsi_cmnd->sense_buffer;
#else
	pacit_io_request->psense_buffer_virtual = &pscsi_cmnd->sense_buffer;
#endif

#endif //__VMKLNX__
	pacit_io_request->sense_buffer_length = SCSI_SENSE_BUFFERSIZE;

	/* 
	 * Since this is a OS request, we dont set post routine. ACIT
	 * behaviour is that if the post routine is NULL, it calls
	 * osm_complete_io that we provided in the API table
	 */
	pacit_io_request->ppost_routine = NULL;
	if (pscsi_cmnd->cmnd[0] == REPORT_LUNS)
		memset(&pacit_io_request->lun.u.lun_complete,0x0,sizeof(pdevice->lun.u.lun_complete)); 
	else
		pacit_io_request->lun.u.lun_complete = pdevice->lun.u.lun_complete;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,5)
	if (pscsi_cmnd->sc_data_direction == SCSI_DATA_WRITE) 
	       pacit_io_request->flags.outbound_data = 1;
	else if (pscsi_cmnd->sc_data_direction == SCSI_DATA_READ)
	       pacit_io_request->flags.inbound_data = 1;
#else   /* ChangeLog-2.6.6:2404:  - Stop using scsi_to_pci_dma_dir() */
	if (pscsi_cmnd->sc_data_direction == DMA_TO_DEVICE) 
		pacit_io_request->flags.outbound_data = 1;
	else if (pscsi_cmnd->sc_data_direction == DMA_FROM_DEVICE)
	       pacit_io_request->flags.inbound_data = 1;
#endif


	pacit_io_request->timeout = timeout;

	/* TODO - set queued option */
	/* For disk enable tagged queueing and for tapes disable it?? */
	if (pscsi_cmnd->device->tagged_supported) {
		pacit_io_request->task_attribute = ACIT_TASK_ATTRIBUTE_SIMPLE;
	} else {
		pacit_io_request->task_attribute = ACIT_TASK_ATTRIBUTE_UNTAGGED;
	}

	/* Workspace settings */
	pacit_io_request->pio_key->acit_key = 0;
	pacit_io_request->pio_key->osm_key = prequest;
	pacit_io_request->plun_key = &pdevice->lun_key;
	pacit_io_request->ptarget_key = ptarget->posm_target_key;
	pacit_io_request->pcontroller_key = &phba->controller_key;
#ifdef VIRT_TO_BUS_NOTSUPPORTED
	pacit_io_request->posm_reference = se_get_dma_mem_profile( phba, (unsigned long)prequest);
#endif

	if (phba->acit_io_ws_size) {
		acit_status = acitapi_init_io_ws(phba, &phba->controller_key,
				pacit_io_request,
				pacit_io_request->pio_key,
				ptarget,
				&pio_request->io_ws);

		if (acit_status != ACIT_STATUS_SUCCESS) {
			OSM_LOG_IO(phba, KERN_ERR,"could not initialize io ws %d\n",
					acit_status);

			return acit_status;
		}
	}

	/* save pointer to ACIT_IO_REQUEST in struct scsi_cmnd. We will 
	   need this if os calls us with this scsi_cmnd for abort */
	pscsi_cmnd->host_scribble = (unsigned char *)pio_request; 

	return ACIT_STATUS_SUCCESS;
}

/*++
Function:
	se_queue

Parameter:
	kernels before 2.6.36 entering se_queue with host_lock
	pscsi_cmnd:   [in] pointer to scsi_cmnd
	scsi_done:    [in] completion routine

	kernels 2.6.36+ entering se_queue no host_lock
	Scsi_Host ptr
	scsi_cmnd ptr

Description:
	This is the queue_command entry point for the driver. It does some
	checking if the command can be sent down to the specified device, and
	if so will submit the request to ACIT.

--*/

ACIT_SEXACT32 
se_queue(
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)) || defined(NEW_QUEUECMD)
	struct Scsi_Host *phost,
	struct scsi_cmnd *pscsi_cmnd)
#else
	struct scsi_cmnd *pscsi_cmnd, 
	void (*done_fn)(struct scsi_cmnd *))
#endif
{
	se_hba        *phba = NULL;
	se_target     *ptarget = NULL;
	se_device     *pdevice = NULL;        
	se_io         *pio_request = NULL;
	se_req        *prequest = NULL;
	ACIT_UEXACT32   acit_status;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)) && !defined(NEW_QUEUECMD)
	struct Scsi_Host *phost;
#endif
	unsigned int channel;
	unsigned int target;
	unsigned int lun;
#if defined(__VMKERNEL_MODULE__) 
	unsigned	timeout;
#endif

	BEGIN;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)) && !defined(NEW_QUEUECMD)
	phost = pscsi_cmnd->device->host;
#endif
	channel = pscsi_cmnd->device->channel;
	target = pscsi_cmnd->device->id;
	lun = pscsi_cmnd->device->lun;

#if defined(__VMKERNEL_MODULE__)
	timeout = pscsi_cmnd->timeout_per_command;

	if (timeout < 10 * HZ) {
	       timeout = 30 * HZ;
	}
#endif


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)) && !defined(NEW_QUEUECMD)
	pscsi_cmnd->scsi_done = done_fn;
#endif
	pscsi_cmnd->result = (DID_ERROR << 16);

#if 0
#if !defined(__VMKERNEL_MODULE__)
	/*
	 * ESX 3.5 does a full bus scan which is way too much debug output
	 */
	PDEBUG(DBG_SCSI, "BTL = %d %d %d CMd = %x\n",channel, target,
		lun, pscsi_cmnd->cmnd[0]); 
#endif
#endif

	/*
	 * SCSI REQUEST_SENSE commands will be executed automatically by the 
	 * Host Adapter for any errors, so they should not be executed 
	 * explicitly unless the Sense Data is zero indicating that no error 
	 * occurred.
	 */
	if ((pscsi_cmnd->cmnd[0] == REQUEST_SENSE) && 
		(pscsi_cmnd->sense_buffer[0] != 0)) {

		OSM_LOG_IO(phba, KERN_ERR,"Request Sense\n");
		pscsi_cmnd->result = (DID_OK << 16);
		goto done;
	}

	phba = (se_hba *) phost->hostdata[0];

	/*
	 * We are still on the Bus Reset Stage.
	 * Return the IO to be retried later on.
	 */   
	if (phba->state & SE_STATE_BUS_RESET) {
		pscsi_cmnd->result = (DID_ERROR << 16);
		OSM_LOG_IO(phba, KERN_ERR,"BUS_RESET\n");
		goto done;
	}

	pdevice = pscsi_cmnd->device->hostdata;

	if (pdevice == NULL) {
		pscsi_cmnd->result = (DID_NO_CONNECT << 16);
		OSM_LOG_IO(phba, KERN_ERR,"DID_NO_CONNECT, pdevice is NULL\n");
		goto done;
	}

#if defined(__VMKERNEL_MODULE__)
	if (pdevice->state & SE_DEV_DESTROY) {  
		/* 
		 * Failing I/O IF Device is marked for deletion.
		 */
		pscsi_cmnd->result = (DID_NO_CONNECT << 16); 
		OSM_LOG_IO(phba, KERN_ERR,"bad target - BTL = %d %d %d CMd = %x\n",
				channel, target, lun, pscsi_cmnd->cmnd[0]); 
		goto done;
	}
#endif

	pdevice->pScsi_dev = pscsi_cmnd->device;


	ptarget = pdevice->pparent_target;

	/*workaround to update timeout for kernels < (2,6,28)*/
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,27)  
	if(ptarget != NULL){
		if(((ptarget->extended_timeout + 30) * HZ)!= pdevice->pScsi_dev->timeout){
			pdevice->pScsi_dev->timeout = (ptarget->extended_timeout + 30) * HZ;
		}
	}
#endif
	/*
	 * Returned the IO back to the OS for that device that is gone for 
	 * good.  
	 */
	if (ptarget->state & SE_TARGET_OFFLINE) {
		if (pscsi_cmnd->device != NULL)  
		{
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,6)
			if (0 == ddm)
				pscsi_cmnd->device->online = FALSE;
#endif
		}

		pscsi_cmnd->result = (DID_NO_CONNECT << 16);
		OSM_LOG_IO(phba, KERN_ERR,"offline or session recovery\n");
		goto done;
	}

	/*
	 * Check if we can send more IOs down to the  Adapter.
	 */ 
	if ((phba->aborted_cmd_count > BE2ISCSI_MAX_Q_DEPTH)
	     || (phba->active_cmds > BE2ISCSI_MAX_Q_DEPTH)) {
		OSM_LOG_IO(phba, KERN_ERR,"Reached MAX IOs limit\n");

		pscsi_cmnd->result = (DID_BUS_BUSY << 16);
		goto done;
	}

	 /* Get a free io request from the queue */

	prequest = se_get_free_cmd(phba);

	if (!prequest) {
		OSM_LOG_IO(phba, KERN_ERR,"Did not get a free command\n");
		pscsi_cmnd->result = (DID_BUS_BUSY << 16);
		goto done;
	}
	prequest->u.io.curr_ser_num = pscsi_cmnd->serial_number;
	pio_request = &prequest->u.io;


	/* Formulate a acit layer request from osm request */
	acit_status = se_build_io_request(phba, 
				    ptarget,
				    pdevice,
				    prequest, 
				    pscsi_cmnd,
#if defined(__VMKERNEL_MODULE__)
				    timeout / HZ);
#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
				    ((pscsi_cmnd->timeout_per_command/HZ) -2));
#else
				    ((pscsi_cmnd->request->timeout/HZ)-2)); 
#endif
#endif

	if (acit_status != ACIT_STATUS_SUCCESS) {
		/* free the request back to the free_io queue */
		prequest->u.io.old_ser_num = pscsi_cmnd->serial_number;
		se_dma_unmap(phba, pio_request, pscsi_cmnd);
		se_return_cmd(phba, prequest);
		if (acit_status == SHTDWN_EVENT) {
			if (SYNCHRONIZE_CACHE == pscsi_cmnd->cmnd[0]) {
				pscsi_cmnd->result = (DID_OK << 16);
			} else {
				pscsi_cmnd->result = (DID_NO_CONNECT << 16);
			}
		}
		else {
			if ((0 == eto) && (0 == ldto)) {
				pscsi_cmnd->result = (DID_NO_CONNECT << 16);
			} else {
				pscsi_cmnd->result = (DID_SOFT_ERROR << 16);
			}
		}
		goto done;
	}

#ifdef BE2ISCSI_DEBUG
	if (pdevice != NULL) {
		pdevice->num_ios_outstanding++;
	}
#endif

	acit_status = acitapi_start_io(phba, &phba->controller_key,
		ptarget,&pio_request->io_request);

	if (acit_status == SHTDWN_EVENT ) {
		OSM_LOG_IO(phba, KERN_ERR,"se_queue failed due to shutdown or device "
				"being removed\n");
		se_dma_unmap(phba, pio_request, pscsi_cmnd);
		se_return_cmd(phba, prequest);
		pscsi_cmnd->result = (DID_NO_CONNECT << 16);
		goto done;
	}


	return 0;
done: 
	if (pscsi_cmnd->scsi_done != NULL) {
		pscsi_cmnd->scsi_done(pscsi_cmnd);
	}

	return 0;
}

/* ---------------------------------------------------------------- */
#ifdef REGISTER_ISCSI
#ifdef __VMKLNX__
struct scsi_transport_template *be2iscsi_scsi_transport;

static int
be2iscsi_host_get_param(
struct Scsi_Host		*shost,
enum iscsi_host_param		param,
char				*buf
)
{
	int len = 0;
	int status = 0;
 	se_hba	*phba = NULL;	
	
	BEGIN;
	
	phba = (se_hba *)shost->hostdata[0];

	if (phba == NULL) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"phba NULL\n");
		END;
		return -1;
	}

	switch (param) {
	case ISCSI_HOST_PARAM_IPADDRESS:
		len = sprintf(buf, "%s", phba->buf_ip);

		OSM_LOG_CONFIG(phba, KERN_ERR,"Initiator IP Address : %s\n", buf);

		break;

	case ISCSI_HOST_PARAM_INITIATOR_NAME:

		len = sprintf(buf,"%s", phba->buf_name);

		OSM_LOG_CONFIG(phba, KERN_ERR,"Initiator Name:%s\n", buf);

		break;

	default:
		END;
		return -ENOSYS;
	}

	END;
	return len;
}

static int
be2iscsi_conn_get_param(
struct iscsi_cls_conn		*conn,
enum iscsi_param		param,
char				*buf
)
{
	int len = 0;
	int status = 0;
	struct iscsi_cls_session *sess;
	se_hba *phba = NULL;
	se_target *ptarget =NULL;
	
	BEGIN;

	sess = conn->session;

	if (sess == NULL) {
		END;
		return -1;
	}

	phba = (se_hba *)sess->host->hostdata[0];

	if (phba == NULL) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"phba NULL\n");
		END;
		return -1;
	}
	
	ptarget = (se_target *)sess->dd_data;

	if (ptarget == NULL) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"ptarget NULL\n");
		END;
		return -1;
	}

	switch (param) {
	case ISCSI_PARAM_CONN_PORT:
		len = sprintf(buf, "%s", ptarget->buf_port);
		OSM_LOG_CONFIG(phba, KERN_ERR,"Destination Port: %s\n", buf);
		break;

	case ISCSI_PARAM_CONN_ADDRESS:
		len = sprintf(buf, "%s", ptarget->buf_ip);
		OSM_LOG_CONFIG(phba, KERN_ERR,"Destination IP: %s\n", buf);
		break;

	default:
		END;
		return -ENOSYS;
	}

	END;
	return len;
}

static int
be2iscsi_sess_get_param(
struct iscsi_cls_session	*sess,
enum iscsi_param		param,
char				*buf
)
{
	int len = 0, status = 0;
	se_hba *phba = NULL;
	se_target *ptarget =NULL;

	BEGIN;

	phba = (se_hba *)sess->host->hostdata[0];

	if (phba == NULL) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"phba NULL\n");
		END;
		return -1;
	}

	ptarget = (se_target *)sess->dd_data;

	if (ptarget == NULL) {
		OSM_LOG_CONFIG(phba, KERN_ERR,"ptarget NULL\n");
		END;
		return -1;
	}

	switch (param) {
		case ISCSI_PARAM_TARGET_NAME:
		len = snprintf(buf, PAGE_SIZE_4K - 1, "%s", ptarget->buf_name);
		OSM_LOG_CONFIG(phba, KERN_ERR,"iscsi Target Name = %s\n", buf);
		break;

	case ISCSI_PARAM_TPGT:
		len = sprintf(buf, "%u", ptarget->portal_group_tag);
		OSM_LOG_CONFIG(phba, KERN_ERR,"portal group tag = %s\n", buf);
		break;

	case ISCSI_PARAM_ISID:
		len = sprintf(buf, "%s", ptarget->buf_isid);
		OSM_LOG_CONFIG(phba, KERN_ERR,"iscsi Target ISID = %s\n", buf);
		break;

	default:
		END;
		return -ENOSYS;
	}

	END;
	return len;
}

static void
be2iscsi_recovery_timedout(
struct iscsi_cls_session	*session
)
{
	printk("%s:%d\n", __FUNCTION__, __LINE__);
}
 
struct iscsi_transport be2iscsi_iscsi_transport = {
	.owner			= THIS_MODULE,
	.name			= "be2iscsi",
	.description		= "Emulex iSCSI Adapter",
	.param_mask		= ISCSI_CONN_PORT |
				  ISCSI_CONN_ADDRESS |
				  ISCSI_TARGET_NAME |
				  ISCSI_TPGT,
	.sessiondata_size	= sizeof(unsigned),
	.host_template		= &driver_template,

	.get_host_param		= be2iscsi_host_get_param,
	.get_conn_param		= be2iscsi_conn_get_param,
	.get_session_param	= be2iscsi_sess_get_param,
	.session_recovery_timedout
				= be2iscsi_recovery_timedout,

	.caps = (CAP_DATA_PATH_OFFLOAD | CAP_FW_DB | CAP_RECOVERY_L0 |
		CAP_HDRDGST | CAP_DATADGST | CAP_SENDTARGETS_OFFLOAD |
		CAP_CONN_CLEANUP),
};
#endif
#endif

/* ---------------------------------------------------------------- */

/*++
Function:
	se_scsi_register

Parameter:
	phba:  [in] pointer to hba structure
	sht:   [in] pointer to struct scsi_host_template

Return:
	0 on success
	non-zero on error

Context:
	Init

Description:
	This routine is called from detect entry point to register with
	SCSI Midlayer.

--*/

ACIT_SEXACT32 
se_scsi_register(
		se_hba *phba, 
		struct scsi_host_template *sht)
{
	struct Scsi_Host         *host = NULL;
	ACIT_SEXACT32		ret;
#ifdef __VMKLNX__
	struct be_msix_entry  	*be_entry;
	uint16_t  		irq_save;
#endif

	BEGIN;

	host = scsi_host_alloc(sht, sizeof(se_hba *));
	if (host == NULL) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"%s:  scsi_register returned NULL\n", 
				phba->name);
		return -1;
	}

	host->hostdata[0] = (unsigned long)phba;
	phba->host = host;
	host->irq = phba->pcidev->irq;
	/* 
	 * no IO ports, so don't have to set host->io_port and 
	 * host->n_io_port
	 */
	host->io_port = 0;
	host->n_io_port = 0;

	/* 
	 * se_scsi_register is the first call,
	 * So there will be no targets present at this point
	 * Hence initializing these values to zeroes instead of
	 * host->max_id = phba->top_scsi_id+1;
	 * host->max_lun = phba->top_scsi_lun + 1;
	 * host->max_channel = phba->top_scsi_channel;
	 * TODO: SCSI 2 Targets does not support Report LUNS,
	 * this will not increment max_id/max_lun in scan for luns
	 * FIXME: Luns cannot be found in that case.
	 */
	host->max_id = 256;
	host->max_lun = 256;
	host->max_cmd_len = 16;
	host->max_channel = 0;

	/* see comments in hosts.h */
	host->cmd_per_lun = BE2ISCSI_CMD_PER_LUN;

	host->unique_id = phba->hba_unique_id;

	host->can_queue = BE2ISCSI_MAX_Q_DEPTH;
	host->sg_tablesize = sge -2;   /*1 SGE is taken up for iSCSI PDU header and one more is not used for legacy reasons.*/

	host->max_sectors = max_io_size * 2;
#ifdef REGISTER_ISCSI
#ifdef __VMKLNX__
	host->transportt = be2iscsi_scsi_transport;
#endif
#endif



#ifdef __VMKLNX__
	/*
	 * Hack:  When our driver calls pci_enable_msix(), the pcidev->irq
	 * value is set to -1 by LinuxPCIIntrVectorFree().  When the driver
	 * then calls:
	 *    request_irq() -> vmklnx_request_irq() -> AddIRQInfo()
	 * the MSI-X interrupt vector is added to the IRQList.
	 * The call to scsi_add_host() below will end up calling 
	 * vmk_SCSIRegisterIRQ() with pcidev->irq instead of one of the vectors
	 * that was added to the IRQList by request_irq().  This causes the
	 * dump path:
	 *
	 *    vmk_SCSIIssueSyncDumpCommand() ->
	 *      adapter->intrHandler() ... Linux_IRQHandler()
	 *
	 * To not find a match for our adapter, because it tries to find
	 * the value that was in pcidev->irq (which was -1) in the IRQList 
	 * for the module ... and fails.
	 *
	 * The solution is to save the pcidev->irq value and replace it with
	 * the MSI-X vector for the duration of the call to scsi_add_host().
	 */
	be_entry = &phba->msix_entries[def_msix_entries[0].index];

	if (enable_msix != 0) {
		irq_save = phba->pcidev->irq;
		phba->pcidev->irq = be_entry->msix_vector;
	}
#endif
	ret = scsi_add_host(host, &phba->pcidev->dev);



#ifdef __VMKLNX__
	if (enable_msix != 0) {
		phba->pcidev->irq = irq_save;
	}
#endif

	return ret;
}

/*++
Function:
	se_bios_param

Parameter:
	disk:  [in] pointer to SCSI Disk
	dev:   [in] device major, minor number
	geom:  [in] heads, sectors and cylinders

Return:
	0 

Context:
	Entry point

Description:
	This is the entry point called by the scsicam layer to get
	drive geometry information.

--*/

ACIT_SEXACT32
se_bios_param(struct scsi_device *sdev,
		struct block_device *bdev,
		sector_t capacity,
		int geom[])

{
	ACIT_SEXACT32 heads = -1;
	ACIT_SEXACT32 sectors = -1;
	ACIT_SEXACT32 cylinders = -1;

	if ((capacity >> 11) >= 1024) {
		heads = 255;
		sectors = 63;
	} else {
		heads = 64;
		sectors = 32;
	}

	cylinders = 100; /*capacity / (heads * sectors);*/
	//cylinders = capacity / (heads * sectors);

	geom[0] = heads;
	geom[1] = sectors;
	geom[2] = cylinders;

	return 0;
}

/*++
Function:
	se_timer

Parameter:
	arg:  [in] context information for the timer

Return:
	None 

Context:
	Timer

Description:
	This routine is the recurring timer for the driver. It calls into
	ACIT to support the 500ms timer that it needs.

--*/

void
se_timer (ACIT_U32_64 arg)
{
	se_hba                 *phba = (se_hba *)arg;
	ACIT_UEXACT32   acit_status;

	/* Let ACIT do its things */
	acit_status = acitapi_process_timer_tick(phba, &phba->controller_key);
	if (acit_status == SHTDWN_EVENT) {
		return;
	}
	/* Reschedule the timer */
	if (!phba->timer_shutdown)
		mod_timer(&phba->hba_timer, jiffies + TIMER_INTERVAL);
	return;
}



void se_schedule_job(se_hba *phba, struct se_timer_queue *job)
{
	ACIT_U32_64  flags = 0;
	spin_lock_irqsave(&phba->job_queue_lock, flags);
	list_add_tail(&job->job_queue, &phba->job_queue);
	spin_unlock_irqrestore(&phba->job_queue_lock, flags);
#if !defined(__VMKERNEL_MODULE__)
	if (phba->state & SE_STATE_HBA_INITIALIZED)
		up(&phba->kthread_sema);
#else
	complete(&phba->worker_thread.work_queued);
#endif
}                    

// Registers a Character device given the minor number

int register_be_char_dev(unsigned int minor_num)

{
	char be_char_name[15]; // Actual char device name
	dev_t mydev;
	void *pStatus;
	int val;
	se_hba *phba = NULL;
#ifdef __VMKERNEL_MODULE__

	snprintf(be_char_name, sizeof(be_char_name), "%s_%d",
			driver_name, minor_num);
	val = register_chrdev(0,be_char_name, &se_fops);
	if (val < 0) {
		printk("Failed register_chrdrv\n");
		return -1;
	}
	OSM_LOG_INIT_UNLOAD(phba,KERN_INFO," char device registered with major = %d \n",val);

	return val;
#else

	if (hba_count == 1) {

		/* Register our character Device */
		val = register_chrdev(0,driver_name, &se_fops);
		if (val < 0) {
			OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Failed register_chrdrv\n");
			goto char_dev_reg_failure;
		}

		be2iscsi_major = val;
	}

	// Get the dev for this minor num
	mydev = MKDEV( be2iscsi_major, minor_num );

	// Generate corresponding char dev name
	snprintf(be_char_name, sizeof(be_char_name), "%s_%d", 
			driver_name, minor_num);

	/* Let's Start Udev stuff */
	if (vnic_class == NULL) {
		vnic_class = class_create(THIS_MODULE,driver_name);

		if (IS_ERR(vnic_class)) {
			OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Error Creating class\n");
			goto char_dev_reg_failure_1;
		}
	}

	// MKNOD is done by this routine

	pStatus = (void *)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
		device_create
#else
		class_device_create

#endif
		(
		 vnic_class, 
		 NULL,
		 mydev, 
		 NULL, 
		 be_char_name
		);

	if (IS_ERR(pStatus)) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Error device creation\n");
		goto char_dev_reg_failure_2;

	}

//	printk("Char Dev: %s   Major: %d   Minor: %d Registered successfully\n",be_char_name, be2iscsi_major, minor_num);     

	return 0;

char_dev_reg_failure_2:

	class_destroy(vnic_class);

char_dev_reg_failure_1:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	val = unregister_chrdev(be2iscsi_major,driver_name);
#else
	unregister_chrdev(be2iscsi_major,driver_name);
#endif
char_dev_reg_failure:

	return -1;

#endif //VMWARE
}


/* ---------------------------------------------------------------- */

/*++
Function:
se_pci_dev_probe

Parameter:
pci:    [in]    pointer to struct pci
id:     [in]    pointer to pci device id (VID+DID)

Return:
0 - no errors
-ENODEV - could not enable device
- ACIT specific init failed
Context:
Entry point

Description:
This routine is called as a result of LLD calling pci_register_driver.
This is the probe function reported in struct pci_driver in this module.

--*/

ACIT_SEXACT32
se_pci_dev_probe(struct pci_dev *pcidev, const struct pci_device_id *id)
{
	ACIT_DEVICE     acit_device; 
	se_hba *      phba = NULL;
	ACIT_UEXACT32   acit_status = 0;
	ACIT_DISCOVER_PARAM        acit_discover_param;
	pid_t pid;
	u8* real_offset = 0;
	u32 value = 0;
	u32 num_login_retries=0;	
	u32 status=0;	
	int val = 0, hba_timer = 0;
	ACIT_U32_64     flags = 0;
	se_pci_resource *presource;
	PGET_BOOT_TARGET_OUT pattr_buf;
	sigset_t	sigblocked;
	sigset_t	sigsaved;
	uint8_t	port_name = 0xff;

	if (!dma_set_mask(&(pcidev->dev), DMA_64BIT_MASK))
	{
		if (pci_set_consistent_dma_mask(pcidev, DMA_64BIT_MASK))
		{
			if (pci_set_consistent_dma_mask(pcidev, DMA_32BIT_MASK))
			{
				OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Can't set DMA_64BIT_MASK/DMA_32BIT_MASK for pci_set_consistent_dma_mask\n");
				return -ENODEV;
			}
			else
			{
				OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"0x%p dma_set_mask = DMA_64BIT_MASK but pci_set_consistent_dma_mask = DMA_32BIT_MASK\n", pcidev);
			}
		}
		else
		{
	    		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"0x%p dma_set_mask = DMA_64BIT_MASK and pci_set_consistent_dma_mask = DMA_64BIT_MASK\n",pcidev);
		}
	}
	else if (!dma_set_mask(&(pcidev->dev), DMA_32BIT_MASK))
	{
		if (pci_set_consistent_dma_mask(pcidev, DMA_32BIT_MASK))
		{
				OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Can't Set DMA_32BIT_MASK for pci_set_consistent_dma_mask\n");
	    		return -ENODEV;
		}
		else
		{
	    	OSM_LOG_INIT_UNLOAD(phba,KERN_INFO,"0x%p dma_set_mask = DMA_32BIT_MASK and pci_set_consistent_dma_mask = DMA_32BIT_MASK\n", pcidev);
		}
	}
	else
	{
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"0x%p cant set dma_set_mask to either DMA_32BIT_MASK or DMA_64BIT_MASK\n", pcidev);
		return -ENODEV;
	}

	if (pci_enable_device(pcidev)) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"pci_enable_device failed\n");
		return -ENODEV;
	}

	pci_set_master(pcidev);

	/*
	 * Fill up the acit device structure with information from
	 * pci_dev and call common ACIT specific initialization
	 * routine
	 */

	memset(&acit_device, sizeof(acit_device),0);

	acit_device.vendor_id = pcidev->vendor;
	acit_device.device_id = pcidev->device;
	acit_device.subsys_id = pcidev->subsystem_device;
	acit_device.subsys_vendor_id = pcidev->subsystem_vendor;
	acit_device.revision_id = BE2ISCSI_PCI_REVISION_ID; 

	/* Allocate and initialize the data structure */

	if ((phba = kmalloc(sizeof(se_hba), GFP_KERNEL)) == NULL) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Memory Allocation Failure\n");
		goto no_mem;        
	}

	memset(phba, 0, sizeof(se_hba));
	INIT_LIST_HEAD(&phba->job_queue);
	INIT_LIST_HEAD(&phba->target_list);
	spin_lock_init(&phba->job_queue_lock);
#ifdef VIRT_TO_BUS_NOTSUPPORTED
	spin_lock_init(&phba->dma_profile_lock);
#endif

	/* Back pointer from pci_dev to se_hba */
	pci_set_drvdata(pcidev, (void *)phba);

	/* Setup pointer to pcidev so that we can call PCI functions */
	phba->pcidev = pcidev;

	/* Initialize spinlock to call into ACIT */
	spin_lock_init(&phba->acitapi_lock);
	sema_init(&phba->internal_io_lock, 1);

	/* Initialize ioctl_lock used in serialization of ioctls */
	sema_init(&phba->ioctl_lock, 1);

	 /* allocate pci resources */
	if (se_alloc_io_handles(phba) != ACIT_STATUS_SUCCESS) {
		goto free_hba;
	}

	se_verify_parameters();

	if (!hba_count) {
		presource = &phba->resourcemap[1];

		real_offset = (u8 *)presource->address + MPU_EP_SEMAPHORE;
		value = readl((void *)real_offset);
		if (value & 0x00010000) {
			gcrashmode = 1;
			sync_cd_ioc = 1;
			OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Loading Driver in crashdump mode \n");
		} else {
			value |= 0x00010000;
			writel(value,(void *)real_offset);
		}

		acit_initialize(ACIT_OP_MODE_NORMAL, &acitapi_table,
		sizeof(ACIT_API_TABLE));

		acitapi_get_api_version(&acit_api_version);
	}

	if(large_io > 128){
		sge = 256;
		max_io_size = large_io;
	}

	se_init_params();

	acit_status = se_initialize_hba(pcidev, &acit_device, phba);

	if (acit_status != ACIT_STATUS_SUCCESS) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"HBA Initialization failed\n");
		goto free_hba;
	}

#ifdef VIRT_TO_BUS_NOTSUPPORTED
	phba->internal_ioctl = (PACIT_VOID) se_pci_alloc_consistent(phba->pcidev, 
				sizeof(ACIT_IOCTL_REQUEST), &phba->internal_ioctl_pa);
	if (phba->internal_ioctl == NULL) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"ioctl mem alloc failed\n");
		goto free_hba;
	}
#endif

	init_timer(&phba->hba_timer);
	phba->timer_shutdown = 0;
	phba->hba_timer.function = se_timer;
	phba->hba_timer.data = (unsigned long)phba;
	phba->hba_timer.expires = jiffies + TIMER_INTERVAL;
	add_timer(&phba->hba_timer);
	hba_timer++;

	if (se_scsi_register(phba, &driver_template)) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Registering with SML failed\n");
		goto free_hba;
	} 

#if defined (__VMKLNX__) && defined (REGISTER_ISCSI)
	if (iscsi_register_host(phba->host, &be2iscsi_iscsi_transport) != 0) {

		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Registering with iscsi_register_host failed\n");
		goto free_hba;
	}
#endif

#if defined ESX_50
	if(enable_msix){

		vmklnx_scsi_register_poll_handler(phba->host,
				phba->msix_entries[0].msix_vector,
				def_msix_entries[0].handler,
				phba);
	} else {
		vmklnx_scsi_register_poll_handler(phba->host,
				phba->pcidev->irq,
				se_isr,
				phba);
	}
#endif

	acit_discover_param.info = 0; /* undefined */
	acit_discover_param.pptarget_table = &phba->pacit_target_table;
	acit_discover_param.result = 0; /* undefined */
	acit_discover_param.reserved = 0;
	acit_discover_param.opcode = ACIT_DISCOVER_OPCODE_ASYNCHRONOUS;

	acit_status = acitapi_discover_targets(phba,&phba->controller_key,
				&acit_discover_param);

	phba->worker_thread.terminate = 0;
	init_completion(&phba->worker_thread.thread_completion);
	init_waitqueue_head(&phba->worker_thread.queue);

	/*
	 * acit_discover_targets might have put some jobs in
	 * the job queue for deferred execution. Since we are
	 * in the initialization path we want everything to be
	 * synchronous. So rather than processing it inside
	 * timer and wait for it to finish, we will just
	 * process it here synchronously. The jobs we are
	 * expecting here are se_target_add
	 * only 
	 */

	se_process_deferred_jobs(phba);

	val = register_be_char_dev(phba->hba_unique_id);

	if (val < 0) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Failed register_chrdrv\n");
	}

	phba->major = val;

	OSM_LOG_INIT_UNLOAD(phba,KERN_INFO,"max_id = %d, max_lun = %d, max_channel = %d\n",
			phba->host->max_id, phba->host->max_lun, 
			phba->host->max_channel);

	spin_lock_irqsave(&phba->acitapi_lock,flags);
	phba->state |= SE_STATE_HBA_INITIALIZED;
#if !defined(__VMKERNEL_MODULE__)
	sema_init(&phba->kthread_sema, 0);
#endif
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);

#if !defined(__VMKERNEL_MODULE__) && defined(MODULE)
	sigfillset(&sigblocked);
	sigprocmask(SIG_BLOCK, &sigblocked, &sigsaved);
	flush_signals(current);
#endif

	pid = kernel_thread(se_worker_thread, phba, 0);
	if (pid < 0) {
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"kernel_thread failed with status %d\n",
			pid);
		goto free_hba;
	}
#if !defined(__VMKERNEL_MODULE__) && defined(MODULE)
	sigprocmask(SIG_BLOCK, &sigsaved, &sigblocked);
#endif


#if !defined(__VMKERNEL_MODULE__)
	up(&phba->kthread_sema);
#endif
	complete(&phba->worker_thread.work_queued);
	num_login_retries=0;
retry_login:	
	pattr_buf = (PGET_BOOT_TARGET_OUT)kmalloc(sizeof(GET_BOOT_TARGET_OUT),GFP_KERNEL);
	status = se_get_boot_target_info(phba,pattr_buf);
	if (status) {
		kfree(pattr_buf);
		num_login_retries++;
		mdelay(1000);
		if (num_login_retries > 5) {
		  OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"se_get_boot_target_info Failed. Continuing to load \n");
		  goto load_drvr;		
		}else
		  goto retry_login;
	}
	if (pattr_buf->boot_session_count) {
	   if (pattr_buf->boot_session_handle == 0xFFFFFFFF) {
		kfree(pattr_buf);
		num_login_retries++;
		mdelay(1000);
		if (num_login_retries > 10) {
		  OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Could not log into iscsi boot. Continuing to load \n");
    		  goto load_drvr;		
		}else
		  goto retry_login;
	   }else {
		num_login_retries=0;
		while(!phba->boot_tgt_added) {
		   mdelay(1000);
		   num_login_retries++;
		   if (num_login_retries > 10) {
		       OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"FW reported boot Target but Failed to Add it\n");
		       goto load_drvr;		
	  	   } 		
		} 	
	 }
	}

load_drvr:
#ifndef __VMKERNEL_MODULE__
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,4)
	be2iscsi_alloc_sysfs_attr(phba);
#endif
#endif
	se_get_port_name(phba, &port_name);

	OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"scsi%d : be2iscsi port name %c\n", phba->host->host_no, port_name);
	OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"scsi%d : be2iscsi version %s\n", phba->host->host_no, BUILD_STR);

	/* Set teh default logging level for IO */
	//phba->osm_io_loglvl = OSM_LVL_1;
	if(se_proc_create(phba)){
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"se_proc_create failed \n");
	}

#ifndef __VMKLNX__
	if(se_proc_add_hostdir(phba)){
		OSM_LOG_INIT_UNLOAD(phba,KERN_ERR,"Debugfs Entry Creation Failed\n");
	}
#endif			
	return 0;

free_hba:
no_mem:
	if(phba) {
		if (hba_timer)
			del_timer_sync(&phba->hba_timer);

		presource = &phba->resourcemap[1];
		real_offset = (u8 *)presource->address + MPU_EP_SEMAPHORE;

		value = readl((void *)real_offset);

		if (value & 0x00010000) {
			value &= 0xfffeffff;
			writel(value, (void *)real_offset);	
		}

		se_release_io_handles(phba);	

#ifdef VIRT_TO_BUS_NOTSUPPORTED
		if (phba->internal_ioctl)
			se_pci_free_consistent(phba->pcidev, sizeof(ACIT_IOCTL_REQUEST),
				(void *) phba->internal_ioctl, phba->internal_ioctl_pa);
#endif

		kfree(phba);

		phba = NULL;
	}

	pci_set_drvdata(pcidev, NULL);
	pci_disable_device(pcidev);

	LnxLogOSMError(UNI_CFG_LOG_LEVEL_CRITICAL,
			UNI_PHASE_INIT,
			UNI_SUBMODULE_LNXOSM_GENERIC,
			OSM_EC_HW_INIT_FAILED);
	return -ENODEV;
}

// Unregister char device given minor number

void unregister_be_char_dev(unsigned int minor_num)

{
	dev_t mydev;

	mydev = MKDEV(be2iscsi_major, minor_num);

	/* hotplug: only remove chrdev on last hba count */
	if (hba_count == 1) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
		int val;
		val = unregister_chrdev(be2iscsi_major,driver_name);
#else
		unregister_chrdev(be2iscsi_major,driver_name);
#endif
	}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26)
	device_destroy(vnic_class,mydev);
#else
	class_device_destroy(vnic_class, mydev);
#endif
	if (hba_count == 1) {
		class_destroy(vnic_class);
		vnic_class = NULL;
	}


}


/*++
Function:
se_pci_dev_remove

Parameter:
pci:    [in]    pointer to struct pci

Return:
None

Context:
Entry point

Description:
This routine is called as a result of LLD calling pci_unregister_driver.
This is the remove function reported in struct pci_driver.         

--*/

	void 
se_pci_dev_remove(struct pci_dev *pci)
{
	ACIT_U32_64  flags = 0;
	se_hba *phba = NULL;
	u8* real_offset = 0;
	u32 n,value = 0;
	se_pci_resource *presource = NULL;
	struct be_msix_entry  	*be_entry;
	/* clean up acit stuff */
	down(&se_configuration_lock);
	phba = (se_hba *)pci_get_drvdata(pci);
	if (!phba) {
	 	up(&se_configuration_lock);
		return;
	}
	phba->timer_shutdown = 1;
	unregister_be_char_dev(phba->hba_unique_id);
	if (!gcrashmode)
		del_timer_sync(&phba->hba_timer);
	
	spin_lock_irqsave(&phba->acitapi_lock,flags);
	phba->state |= SE_STATE_SHUTDOWN;
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);
	acitapi_disable_irq(phba, &phba->controller_key);

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,5) 

	if (enable_msix) {
		for(n = 0; n < BE_NUM_MSIX_ENTRIES; n++) {
			be_entry = &phba->msix_entries[def_msix_entries[n].index];
			if (be_entry->msix_enabled)
				free_irq(be_entry->msix_vector, phba);
			}
			pci_disable_msix(phba->pcidev);
	} else 
#endif
	{
		msleep(10);
		if (phba->host->irq)
			free_irq(phba->host->irq, phba);
	}

	spin_lock_irqsave(&phba->acitapi_lock,flags);
#ifndef __VMKERNEL_MODULE__
	be2iscsi_free_sysfs_attr(phba);
#endif
	phba->state |= SE_STATE_DELETE_HBA;
	spin_unlock_irqrestore(&phba->acitapi_lock,flags);
	presource = &phba->resourcemap[1];
	real_offset = (u8 *)presource->address + MPU_EP_SEMAPHORE;
	value = readl((void *)real_offset);
	if (value & 0x00010000) {
		value &= 0xfffeffff;
		writel(value,(void *)real_offset);
	}

rem_tgt:
	 msleep(1);
	 spin_lock_irqsave(&phba->acitapi_lock,flags);
	 if (phba->state & SE_STATE_REMOVE_LUNS) {
		 spin_unlock_irqrestore(&phba->acitapi_lock,flags);
		 goto rem_tgt;
	 }else
		 spin_unlock_irqrestore(&phba->acitapi_lock,flags);
	se_proc_destroy(phba);
#ifndef __VMKLNX__
	 /* Remove the debug enteries */
	 se_proc_rm_hostdir(phba);
#endif

	se_delete_hba(phba,ACIT_STOP_TYPE_POWER_MGMT);
	pci_set_drvdata(pci,NULL);
	pci_disable_device(pci);
	up(&se_configuration_lock);
	return;
}
/* ---------------------------------------------------------------- */

/* --------------- cleanup helper functions ----------------------*/
/*++
Function:
se_delete_hba

Parameter:
phba: [in]       pointer to allocated memory for HBA

Return:
None

Context:
Unload

Description:
This routine deletes all resources belonging to an hba entry

--*/
	void 
se_delete_hba(se_hba *phba,ACIT_UEXACT32 stop_type)
{
	/*
	 *  Remove the timer so timer tick stops
	 */
	BEGIN;

#ifndef __VMKERNEL_MODULE__

#if LINUX_VERSION_CODE != KERNEL_VERSION(2,6,5)
	// The problem described below was seen on SLES9 systems only

	/*
	 * We get into this case if the driver is being unloaded (the
	 * system is not being shutdown).  We avoid this call in the
	 * shutdown path because it will potentially trigger a 
	 * hotplug-stopped helper to run and issue I/O's.  If we do 
	 * call these functions during shutdown, we will get device not 
	 * connected errors and in the case of Reiserfs, we will get 
	 * Reiser specific error messages.
	 */
	scsi_remove_host(phba->host);
	scsi_host_put(phba->host);
#else
	if (!(gshtdwn == 1)) {
		scsi_remove_host(phba->host);
		scsi_host_put(phba->host);
	}
#endif
	if (!gcrashmode) {
		acitapi_stop_controller(phba, &phba->controller_key, stop_type);
	}
	phba->worker_thread.terminate = 1;
#if !defined(__VMKERNEL_MODULE__)
	up(&phba->kthread_sema);
#else
	complete(&phba->worker_thread.work_queued);
#endif
	wait_for_completion(&phba->worker_thread.thread_completion);
	se_process_deferred_jobs(phba);

	if (phba->host) {
		if (!(gshtdwn == 1))  {
			aim_free_mem_params(phba);
			se_release_io_handles(phba);
		}
	}

	if (phba->pcontroller_ws) {
		kfree(phba->pcontroller_ws);
		phba->pcontroller_ws = NULL;
	}

	list_del_init(&phba->hba_queue); 

	hba_count--;
	if (phba->hba_unique_id < BE2ISCSI_MAX_HBA)
		se_hba_uid[phba->hba_unique_id] |= SE_HBA_REMOVED;

	if (phba)
		kfree(phba); 
#endif
}


/*++

Function:
    se_delete_all_hba

Context:
	Init, shutdown.

Description:
    This function cleans up all memory allocated by the driver for all
    HBAs. The HBA chain is traversed and all HBA instances are deleted.

--*/

void 
se_delete_all_hba(void)
{
	se_hba         *phba = NULL;
	struct list_head *lh, *lh_sf;
	ACIT_U32_64  flags = 0;
	int n = 0;
	struct be_msix_entry  	*be_entry;
	 
	/* 
	 * Delete all Host Adapter from the controller chain 
	 */

	down(&se_configuration_lock);

	list_for_each_safe(lh, lh_sf, &hba_chain) {
		phba = list_entry(lh, se_hba, hba_queue); 
		phba->timer_shutdown = 1;
		if (!gcrashmode)
		       	 del_timer_sync(&phba->hba_timer);
		
		spin_lock_irqsave(&phba->acitapi_lock,flags);
		phba->state |= SE_STATE_SHUTDOWN;
		spin_unlock_irqrestore(&phba->acitapi_lock,flags);
		acitapi_disable_irq(phba, &phba->controller_key);
		
		if (enable_msix) {
                        for(n = 0; n < BE_NUM_MSIX_ENTRIES; n++) {
                                be_entry = &phba->msix_entries[def_msix_entries[n].index];
                                if (be_entry->msix_enabled)
                                        free_irq(be_entry->msix_vector, phba);
                        }
                        pci_disable_msix(phba->pcidev);
                } else{
                        if (phba->host->irq)
                                free_irq(phba->host->irq, phba);

                }

	}


	list_for_each_safe(lh, lh_sf, &hba_chain) {
		phba = list_entry(lh, se_hba, hba_queue); 
		spin_lock_irqsave(&phba->acitapi_lock,flags);
		phba->state |= SE_STATE_DELETE_HBA;
		spin_unlock_irqrestore(&phba->acitapi_lock,flags);
		se_delete_hba(phba,ACIT_STOP_TYPE_NORMAL);
	}
	up(&se_configuration_lock);

}

void
se_free_dma_buffer(
		se_hba                        *phba, 
		void                             *entry_data)
{
	struct se_free_dma_buf *buff = (struct se_free_dma_buf *) entry_data;

	se_pci_free_consistent(phba->pcidev, 
			buff->psg.length,
			(void *) buff->virt,
			buff->psg.address.u.a64.address);
}


void
se_change_luns(
		se_hba                        *phba, 
		void                             *entry_data)
{
	se_target               *ptarget;
	int                     command;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)
	ACIT_UEXACT8        	lun = 0; 
	se_device 		*pdevice;
	ACIT_UEXACT32 		mod_flag = 0;
	struct list_head        *lh, *lh_sf = NULL;
	struct scsi_device *sdev;
	ACIT_UEXACT32 status = FAILED;        
#endif
	struct se_target_job_data *pdata = 
				(struct se_target_job_data *)entry_data;

	BEGIN;
	ptarget     = pdata->ptarget;
	command     = pdata->command;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,6)

		list_for_each_safe(lh, lh_sf, &ptarget->lun_list) {
		    pdevice = list_entry(lh, se_device, sibling_list);  
		    if (!(pdevice->scsi_lun == lun)) {
		     	mod_flag = 1;
		     	scsi_remove_device(pdevice->pScsi_dev);
		    }
		}

		if (!mod_flag) {
			sdev = scsi_add_device(phba->host,ptarget->bus,ptarget->target,lun);
			if ((sdev != -ENODEV)) {
				status = se_slave_alloc(sdev);
				if (status == -ENXIO)
					printk(KERN_ERR "No Target Found \n");
			}
		}

#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16)
	down(&phba->host->scan_mutex);
	scsi_scan_target(phba->host,ptarget->bus,ptarget->target,-1,1);
	up(&phba->host->scan_mutex);
#else
	scsi_scan_target(&phba->host->shost_gendev,ptarget->bus,ptarget->target,-1,1);
#endif
#endif

}

void
se_delete_target_job(
se_hba	*phba, 
void	*entry_data
)
{
	struct se_target_job_data	*pdata;
	se_target			*ptarget;
	int				command;

	BEGIN;


	pdata = (struct se_target_job_data *)entry_data;

	ptarget = pdata->ptarget;

	command = pdata->command;

	ptarget->state |= SE_TARGET_DESTROY;
	
	se_delete_target(phba, ptarget);
}

void
se_destroy_target(
se_hba		*phba, 
se_target	*ptarget
)
{
	ACIT_U32_64		flags;

	spin_lock_irqsave(&phba->acitapi_lock,flags);

	phba->state &= ~SE_STATE_REMOVE_LUNS;

	spin_unlock_irqrestore(&phba->acitapi_lock,flags);
#if defined (__VMKLNX__) && defined (REGISTER_ISCSI)
	if (ptarget->conn != NULL) {
		iscsi_if_destroy_session_done(ptarget->conn);
		iscsi_destroy_conn(ptarget->conn);
		ptarget->conn = NULL;
	}

	if (ptarget->sess != NULL) {
		iscsi_offline_session(ptarget->sess);
		iscsi_remove_session(ptarget->sess);
	}
#else
	se_return_to_bt_slot(phba, ptarget->bus, ptarget->target);
	list_del_init(&ptarget->sibling_list);
	kfree(ptarget->pbtl_mapping);
	kmem_cache_free(phba->posm_target_cache, ptarget);

#endif

	return;
}

void
se_delete_target(
se_hba		*phba, 
se_target	*ptarget
)
{
	se_device		*pdevice;
	struct list_head        *lh;
	struct list_head        *lh_sf;
	ACIT_U32_64		flags;
	BEGIN;

	lh_sf = NULL;
	flags = 0;

	/* TODO: cleanup whatever acit_init_target_ws did */
	if (ptarget->pacit_target_ws) {
		kmem_cache_free(phba->pacit_target_ws_cache,
				ptarget->pacit_target_ws);
	}

	/*
	 * cleanup any LUN specific stuff
	 */
	list_for_each_safe(lh, lh_sf, &ptarget->lun_list) {

		pdevice = list_entry(lh, se_device, sibling_list);  

		if (pdevice == NULL) {
			continue;
		}

		OSM_LOG_CONFIG(phba, KERN_ERR,"BTL:%d:%d:%d\n",
			pdevice->scsi_channel,
			pdevice->scsi_id,
			pdevice->scsi_lun);

		spin_lock_irqsave(&phba->acitapi_lock,flags);

		phba->state |= SE_STATE_REMOVE_LUNS;

#if defined(__VMKLNX__)
		pdevice->state |= SE_DEV_DESTROY;
#endif

		if (!(phba->state & SE_STATE_DELETE_HBA)) {
			spin_unlock_irqrestore(&phba->acitapi_lock,flags);
#ifndef REGISTER_ISCSI
			scsi_remove_device(pdevice->pScsi_dev);
#endif
		} else {
			spin_unlock_irqrestore(&phba->acitapi_lock,flags);
		}
	}

		se_destroy_target(phba, ptarget);


	return;
}


/*++
Function:
	se_shutdown_event

Parameter:
	np:     [in]    pointer to notifier block
	code:   [in]    event code (restart/halt/poweroff)
	unused: [in]    pointer to context, not used

Return:
	NOTIFY_DONE - the notification was handled

Context:
	Init/unload

Description:
	This routine is the shutdown event handler for the driver.
	It flushes all commands outstanding and frees up the memory
	allocated.
--*/

ACIT_SEXACT32
se_shutdown_event(
		struct notifier_block *nb, 
		ACIT_U32_64 code, 
		void *unused)
{
	gshtdwn = 1; //  Needs no lock as it is modified only during shutdown
	if ((code == SYS_RESTART) || (code == SYS_HALT) ||
		(code == SYS_POWER_OFF)) {
		/*
		 * Shutdown Host Adapter and perform House Cleaning. 
		 */  
		se_delete_all_hba();
	}

	return NOTIFY_DONE;
}

void 
be2iscsi_dump_poll(struct scsi_device *device)
{
	se_hba    *phba = NULL;

	phba = (se_hba *) device->host->hostdata[0];

	if (enable_msix)
		acitapi_process_irq_msix(phba, &phba->controller_key);
	else
		acitapi_process_irq(phba, &phba->controller_key);
}

int
be2iscsi_sanity_check(struct scsi_device *device)
{
	se_hba    *phba = NULL;

	be2iscsi_dump_in_progress = 1;

	phba = (se_hba *) device->host->hostdata[0];

	if (spin_is_locked(&phba->acitapi_lock)) 
		return -EBUSY;

	return 0;
}


#ifndef __VMKERNEL_MODULE__
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,4)

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_fw_version(struct device *class_dev,
				 struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_fw_version(struct class_device *class_dev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	se_hba *phba = (se_hba *)shost->hostdata[0];
	int len,status;
	ACIT_U32_64  flags = 0;

	PMGMT_HBA_ATTRIBUTES pattr_buf;


	pattr_buf = (PMGMT_HBA_ATTRIBUTES)kmalloc(sizeof(MGMT_HBA_ATTRIBUTES),GFP_KERNEL);
	status = se_get_cntlr_info(phba,pattr_buf);
	if (status) {
		kfree(pattr_buf);
		return 0;
	}
	spin_lock_irqsave(shost->host_lock, flags);
	len = snprintf(buf, PAGE_SIZE_4K, "%s \n",pattr_buf->firmware_version_string);
	spin_unlock_irqrestore(shost->host_lock, flags);
	kfree(pattr_buf);
	return len;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_port_name(struct device *class_dev,
				 struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_port_name(struct class_device *class_dev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	se_hba *phba = (se_hba *)shost->hostdata[0];
	int len,status;
	ACIT_U32_64  flags = 0;
	uint8_t	port_name;


	status = se_get_port_name(phba, &port_name);
	if (status) {
		return 0;
	}
	spin_lock_irqsave(shost->host_lock, flags);

	len = snprintf(buf, PAGE_SIZE_4K, "%c \n", port_name);

	spin_unlock_irqrestore(shost->host_lock, flags);
	return len;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_driver_version(struct device *class_dev,
				     struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_driver_version(struct class_device *class_dev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	int len;
	ACIT_U32_64  flags = 0;

	spin_lock_irqsave(shost->host_lock, flags);
	len = snprintf(buf, PAGE_SIZE_4K, "%s \n", be2iscsi_driver_version);
	spin_unlock_irqrestore(shost->host_lock, flags);
	return len;

}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_ip_address(struct device *class_dev,
				 struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_ip_address(struct class_device *class_dev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	se_hba *phba = (se_hba *)shost->hostdata[0];
	int len,status;
	ACIT_U32_64  flags = 0;

	PGET_IF_INFO_PARAMS pattr_buf;

	pattr_buf = (PGET_IF_INFO_PARAMS) kmalloc(sizeof(GET_IF_INFO_PARAMS),GFP_KERNEL);
	status = se_get_nw_info(phba,pattr_buf);
	if (status) {
		kfree(pattr_buf);
		return 0;
	}
	spin_lock_irqsave(shost->host_lock, flags);
	len = snprintf(buf, PAGE_SIZE_4K, "%d.%d.%d.%d \n",pattr_buf->response.ip_address[0].IPAddress[0],pattr_buf->response.ip_address[0].IPAddress[1],pattr_buf->response.ip_address[0].IPAddress[2],pattr_buf->response.ip_address[0].IPAddress[3] );
	spin_unlock_irqrestore(shost->host_lock, flags);
	kfree(pattr_buf);
	return len;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_gateway(struct device *class_dev,
			      struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_gateway(struct class_device *class_dev, char *buf)
#endif
{

	struct Scsi_Host *shost = class_to_shost(class_dev);
	se_hba *phba = (se_hba *)shost->hostdata[0];
	int len,status;
	ACIT_U32_64  flags = 0;

	PGET_DEFAULT_ROUTE_PARAMS pattr_buf;

	pattr_buf = (PGET_DEFAULT_ROUTE_PARAMS ) kmalloc(sizeof(GET_DEFAULT_ROUTE_PARAMS),GFP_KERNEL);
	status = se_get_default_gateway(phba,pattr_buf);
	if (status) {
		kfree(pattr_buf);
		return 0;
	}
	spin_lock_irqsave(shost->host_lock, flags);
	len = snprintf(buf, PAGE_SIZE_4K, "%d.%d.%d.%d \n",pattr_buf->response.gateway.IPAddress[0],pattr_buf->response.gateway.IPAddress[1],pattr_buf->response.gateway.IPAddress[2],pattr_buf->response.gateway.IPAddress[3]);
	spin_unlock_irqrestore(shost->host_lock, flags);
	kfree(pattr_buf);
	return len;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_netmask(struct device *class_dev,
			      struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_netmask(struct class_device *class_dev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	se_hba *phba = (se_hba *)shost->hostdata[0];
	int len,status;
	ACIT_U32_64  flags = 0;

	PGET_IF_INFO_PARAMS pattr_buf;

	pattr_buf = (PGET_IF_INFO_PARAMS) kmalloc(sizeof(GET_IF_INFO_PARAMS),GFP_KERNEL);
	status = se_get_nw_info(phba,pattr_buf);
	if (status) {
		kfree(pattr_buf);
		return 0;
	}
	spin_lock_irqsave(shost->host_lock, flags);
	len = snprintf(buf, PAGE_SIZE_4K, "%d.%d.%d.%d \n",pattr_buf->response.ip_address[0].SubNetMask[0],pattr_buf->response.ip_address[0].SubNetMask[1],pattr_buf->response.ip_address[0].SubNetMask[2],pattr_buf->response.ip_address[0].SubNetMask[3] );

	spin_unlock_irqrestore(shost->host_lock, flags);
	kfree(pattr_buf);
	return len;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_mac_address(struct device *class_dev,
				  struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_mac_address(struct class_device *class_dev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	se_hba *phba = (se_hba *)shost->hostdata[0];
	int len,status;
	ACIT_U32_64  flags = 0;

	PNIC_CONFIG_PARAMS pattr_buf;

	pattr_buf = (PNIC_CONFIG_PARAMS) kmalloc(sizeof(NIC_CONFIG_PARAMS),GFP_KERNEL);
	status = se_get_nic_info(phba,pattr_buf);
	if (status) {
		kfree(pattr_buf);
		return 0;
	}
	spin_lock_irqsave(shost->host_lock, flags);
	len = snprintf(buf, PAGE_SIZE_4K, "%x:%x:%x:%x:%x:%x \n",pattr_buf->nic_ports[0].mac_address.MACAddress[0],pattr_buf->nic_ports[0].mac_address.MACAddress[1],pattr_buf->nic_ports[0].mac_address.MACAddress[2],pattr_buf->nic_ports[0].mac_address.MACAddress[3],pattr_buf->nic_ports[0].mac_address.MACAddress[4],pattr_buf->nic_ports[0].mac_address.MACAddress[5]);

	spin_unlock_irqrestore(shost->host_lock, flags);
	kfree(pattr_buf);
	return len;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_adapter_iqn(struct device *class_dev,
				  struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_adapter_iqn(struct class_device *class_dev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	se_hba *phba = (se_hba *)shost->hostdata[0];
	int len,status;
	ACIT_U32_64  flags = 0;

	PISCSI_HBA_NAME_PARAMS pattr_buf;

	pattr_buf = (PISCSI_HBA_NAME_PARAMS) kmalloc(sizeof(ISCSI_HBA_NAME_PARAMS),GFP_KERNEL);
	status = se_get_iqn_info(phba,pattr_buf);
	if (status) {
		kfree(pattr_buf);
		return 0;
	}
	spin_lock_irqsave(shost->host_lock, flags);
	len = snprintf(buf, PAGE_SIZE_4K, "%s \n",pattr_buf->iscsiname);
	spin_unlock_irqrestore(shost->host_lock, flags);
	kfree(pattr_buf);
	return len;
}

#ifndef __VMKERNEL_MODULE__
#ifdef NEW_BLK_API
ssize_t be2iscsi_read_login_state(struct file *filp,
		struct kobject *kobj,
		struct bin_attribute *bin_attr,
		char *buf, loff_t off,
		size_t count)
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_read_login_state(struct kobject *kobj,
		struct bin_attribute *bin_attr,
		char *buf, loff_t off,
		size_t count)
#else
ssize_t be2iscsi_read_login_state(struct kobject *kobj, char *buf, loff_t off,
		size_t count)
#endif
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
	struct Scsi_Host *shost = class_to_shost(container_of(kobj,struct class_device,kobj));
#else
	struct Scsi_Host *shost = dev_to_shost(container_of(kobj, struct device, kobj));
#endif
	se_hba *phba = (se_hba *)shost->hostdata[0];
	unsigned int len,status,cnt,i;
	ACIT_U32_64  flags = 0;
	char *tempbuff = NULL;
	
	PGET_ALL_SESSION_ID_PARAMS pattr_buf = NULL;
	PGET_A_SESSION_PARAMS	   psess_buf = NULL;
	DISP_LG_TIQN 		   *pdisp_lgtiqn = NULL;	

	/* 128 session per port and 4 bytes for each handle */
	len = (128 * 4);
	pattr_buf = (PGET_ALL_SESSION_ID_PARAMS)kmalloc(len,GFP_KERNEL);
	if(!pattr_buf) {
		printk(KERN_ERR"Memory Allocation Failure\n");
		return -ENOMEM;
	}
	memset(pattr_buf,0x0,len);	

	status = se_get_allsess_info(phba,pattr_buf,len);
	if (status) {
		kfree(pattr_buf);
		if (status <8) {
			return 0;
		} else {
			pattr_buf = (PGET_ALL_SESSION_ID_PARAMS) kmalloc(status,GFP_KERNEL);
			memset(pattr_buf,0x0,sizeof(GET_ALL_SESSION_ID_PARAMS));	
			status = se_get_allsess_info(phba,pattr_buf,status);
			if (status) {
				kfree(pattr_buf);
				return 0;
			}
		}
	}
	cnt = pattr_buf->response.session_id_count;
	pdisp_lgtiqn = (PDISP_LG_TIQN) kmalloc(sizeof(SE_SESSION_DATA)*cnt,GFP_KERNEL);
	if(!pdisp_lgtiqn) {
		return -ENOMEM;
	}
	psess_buf = (PGET_A_SESSION_PARAMS) kmalloc(sizeof(GET_A_SESSION_PARAMS),GFP_KERNEL);
	if(!psess_buf) {
		return -ENOMEM;
	}
	for(i = 0 ; i < cnt ; i++) {
			memset(psess_buf,0x0,sizeof(GET_A_SESSION_PARAMS));
			status = se_get_asession_info(phba,psess_buf,pattr_buf->response.session_id_list[i].session_handle);
				
			if (status) {
				kfree(psess_buf);
				kfree(pattr_buf);
				kfree(pdisp_lgtiqn);
				return 0;
			}
		pdisp_lgtiqn->session_list[i].login_state = psess_buf->response.session_info.status;
		memcpy(&pdisp_lgtiqn->session_list[i].target_name,&psess_buf->response.session_info.target_name,224);

	}
	
	len = 0;
	spin_lock_irqsave(shost->host_lock, flags);
	if (0 == cnt) {
		if (off == 0)
			count = snprintf(buf+len , PAGE_SIZE_4K - len, "Login_state = Not Logged In\n");
		else count = 0;
	} else {
		unsigned int alloc_size = 0;
		
		alloc_size = cnt * 512;
		tempbuff = (char *) kmalloc(alloc_size,GFP_KERNEL);

		if (!tempbuff) { 
			count = 0;
			goto out;
		}
		memset(tempbuff,0x0,alloc_size);

		for (i = 0 ; i < cnt ; i++) {
			if (pdisp_lgtiqn->session_list[i].login_state & MGMT_SESSION_STATUS_OPENED)
				len += snprintf(tempbuff+len , alloc_size - len, "Login_state = Logged In,      Target iSCSI Name = %s \n ",pdisp_lgtiqn->session_list[i].target_name);
			else
				len += snprintf(tempbuff+len , alloc_size - len, "Login_state = Not Logged In,  Target iSCSI Name = %s \n ",pdisp_lgtiqn->session_list[i].target_name);
		}

		if (off > len) {
			count = 0;
			kfree(tempbuff);
			goto out;
		}

		if (off + count > len)
			count = len - off;

		memcpy(buf, &tempbuff[off], count);
		kfree(tempbuff);
		
	}
out:
	spin_unlock_irqrestore(shost->host_lock, flags);
	kfree(psess_buf);
	kfree(pattr_buf);
	kfree(pdisp_lgtiqn);

	return count;
}


#ifdef NEW_BLK_API
ssize_t be2iscsi_read_target_iqn(struct file *filp,
		struct kobject *kobj,
		struct bin_attribute *bin_attr,
		char *buf, loff_t off,
		size_t count)
{
	ssize_t status;
	status = be2iscsi_read_login_state(filp, kobj, bin_attr, buf, off, count);
	return status;
}
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_read_target_iqn(struct kobject *kobj,
		struct bin_attribute *bin_attr,
		char *buf, loff_t off,
		size_t count)
{
	ssize_t status;
	status = be2iscsi_read_login_state(kobj, bin_attr, buf, off, count);
	return status;
}
#else
ssize_t be2iscsi_read_target_iqn(struct kobject *kobj, char *buf, loff_t off,
		size_t count)
{
	ssize_t status;
	status = be2iscsi_read_login_state(kobj, buf, off, count);
	return status;
}
#endif
#endif


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_adapter_state(struct device *class_dev,
				    struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_adapter_state(struct class_device *class_dev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	se_hba *phba = (se_hba *)shost->hostdata[0];
	int len,status;
	ACIT_U32_64  flags = 0;

	PMGMT_HBA_ATTRIBUTES pattr_buf;

	pattr_buf = (PMGMT_HBA_ATTRIBUTES)kmalloc(
		sizeof(MGMT_HBA_ATTRIBUTES),GFP_KERNEL);

	status = se_get_cntlr_info(phba,pattr_buf);

	if (status) {
		kfree(pattr_buf);
		return 0;
	}

	spin_lock_irqsave(shost->host_lock, flags);

	if (pattr_buf->hba_status)
		len = snprintf(buf, PAGE_SIZE_4K, "Need to upgrade Firmware\n");
	else 
		len = snprintf(buf, PAGE_SIZE_4K, "Active\n");

	spin_unlock_irqrestore(shost->host_lock, flags);

	kfree(pattr_buf);

	return len;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,25)
ssize_t be2iscsi_show_adapter_info(struct device *class_dev,
				   struct device_attribute *attr, char *buf)
#else
ssize_t be2iscsi_show_adapter_info(struct class_device *class_dev, char *buf)
#endif
{
	struct Scsi_Host *shost = class_to_shost(class_dev);
	se_hba *phba = (se_hba *)shost->hostdata[0];
	int len;
	ACIT_U32_64  flags = 0;

	spin_lock_irqsave(shost->host_lock, flags);

	if (phba->pcidev->device == 0x0222) {
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe11100 10Gb iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
		spin_unlock_irqrestore(shost->host_lock, flags);
		return len;
	}

	if (phba->pcidev->device == 0x0712) {
		if (phba->pcidev->subsystem_device == 0x304b)
			len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe11104-F-HI 10Gb iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
		else
			len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe11100, iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
		spin_unlock_irqrestore(shost->host_lock, flags);
		return len;
	}

	if (phba->pcidev->subsystem_device == 0x1000)
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x1001) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x1010) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb CNA iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x1011) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb CNA iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x2000 ) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x2001 ) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x2010 ) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb CNA iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x2011 ) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb CNA iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x1020 ) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x1021 ) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x1030 ) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb CNA iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else if (phba->pcidev->subsystem_device == 0x1031 ) 
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb CNA iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);
	else
		len = snprintf(buf, PAGE_SIZE_4K, "Emulex OneConnect OCe10100 10Gb CNA iSCSI Initiator Port %d\n",
				phba->pcidev->devfn % 2);

	spin_unlock_irqrestore(shost->host_lock, flags);
	return len;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_fw_version_attr = {
#else
static struct device_attribute be2iscsi_fw_version_attr = {
#endif
	.attr = {
		.name =		"fw_version",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_fw_version,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_port_name_attr = {
#else
static struct device_attribute be2iscsi_port_name_attr = {
#endif
	.attr = {
		.name =		"port_name",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_port_name,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_driver_version_attr = {
#else
static struct device_attribute be2iscsi_driver_version_attr = {
#endif
	.attr = {
		.name =		"driver_version",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_driver_version,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_ip_address_attr = {
#else
static struct device_attribute be2iscsi_ip_address_attr = {
#endif
	.attr = {
		.name =		"ip_address",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_ip_address,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_gateway_attr = {
#else
static struct device_attribute be2iscsi_gateway_attr = {
#endif
	.attr = {
		.name =		"gateway",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_gateway,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_netmask_attr = {
#else
static struct device_attribute be2iscsi_netmask_attr = {
#endif
	.attr = {
		.name =		"netmask",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_netmask,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_mac_address_attr = {
#else
static struct device_attribute be2iscsi_mac_address_attr = {
#endif
	.attr = {
		.name =		"mac_address",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_mac_address,
};


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_adapter_iqn_attr = {
#else
static struct device_attribute be2iscsi_adapter_iqn_attr = {
#endif
	.attr = {
		.name =		"adapter_iqn",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_adapter_iqn,
};


#ifndef __VMKERNEL_MODULE__
static struct bin_attribute be2iscsi_login_state_attr = {
	.attr = {
		.name = "login_state",
		.mode = S_IRUSR,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0)
		.owner = THIS_MODULE,
#endif
	},
	.size = 0,
	.read = be2iscsi_read_login_state,
};

static struct bin_attribute be2iscsi_target_iqn_attr = {
	.attr = {
		.name = "target_iqn",
		.mode = S_IRUSR,
#if LINUX_VERSION_CODE < KERNEL_VERSION(3,0,0)
		.owner = THIS_MODULE,
#endif
	},
	.size = 0,
	.read = be2iscsi_read_target_iqn,
};

static struct sysfs_entry {
	char *name;
	struct bin_attribute *attr;
} bin_file_entries[] = {
	{ "login_state", &be2iscsi_login_state_attr },
	{ "target_iqn", &be2iscsi_target_iqn_attr },
	{ 0 },
};

void
be2iscsi_alloc_sysfs_attr(se_hba *ha)
{
	struct Scsi_Host *host = ha->host;
	struct sysfs_entry *entry;
	int ret;

	for (entry = bin_file_entries; entry->name; entry++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
		ret = sysfs_create_bin_file(&host->shost_classdev.kobj,
#else
		ret = sysfs_create_bin_file(&host->shost_dev.kobj,
#endif
		    entry->attr);
		if (ret)
			printk("Unable to create sysfs %s binary attribute "
			    "(%d).\n", entry->name, ret);
	}
}

void
be2iscsi_free_sysfs_attr(se_hba *ha)
{
	struct Scsi_Host *host = ha->host;
	struct sysfs_entry *entry;

	for (entry = bin_file_entries; entry->name; entry++) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
		sysfs_remove_bin_file(&host->shost_classdev.kobj,
#else
		sysfs_remove_bin_file(&host->shost_dev.kobj,
#endif
		    entry->attr);
	}
}
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_adapter_state_attr = {
#else
static struct device_attribute be2iscsi_adapter_state_attr = {
#endif
	.attr = {
		.name =		"adapter_state",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_adapter_state,
};
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
static struct class_device_attribute be2iscsi_adapter_info_attr = {
#else
static struct device_attribute be2iscsi_adapter_info_attr = {
#endif
	.attr = {
		.name =		"adapter_info",
		.mode =		S_IRUGO,
	},
	.show = be2iscsi_show_adapter_info,
};


#ifndef __VMKERNEL__
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22)
struct class_device_attribute *be2iscsi_attrs[] = {
#else
struct device_attribute *be2iscsi_attrs[] = {
#endif
	&be2iscsi_fw_version_attr,
	&be2iscsi_driver_version_attr,
	&be2iscsi_ip_address_attr,
	&be2iscsi_gateway_attr,
	&be2iscsi_netmask_attr,
	&be2iscsi_mac_address_attr,
	&be2iscsi_adapter_iqn_attr,
	&be2iscsi_adapter_state_attr,
	&be2iscsi_adapter_info_attr,
	&be2iscsi_port_name_attr,
	NULL,
};
#endif
#endif
#endif
void
LnxLogOSMError(
    UEXACT32                   log_level,
    UEXACT32                   phase,
    UEXACT32                   sub_module,
    UEXACT32                   error_code)
/*++

Routine Description:

    This function is used to log OSM internal errors in system's
    event log.

Arguments:

    log_level    -  universal error code log level
    phase        -  universal error code phase
    sub_module   -  OSM execution submodule
    error_code   -  submodule specific error code

Return Value:

    Nothing

--*/
{
	UEXACT32 Code = 0;

	//
	// Set the error code to log in event viewer or VASE
	//

	if (log_level < UNI_CFG_LOG_LEVEL) {
		return;
	}

	Code = (
		((log_level       << UNI_ERROR_LEVEL_BIT_SHIFT) & UNI_ERROR_LEVEL_BIT_MASK) |
		((UNI_MODULE_OSM  << UNI_MODULE_BIT_SHIFT)      & UNI_MODULE_BIT_MASK)      |
		((sub_module      << UNI_SUBMODULE_BIT_SHIFT)   & UNI_SUBMODULE_BIT_MASK)   |
		((phase           << UNI_PHASE_BIT_SHIFT)       & UNI_PHASE_BIT_MASK)       |
		((error_code      << UNI_ERROR_CODE_BIT_SHIFT)  & UNI_ERROR_CODE_BIT_MASK)
	);

	be2iscsi_log_error(NULL, Code);
}
